//
//    Copyright 2002-2010 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "daqapp.h"

#include <getopt.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/stat.h>
#ifndef Q_NO_POSIX_SIGNALS
#include <csignal>
#endif

#include <QtGlobal>
#include <QCoreApplication>
#include <QTimer>
#include <QDebug>
#include <QDateTime>
#include <QDir>
#include <QSharedMemory>

#include <cstddef>
#include <iostream>
#include <ctime>
#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <cerrno>
#include <stdexcept>
#include <string>

#include "colorlog/ansi_col.h"
#include "daqcommon/daqexc.h"
#include "daqcommon/daq_ipc.h"
#include "daqcommon/daq.h"
#include "daqcommon/timer.h"
#include "storage.h"
#include "hal/hw_backend.h"
#include "hal_xml/hw_settings.h"
#include "runctrl.h"
#include "daq_debug.h"
#include "daqcommon/rawdata.h"
#include "hal/hwif_event.h"
#include "spillman.h"
#include "daqcommon/LinkStats.h"
#include "PacketReceiver.h"
#include "decoder/dvs.h"
#include "daqcommon/UnixTimeStamp.h"
#include "daqapp_p.h"
#include "startupconfig.h"

namespace {

#ifndef DEBUG_DAQ
#define DEBUG_DAQ 0
#endif

#define PREFIX "daq: "

    bool sigterm = false;
    //    bool sighup = false;

    bool static_endspill_flag = 0;

    const int eventWaitTimeoutMs = 10;

    QString settingsFileName()
    {
        const QString &homedir = QDir::homePath();
        return homedir + "/.vmedaq";
    }
}

void DaqApp::dataHandler(const QByteArray &data)
{
    const size_t start_byte_off = d->shm_pio_rawdata->count;
    const dma_buf_t *buf = static_cast<const unsigned int *>(static_cast<const void *>(data.constData()));
    const size_t len = data.size()/4;
    put_pio_rawdata_buf(buf, len, d->shm_pio_rawdata);
    daqApp->dma_data_handler(d->shm_pio_rawdata->data,
                             4 * start_byte_off, 4 * d->shm_pio_rawdata->count);
}

// end-of-burst, stop
bool DaqApp::wait_for_data_receive_complete()
{
    size_t last_count = 0;
    double start = UnixTimeStamp::sec();
    while (UnixTimeStamp::sec() < start + 0.1) {
        last_count = packetReceiver->handle_rx_pkt();
    }
    if (last_count) {
        qDebug() << __func__ << "Receive data not finished";
        return false;
    }
    return true;
}

//! decode and store data from dma_buf[vs...ve-1]
void DaqApp::handle_raw_words(const dma_buf_t *buf, size_t vs, size_t ve, bool endspill)
{
    Q_ASSERT(vs <= ve);
    size_t tmp_ve = ve;
    dvs->verify_raw(d->shm_evlist, buf, vs, &tmp_ve, endspill);

    //  if (ve+15 < dmaeng.get_dma_buf_size()) ve += 15;
    //  print_raw_burst_dump (buf, vs, ve);
    //  if (bs == BS_ON)
    bsize += (ve - vs);
    storage->write(buf, vs, ve);
}

// asyncronous data handler, called by dma_engine and by this program
void DaqApp::dma_data_handler(const dma_buf_t *buf, const size_t start_off, const size_t end_off)
{
    spillman->update_rawdata_count(end_off);
    handle_raw_words(buf, start_off/4, end_off/4, static_endspill_flag);
    spillman->update_event_count(dvs->nev);
}

void DaqApp::task_end_of_burst_readout()
{
    if (!d->shm_pio_rawdata) return;

    size_t start_byte_off = d->shm_pio_rawdata->count;
    put_pio_rawdata_word(ADLVME_DTYPE_SHDR | DATA_SHDR_ENDBURST, d->shm_pio_rawdata); // ENDBURST DATA BEGIN
    unsigned int evn = 0;
    unsigned int wcnt = 0;
    put_pio_rawdata_word(ADLVME_DTYPE_EHDR|evn, d->shm_pio_rawdata);

    task_end_of_burst();

    wcnt = d->shm_pio_rawdata->count - 2 - start_byte_off;
    put_pio_rawdata_word(ADLVME_DTYPE_ETRL|wcnt, d->shm_pio_rawdata);
    put_pio_rawdata_word(ADLVME_DTYPE_STRL | DATA_STRL_ENDBURST, d->shm_pio_rawdata); // ENDBURST DATA END
    static_endspill_flag = 1;
    daqApp->dma_data_handler(d->shm_pio_rawdata->data, 4 * start_byte_off, 4 * d->shm_pio_rawdata->count);
    static_endspill_flag = 0;
}

int DaqApp::decode_data_at_end_of_burst()
{
    if (dvs->nv >= 1) {
        unsigned int spill_count = spillman->get_spill_count();
        if (dvs->spill_number+1 != spill_count) {
            qDebug().nospace() << "Spill number: hardware count "
                    <<  dvs->spill_number+1 << ", counted " << spill_count;
            dvs->spill_number = spill_count-1;
        }
    }
    if (dvs->aborted) {
        qWarning() << "Data verification aborted";
    }
    qDebug().nospace() << "Verified: " << dvs->nv << " words, "
            << dvs->nev << " events, "
            << dvs->modnum << " of " << dvs->pnmod << " modules";
    {
        qDebug().nospace() << "Spill " << spillman->get_spill_count();
        {
            int nxoff = dvs->sptrcnt - dvs->nev;
            if (nxoff < 0)
                nxoff = 0;
            // Burst counts
            runcounts_t rc;
            rc.burst = 1;
            rc.xoff = nxoff;
            rc.event = dvs->nev;
            rc.byte = bsize * 4;
            rc.daqtime = spillman->getSpillTimeSeconds();
//            rc.walltime = wtime;
            d->shm_status->bc = rc;
            // Run counts
            runcounts_add(&d->shm_status->rc, rc);
        }
    }

    dvs->print_verify_summary();

    qDebug().nospace() << "Data size: " << 1. * dvs->nv / 256 / 1024 << " MB";

    storage->flush();
    return 0;
}

//------------------------------------------------------------------------------


void DaqApp::clearstats()
{
    //  runstats_clear(&rst);
    spillman->reset();
    runTimeSeconds = 0;
    bsize = 0;
    dvs->clear();
}


void DaqApp::initialize_database()
{
    Q_ASSERT(hwb);
    hwb->initialize_database();
}

void DaqApp::task_init()
{
    clearstats();
    if (!hwb) {
        try {
            hwb = new hw_backend(d->shm_status, d->shm_pio_rawdata);
        }
        catch (std::runtime_error &e) {
//            qCritical() << e.what();
            qFatal(e.what());
//            QTimer::singleShot(1000, this, SLOT(performRestart()));
            runcontrol->go_error();
            return;
        }
        catch (std::exception &e) {
//            qCritical() << e.what();
            qFatal(e.what());
            //          QCoreApplication::exit(EXIT_FAILURE);
//            QTimer::singleShot(1000, this, SLOT(performRestart()));
            runcontrol->go_error();
            return;
        }
    }
//    if (startupConfig->erase_ipc_memory) {
//        initialize_database();
//        startupConfig->erase_ipc_memory = false;
//    }
//    runcontrol->go_init();
    Q_ASSERT(hwb);
    hwb->start_communication();
    //  if (!no_startup_selftest)
    //	hwb->selftest();
    hwb->autodetect();
}

void DaqApp::task_initialize_database()
{
    initialize_database();
}

void DaqApp::task_reset()
{
    clearstats();
    daqApp->spillman->reset();
    d->run_reset(); // FIXME
    Q_ASSERT(hwb);
    hwb->reset();
    if (!startupConfig->no_startup_selftest)
        hwb->selftest();
}

void DaqApp::task_preparerun()
{
    d->clear_shm_pio_rawdata();

    Q_ASSERT(hwb);
    hwb->update_status();
    //  read_hardware_settings(settingsFileName(), *hwb);
    hwb->preparerun();
    if (startupConfig->write_calibration_data) {
        hwb->calibration();
        write_hardware_settings(settingsFileName(), *hwb);
        hwb->preparerun();
        qDebug() << __func__ << "Calibration complete, exiting";
        exit(1);
    }
    daqApp->storage->preparerun();
    if (d->shm_status) {
        if (qstrlen(d->shm_status->rp.fn)) {
            // sync_to_db is called in hwb->preparerun
            write_hardware_settings(QString::fromLocal8Bit(d->shm_status->rp.fn), *hwb);
        }
    }
}

void DaqApp::task_run()
{
    Q_ASSERT(hwb);
    hwb->start_run();
    d->clear_shm_pio_rawdata();
}


void DaqApp::task_pause()
{
    Q_ASSERT(hwb);
    hwb->stop_run();
    daqApp->wait_for_data_receive_complete();
}

void DaqApp::task_start_of_burst()
{
    dvs->clear();
    Q_ASSERT(d->shm_evlist);
    d->shm_evlist->clear();
    //  reset_dump_burst ();
    bsize = 0;
    Q_ASSERT(hwb);
    hwb->start_of_burst_signal();
}

void DaqApp::task_end_of_burst()
{
    runTimeSeconds += spillman->getSpillTimeSeconds();
    Q_ASSERT(hwb);
    hwb->endburst();
}

void DaqApp::task_interspill_test()
{
    Q_ASSERT(hwb);
    hwb->interspill_test();
}

bool DaqApp::get_hardware_events()
{
    Q_ASSERT(hwb);
    hwb->get_events(&hweq, eventWaitTimeoutMs);
    if (hweq.empty())
        return false;

    while (!hweq.empty()) {
        const hwifEvent &ev = hweq.top();
        switch (ev.getId()) {
        case hwifEvent::SPILL_ON:
            qDebug() << "--- Begin Spill ---";
            spillman->spill_start();
            //       d->shm_status->bs = bs;
            task_start_of_burst(); // clear counters, write d->shm_status
            storage->start_of_spill();
            //   conv_reset();
            break;

        case hwifEvent::SPILL_OFF:
            qDebug() << "--- End Spill ---";
            if (!spillman->spill_started()) break;      // skip first E-O-S signal
            spillman->spill_stop();
            // get remaining data, if available
            wait_for_data_receive_complete();
            //      runcontrol->;
            //	  usleep(100000); // flush data
            hwb->stop_run(); // required for ADLVME
            //       d->shm_status->bs = bs;

            hwb->end_of_burst_signal();
            dvs->hwif_spwc = hwb->get_spwc();
            dvs->check_endspill_data();
            task_end_of_burst_readout();
            decode_data_at_end_of_burst();

            task_interspill_test();
            {
                bool startNewRun = (ST_RUN == runcontrol->get_state());
                if (startNewRun) {
                    hwb->start_run(); // required for ADLVME
                    packetReceiver->printLinkStats();
                    packetReceiver->reset();
                    d->clear_shm_pio_rawdata();
                }
            }
            break;
        case hwifEvent::DATA:
            break;
        case hwifEvent::RXPKT:
            packetReceiver->handle_rx_pkt();
            break;
        }
        hweq.pop();
    }

    return true;
}

namespace {

#ifndef Q_NO_POSIX_SIGNALS
#ifndef RETSIGTYPE
# define RETSIGTYPE void
#endif
    static RETSIGTYPE master_sig_terminate(int sig)
    {
        qDebug().nospace() << "Terminating on signal " << sig
#if defined(Q_OS_UNIX)
                << " (" << strsignal(sig) << ")"
#endif
                ;
        sigterm = true;
        //  QCoreApplication::quit();
        //  exit (1);                     /* or _exit(1)? */
    }

    //static RETSIGTYPE signal_hup (int sig)
    //{
    //  log_notice("signal %d (%s)", sig, strsignal(sig));
    //  sighup = true;
    //}

#ifdef SIGUSR1
    static RETSIGTYPE signal_usr1(int sig)
    {
        qDebug().nospace() << "Caught signal " << sig
#if defined(Q_OS_UNIX)
                << " (" << strsignal(sig) << ")"
#endif
                ;
    }
#endif

    static void install_signal_handlers()
    {
#ifdef SIGHUP
        signal(SIGHUP, master_sig_terminate);
#endif
        //  signal(SIGHUP, signal_hup);
#ifdef SIGUSR1
        signal(SIGUSR1, signal_usr1);
#endif

        signal(SIGINT, master_sig_terminate);
#ifdef SIGQUIT
        signal(SIGQUIT, master_sig_terminate);
#endif
        signal(SIGTERM, master_sig_terminate);
    }
    static void restore_signal_handlers()
    {
#ifdef SIGHUP
        signal(SIGHUP, SIG_DFL);
#endif
#ifdef SIGUSR1
        signal(SIGUSR1, SIG_DFL);
#endif
        signal(SIGINT, SIG_DFL);
#ifdef SIGQUIT
        signal(SIGQUIT, SIG_DFL);
#endif
        signal(SIGTERM, SIG_DFL);
    }

#else
    static void install_signal_handlers() {}
    static void restore_signal_handlers() {}
#endif /* Q_NO_POSIX_SIGNALS */

    //const char *svn_str_get_rev(const char *rev) {
    //  char *result;
    //  char tmp[101];
    //  if (!rev) return ("");
    //  qstrncpy(tmp, rev, 100);
    //  tmp[100] = 0;
    //  if ((strlen(rev)>6) && !strncmp(rev, "$Rev: ", 6)) qstrncpy(tmp, &rev[6], 100);
    //  if (strlen(tmp)>2) {
    //        if (!strncmp(&tmp[strlen(tmp)-2], " $", 1)) tmp[strlen(tmp)-2] = 0;
    //  }
    //  result = strdup(tmp);
    //  return(result);
    //}

    void banner()
    {
        std::string str;

    #ifdef DEBUG
        str += " (debug)";
    #endif
        std::cerr << "Data Readout Program, ver " << VERSION << str << " rev." << SVN_REV
                << ", compiled " <<  __DATE__ << " " << __TIME__ << std::endl;
    }
}


DaqApp::DaqApp(int argc, char **argv, QObject *parent)
    : QObject(parent),
    startupConfig(new StartupConfig(argc, argv)),
    d(new DaqAppPrivate(*startupConfig)),
    packetReceiver(new PacketReceiver),
    runcontrol(new Run_State_Machine(d->shm_status, d->shm_ctl)),
    dvs(new dvs_t),
    runTimeSeconds(0.),
    bsize(0),
    storage(new Storage(d->shm_status)),
    spillman(new SpillManager(d->shm_status))
{
    banner();
    //    init_timer();
    //    install_exit_handler();
    install_signal_handlers();

    //    init_ipc(erase_ipc_memory);
    //  log_notice("********** Startup **********************************************");
    std::cerr << "*** VMEDAQ ver. " << VERSION << " rev. " << SVN_REV
            << ", compiled " << __DATE__ << " " << __TIME__ << " ***" << std::endl;

    connect(packetReceiver, SIGNAL(bytesRead(const QByteArray &)), this, SLOT(dataHandler(const QByteArray &)));
    connect(runcontrol, SIGNAL(task_init()), SLOT(task_init()));
    connect(runcontrol, SIGNAL(task_initialize_database()), SLOT(task_initialize_database()));
    connect(runcontrol, SIGNAL(task_pause()), SLOT(task_pause()));
    connect(runcontrol, SIGNAL(task_preparerun()), SLOT(task_preparerun()));
    connect(runcontrol, SIGNAL(task_reset()), SLOT(task_reset()));
    connect(runcontrol, SIGNAL(task_run()), SLOT(task_run()));
    runcontrol->set_automode(startupConfig->runcontrol_automode);
    hwb = 0;
//    QTimer::singleShot(0, this, SLOT(performRestart()));
}

DaqApp::~DaqApp()
{
    delete dvs;
    delete runcontrol;
    delete packetReceiver;
    delete spillman;
    delete storage;
    delete d;
    delete startupConfig;
}

void DaqApp::periodicTask()
{
    if (sigterm) {
        QCoreApplication::quit();
        return;
    }
    try {
        runcontrol->poll();
        if (hwb)
            get_hardware_events(); // sleeps
    }
    catch (std::runtime_error &e) {
        qFatal(e.what());
        qCritical() << e.what();
//        QTimer::singleShot(1000, this, SLOT(performRestart()));
        runcontrol->go_error();
        return;
    }
    catch (std::exception &e) {
        qFatal(e.what());
        qCritical() << e.what();
//        QTimer::singleShot(1000, this, SLOT(performRestart()));
        runcontrol->go_error();
        return;
    }
}

//void DaqApp::performRestart()
//{
//    if (!hwb) {
//        try {
//            hwb = new hw_backend(d->shm_status, d->shm_pio_rawdata);
//        }
//        catch (std::runtime_error &e) {
//            qCritical() << e.what();
////            QTimer::singleShot(1000, this, SLOT(performRestart()));
//            runcontrol->go_error();
//            return;
//        }
//        catch (std::exception &e) {
//            qCritical() << e.what();
//            //          QCoreApplication::exit(EXIT_FAILURE);
////            QTimer::singleShot(1000, this, SLOT(performRestart()));
//            runcontrol->go_error();
//            return;
//        }
//    }
////    if (startupConfig->erase_ipc_memory) {
////        initialize_database();
////        startupConfig->erase_ipc_memory = false;
////    }
//    runcontrol->go_init();
//}

void DaqApp::cleanup()
{
    qDebug() << "cleanup";
    restore_signal_handlers();
    delete hwb;
    hwb = 0;
}
