//
//    Copyright 2010 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#ifndef DAQAPP_P_CPP
#define DAQAPP_P_CPP

#include "daqapp_p.h"

#include <QDebug>

#include <stdexcept>

#include "daqcommon/daq_ipc.h"
#include "decoder/dvs.h"
#include "storage.h"
#include "startupconfig.h"

DaqAppPrivate::DaqAppPrivate(const StartupConfig &config)
    :
    startupConfig(new StartupConfig(config)),
    shmStatus("VMEDAQ_STATUS"),
    shmCtl("VMEDAQ_CONTROL"),
    shmEvlist("VMEDAQ_EVENTLIST"),
    shmRawdata("VMEDAQ_RAWDATA"),
    shm_status(NULL),
    shm_pio_rawdata(NULL),
    shm_evlist(NULL),
    shm_ctl(NULL)
{
    init_ipc();
}

DaqAppPrivate::~DaqAppPrivate()
{
    if (shm_ctl)
        shm_ctl->daq_pid = 0;
    delete startupConfig;
}

void DaqAppPrivate::clear_shm_pio_rawdata()
{
  if (!shm_pio_rawdata) return;
  shm_pio_rawdata->count = 0;
  shm_pio_rawdata->magic = SHM_RAWDATA_MAGIC;
//  memset(shm_pio_rawdata->data, 0, shm_pio_rawdata->size * sizeof(shm_pio_rawdata->data[0]));
}

void DaqAppPrivate::run_reset()
{                               // clear counters at new run
  if (shm_status) {
        shm_status->bc.clear();
        shm_status->rc.clear();
  }
  clear_shm_pio_rawdata();
}

shm_status_t *DaqAppPrivate::createStatus()
{
    createShm(shmStatus, sizeof(shm_status_t));
    return static_cast<shm_status_t *>(shmStatus.data());
}

shm_ctl_t *DaqAppPrivate::createCtl()
{
    createShm(shmCtl, sizeof(shm_ctl_t));
    return static_cast<shm_ctl_t *>(shmCtl.data());
}

shm_evlist_t *DaqAppPrivate::createEvlist()
{
    createShm(shmEvlist, sizeof(shm_evlist_t));
    return static_cast<shm_evlist_t *>(shmEvlist.data());
}

shm_pio_rawdata_t *DaqAppPrivate::createRawdata()
{
    size_t tmp_shm_size = startupConfig->requested_rawdata_size;
    while(true) {
        try {
            createShm(shmRawdata, tmp_shm_size);
            break;
        }
        catch (std::runtime_error &) {
            tmp_shm_size = (size_t)(tmp_shm_size*0.8);
            if (tmp_shm_size <= 1024*1024) { // absolute minimum
                break;
            }
        }
    }
    if (tmp_shm_size < startupConfig->requested_rawdata_size) {
        qDebug() << "Running with limited shared memory segment size: " << tmp_shm_size;
    }

    return static_cast<shm_pio_rawdata_t *>(shmRawdata.data());
}


void *DaqAppPrivate::createShm(QSharedMemory &shm, size_t size)
{
    bool ok = shm.attach(QSharedMemory::ReadWrite);
    if (!ok) {
        ok = shm.create(size, QSharedMemory::ReadWrite);
        if (!ok) {
            qCritical().nospace() << "createShm: Can't create shared memory segment " << shm.key() << ", size " << size << ": " << shm.errorString();
            throw std::runtime_error("shared memory error");
        }
    }
    if (!ok) {
        qCritical() << "createShm: Can't attach shared memory segment " << shm.key() << ":" << shm.errorString();
        throw std::runtime_error("shared memory error");
    }
    if (size_t(shm.size()) < size) {
        qCritical() << "createShm: Shared memory segment size mismatch, key " << shm.key() << ", requested size " << size << ", existing size " << shm.size();
        throw std::runtime_error("shared memory error");
    }
    Q_ASSERT(shm.data());
    Q_ASSERT(size_t(shm.size()) >= size);
    return shm.data();
}

void DaqAppPrivate::init_ipc(bool erase)
{
    shm_status = createStatus();
    Q_ASSERT(shm_status);
    if (erase || shm_status->magic != SHM_STATUS_MAGIC) {
        shm_status_reset(shm_status);
        shm_status->rp.cyclic = startupConfig->stor_cyclic;
        qstrncpy(shm_status->rp.fn, Storage::DEFAULT_STORAGE_FILENAME().toLocal8Bit(), 1024);
//        runcontrol->attach_shm_status(shm_status);
//        spillman->attach_shm_status(shm_status);
//        storage->attach_shm_status(shm_status);
    }

    shm_ctl = createCtl();
    Q_ASSERT(shm_ctl);

    shm_evlist = createEvlist();
    Q_ASSERT(shm_evlist);
    memset(shm_evlist, 0, sizeof(shm_evlist_t));
    shm_evlist->magic = SHM_EVLIST_MAGIC;

    shm_pio_rawdata = createRawdata();
    Q_ASSERT(shm_pio_rawdata);
    size_t wordsize = sizeof(shm_pio_rawdata->data[0]);
    size_t avail_data_bytes = shmRawdata.size() - sizeof(shm_pio_rawdata_t) + wordsize;
    shm_pio_rawdata->size = avail_data_bytes/wordsize;
    clear_shm_pio_rawdata();


//    size_t tmp_shm_size = requested_rawdata_size;
//#if defined(Q_OS_UNIX)
//// get available shm size
//  size_t shmmax = get_shmmax();
//  qDebug() << "Maximum shared memory segment size:" << shmmax / 1024/1024 << "MB";
//  if (shmmax < tmp_shm_size) {
//    qWarning() << "Not enough shared memory segment size, required" << tmp_shm_size/1024/1024 << "MB";
//    qWarning() << "Run: echo -n " << tmp_shm_size << " >/proc/sys/kernel/shmmax";
//    tmp_shm_size = shmmax;
//  }
//#endif
//// try to allocate in decreasing manner until success
//  shm_pio_rawdata = 0;
//  while (!shm_pio_rawdata) {
//        shm_pio_rawdata = static_cast<shm_pio_rawdata_t *>(shm_attach(SHM_RAWDATA_KEY, tmp_shm_size, 1, 0));
//        if (shm_pio_rawdata) {// success
//          break;
//        }
//        if (!shm_pio_rawdata && !(errno == ENOMEM || errno == EINVAL)) { // not a size problem
//          break;
//        }
//        tmp_shm_size = (size_t)(tmp_shm_size*0.8);
//        if (tmp_shm_size <= 1024*1024) { // absolute minimum
//          break;
//        }
//  }
//  if (shm_pio_rawdata) {
//        size_t wordsize = sizeof(shm_pio_rawdata->data[0]);
//        size_t avail_data_bytes = tmp_shm_size - sizeof(shm_pio_rawdata_t) + wordsize;
//        shm_pio_rawdata->size = avail_data_bytes/wordsize;
//        clear_shm_pio_rawdata();
//  }

//  if (tmp_shm_size < requested_rawdata_size) {
//    qDebug() << "Running with limited shared memory segment size";
//  }
}

#endif // DAQAPP_P_CPP
