
#include "runctrl.h"

#include <QDebug>
#include <QTimer>

#include <stdexcept>

#include "daqcommon/daq_ipc.h"
#include "daqcommon/RunStateTransitionRules.h"

const int errorRecoveryTimeMs = 3000;

RunControl::RunControl(shm_ctl_t *control)
    : shm_ctl(control),
    prev_ctl_reset_magic(0)
{
    set_default_ctl();
}


Run_State_Machine::Run_State_Machine(shm_status_t *status, shm_ctl_t *control)
    : IpcShmStatus(status),
    rc(new RunControl(control))
{
    connect(rc, SIGNAL(controlChanged(int)), SLOT(onControlChange(int)));
    ds = ST_OFF;
    automode = false;
//    errorRecoveryTimer.setSingleShot(true);
//    errorRecoveryTimer.setInterval(errorRecoveryTimeMs);
//    connect(&errorRecoveryTimer, SIGNAL(timeout()), SLOT(go_init()));
}

Run_State_Machine::~Run_State_Machine()
{
}

bool Run_State_Machine::isTransitionAllowed(daq_state_t from, daq_state_t to)
{
    return RunStateTransitionRules::isAllowed(from, to);
}

void Run_State_Machine::changestate(daq_state_t st)
{
    try {
//    qDebug() << QString("changestate: %1 -> %2").arg(daq_state_str(ds)).arg(daq_state_str(st));
    if (!isTransitionAllowed(ds, st)) {
        qWarning() << QString("changestate: invalid state transition %1 to %2")
                .arg(daq_state_str(ds)).arg(daq_state_str(st));
        return;
    }
    ds = st;
    if (shm_status)
        shm_status->ds = ds;
    // run task for state st
    switch (st) {
    case ST_OFF:
        break;
    case ST_INIT:
        task_init();
        break;
    case ST_RESET:
        task_reset();
        break;
    case ST_PREPARERUN:
        task_preparerun();
        break;
    case ST_RUN:
        task_run();
        break;
    case ST_PAUSE:
        task_pause();
        break;
    case ST_FAIL:
        QTimer::singleShot(errorRecoveryTimeMs, this, SLOT(go_init()));
//        errorRecoveryTimer.start();
        break;
    default:
        break;
    }
}
    catch(std::runtime_error &e) {
        qFatal(QString("Exception in task %1: %2").arg(daq_state_str(st)).arg(e.what()).toAscii());
//        qCritical() << QString("Exception in task %1: %2").arg(daq_state_str(st)).arg(e.what());
//        go_error();
    }
}

void Run_State_Machine::go_off()
{
    changestate(ST_OFF);
}

void Run_State_Machine::go_init()
{
    changestate(ST_INIT);
}

void Run_State_Machine::go_reset()
{
    changestate(ST_RESET);
}

void Run_State_Machine::go_run()
{
    changestate(ST_RUN);
}

void Run_State_Machine::go_preparerun()
{
    changestate(ST_PREPARERUN);
}

void Run_State_Machine::go_pause()
{
    changestate(ST_PAUSE);
}

void Run_State_Machine::go_error()
{
    changestate(ST_FAIL);
}

void Run_State_Machine::autoChangeState()            // automatic state change
{
    // INIT and RESET are performed automatically
    if (ds == ST_OFF)
        go_init();
    if (ds == ST_INIT)
        go_reset();

    if (automode) {
        if (ds == ST_RESET)
            go_preparerun();
        if (ds == ST_PREPARERUN)
            go_run();
    }
}

void Run_State_Machine::onControlChange(int runctl)
{
    switch (runctl) {
    case DAQCTL_CMD_INITIALIZE_DATABASE:
        task_initialize_database();
        break;
    case DAQCTL_CMD_INIT:
        go_init();
        break;
    case DAQCTL_CMD_RESET:
        go_reset();
        break;
    case DAQCTL_CMD_PREPARERUN:
        //    nburst = 0;
        go_preparerun();
        break;
    case DAQCTL_CMD_RUN:
        go_run();
        break;
    case DAQCTL_CMD_PAUSE:
        go_pause();
        break;
    case DAQCTL_CMD_ACK:
        // do nothing
        break;
    }
}

void RunControl::handle_reset_magic()
{
    ctl_magic_t runctl = shm_ctl->reset_magic;

    if (!shm_ctl)
        return;

    Q_EMIT controlChanged(runctl);
    shm_ctl->reset_magic = DAQCTL_CMD_ACK;
    prev_ctl_reset_magic = runctl;
}

void RunControl::handle_ctl_change()
{
//    qDebug() << QString("Control command %1").arg(shm_ctl->reset_magic, 8, 16, QChar('0'));

    if (shm_ctl->magic != DAQCTL_HDR_MAGIC) {
        qWarning() << "Bad control magic - set defaults";
        set_default_ctl();
        prev_ctl_reset_magic = shm_ctl->reset_magic;
    }

    handle_reset_magic();

    prev_ctl_reset_magic = shm_ctl->reset_magic;
}

void RunControl::set_default_ctl()
{
    *shm_ctl = shm_ctl_t();
    shm_ctl->magic = DAQCTL_HDR_MAGIC;
    shm_ctl->reset_magic = DAQCTL_CMD_ACK;
#if defined(Q_OS_UNIX)
    shm_ctl->daq_pid = getpid();
#endif
}

bool RunControl::control_changed()
{
    return shm_ctl->reset_magic != prev_ctl_reset_magic;
}

void RunControl::checkcontrol()
{
    if (control_changed())
        handle_ctl_change();
}

void Run_State_Machine::poll()
{
    autoChangeState();
    rc->checkcontrol();
}
