
#include "spillman.h"

#include <sys/time.h>

#include <cstdio>
#include <cstring>
#include <list>

#include "daqcommon/daq_ipc.h"
#include "daqcommon/sysutil.h"

namespace {
    const size_t sp_starts_size = 10;
    double tv_dt(const struct timeval t1, const struct timeval t2) {
        return (((t2.tv_sec)-(t1.tv_sec)) + 1e-6*((t2.tv_usec)-(t1.tv_usec)));
    }
}

SpillManager::SpillManager(shm_status_t *status)
    :
    IpcShmStatus(status),
    bs(BS_OFF),
    spr_p(0),
    raw_size(0),
    bsizep(0),
    ev_count(0),
    ev_countp(0),
    spill_count(0),
    spillTimeSeconds(0.)
{
  reset();
}

void SpillManager::update_shm_status()
{
  if (!shm_status) return;
  shm_status->bs = bs;
  shm_status->raw_size = raw_size;
  shm_status->spill_count = spill_count;
  shm_status->t_burstreset = tv_burstreset;
  shm_status->t_endspill = tv_endspill;
  update_spr();
}

void SpillManager::reset()
{
  bs = BS_OFF;
  raw_size = 0;
  ev_count = 0;
  spillTimeSeconds = 0.;
  sp_starts.clear();
  clear_spr();
  memset(&tv_burstreset, 0, sizeof(struct timeval));
  memset(&tv_endspill, 0, sizeof(struct timeval));
  update_shm_status();
}

void SpillManager::spill_start()
{
  bs = BS_ON;
  spill_count++;
  spillTimeSeconds = 0.;
  gettimeofday(&tv_burstreset, NULL);
  while (sp_starts.size() >= sp_starts_size)
        sp_starts.pop_front();
  sp_starts.push_back(tv_burstreset);
  update_shm_status();
}

void SpillManager::spill_stop()
{
  bs = BS_OFF;
  gettimeofday(&tv_endspill, NULL);
  spillTimeSeconds += tv_dt(tv_burstreset, tv_endspill);
  update_shm_status();
}

void SpillManager::update_rawdata_count(size_t bytes)
{
  raw_size = bytes;
  update_shm_status();
}

void SpillManager::update_event_count(size_t events)
{
  ev_count = events;
  update_shm_status();
}

daq_spr_t SpillManager::getSpr() const
{
    return shm_status->spr;
}

void SpillManager::setSpr(const daq_spr_t &spr)
{
    shm_status->spr =  spr;
}

void SpillManager::clear_spr()
{
//  bsize = 0;
  bsizep = 0;
  if (!shm_status) return;

  daq_spr_t spr;
  memset(&spr, 0, sizeof(spr));
  setSpr(spr);
/*
//  shm_status->spr_time = DAQ_SPR_TIME;
  for (int n = 0; n < DAQ_SPR_NUM; n++)
    for (int i = 0; i < DAQ_SPR_LEN; i++)
      shm_status->spr.v[n][i] = 0;
  shm_status->spr.end = 0;
*/
  spr_p = 0;
}

void SpillManager::update_spr()
{
  if (!shm_status) return;

  shm_status->spr.time = DAQ_SPR_TIME;

      int spri = 0;
      double bt = 0;
      struct timeval tnow;

      gettimeofday(&tnow, NULL);
      bt = tv_dt(tv_burstreset, tnow);
      spri = static_cast<int>(DAQ_SPR_LEN * bt / DAQ_SPR_TIME);
//      log_debug(0, "*** spri=%i, bt=%f", spri, bt);
      if (spri < 0)
        spri = 0;
      if (spri >= DAQ_SPR_LEN)
        spri = DAQ_SPR_LEN - 1;
      if (spri < spr_p) {       // new cycle. clear till end,
        for (int n = 0; n < DAQ_SPR_NUM; n++)
          for (int i = spr_p + 1; i < DAQ_SPR_LEN; i++)
            shm_status->spr.v[n][i] = 0;
        spr_p = 0;
      }
      if (spri >= spr_p) {      // old or new point
        for (int n = 0; n < DAQ_SPR_NUM; n++)
          for (int i = spr_p + 1; i < spri - 1; i++)
            shm_status->spr.v[n][i] = 0;
        shm_status->spr.v[0][spri] = bt;
      }
      if (spri > shm_status->spr.end)
        shm_status->spr.end = spri;
      if (spri > spr_p) {       // new point
// read-and-clear counters
        shm_status->spr.v[1][spri] = 1;
//       shm_status->spr[2][spri] = (int)(5*(double)rand()/RAND_MAX);
                size_t sz = raw_size;
        int bsinc = sz - bsizep;
        int evinc = ev_count - ev_countp;
//       log_debug(1, "bsize=%d", bsize);
        bsizep = sz;
        ev_countp = ev_count;
        if (bsinc < 0)
          bsinc = 0;
        if (evinc < 0)
          evinc = 0;
        shm_status->spr.v[2][spri] = bsinc;   // burst size, bytes
        shm_status->spr.v[3][spri] = evinc;
// ...
//       shm_status->spr.v[7][spri] = 1;
      }
  spr_p = spri;
// print
/*
  log_debug(0, "spr_time=%f, shm_status->spr_end=%d", shm_status->spr_time, shm_status->spr_end);
  for (int n=0; n<shm_status->spr_end; n++) {
        log_debug(0, "%e\t%e\t%e", shm_status->spr[0][n], shm_status->spr[1][n], shm_status->spr[2][n]);
        if (n>20) {
          log_debug(0, "...");
          break;
        }
  }
*/
}
