//
//    Copyright 2010 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "startupconfig.h"

#include <getopt.h>

#include <QCoreApplication>
#include <QFileInfo>

#include <cstring>
#include <cstdio>
#include <cstdlib>
#include <iostream>

#include "daqcommon/daq_ipc.h"
#include <colorlog/log.h>

StartupConfig::StartupConfig(int argc, char *const argv[])
    :
    no_startup_selftest(false),
    stor_cyclic(1),
    erase_ipc_memory(false),
    write_calibration_data(false),
    runcontrol_automode(true),
    requested_rawdata_size(DEFAULT_PIO_RAWDATA_BYTES)
{
    parse_cmdline(argc, argv);
}

void StartupConfig::parse_cmdline(int argc, char *const argv[])
{
    int c;
#ifdef DEBUG
    int dl = 0;
#endif
    size_t set_dma_size = 0;
    bool set_stor_keep = false;
    bool set_auto = false;
    bool set_noauto = false;
    bool set_notest = false;
    int go_exit = 0;
    bool show_help = false;
    char str[20];
    //  long d;
    //  char *endptr;
    //  struct option opt;
    int option_index = 0;

    static struct option long_options[] = {
#ifdef DEBUG
        {"debug", optional_argument, 0, 'd'},
#endif
        {"help", no_argument, 0, 'h'},
        {"buf-size", required_argument, NULL, 'b'},
        {"stor-keep", no_argument, 0, 'k'},
        {"auto", no_argument, 0, 'a'},
        {"noauto", no_argument, 0, 'm'},
        {"notest", no_argument, 0, 'n'},
        {"initialize", no_argument, 0, 'z'},
        {"calibrate", no_argument, 0, 'c'},
        {NULL, 0, NULL, 0}
    };
    str[0] = 0;
#ifdef DEBUG
    strcpy(str, "d::");
#endif
    strcat(str, "hb:kanmzc");
    while (1) {
        c = getopt_long(argc, argv, str, long_options, &option_index);
        if (c == -1)
            break;
        /*
   if ((c != ':')&&(c != '?')) {
      printf("option %s", long_options[option_index].name);
      if (optarg)
        printf(" with arg %s", optarg);
      printf("\n");
  }
*/
        switch (c) {
#ifdef DEBUG
        case 'd':
            if (optarg) {
                char *endptr = 0;
                long d = strtol(optarg, &endptr, 10);
                if ((*endptr == 0) && (d >= 0)) {
                    dl = d;
                } else {
                    fprintf(stderr, "bad option 'd' value: %s\n", optarg);
                }
            } else {
                dl++;
            }
            break;
#endif
    case 'h':
            show_help = 1;
            go_exit = 1;
            break;

    case 'z':
            erase_ipc_memory = true;
            break;

    case 'c':
            write_calibration_data = true;
            erase_ipc_memory = true;
            set_auto = 1;
            break;

    case 'k':
            set_stor_keep = true;
            break;

    case 'a':
            set_auto = true;
            break;

    case 'm':
            set_noauto = true;
            break;

    case 'n':
            set_notest = true;
            break;
    case 'b': {
            char *endptr = 0;
            long d = strtol(optarg, &endptr, 10);
            if ((*endptr == 0) && (d > 0)) {
                set_dma_size = d;
            } else {
                fprintf(stderr, "bad option '%c' value: %s\n", c, optarg);
            }
            break;
        }
    case '?':
        show_help = true;
        go_exit = 2;
        break;
    default:
        printf("?? getopt returned character %c (code 0%o) ??\n", c, c);
        show_help = 1;
        go_exit = 2;
        break;
    }
    }

    if (optind < argc) {
        printf("non-option ARGV-elements: ");
        while (optind < argc)
            printf("%s ", argv[optind++]);
        printf("\n");
        show_help = 1;
        go_exit = 2;
    }

    if (show_help) {
        usage();
        if (!go_exit)
            go_exit = 1;
        exit(go_exit - 1);
    }

    if (go_exit)
        exit(go_exit - 1);

#ifdef DEBUG
    if (dl) {
        printf("debug level: %d\n", dl);
        colorlog::set_debug_level(dl);
    }
#endif
    if (set_dma_size > 0) {
        size_t size_max = size_t(-1);
        size_t size_mb_max = size_max/1024/1024;
        if (set_dma_size > size_mb_max) {
            std::cerr << "invalid buf size specified: " << set_dma_size
                    << ", maximum is " << size_mb_max << "\n";
            set_dma_size = size_mb_max;
        }
        requested_rawdata_size = set_dma_size*1024*1024;
    }

    if (set_stor_keep) {
        stor_cyclic = 0;
    }
    if (set_auto && set_noauto) {
        std::cerr <<  "Use either --auto or --noauto, not both. Implied noauto mode." << std::endl;
    }
    if (set_auto)
        runcontrol_automode = set_auto;
    if (set_noauto)
        runcontrol_automode = !set_noauto;
    if (set_notest)
        no_startup_selftest = set_notest;

    if (go_exit)
        exit(go_exit - 1);
    //  fprintf(stderr, "getopts: dma: frag_size %d, buf_size %d\n", dma_frag_size, dma_buf_size);
}

void StartupConfig::usage()
{
    QFileInfo fi(qApp->argv()[0]);
    const QString &progname = fi.baseName();

    std::string str;

#ifdef DEBUG
  str += "[-d] ";
#endif
  std::cerr << "\nUsage: " << progname.toStdString() << " " << str << " [-t number]\n\n";
  std::cerr << "   -h, --help              show this help\n";
  std::cerr << "   -r, --raw-dump[=NUM]    enable raw dump (show first NUM lines)\n";
  std::cerr << "   -b, --buf-size=NUM      set rawdata buffer size to NUM megabytes (default " << DEFAULT_PIO_RAWDATA_BYTES/1024/1024 << ")\n";
  std::cerr << "   -k, --storage-keep      do not recycle storage data file every burst\n";
  std::cerr << "   -a, --auto              automatic mode (hardware constants are loaded default\n";
  std::cerr << "   -n, --notest            disable system selftest at startup\n";
  std::cerr << "   -z, --initialize        initialize configuration, all settings will be lost\n";
  std::cerr << "   -c, --calibrate         perform calibration and overwrite calibration file\n";
#ifdef DEBUG
  std::cerr << "   -d, --debug[=LEVEL]     set debug level to LEVEL, or increase by 1 if not specified\n";
#endif
  std::cerr << "\n";
}
