
#define _LARGEFILE64_SOURCE
#define _FILE_OFFSET_BITS 64

#include "storage.h"

#include <fcntl.h>
#include <unistd.h>

#include <QFileInfo>
#include <QDir>
#include <QDebug>

#include <cerrno>
#include <cstdlib>

Storage::Storage(shm_status_t *status)
    : IpcShmStatus(status)
{
    storage_initialized = false;
    enable_storage = true;
    storage_cyclic = true;
}

Storage::~Storage()
{
}

QString Storage::DEFAULT_STORAGE_FILENAME()
{
    return "/data/vmedaq.dat";
}

qint64 Storage::getnb() const
{
    QFileInfo fi(STORAGE_FILENAME);
    return fi.size();
}

void Storage::setopt(bool store, bool cyclic)
{
    enable_storage = store;
    storage_cyclic = cyclic;
}

void Storage::setfn(const QString &fn)
{
    if (!fn.isEmpty()) {
        STORAGE_FILENAME = fn;
    } else {
        STORAGE_FILENAME = DEFAULT_STORAGE_FILENAME();
    }
    STORAGE_FILENAME_TMP = STORAGE_FILENAME + ".tmp";

    qDebug() << QString("storage: data file '%1'").arg(STORAGE_FILENAME);
    qDebug() << QString("storage: tmp  file '%1'").arg(STORAGE_FILENAME_TMP);
}

void Storage::preparerun()
{
    update_setup();

    if (storage_cyclic) {
        int ret = unlink(STORAGE_FILENAME.toLocal8Bit());
        if (ret) {
            if (errno != ENOENT) {
                ret = -errno;
                goto err;
            }
        }
    }
    {
        int ret = unlink(STORAGE_FILENAME_TMP.toLocal8Bit());
        if (ret) {
            if (errno != ENOENT) {
                ret = -errno;
                goto err;
            }
        }
    }
    storage_initialized = true;
    update_statistics();
    return;
    err:
    qWarning() << QString("storage_preparerun: %1").arg(strerror(errno));
}

int Storage::start_of_spill()
{
    int ret;

    if (!storage_initialized)
        return (-EINVAL);

    if (!enable_storage)
        return 0;

    if (storage_cyclic) {
        ret = unlink(STORAGE_FILENAME_TMP.toLocal8Bit());
        if (ret) {
            if (errno != ENOENT) {
                ret = -errno;
                goto err;
            } else {
                ret = 0;
            }
        }
    }
    update_statistics();
    return 0;
    err:
    return (ret);
}

bool Storage::flush()
{
    if (!enable_storage)
        return true;

    if (!storage_initialized) {
        qWarning() << "storage_flush: not initialized";
        return false;
    }

    if (storage_cyclic) {
        unlink(STORAGE_FILENAME.toLocal8Bit());
        QFileInfo fi(STORAGE_FILENAME_TMP);
        if (!fi.exists())
            return false;
        QDir dir(fi.dir());
        if (!dir.rename(STORAGE_FILENAME_TMP, STORAGE_FILENAME)) {
            qWarning() << QString("Failed to rename %1 to %2")
                    .arg(STORAGE_FILENAME_TMP).arg(STORAGE_FILENAME);
            return false;
        }
    }
    update_statistics();
    return true;
}

int Storage::write(const dma_buf_t *buf, int start, int end)
{
    int fd, ret;
    ssize_t towr;
    ssize_t nwr;
    QString fn;

    if (!enable_storage)
        return 0;

    if (start < 0) {
        ret = -EINVAL;
        goto err;
    }
    if (start >= end)
        return 0;
    towr = 4 * (end - start);

//    qDebug() << QString("storage_write: start %1, end %2, towr %3 bytes")
//              .arg(start * 4).arg(end * 4).arg(towr);

    if (!storage_initialized) {
        qWarning() << "storage_write: not initialized";
        return 0;
    }

    if (storage_cyclic) {
        fn = STORAGE_FILENAME_TMP;
    } else {
        fn = STORAGE_FILENAME;
    }
    fd = open(fn.toLocal8Bit(), O_WRONLY | O_CREAT | O_APPEND, 0644);
    if (fd < 0) {
        ret = -errno;
        goto err;
    }

    if ((nwr = ::write(fd, &buf[start], towr)) != towr) {
        qWarning() << QString("storage_write: Can't write data file '%1': %2").arg(fn).arg(strerror(errno));
        ret = -errno;
        goto errclose;
    }
    if (::close(fd)) {
        ret = -errno;
        goto err;
    }
    update_statistics();
    return 0;
    err:
    qWarning() << QString("storage_write: %1").arg(strerror(-ret));
    return (ret);
    errclose:
    ::close(fd);
    qWarning() << QString("storage_write: %1").arg(strerror(-ret));
    return ret;
}

void Storage::update_setup()
{
    const runparam_t &rp = getRunParams();
    setopt(rp.store, rp.cyclic);
    setfn(QString::fromLocal8Bit(rp.fn));
}

void Storage::update_statistics()
{
    shm_status->stor_nb = getnb();
}
