
#include "histmgr.h"

#include <QDebug>
#include <QMetaType>
#include <QVector>
#include <QList>
#include <QTimer>
#include <qstringlist.h>
#include <qfile.h>
#include <qdom.h>

#include <cerrno>
#include <cstring>
#include <cstdlib>
#include <cmath>
#include <list>
#include <vector>
#include <map>
#include <string>
#include <sstream>

#include "daqcommon/daqexc.h"
#include "daqcommon/th1.h"
#include "hitsorter.h"
#include "hal/hw_frontend.h"
#include "daqcommon/dsplib.h"
#include "hptdc/hptdc.h"
#include "hptdc/hptdc_types.h"
#include "daqcfg.h"
#include "decoder/decoder.h"
#include "decoder/decoder_modlist_inc.h"
#include "modules/tqdc_types.h"
#include "modules/trig_types.h"
#include "decoder/tqdc_calibration_set.h"
#include "hal/base_mod.h"
#include "decoder/CalibrationData.h"
#include "foreach.h"
#include "decoder/inlcorr.h"


namespace {
//enum { DECODER_TDC_MAXHITS = 2050 }; /* 768 maximum for one tdc-96 */
enum { DECODER_TDC_MAXHITS = 3072 }; /* 768 maximum for one tdc-96 */

enum { DECODER_TDC_MAXCNT = 64 }; /* used only in MTDC64, maximum 2048 (encoder limit) */

const unsigned int DECODER_TQDC_MAX_T_HITS = 1024; /* 512 maximum for one tqdc-16 */
const unsigned int DECODER_TQDC_MAX_Q_HITS = 1024; /* 512 maximum for one tqdc-16 */

const int TQDC_Q_BITS = 14;
const int TQDC_Q_RANGE = 1 << TQDC_Q_BITS;
const int TQDC_Q_MIN = - (1 << (TQDC_Q_BITS-1));
const int TQDC_Q_MAX = (1 << (TQDC_Q_BITS-1)) - 1;

// BRICBOX DCH
enum { BRIC_BRANCHES = 16 };
enum { BRIC_WINDOWS = 16 };
enum { BRIC_BRANCH_CARDS = 32 };
enum { BRIC_CARDS = BRIC_BRANCHES*BRIC_BRANCH_CARDS };
enum { BRIC_WINDOW_BITS = 32*BRIC_BRANCH_CARDS };
enum { BRIC_WINDOW_BYTES = (BRIC_WINDOW_BITS+7)/8 };

#ifdef DEBUG
bool enableCalibration = true;
#else
bool enableCalibration = false;
#endif

}

/**
 * histmgr class implementation
 */
histmgr::histmgr (const hw_frontend &p, QObject *parent)
    :
      QObject(parent),
      hwf(p),
      hs(new hitsorter (16 * 4)),
      tqdc_cal(new tqdc_calibration_set()),
      calibrationData(GlobalCalibrationData::Instance())
{
    tdc_common_stop_channel = 0;
    save_trig_ts = 0;
    read_corrections();
#ifdef DEBUG
    QTimer *timer = new QTimer(this);
    timer->setInterval(20);
    timer->start();
    connect(timer, SIGNAL(timeout()), SLOT(fillDebugHist()));
//    QTimer *timer2 = new QTimer(this);
//    timer2->setInterval(500);
//    timer2->start();
//    connect(timer2, SIGNAL(timeout()), SLOT(sendDebugHist()));
#endif
}

histmgr::~histmgr()
{
    delete tqdc_cal;
    delete hs;
}

Q_DECLARE_METATYPE(QList<QVariant>)

bool histmgr::isCalibrationEnabled() const
{
    return enableCalibration;
}

//void histmgr::saveTdcInlCorr(QString tag, const hptdc_inl_corr_t &data)
//{
//    QList<QVariant> v;
//    for(int i=0; i<1024; i++) {
//        v.push_back(data[i]);
//    }
//    calibrationData->setValue(QString("%1-*-inl_corr/%2").arg(tag).arg(0), v);
//}

void histmgr::saveCalibrations()
{
    qDebug() << "saveCalibrations";
    if (DECODE_TDC) {
        for(int ch=0; ch< hwf.get_nch_byfam (MF_TDC); ch++) {
            HistId id = QString("tdc_dnl_%1").arg(ch);
            if (th1Ptr h = find (id)) {
                Q_ASSERT(h);
                qint64 nent = h->get_nent() - h->get_over() - h->get_under();
                if (nent <= 0)
                    continue;
                if (nent > 0 && nent < 100000) {
                    qDebug() << "Not enough hits for TDC DNL calibration o nchannel " << ch;
                    continue;
                }
                QVector<double> v(1024);
                for(int i=0; i<1024; i++) {
                    v[i] = h->get_bin(i);
                }
                QVector<double> ct = InlCorr::makeDNLCorrectionTable(v);
                int mch = 0;
                const base_mod *m = hwf.find_mod_by_fam_ch_const(MF_TDC, ch, &mch);
                Q_ASSERT(m);
                quint64 serial = m->get_serial();
                setInlCorr("TDC", serial, mch, ct);
            }
        }
    } // end if (DECODE_TDC)
    if (DECODE_TQDC) {
        for(int ch=0; ch< hwf.get_nch_byfam (MF_TQDC); ch++) {
            HistId id = QString("tqdc_dnl_%1").arg(ch);
            if (th1Ptr h = find (id)) {
                Q_ASSERT(h);
                qint64 nent = h->get_nent() - h->get_over() - h->get_under();
                if (nent <= 0)
                    continue;
                if (nent > 0 && nent < 10000) {
                    qDebug() << "Not enough hits for TQDC DNL calibration o nchannel " << ch;
                    continue;
                }
                QVector<double> v(1024);
                for(int i=0; i<1024; i++) {
                    v[i] = h->get_bin(i);
                }
                QVector<double> ct = InlCorr::makeDNLCorrectionTable(v);
                int mch = 0;
                const base_mod *m = hwf.find_mod_by_fam_ch_const(MF_TQDC, ch, &mch);
                Q_ASSERT(m);
                quint64 serial = m->get_serial();
                setInlCorr("TQDC", serial, mch, ct);
            }
        }
    } // end if (DECODE_TQDC)
}

void histmgr::setInlCorr(const QString &boardName, quint64 serialId, int channel, const QVector<double> &data)
{
    if (data.size() != 1024)
        return;
    QString key = QString("%1-%2-inl_corr/%3").arg(boardName).arg(serialId).arg(channel);
    QList<QVariant> v;
    for(int i=0; i<1024; i++)
        v.push_back(data[i]);
    calibrationData->setValue(key, v);
}

hptdc_inl_corr_t histmgr::getInlCorr(const QString &boardName, quint64 serialId, int channel)
{
    QList<QVariant> v;
    QString key = QString("%1-%2-inl_corr/%3").arg(boardName).arg(serialId).arg(channel);
    if (calibrationData->contains(key)) {
        v = calibrationData->value(key).toList();
    } else {
        key = QString("%1-%2-inl_corr/%3").arg(boardName).arg("").arg(channel);
        if (calibrationData->contains(key)) {
            v = calibrationData->value(key).toList();
        } else {
            key = QString("%1-%2-inl_corr/%3").arg(boardName).arg("").arg(0);
            if (calibrationData->contains(key)) {
                v = calibrationData->value(key).toList();
            } else {
                return hptdc_inl_corr_t();
            }
        }
    }
    qDebug() << "Read INL corrections " << key;
    QMap<int, double> m;
    for(int i=0; i<v.size(); i++)
        m[i] = v.at(i).toDouble();
    return m;
}

void histmgr::read_corrections()
{
    for(int ch=0; ch< hwf.get_nch_byfam (MF_TDC); ch++) {
        int mch = 0;
        const base_mod *m = hwf.find_mod_by_fam_ch_const(MF_TDC, ch, &mch);
        Q_ASSERT(m);
        quint64 serial = m->get_serial();
        tdc_inl_corrections[ch] = getInlCorr("TDC", serial, mch);
    }
    for(int ch=0; ch< hwf.get_nch_byfam (MF_TQDC); ch++) {
        int mch = 0;
        const base_mod *m = hwf.find_mod_by_fam_ch_const(MF_TQDC, ch, &mch);
        Q_ASSERT(m);
        quint64 serial = m->get_serial();
        tqdc_inl_corrections[ch] = getInlCorr("TQDC", serial, mch);
    }
}

namespace {
QDomDocument read_any_xml_file(const QString &filename, bool verbose = true)
{
    //  qlog_debug(QString("read: ")+filename+QString());
    bool st;
    QDomDocument doc;
    QFile file (filename);
    st = file.open (QIODevice::ReadOnly);
    if (!st) {
        if (verbose)
            qDebug (QString("Can't open file ") + filename);
        return QDomDocument();
    }
    if (!doc.setContent (&file)) {
        file.close ();
        if (verbose)
            qWarning (QString("Syntax error in file %1").arg(filename));
        return QDomDocument();
    }
    file.close ();
    return doc;
}
}

void histmgr::get_tdc_params(mfam_t mf, TdcResIndex *fres, TdcResIndex *fpres, int *fwin) const
{
    hptdc_setup_t tmp_hptdc_setup;
    hptdc_set_setup_defaults (&tmp_hptdc_setup);
    TdcResIndex res = TDC_RES_INDEX_100;
    hptdc_setup_get_resindex(tmp_hptdc_setup, &res); // default
    TdcResIndex pres = static_cast<TdcResIndex>(tmp_hptdc_setup.bit.width_select);
    int win = tmp_hptdc_setup.bit.match_window;
    try {
        res = static_cast<TdcResIndex>(hwf.get_int ("res", mf));
        pres = static_cast<TdcResIndex>(hwf.get_int ("pres", mf));
        win = hwf.get_int ("win", mf) + 1;
    }
    catch (exc &e) {
        // ignore, use defaults
    }
    if (fres) *fres = res;
    if (fpres) *fpres = pres;
    if (fwin) *fwin = win;
}

void histmgr::update_descriptions()
{
    QString fileBase = "vmedaq_histmgr.xml";
    QString homeDir = getenv("HOME");
    QStringList searchPath;
    searchPath += homeDir+"/.";
    searchPath += "/etc/";
    searchPath += "/usr/local/etc/";
    QString fileName;
    Q_FOREACH(const QString &basePath, searchPath) {
        QString path = basePath + fileBase;
        if (QFile::exists(path)) {
            fileName = path;
            break;
        }
    }
    if (fileName.isNull()) return;
    QDomDocument doc = read_any_xml_file(fileName);
    if (doc.isNull()) return;
    QString strInval = "Invalid file: ";
    QDomElement root = doc.documentElement();
    if (root.tagName() != "histmgr") {
        qWarning(strInval + "root node is not histmgr (found "+root.tagName()+")");
        return;
    }
    QDomNodeList nl = root.elementsByTagName("hist");
    for (int i=0; i<nl.count(); i++) {
        QDomElement hist = nl.item(i).toElement();
        QString id = hist.attribute("id");
        th1Ptr h = find(id);
        if (h.isNull())
            continue;
        QString s;
        s = hist.attribute("title");
        if (!s.isNull())
            h->setname(s);
        s = hist.attribute("description");
        if (!s.isNull())
            h->setDescription(s);
        s = hist.attribute("xlabel");
        if (!s.isNull())
            h->setXLabel(s);
        s = hist.attribute("ylabel");
        if (!s.isNull())
            h->setYLabel(s);
        s = hist.attribute("access");
        if (!s.isNull())
            h->setAccessLevel(s.toInt());
    }
}

QList<HistDescr> histmgr::getHistDescrs() const
{
    bool CREATE_ALWAYS = false;
    QList<HistDescr> dl;
//    qDebug() << "Creating histograms - common";
    dl.push_back(HistDescr("event_size", "Event Size", 10240, 0, 10240));
    dl.push_back(HistDescr("event_size_history", "Event Size History", 1000, 0, 1000));
    dl.push_back(HistDescr("burst_history", "Spill Size History", 1000, 0, 1000));
    //  dl.push_back(HistDescr("event_time_dist", "Event Time Distance, us", 1000, 0, 10000));
#ifdef DEBUG
    dl.push_back(HistDescr("debug_hist_1", "debug histogram 1", 50, 0, 50));
    dl.push_back(HistDescr("debug_hist_2", "debug histogram 2", 4, 0, 4));
#endif
    if (DECODE_BRIC) {
        int nch = hwf.get_nch_byfam (MF_DCH);
        if (CREATE_ALWAYS || nch > 0) {
            const int channelCount = BRIC_BRANCHES*BRIC_WINDOW_BITS;
//            qDebug() << QString("Creating histograms - BRIC, %1 branches").arg(BRIC_BRANCHES);
            //  	dl.push_back(HistDescr("dch_win_b0", "DCH window counts, Branch 1", BRIC_WINDOWS, 0, BRIC_WINDOWS));
            //  	dl.push_back(HistDescr("dch_win_b1", "DCH window counts, Branch 2", BRIC_WINDOWS, 0, BRIC_WINDOWS));
            //  	dl.push_back(HistDescr("dch_win_count", "DCH event: window count", BRIC_WINDOWS*BRIC_BRANCHES, 0, BRIC_WINDOWS*BRIC_BRANCHES));
            dl.push_back(HistDescr("dch_card_event_count", "DCH card event count", BRIC_CARDS, 0, BRIC_CARDS));
            dl.push_back(HistDescr("dch_win", "DCH hits by window", BRIC_WINDOWS+1, 0, BRIC_WINDOWS+1));
            dl.push_back(HistDescr("dch_event_hits", "DCH hits in event", channelCount, 0, channelCount));
            dl.push_back(HistDescr("dch_ch", "DCH channels", channelCount, 0, channelCount));
            dl.push_back(HistDescr("dch_ch_real", "DCH mapped channels", channelCount, 0, channelCount));
            dl.push_back(HistDescr("dch_rpk", "DCH cards", BRIC_BRANCHES*BRIC_WINDOW_BITS/32, 0, BRIC_BRANCHES*BRIC_WINDOW_BITS/32));
            dl.push_back(HistDescr("dch_hits_rpk", "DCH hits by card", BRIC_BRANCHES*BRIC_WINDOW_BITS/32, 0, BRIC_BRANCHES*BRIC_WINDOW_BITS/32));
            dl.push_back(HistDescr("dch_branch_len", "DCH branch length", 1024, 0, 1024)); // use 1024 instead of BRIC_BRANCH_CARDS
        }
    }
    if (DECODE_TRIG) {
        int nch = hwf.get_nch_byfam (MF_TRIG);
        if (CREATE_ALWAYS || nch > 0) {
//            qDebug() << QString("Creating histograms - TRIG, nch %1").arg(nch);
            dl.push_back(HistDescr("trig_incnt", "Trigger input count (all)", TRIG_N_INPUTS, 0, TRIG_N_INPUTS));
            dl.push_back(HistDescr("trig_bits", "Trigger bits", TRIG_N_BITS, 0, TRIG_N_BITS));
            dl.push_back(HistDescr("trig_states", "Trigger states", 1<<TRIG_LOGIC_WIRES, 0, 1<<TRIG_LOGIC_WIRES));
            dl.push_back(HistDescr("trig_states_read", "Trigger states - read", 1<<TRIG_LOGIC_WIRES, 0, 1<<TRIG_LOGIC_WIRES));
            dl.push_back(HistDescr("trig_states_all", "Trigger states - all", 1<<TRIG_LOGIC_WIRES, 0, 1<<TRIG_LOGIC_WIRES));
            dl.push_back(HistDescr("trig_evd", "Event Distance (coarse), us", 1000, 0, 100000));
            dl.push_back(HistDescr("trig_evd_fine", "Event Distance (fine), us", 1000, 0, 1000));
        }
    }
    if (DECODE_TDC) {
        int nch = hwf.get_nch_byfam (MF_TDC);
        // TDC
        if (CREATE_ALWAYS || nch > 0) {
//            qDebug() << QString("Creating histograms - TDC, nch %1").arg(nch);
            mfam_t mf = MF_TDC;
            // prepare defaults
            if (nch < 32) nch = 32; // meaningful defaults to prevent null-sized histograms
            TdcResIndex res, pres;
            int win;
            get_tdc_params(mf, &res, &pres, &win);
            int tdcrawmax = (win * tdc_rawfbins (res));
            double tdcrealbin = tdc_res_ns (res);
            double tdcrealmax = tdcrawmax * tdcrealbin;
            dl.push_back(HistDescr("nhits_tdc", "TDC Hits per Event", DECODER_TDC_MAXHITS, 0, DECODER_TDC_MAXHITS));
            dl.push_back(HistDescr("tdc_ch", "TDC Hits by Channel", nch, 0, nch));
            dl.push_back(HistDescr("tdc_input_ch", "TDC Input Counts by Channel", nch, 0, nch));
            dl.push_back(HistDescr("tdc_rawtime", "TDC RAW Time", 1024, 0, tdcrawmax));
            dl.push_back(HistDescr("tdc_realtime", "TDC Time, ns", 1024, 0, tdcrealmax));
            int raw_fbins = win*tdc_rawfbins(res);
            //            if (raw_fbins > 8192) raw_fbins = 8192;
            if (raw_fbins > 4*8192) raw_fbins = 4*8192;
            dl.push_back(HistDescr("tdc_startstopraw", "TDC Start-Stop RAW Time", 2*raw_fbins, -raw_fbins, raw_fbins));
            dl.push_back(HistDescr("tdc_startstop", "TDC Start-Stop Time, ns", 2*raw_fbins, -tdcrealbin * raw_fbins, tdcrealbin * raw_fbins));
            //    dl.push_back(HistDescr("tdc_width", "TDC Pulse Width, ns", 0x7E, 0, 0x7E*25./256.*(1<<pres)));
            dl.push_back(HistDescr("tdc_width", "TDC Pulse Width, ns", raw_fbins, 0, tdcrealmax));
            dl.push_back(HistDescr("tdc_errors", "TDC Error Bits", 32, 0, 32));
            //    dl.push_back(HistDescr("tdc_dlltime", "TDC DLL Time", 32, 0, 32));
            dl.push_back(HistDescr("tdc_cd", "TDC Code Density", 1024, 0, 1024));
            //            for(int i=0; i<nch; i++) {
            //                QString s = QString::number(i);
            //                dl.push_back(HistDescr("tdc_cd_"+s, "TDC Code Density (channel "+s+")", 1024, 0, 1024));
            //            }
            if (enableCalibration)
                for(int n=0; n<nch; n++)
                    dl.push_back(HistDescr(QString("tdc_dnl_%1").arg(n), QString("TDC INL %1").arg(n), 1024, 0, 1024));
        }
    }
    if (DECODE_TQDC) {
        // TQDC
        int nch = hwf.get_nch_byfam (MF_TQDC);
        if (CREATE_ALWAYS || nch > 0) {
//            qDebug() << QString("Creating histograms - TQDC, nch %1").arg(nch);
            mfam_t mf = MF_TQDC;
            dl.push_back(HistDescr("nhits_tqdc_t", "TQDC NHits T", DECODER_TQDC_MAX_T_HITS, 0, DECODER_TQDC_MAX_T_HITS));
            dl.push_back(HistDescr("nhits_tqdc_q", "TQDC NHits Q", DECODER_TQDC_MAX_Q_HITS, 0, DECODER_TQDC_MAX_Q_HITS));
            dl.push_back(HistDescr("nhits_tqdc_x", "TQDC NHits T - NHits Q, != 0", 200, -100, 100));
            dl.push_back(HistDescr("tqdc_ch_rawhits", "TQDC RAW Hits", nch, 0, nch));
            dl.push_back(HistDescr("tqdc_ch_t", "TQDC T Channels", nch, 0, nch));
            dl.push_back(HistDescr("tqdc_ch_q", "TQDC Q Channels", nch, 0, nch));
            dl.push_back(HistDescr("tqdc_ch_sumq", "TQDC Channels: SumQ", nch, 0, nch));
            dl.push_back(HistDescr("tqdc_ch_sumq_raw", "TQDC Channels: SumQ RAW", nch, 0, nch));
            TdcResIndex res, pres;
            int win;
            get_tdc_params(mf, &res, &pres, &win);
            int tdcrawmax = (win + 1) * tdc_rawfbins (res);
            double tdcrealbin = tdc_res_ns (res);
            double tdcrealmax = tdcrawmax * tdcrealbin;
            int raw_fbins = win*tdc_rawfbins(res);
            //            if (raw_fbins > 8192) raw_fbins = 8192;
            if (raw_fbins > 4*8192) raw_fbins = 4*8192;
            //    printf("tqdc: res=%d, tdcrawmax=%d, tdcrealmax=%f, tdcrealbin=%f ns\n", res, tdcrawmax, tdcrealmax, tdcrealbin);
            dl.push_back(HistDescr("tqdc_rawtime", "TQDC RAW Time", 1024, 0, tdcrawmax));
            dl.push_back(HistDescr("tqdc_rawtime_2p", "TQDC Double-Pulse RAW Time", tdcrawmax, 0, tdcrawmax));
            dl.push_back(HistDescr("tqdc_realtime", "TQDC Time, ns", 1024, 0, tdcrealmax));
            //            dl.push_back(HistDescr("tqdc_realtime_2p", "TQDC Double-Pulse Time, ns", 1024, 0, tdcrealmax));
            dl.push_back(HistDescr("tqdc_realtime_2p", "TQDC Double-Pulse Time, ns", 2*raw_fbins, -tdcrealbin * raw_fbins, tdcrealbin * raw_fbins));
            dl.push_back(HistDescr("tqdc_dlltime", "TQDC DLL Time", 32, 0, 32));
            dl.push_back(HistDescr("tqdc_dnl", "TQDC INL", 1024, 0, 1024));
            dl.push_back(HistDescr("tqdc_rctime", "TQDC RC Time", 5, 0, 5));
            dl.push_back(HistDescr("tqdc_charge", "TQDC Charge", TQDC_Q_RANGE, TQDC_Q_MIN, TQDC_Q_MAX)); // default
            dl.push_back(HistDescr("tqdc_ts", "TQDC TS", 2048, -1024, 1024));

            dl.push_back(HistDescr("tqdc_tq_time", "TQDC T-Q Time", 1000, 0, 100));
            dl.push_back(HistDescr("tqdc_trig_ts", "TQDC Trig TS % 5", 6, 0, 6));
            dl.push_back(HistDescr("tqdc_hit_ts", "TQDC Hit TS % 5", 6, 0, 6));
            dl.push_back(HistDescr("tqdc_errors", "TQDC Error Bits", 32, 0, 32));
            if (enableCalibration)
                for(int n=0; n<nch; n++)
                    dl.push_back(HistDescr(QString("tqdc_dnl_%1").arg(n), QString("TQDC INL %1").arg(n), 1024, 0, 1024));
        }
    }
    return dl;
}

void histmgr::create_hists ()
{
    const QList<HistDescr> &dl = getHistDescrs();
    hl.clear ();
    Q_FOREACH(HistDescr hd, dl) {
        hl.insert(hd.id, th1Ptr(new th1(hd)));
    }
    HistList histList;
    Q_FOREACH(HistDescr hd, dl) {
        HistListItem item;
        item.id = hd.id;
        item.title = hd.title;
        histList.push_back(item);
    }
    emit histListChanged(histList);

    rescale_hists();
    read_corrections();
    update_descriptions();
}

void histmgr::delete_hists ()
{
    hl.clear ();
}

void histmgr::rescale_hists ()
{
    //    try {
    //        int n_int = 32; // default
    //        n_int = hwf.get_int ("adc_match", MF_TQDC);
    //        rescale_hist("tqdc_charge", (1024 * n_int), -512, 512);
    //    } catch (exc &e) { }
}

void histmgr::rescale_hist(const HistId &name, size_t new_nbins, double new_xmin, double new_xmax)
{
    th1Ptr h = find(name);
    if (!h) return;
    h->setrange(new_xmin, new_xmax);
    h->rebin(new_nbins);
}

void histmgr::burstreset (const DaqCfg &cfg)
{
    //  printf("histmgr::burstreset\n");
    /*
  if (tmpfile)
        fclose(tmpfile);
  tmpfile = fopen ("tmp.txt", "w");
//    rewind (tmpfile);
    tqdc_cal->print();
*/

    if (cfg.get_hist_clear_by_burst()) {
        hold(true);
        clear();
    } else {
        if (th1Ptr h = find ("event_size_history"))
            h->shift_right ();
        if (th1Ptr h = find ("burst_history"))
            h->shift_right ();
        hold(false);
    }
}

void histmgr::endburst (const DaqCfg &cfg)
{
    //  printf("histmgr::endburst\n");
    if (cfg.get_hist_clear_by_burst())
        hold(false);
}

//! Freeze current state: disable update, but continue fill
void histmgr::hold (bool state)
{
    Q_FOREACH(const th1Ptr &p, hl)
        if (p) p->hold(state);
}

//! Perform full initialization of histograms, discarding temporary state
void histmgr::reset ()
{
    Q_FOREACH(const th1Ptr &p, hl)
        if (p) p->reset ();
    read_corrections();
}

//! Reset only current data, preserving state
void histmgr::clear ()
{
    Q_FOREACH(const th1Ptr &p, hl)
        if (p) p->clear ();
}

const th1Ptr histmgr::getHist (const HistId &id) const
{
    return hl.value(id);
}

const th1Ptr histmgr::getHistByName (const QString &name) const
{
    Q_FOREACH(th1Ptr p, hl) {
        if (p && (p->getname () == name))
            return p;
    }
    return th1Ptr();
}


/*
int histmgr::write ()
{

  if (!hl)
    return -1;
  for (int i = 0; i < hl->n; i++)
    hl->items[i].h->save_to_file ();
  return 0;
}
*/
th1_list_t histmgr::get_hist_list () const
{
    return (hl);
}

bool histmgr::get_hist_enable (const HistId &id) const
{
    Q_UNUSED(id);
    /*
  if (!cfg) return(true);
  bool en_dt = false;
  bool en_ch[ADC_NCH];
  bool en_any_m= false, en_any_s = false, en_any;

  for (int i=0;i<ADC_NCH;i++) {
        int trig = cfg->adc.trig[i];
        int mask_m = (ADCM_CCTRL_BIT_M | ADCM_CCTRL_BIT_CMPEN);
        int mask_s = (ADCM_CCTRL_BIT_S | ADCM_CCTRL_BIT_CMPEN);
        bool en_m = ((trig & mask_m) == mask_m);
        bool en_s = ((trig & mask_s) == mask_s);
// bit ADCM_CCTRL_BIT_TEST is ignored
        if (en_m) en_any_m = true;
        if (en_s) en_any_s = true;
        en_ch[i] = (en_m||en_s);
  }
  en_any = (en_any_m || en_any_s);
  en_dt = (en_any_m && en_any_s);
  if (index == h_dT_index) return en_dt;
  if (index == h_Tm_index) return en_any_m;
  if (index == h_Ts_index) return en_any_s;
  if (index == h_Ti_index) return en_any;
  if (index == h_Tff_index) return en_any;
  if (index == h_chi2_index) return en_any;
  if (index == h_niter_index) return en_any;
  if (index == h_w_index) return en_any;
  if ((index >= h_Int_index) && (index < h_Int_index+ADC_NCH)) return en_ch[index-h_Int_index];
*/
    return true;
}

//void histmgr::fill (const DaqCfg &cfg, const std::string &name, double value, size_t count)
//{
//    int dsf = cfg.get_hist_downscale();
//    if (dsf > 1) {
//        if (ranuni (0, dsf) >= 1)
//            return;                   // downscale it
//    }
//    if (th1Ptr h = find_by_name (name))
//        h->filln (value, dsf * count);
//}

void histmgr::fillHist(const HistId &id, double value, double count)
{
    if (th1Ptr h = find (id))
        h->filln(value, count);
}

void histmgr::fillHist(const HistId &id, const std::vector<double> &values, double count)
{
    if (th1Ptr h = find (id))
        h->fill(values, count);
}

void histmgr::fillHist(const HistId &id, const std::map<double, double> &values)
{
    if (th1Ptr h = find (id))
        h->fill(values);
}

void histmgr::fillGeneral (const event_t &ev, int dsf)
{
    //  adc_event_t *adc = &ev->adc;

    //  log_log_debug(1, "Event %d: TDC: %d hits, TQDC: %d T, %d Q hits", ev->en, tdc->n, tqdc->n_t, tqdc->n_q);

    fillHist("event_size", ev.size, dsf);
    //  if (th1Ptr h = find ("event_time_dist"))
    //    h->filln (ev->time_dist, dsf);

    fillHist("event_size_history", 0, ev.size*dsf);
    fillHist("burst_history", 0, dsf);
}

void histmgr::fillTDC (const event_t &ev, int dsf)
{
    const tdc_event_t &tdc = ev.tdc;
    const tdc_hits_t &hits = tdc.hits;
    const tdc_cnts_t &cnts = tdc.cnts;

    size_t nhits = 0;
    Q_FOREACH(int ch, hits.keys()) {
        nhits += hits[ch].size();
    }

    fillHist("nhits_tdc", nhits, dsf);
    if (nhits) {
        // Fill TDC histograms
        TdcResIndex tdc_resindex = static_cast<TdcResIndex>(hwf.get_int ("res", MF_TDC));
        int dnl_scaler = 1024 / tdc_rawfbins(tdc_resindex);
        bool vh = tdc_res_vh(tdc_resindex);
        int tdc_edge = hwf.get_int ("edge", MF_TDC);
        bool has_width = (tdc_edge == 3);
        //      int tdc_pres = hwf.get_int ("pres", MF_TDC);
        //      int dllbins = 32;
        //      if (tdc_res<0) dllbins*=4;
        double tbin = tdc_res_ns (tdc_resindex);   // FIXME
        //      log_info("res: %d, edge: %d, has_wid: %d, tbin: %f", tdc_res, tdc_edge, has_width, tbin);
        typedef std::vector<double> hc_ent_t;
        hc_ent_t hc[7];    // histogram cache

        //  log_log_debug(2, "Event %d: %d hits", ev->en, tdc->n);
        //      if (tdc->n <= 0) return;
        //  log_log_debug(1, "%d hits", tdc->n);

        // populate cache
        const int trig_ch = tdc_common_stop_channel; // trigger channel
        int t0_rawtime = 0;
        double t0_time = 0;
        bool t0_isset = false;

        Q_FOREACH(int ch, hits.keys()) {
            if (ch != trig_ch)
                continue;
            const tdc_hitv_t &hit = hits.value(ch);
            // get first hit
            if (hit.size() < 1)
                break;
            t0_rawtime = hit[0].time;
            int dnl_code = (t0_rawtime * dnl_scaler) % 1024;
            t0_rawtime += tdc_inl_corrections[ch][dnl_code] / dnl_scaler;
            t0_time = t0_rawtime * tbin;
            t0_isset = true;
            break;
        }

        std::vector<double> tdc_startstopraw_values;
        std::vector<double> tdc_startstop_values;
        Q_FOREACH(int ch, hits.keys()) {
            const tdc_hitv_t &hitv = hits.value(ch);
            Q_FOREACH(const tdc_hit_t &hit, hitv) {
                int raw_time = hit.time;
                int wid = hit.wid;
                int dlltime = (vh?(raw_time>>2):raw_time)%32;
                int rctime = raw_time % 4;
                int dnl_code = (raw_time * dnl_scaler) % 1024;
                raw_time += tdc_inl_corrections[ch][dnl_code] / dnl_scaler;
                double real_time = raw_time * tbin; // time in ns
                //      	  log_info("time: %d, real_time: %f", time, real_time);
                hc[0].push_back(ch);
                hc[1].push_back(raw_time);
                hc[2].push_back(dlltime);
                hc[3].push_back(rctime);
                hc[4].push_back(real_time);
                if (wid > 0)
                    hc[5].push_back(wid * tbin);
                hc[6].push_back(dnl_code);
                if (t0_isset && ch != trig_ch) {
                    tdc_startstopraw_values.push_back(raw_time - t0_rawtime);
                    tdc_startstop_values.push_back(real_time - t0_time);
                }
                if (enableCalibration) {
                    fillHist(QString("tdc_dnl_%1").arg(ch), dnl_code, dsf);
                }
            }
        }
        fillHist("tdc_startstop", tdc_startstop_values);
        fillHist("tdc_startstopraw", tdc_startstopraw_values);

        if (nhits > 0) {
            // FIXME: CPU consuming
            fillHist("tdc_ch", hc[0], dsf);
            fillHist("tdc_rawtime", hc[1], dsf);

            //                    fillHist("tdc_dlltime", hc[2], dsf);
            //        if (tdc_res_vh (tdc_res)) {  // FIXME
            //          fillHist("tdc_rctime", hc[3], dsf);
            //		}
            fillHist("tdc_realtime", hc[4], dsf);
            if (has_width) {
                fillHist("tdc_width", hc[5], dsf);
            }
            fillHist("tdc_cd", hc[6], dsf);
        }
    } // end of if (tdc->n)

    std::map<double, double> tdc_input_ch_values;
    Q_FOREACH(int ch, cnts.keys()) {
        tdc_cnt_t count = cnts.value(ch);
        tdc_input_ch_values[ch] = count;
    }
    fillHist("tdc_input_ch", tdc_input_ch_values);

    fillHist("tdc_errors", 31); // total count
    for (int b=0; b<31; b++) {
        if (tdc.error & (1<<b))
            fillHist("tdc_errors", b);
    }
}

void histmgr::fillBRIC (const event_t &ev, int dsf)
{
    const dch_event_t &dch = ev.bric;
    const dch_branches_t &branches = dch.branches;
    QMap<int, bool> rpk_present; // index: rpk number

    int event_hits = 0;
    std::vector<double> dch_ch_values;
    std::vector<double> dch_ch_real_values;
    std::vector<double> dch_rpk_values;
    std::vector<double> dch_hits_rpk_values;
    std::vector<double> dch_win_values;
    std::vector<double> dch_branch_len_values;

    Q_FOREACH(int branch_id, branches.keys()) {
        const dch_branch_t &branch = branches.value(branch_id);
        const dch_windows_t &windows = branch.wins;
        Q_FOREACH(int window_id, windows.keys()) {
            const dch_window_t &win = windows.value(window_id);
            unsigned int num_bits = win.bits.size()*8;
            if (num_bits>0) {
                unsigned int window_hits = 0;
                //        	log_debug(1, "b %u, w %u, num_bits %u", b, w, num_bits);
                for (unsigned int i=0; i<num_bits; i++) {
                    //      		  unsigned int ch = /*b*BRIC_WINDOW_BITS +*/ i;
                    unsigned int rpk = i/32; // card number in branch
                    int global_rpk = rpk + BRIC_BRANCH_CARDS*branch_id;
                    rpk_present[global_rpk] = true;
                    //      		  unsigned int mch = i%32; // channel in card
                    //      		  unsigned int ch = mch + rpk*64; // absolute channel in branch
                    //      		  if (b%2 != 1) ch += 32;
                    //      		  unsigned int ch = i + b*BRIC_WINDOW_BITS; // absolute channel
                    unsigned int ch = i/32*32 + (31-i%32) + branch_id*BRIC_WINDOW_BITS; // absolute channel
                    unsigned int ch_r = 2*(i/32*32 + (31-i%32)) + branch_id;

                    bool hit = win.bits[i/8] & (1<<(i%8));
                    if (hit) {
                        window_hits++;
                        dch_ch_values.push_back(ch);

                        dch_hits_rpk_values.push_back(rpk);
                        dch_ch_real_values.push_back(ch_r);
                    }
                    dch_rpk_values.push_back(rpk + BRIC_WINDOW_BITS/32*branch_id);
                    if (hit) dch_win_values.push_back(window_id);
                }
                //        	log_debug(0, "b %u, w %u, num_bits %u, hits %u", b, w, num_bits, window_hits);
                dch_branch_len_values.push_back(num_bits/32);
                event_hits += window_hits;
            } // end if (num_bits>0) //
        } // end of windows
    } // end for (b) //
    fillHist("dch_ch", dch_ch_values);
    fillHist("dch_ch_real", dch_ch_real_values);
    fillHist("dch_rpk", dch_rpk_values);
    fillHist("dch_hits_rpk", dch_hits_rpk_values);
    fillHist("dch_win", dch_win_values);
    fillHist("dch_branch_len", dch_branch_len_values);
    fillHist("dch_event_hits", event_hits);

    std::vector<double> dch_card_event_count_values;
    Q_FOREACH(int rpk, rpk_present.keys()) {
        bool is_rpk_present = rpk_present.value(rpk);
        if (is_rpk_present) {
            dch_card_event_count_values.push_back(rpk);
        }
    }
    fillHist("dch_card_event_count", dch_card_event_count_values);

}

void histmgr::fillTRIG (const event_t &ev, int dsf)
{
    const trig_event_t &trig = ev.trig;
    if (trig.ts != 0) {
        quint64 tsdiff = trig.ts - save_trig_ts;
        double evd = tsdiff* 1./80.; // 80MHz step, measure us
        //  	  log_log_debug(0, "evd=%e", evd);
        fillHist("trig_evd", evd);
        fillHist("trig_evd_fine", evd);
        save_trig_ts = trig.ts;
    }
    {
        std::vector<double> trig_bits_values;
        for (unsigned int i=0; i<TRIG_N_BITS; i++) {
            if (trig.trigbits & (1<<i))
                trig_bits_values.push_back(i);
        }
        fillHist("trig_bits", trig_bits_values);
    }
    {
        //  	  double hbuf_trig_in[TRIG_N_BITS];
        unsigned int logic_state = 0;
        for (unsigned int i=0; i<TRIG_N_BITS; i++) {
            //  		hbuf_trig_in[i] = (trig->trigbits & (1<<i));
            unsigned int this_bit_on = (trig.trigbits & (1<<i));
            if (this_bit_on) {
                int this_state_index = trig_get_state_by_bit(i);
                if (this_state_index >= 0) logic_state |= (1<<this_state_index);
            }
        }
        fillHist("trig_states", logic_state);
        //  	  h->fillbuf(hbuf_trig_in, 20);
    }
    //  	  log_log_debug(3, "trig->n = %d", trig->n);
    {
        std::map<double, double> trig_incnt_values;
        const trig_cnts_t &cnts = trig.incnt;
        Q_FOREACH (int ch, cnts.keys()) {
            trig_incnt_values[ch] = cnts.value(ch) * dsf;
        }
        fillHist("trig_incnt", trig_incnt_values);
    }
    {
        std::map<double, double> trig_states_all_values;
        const trig_cnts_t &cnts = trig.logic_all;
        Q_FOREACH (int ch, cnts.keys()) {
            trig_states_all_values[ch] = cnts.value(ch) * dsf;
        }
        fillHist("trig_states_all", trig_states_all_values);
    }
    {
        std::map<double, double> trig_states_read_values;
        const trig_cnts_t &cnts = trig.logic_read;
        Q_FOREACH (int ch, cnts.keys()) {
            trig_states_read_values[ch] = cnts.value(ch) * dsf;
        }
        fillHist("trig_states_read", trig_states_read_values);
    }
    //  } else {
    //    log_log_debug(0, "TRIG not found!");
}

QDebug operator<<(QDebug dbg, const tqdc_hits_q_t &hits)
{
    dbg.nospace() << "hits_q ";
    QStringList sl;
    Q_FOREACH(int ch, hits.keys()) {
        sl << QString("%1: %2").arg(ch).arg(hits[ch].size());
    }
    dbg.nospace() << sl.join(", ");
    return dbg;
}

void histmgr::fillTQDC (const event_t &ev, int dsf)
{
    hs->clear ();

    const tqdc_event_t &tqdc = ev.tqdc;
    const tqdc_hits_q_t &hits_q = tqdc.hits_q;
    const tqdc_hits_t_t &hits_t = tqdc.hits_t;
    //    const tqdc_wfs_t &wfs = tqdc.wfs;
    const tqdc_cnts_t &cnts = tqdc.cnts;
    int adc_mode = hwf.get_int ("adc_mode", MF_TQDC);
    /*
int nscale = 1;
if (adc_mode == 1 || adc_mode == 3)
  nscale = hwf.get_int ("adc_match", MF_TQDC);
if (nscale < 1)
  nscale = 1;
*/
    // Counters
    std::map<double, double> tqdc_ch_rawhits_values;
    Q_FOREACH (int ch, cnts.keys()) {
        quint32 count = cnts.value(ch);
        if (count > 0) {
            tqdc_ch_rawhits_values[ch] = count * dsf;
        }
    }
    fillHist("tqdc_ch_rawhits", tqdc_ch_rawhits_values);

    if (adc_mode == 1) { // calibration
        tqdc_cal->add(hits_q);
    }
    // sorter (hack)

    // count hits
    int num_hits_q = 0;
    Q_FOREACH(const tqdc_hitv_q_t &v, hits_q) num_hits_q += v.size();
    int num_hits_t = 0;
    Q_FOREACH(const tqdc_hitv_t_t &v, hits_t) num_hits_t += v.size();

    // Fill TQDC histograms

    std::map<double, double> nhits_tqdc_x_values;
    int size_diff = num_hits_t - num_hits_q;
    if (size_diff)
        nhits_tqdc_x_values[size_diff] = dsf;
    fillHist("nhits_tqdc_x", nhits_tqdc_x_values);

    int n_t = 0, n_q = 0;
    // TDC part
    fillHist("nhits_tqdc_t", num_hits_t, dsf);
    if (!hits_t.empty()) {
        TdcResIndex tdc_res = static_cast<TdcResIndex>(hwf.get_int ("res", MF_TQDC));
        bool vh = tdc_res_vh(tdc_res);
        int dnl_scaler = 1024 / tdc_rawfbins(tdc_res);
        double tbin = tdc_res_ns (tdc_res); // FIXME
        std::vector<double> hc[7];    // histogram cache
        //        unsigned int nh = tqdc->n_t;
        //        if (nh > DECODER_TQDC_MAX_T_HITS)
        //          nh = DECODER_TQDC_MAX_T_HITS;

        // populate cache
        const int tmp_t_sz = 2;
        double tmp_t[tmp_t_sz], tmp_raw_t[tmp_t_sz];
        int n_tmp_t=0;
        Q_FOREACH(int ch, hits_t.keys()) {
            const tqdc_hitv_t_t &hitv_t = hits_t.value(ch);
            Q_FOREACH(const tqdc_hit_t_t &hit, hitv_t) {
                int time = hit.time;
                int wid = hit.wid;
                int dlltime = (vh ? (time>>2) : time)%32;
                int dnl_code = (time * dnl_scaler) % 1024;
                time += tqdc_inl_corrections[ch][dnl_code] / dnl_scaler;
                double time_ns = time * tbin; // time in ns
                hc[0].push_back(ch);
                hc[1].push_back(time);
                hc[2].push_back(dlltime);
                hc[3].push_back(time % 4);
                hc[4].push_back(time_ns);
                hc[5].push_back(wid * 25./256.);
                hc[6].push_back(dnl_code);
                n_t++;
                hs->add_hit_t (ch, time_ns);
                if (n_tmp_t < tmp_t_sz) {
                    tmp_raw_t[n_tmp_t] = time;
                    tmp_t[n_tmp_t] = time_ns;
                    n_tmp_t++;
                }
                if (enableCalibration) {
                    fillHist(QString("tqdc_dnl_%1").arg(ch), dnl_code, dsf);
                }
            }
        }

        // FIXME: CPU consuming
        fillHist("tqdc_ch_t", hc[0], dsf);
        fillHist("tqdc_rawtime", hc[1], dsf);
        fillHist("tqdc_dlltime", hc[2], dsf);
        fillHist("tqdc_dnl", hc[6], dsf);
        fillHist("tqdc_realtime", hc[4], dsf);
        if (tdc_res_vh (tdc_res))   // FIXME
            fillHist("tqdc_rctime", hc[3], dsf);
        if (n_tmp_t == 2) {
            double tmp_raw_dt = tmp_raw_t[1]-tmp_raw_t[0];
            if (tmp_raw_dt<0) tmp_raw_dt *= -1;
            double tmp_dt = tmp_t[1]-tmp_t[0];
            if (tmp_dt<0) tmp_dt *= -1;
            //        	printf("tmp_raw_dt=%f, tmp_dt=%f\n", tmp_raw_dt, tmp_dt);
            fillHist("tqdc_rawtime_2p", tmp_raw_dt, dsf);
            fillHist("tqdc_realtime_2p", tmp_dt, dsf);
        }
    }
    {
        fillHist("tqdc_errors", 31); // total count
        for (int b=0; b<31; b++) {
            if (tqdc.tdc_error & (1<<b))
                fillHist("tqdc_errors", b);
        }
    }

    // QDC part
    fillHist("nhits_tqdc_q", num_hits_q, dsf);
    if (!hits_q.empty()) {
        typedef std::vector<double> hc_ent_t;
        hc_ent_t hc[3];    // histogram cache

        // populate cache
        Q_FOREACH(int ch, hits_q.keys()) {
            const tqdc_hitv_q_t &hitv_q = hits_q.value(ch);
            Q_FOREACH(const tqdc_hit_q_t &hit, hitv_q) {
                double rawcharge = hit.charge;
                int ts = hit.ts;
                short ts_t = (ts >> 16) & 0xFFFF;
                short ts_h = ts & 0xFFFF;
                //          int ts_d = (ts_h - ts_t) & 0xFFFF;
                int ts_d = ts_h - ts_t;
                //          printf("ts_t=%d, ts_h=%d, ts_d=%d\n", ts_t, ts_h, ts_d);
                //          double qraw = rawcharge * 1. / nscale;
                double qraw = rawcharge;
                double q = qraw - tqdc_cal->get_baseline(ch);
                //          if (th1Ptr h = find ("tqdc_ch_q"))
                //            h->filln (ch, dsf);

                hc[0].push_back(ch);
                hc[1].push_back(q);
                hc[2].push_back(ts_d);
                //          double t_q = ts_d * 10;
                n_q++;
                hs->add_hit_q (ch, ts_d * 10, q);
                fillHist("tqdc_trig_ts", ts_t % 5, dsf);
                fillHist("tqdc_hit_ts", ts_h % 5, dsf);
                fillHist("tqdc_ch_sumq", ch, dsf*q);
                fillHist("tqdc_ch_sumq_raw", ch, dsf*qraw);
                //        printf("ts:%08X ts_t:%d, ts_h:%d, ts_d:%d\n", ts, ts_t, ts_h, ts_d);
            }
        }

        fillHist("tqdc_ch_q", hc[0], dsf);
        fillHist("tqdc_charge", hc[1], dsf);
        fillHist("tqdc_ts", hc[2], dsf);
    }
    // sorter here
    //  double s_t[num_tqdc_ch], s_q[num_tqdc_ch];
    //  int s_n[num_tqdc_ch];
    // find only hits with |t_t-t_q| < T


    if (n_t == 1 && n_q == 1) {
        hs->sort_hits ();
        std::vector<double> tqdc_tq_time_values;
        hit_stq_buf_t *stqbuf = hs->get_stq_buf ();
        for (int c = 0; c < hs->get_nch (); c++) {
            int nh = stqbuf[c].n;
            for (int n = 0; n < nh; n++) {
                double t = stqbuf[c].hit[n].t;
                double s = stqbuf[c].hit[n].s;
                //                                double q = stqbuf[c].hit[n].q;
                tqdc_tq_time_values.push_back(t - s);
            }
        }
        fillHist("tqdc_tq_time", tqdc_tq_time_values);
    }

}

void histmgr::fill (const event_t &ev, const DaqCfg &cfg)
{
    if (!cfg.get_hist_en()) {
        qWarning() << "histmgr::fill - disabled!";
        return;
    }

    int dsf = cfg.get_hist_downscale();
    if (! ev.no_downscale) {
        if (dsf > 1) {
            if (ranuni (0, dsf) >= 1) {
                //      log_log_debug(0, "Downscaled!");
                return;               // downscale it
            }
        }
    }
//    qDebug() << "Histogram event" << ev.en;

//    hs->clear ();
    fillGeneral(ev, dsf);

    if (DECODE_TDC) {
        if (hwf.get_nch_byfam (MF_TDC) > 0) {
            fillTDC(ev, dsf);
        }
    } // end if (DECODE_TDC)

    if (DECODE_BRIC) {
        if (hwf.get_nch_byfam (MF_DCH) > 0) {
            fillBRIC(ev, dsf);
        } // end if (hwf.get_nch_byfam (MF_DCH) > 0) //
    } // end if (DECODE_BRIC)

    if (DECODE_TRIG) {
        // TRIG
        if (hwf.get_nch_byfam (MF_TRIG) > 0) {
            fillTRIG(ev, dsf);
        }
    } // end if (DECODE_TRIG)

    if (DECODE_TQDC) {
        if (hwf.get_nch_byfam (MF_TQDC) > 0) {
            fillTQDC(ev, dsf);
        }
    } // end if (DECODE_TQDC)
}

double histmgr::getbincontent (const HistId &id, int bin) const
{
    const th1Ptr h = getHist (id);
    if (!h) {
        //    log_log_debug(1, "histmgr: getbincontent('%s', %d): id not found", id, bin);
        return 0;
    }
    return (h->get_bin (bin));
}

double histmgr::get_calibration_constant(const std::string &id, int ch) const
{
    if (DECODE_TQDC) {
        if (id == "tqdc_baseline") {
            int mch = 0;
            const base_mod * mod = hwf.find_mod_by_fam_ch_const(MF_TQDC, ch, &mch);
            if (mod) {
                int id = mod->get_serial();
                QString idstr = QString("%1").arg(id, 8, 16, QChar('0'));
                QString key = QString("TQDC-%1").arg(idstr);
                double val = calibrationData->value(QString("%1-adc_baseline/%2").arg(key).arg(mch)).toDouble();
                return val;
            }

            //            return tqdc_cal->get_baseline(ch);
        }
    }
    return 0;
}

void histmgr::dataFormatChanged()
{
    rescale_hists();
}

/*
hptdc_inl_corr_t histmgr::read_hptdc_inl_corr(const QString &filename)

{
    hptdc_inl_corr_t data;
    QFile f(filename);
    if (!f.open(QIODevice::ReadOnly)) {
        qWarning().nospace() << "No TDC INL correction data " << filename;
        return hptdc_inl_corr_t();
    }
    QTextStream ts(&f);
    int i = 0;
    while(true) {
        double d;
        ts >> d;
        if (ts.status() != QTextStream::Ok) break;
        data[i] = d;
        i++;
    }
    f.close();
    if (i != 1024) {
        qWarning().nospace() << "Invalid TDC INL correction data " << filename << ": " << i << " entries, expected 1024";
        return hptdc_inl_corr_t();
    }
    qDebug().nospace() << "Loaded TDC INL corrections: " << filename;
    return data;
}
*/

void histmgr::requestHist(const QString &id)
{
    const th1 * const h = find(id).data();
    if (h)
        emit histAvailable(*h);
    else
        emit histAvailable(th1());
}

#ifdef DEBUG
void histmgr::fillDebugHist()
{
    for (uint i=0;i<10;i++)
        fillHist("debug_hist_1", 25+rangaus(2));
    fillHist("debug_hist_2", rand()%4);
}

//void histmgr::sendDebugHist()
//{
//    requestHist("debug_hist");
//}
#endif
