//
//    Copyright 2010 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "shmslave.h"
#include "decoder/dvs.h"

#include <QDebug>
#include <QDateTime>

#include <stdexcept>

namespace {
    double tv_dt(const struct timeval t1, const struct timeval t2)
    {
        return (((t2.tv_sec)-(t1.tv_sec)) + 1e-6*((t2.tv_usec)-(t1.tv_usec)));
    }
}

ShmSlaveStatus::ShmSlaveStatus()
    :
    shmStatus("VMEDAQ_STATUS"),
    _runparam_writeable(false),
    _status_writeable(false),
    shm_status(allocShmStatus())
{
}

ShmSlaveControl::ShmSlaveControl()
    :
    shmCtl("VMEDAQ_CONTROL"),
    _ctl_writeable(false),
    shm_ctl(allocShmCtl())
{
}

ShmSlave::ShmSlave()
    :
    ShmSlaveStatus(),
    ShmSlaveControl(),
    shmEvlist("VMEDAQ_EVENTLIST"),
    shmRawdata("VMEDAQ_RAWDATA"),
    shm_pio_rawdata(attachRawdata()),
    shm_evlist(attachEvlist())
{
}

shm_status_t *ShmSlaveStatus::allocShmStatus()
{
    shm_status_t *status = NULL;
    bool ok = shmStatus.attach(QSharedMemory::ReadWrite);
    if (ok) {
        _status_writeable = true;
    } else {
        ok = shmStatus.attach(QSharedMemory::ReadOnly);
    }
    if (ok) {
        status = static_cast<shm_status_t *>(shmStatus.data());
    } else {
        qWarning() << "Can't attach shared memory segment " << shmStatus.key() << ":" << shmStatus.errorString();
        status = new shm_status_t;
        _status_writeable = true;
        qWarning() << "Created non-shared data " << shmStatus.key();
    }
    Q_ASSERT(status);

    _runparam_writeable = _status_writeable;
    if (!_status_writeable)
        qDebug() << "shm_status readonly";
    return status;
}

shm_ctl_t *ShmSlaveControl::allocShmCtl()
{
    shm_ctl_t *control = NULL;
    bool ok = shmCtl.attach(QSharedMemory::ReadWrite);
    if (ok) {
        _ctl_writeable = true;
    } else {
        ok = shmCtl.attach(QSharedMemory::ReadOnly);
    }
    if (ok) {
        control = static_cast<shm_ctl_t *>(shmCtl.data());
    } else {
        qWarning() << "Can't attach shared memory segment " << shmCtl.key() << ":" << shmCtl.errorString();
        control = new shm_ctl_t;
        _ctl_writeable = true;
        qWarning() << "Created non-shared data " << shmCtl.key();
    }
    Q_ASSERT(control);

    if (!_ctl_writeable)
        qDebug() << "shm_ctl readonly";
    return control;
}

shm_evlist_t *ShmSlave::attachEvlist()
{
    shm_evlist_t *evlist = NULL;
    bool ok = shmEvlist.attach(QSharedMemory::ReadOnly);
    if (ok) {
        evlist = static_cast<shm_evlist_t *>(shmEvlist.data());
    } else {
        qWarning() << "Can't attach shared memory segment " << shmEvlist.key() << ":" << shmEvlist.errorString();
        evlist = new shm_evlist_t;
        qWarning() << "Created non-shared data " << shmEvlist.key();
    }
    Q_ASSERT(evlist);
    return evlist;
}

shm_pio_rawdata_t *ShmSlave::attachRawdata()
{
    shm_pio_rawdata_t *rawdata = NULL;
    bool ok = shmRawdata.attach(QSharedMemory::ReadOnly);
    if (ok) {
        rawdata = static_cast<shm_pio_rawdata_t *>(shmRawdata.data());
    } else {
        qWarning() << "Can't attach shared memory segment " << shmRawdata.key() << ":" << shmRawdata.errorString();
        rawdata = new shm_pio_rawdata_t;
        qWarning() << "Created non-shared data " << shmRawdata.key();
    }
    Q_ASSERT(rawdata);
    return rawdata;
}

bool ShmSlaveImpl::controlBusy() const
{
    return getCurrentControlState() != DAQCTL_CMD_ACK;
}

ctl_magic_t ShmSlaveControl::getCurrentControlState() const
{
    return shm_ctl->reset_magic;
}

bool ShmSlaveControl::sendControlCommand(ctl_magic_t cmd)
{
    if (!_ctl_writeable)
        return false;
    shm_ctl->reset_magic = cmd;
    return true;
}

pid_t ShmSlaveControl::getDaemonPid() const
{
    return shm_ctl->daq_pid;
}

QString ShmSlaveStatus::getRunparamFilename() const
{
    return QString::fromLocal8Bit(shm_status->rp.fn);
}

void ShmSlaveStatus::setRunparamFilename(const QString &fileName)
{
    if (!runparam_writeable())
        return;
    qstrncpy(shm_status->rp.fn, fileName.toLocal8Bit().data(), sizeof(shm_status->rp.fn) - 1);
}

QString ShmSlaveStatus::getRunparamComment() const
{
    return QString::fromLocal8Bit(shm_status->rp.rundesc);
}

void ShmSlaveStatus::setRunparamComment(const QString &text) const
{
    if (!runparam_writeable())
        return;
    qstrncpy(shm_status->rp.rundesc, text.toLocal8Bit().data(), sizeof(shm_status->rp.rundesc) - 1);
}

runparam_t ShmSlaveStatus::getRunparam() const
{
    return shm_status->rp;
}

void ShmSlaveStatus::setRunparam(const runparam_t &rp)
{
    if (!runparam_writeable())
        return;
    shm_status->rp = rp;
}

bool ShmSlaveStatus::runparamStore() const
{
    return shm_status->rp.store && (qstrlen(shm_status->rp.fn) > 0);
}

bool ShmSlaveStatus::runparamCyclic() const
{
    return shm_status->rp.cyclic != 0;
}

qint64 ShmSlaveStatus::getStorageBytes() const
{
    return shm_status->stor_nb;
}

QDateTime ShmSlaveStatus::getRunStartTime() const
{
    if (shm_status->rp.tstart.tv_sec != 0)
        return QDateTime::fromTime_t(shm_status->rp.tstart.tv_sec);
    else
        return QDateTime();
}

double ShmSlaveStatus::getLastBurstDuration() const
{
    if ((shm_status->t_burstreset.tv_sec > 0)
        && (shm_status->t_endspill.tv_sec > 0))
        return tv_dt (shm_status->t_burstreset, shm_status->t_endspill);
    return 0;
}

daq_state_t ShmSlaveStatus::getDaqState() const
{
    return shm_status->ds;
}

burst_state_t ShmSlaveStatus::getBurstState() const
{
    return shm_status->bs;
}

int ShmSlaveStatus::getSpillCount() const
{
    return shm_status->spill_count;
}

runcounts_t ShmSlaveStatus::getRunCounts() const
{
    return shm_status->rc;
}

runcounts_t ShmSlaveStatus::getBurstCounts() const
{
    return shm_status->bc;
}

daq_spr_t ShmSlaveStatus::getSpr() const
{
    return shm_status->spr;
}

double ShmSlaveStatus::getBurstTime() const
{
    struct timeval tv;
    gettimeofday(&tv, NULL);
    return tv_dt(shm_status->t_burstreset, tv);
}
