
#include "th1.h"

#include <string>
#include <cmath>
#include <vector>
#include <QFile>
#include <QTextStream>
#include <QDebug>

#include "fitter.h"

th1::th1()
{
    init_pvt();
    setrange(-10, 10);
    rebin(100);
    reset();
}

th1::th1 (const HistDescr &descr)
{
    init_pvt();
    setup(descr.id, descr.title, descr.nbins, descr.min, descr.max);
}

th1::th1(const HistId &id, const QString &name, int fnbins, double fmin, double fmax)
{
    init_pvt();
    setup(id, name, fnbins, fmin, fmax);
}

th1::~th1()
{
}

void th1::init_pvt()
{
    mod = 0;
    hold_mod = 0;
    do_hold = false;
    hold_under = 0;
    hold_over = 0;
    hold_nent = 0;
    setid("h_unnamed");
    setname("unnamed");
    setAccessLevel(0);
}

void th1::setup(const HistId &id, const QString &name, int fnbins, double fmin, double fmax)
{
    nbins = fnbins;
    setid(id);
    setname(name);
    setrange(fmin, fmax);
    rebin(nbins);
    //  print();
    clear();
}

//QDebug operator<<(QDebug dbg, const HistId &id)
//{
//    dbg.nospace() << QString::fromStdString(id);
//    return dbg;
//}

QDebug operator<<(QDebug dbg, const th1 &h)
{
    dbg.nospace() << "Histogram '" << h.getname()
                  << "', id " << h.getid()
                  << ", nbins=" << h.get_nbins()
                  << ", nent=" << h.get_nent()
                  << ", under=" << h.get_under()
                  << ", over=" << h.get_over();
    return dbg;
}

void th1::rebin(int nb)
{
    /*
TODO: new rebin
  if (nb<2) nb=2; // fail-safe
  th1 tmp(getid(), getname(), nb, xmin, xmax);
  for(size_t i=0; i<nbins; i++)
        tmp.filln(bincenter(i), get_bin(i));
  nbins = nb;
  py.clear();
  py = tmp.py;
  hold_py.clear();
  hold_py.resize(nbins);
  recalc_internals();
*/
    // old rebin
    if (nb<2) nb=2; // fail-safe
    nbins = nb;
    py.resize(nbins);
    hold_py.resize(nbins);
    recalc_internals();
}

double th1::bincenter(int n) const
{
    //  double binwidth = (xmax - xmin) / nbins;
    double result = xmin + (0.5 + n) * s_binwidth;
    // LOG("bin %d: [%f, %f]  %f\n", n, xmin, xmax, result);
    return result;
}
void th1::recalc_internals()
{
    s_binwidth = (xmax - xmin) / nbins;
}

// speed optimization - inline
int th1::findbin(double x) const
{
    // formula: int ( (x - xmin) / (xmax - xmin) * nbins )
    return static_cast<int>((x - xmin) / s_binwidth);
}

// inmin - first bin left edge
// inmax - last bin right edge
void th1::setrange(double inmin, double inmax)
{
    if (inmin < inmax) {
        xmin = inmin;
        xmax = inmax;
    } else {
        xmin = 0;
        xmax = 1;
    }
    recalc_internals();
    mod++;
}

void th1::fill(double x)
{
    mod++;
    // check range first
    if (x<xmin) {
        under++;
        return;
    }
    if (!(x < xmax)) { // note: catch NaN
        over++;
        return;
    }
    // only if in range, calculate bin number
    int bin = findbin(x);
    py[bin]++;
    nent++;
    //  printf("fill(%p, %s), mod=%d\n", this, getname(), mod);
}

void th1::filln(double x, double count)
{
    if (count==0) return;
    mod++;
    if (x<xmin) {
        under += count;
        return;
    }
    if (!(x < xmax)) {
        over += count;
        return;
    }
    int bin = findbin(x);
    py[bin] += count;
    nent += count;
    //  printf("filln(%p, %s), mod=%d\n", this, getname(), mod);
}

void th1::fill(const std::vector<double> &v, double count)
{
    for(std::vector<double>::const_iterator it = v.begin(); it != v.end(); ++it) {
        double x = (*it);
        filln(x, count);
    }
    /*
  mod++;
  foreach_const(v, it) {
        double x = (*it);
    if (x<xmin) {
          under += count;
          continue;
        }
        if (! (x < xmax)) {
          over += count;
          continue;
        }
        static int bin = findbin (x);
        py[bin] += count;
        nent += count;
  }
//  printf("fillv(%p, %s), mod=%d\n", this, getname(), mod);
*/
}

void th1::fill(const std::map<double, double> &m)
{
    for(std::map<double, double>::const_iterator it = m.begin(); it != m.end(); ++it) {
        double x = (*it).first;
        double count= (*it).second;
        filln(x, count);
    }
}

//void th1::fillbufn(double x[], int len, double count)
//{
//    int bin = 0;

//    double hw = xmax - xmin;
//    double binwidth = hw / nbins;
//    if (binwidth <= 0)
//        binwidth = 1;

//    for (int i = 0; i < len; i++) {
//        if (x[i] < xmin) {
//            under += count;
//        } else {
//            if (x[i] > xmax) {
//                over += count;
//            } else {
//                bin = static_cast<int>((x[i] - xmin) / binwidth);
//                py[bin] += count;
//            }
//        }
//    }
//    nent += len*count;
//    mod++;
//    // log_debug(3, "hist::fillbuf");
//    //  printf("fillbufn(%p, %s), mod=%d\n", this, getname(), mod);
//}

void th1::reset()
{
    do_hold = false;
    clear();
}

void th1::clear()
{
    //    printf("hist %d::clear...\n", myid);
    under = 0;
    over = 0;
    nent = 0;
    //  py.clear();
    for (unsigned int i = 0; i < nbins; i++) {
        py[i] = 0;
    }
    mod++;
}

void th1::hold(bool state)
{
    if (do_hold == state) return;
    //  printf("th1::hold(%s) = %d\n", getname(), state);
    if (!do_hold && state) { // hold
        for (unsigned int i = 0; i < py.size(); i++) {
            hold_py[i] = py[i];
        }
        hold_nent = nent;
        hold_over = over;
        hold_under = under;
        hold_mod = mod;
    }
    if (do_hold && !state) { // resume
    }
    do_hold = state;
}

void th1::shift_right()
{
    //    printf("hist %d::clear...\n", myid);
    for (unsigned int i = nbins-1; i > 0; i--) {
        py[i] = py[i-1];
        hold_py[i] = hold_py[i-1];
    }
    py[0]=0;
    hold_py[0]=0;
    mod++;
}

bool th1::fitgaus(double *fvalue, double *fmean, double *fsigma, int bin1, int bin2) const
{
    if (nbins < 2)
        return false;
    Q_ASSERT(bin1 >= 0);
    Q_ASSERT(bin2 < int(nbins));
    int fitnbins = bin2-bin1+1;
    if (fitnbins < 2)
        return false;

    fitpy.resize(fitnbins);
    double par[3] = {0,0,0};
    //  double err[3], chi2;

    for (int i = 0; i < fitnbins; i++)
        fitpy[i] = py[i+bin1];
    // select fit method: simple or with MIGRAD

    //  if (fitnbins > 100)
    //	FindGausApprox(&fitpy[0], fitnbins, par);
    //  else
    QuickFitGaus(&fitpy[0], fitnbins, par, NULL);

    //  LOG ("Fit: %e, %e, %e\n", par[0], par[1], par[2]);
    double value = par[0];

    double binwidth = (xmax - xmin) / nbins;
    double mean = xmin + (bin1 + par[1]) * binwidth; // (0.5 + par[1])???
    double sigma = par[2] * binwidth;

    if (fvalue)
        *fvalue = value;
    if (fmean)
        *fmean = mean;
    if (fsigma)
        *fsigma = sigma;

    return true;
}

void th1::setname(const QString &name)
{
    fname = name;
}

void th1::setid(const HistId &id)
{
    fid = id;
//    if (fid.empty()) {
//        std::ostringstream os;
//        os << "h_" << std::hex << rand();
//        fid = os.str();
//    }
}

QString th1::getname() const
{
    return(fname);
}

const HistId &th1::getid() const
{
    return(fid);
}

void th1::save_to_file() const
{
//    QString filename = QString("%1.txt").arg(QString::fromStdString(fid));
    QString filename = QString("%1.txt").arg(fid);
    QFile f(filename);
    if (!f.open(QIODevice::WriteOnly)) {
        qWarning() << QString("Can't write file %1: %2").arg(filename).arg(f.errorString());
        return;
    }
    QTextStream s(&f);
    for (unsigned int i = 0; i < nbins; i++)
        s << QString("%1\t%2\n").arg(bincenter(i)).arg(py[i]); // static_cast<int>(py[i]));
    f.close();
}

unsigned int th1::get_nbins() const
{
    return nbins;
}

//std::vector<double> th1::get_all_bins() const
//{
//    if (do_hold)
//        return hold_py;
//    else
//        return py;
//}

double th1::get_bin(unsigned int index) const
{
    if (index >= nbins)
        return 0;
    return (do_hold?hold_py[index]:py[index]);
}

std::vector<double> th1::get_py() const
{
    return py;
}

quint64 th1::get_nent() const
{
    return ((quint64)(do_hold?hold_nent:nent));
}

quint64 th1::get_over() const
{
    return ((quint64)(do_hold?hold_over:over));
}

quint64 th1::get_under() const
{
    return ((quint64)(do_hold?hold_under:under));
}

int th1::modified_count() const
{
    //  printf("modified_count(%p, %s) = %d\n", this, getname(), mod);
    return (do_hold?hold_mod:mod);
}

void th1::clear_modified()
{
    mod = 0;
}

double th1::get_xmin() const
{
    return (xmin);
}

double th1::get_xmax() const
{
    return (xmax);
}

void th1::set_bin_content(int bin, double count)
{
    if (bin<0 || bin >= int(nbins)) return;
    py[bin] = count;
}
