#ifndef VME_TYPES_H
#define VME_TYPES_H

#include <QtCore/QtGlobal>

// VME address modifier codes.
// ANSI/VITA 1-1994 VME64 Standard
// (1.1) - ANSI/VITA 1.1-1997 VME64x Standard
enum VME_AM {
    VME_AM_A24_PRIV_BLT  = 0x3F, // A24 supervisory block transfer (BLT)
    VME_AM_A24_PRIV_PROG = 0x3E, // A24 supervisory program access
    VME_AM_A24_PRIV      = 0x3D, // A24 supervisory data access
    VME_AM_A24_PRIV_MBLT = 0x3C, // A24 supervisory 64-bit block transfer (MBLT)
    VME_AM_A24_BLT       = 0x3B, // A24 non privileged block transfer (BLT)
    VME_AM_A24_PROG      = 0x3A, // A24 non privileged program access
    VME_AM_A24           = 0x39, // A24 non privileged data access
    VME_AM_A24_MBLT      = 0x38, // A24 non privileged 64-bit block transfer (MBLT)
    VME_AM_A40_BLT       = 0x37, // A40BLT
// 0x36 Reserved
    VME_AM_A40_LCK       = 0x35, // A40 lock command (LCK)
    VME_AM_A40           = 0x34, // A40 access
// 0x33 Reserved
    VME_AM_A24_LCK       = 0x32, // A24 lock command (LCK)
// 0x31 Reserved
// 0x30 Reserved
    VME_AM_CRCSR         = 0x2F, // Configuration ROM/Control & Status Register (CR/CSR)
    VME_AM_CRCSR_BUG     = 0x37, // this is incorrect and should be fixed
// 0x2E Reserved
    VME_AM_A16_PRIV      = 0x2D, // A16 supervisory access
    VME_AM_A16_LCK       = 0x2C, // A16 lock command (LCK)
// 0x2B Reserved
// 0x2A Reserved
    VME_AM_A16           = 0x29, // A16 non-privileged access
// 0x22..0x28 Reserved
    VME_AM_3U_2eVME      = 0x21, // 2eVME for 3U bus modules (address size in XAM code) (1.1)
    VME_AM_6U_2eVME      = 0x20, // 2eVME for 6U bus modules (address size in XAM code) (1.1)
// 0x10..0x1F User-defined
    VME_AM_A32_PRIV_BLT  = 0x0F, // A32 supervisory block transfer (BLT)
    VME_AM_A32_PRIV_PROG = 0x0E, // A32 supervisory program access
    VME_AM_A32_PRIV      = 0x0D, // A32 supervisory data access
    VME_AM_A32_PRIV_MBLT = 0x0C, // A32 supervisory 64-bit block transfer (MBLT)
    VME_AM_A32_BLT       = 0x0B, // A32 non privileged block transfer (BLT)
    VME_AM_A32_PROG      = 0x0A, // A32 non privileged program access
    VME_AM_A32           = 0x09, // A32 non privileged data access
    VME_AM_A32_MBLT      = 0x08, // A32 non privileged 64-bit block transfer (MBLT)
// 0x07 Reserved
// 0x06 Reserved
    VME_AM_A32_LCK       = 0x05, // A32 lock command (LCK)
    VME_AM_A64_LCK       = 0x04, // A64 lock command (LCK)
    VME_AM_A64_BLT       = 0x03, // A64 block transfer (BLT)
// 0x02 Reserved
    VME_AM_A64           = 0x01, // A64 single transfer access
    VME_AM_A64_MBLT      = 0x00, // A64 64-bit block transfer (MBLT)

// User defined AM codes
    VME_AM_3U_2e2p       = 0x11, // User defined 2e mode with 2 address phases, 3U
    VME_AM_6U_2e2p       = 0x10, // User defined 2e mode with 2 address phases, 6U
};

// VME64x XAM (extended address modifiers) codes
// VITA 1.5-1999
enum VME_XAM {
    VME_XAM_A32_2eVME       = 0x01, // 6U A32/D64 2eVME, 3U A32/D32 2eVME
    VME_XAM_A64_2eVME       = 0x02, // 6U A64/D64 2eVME
    VME_XAM_A40_2eVME       = 0x02, // 3U A40/D32 2eVME
    VME_XAM_A32_2eSST       = 0x11, // 6U A32/D64 2eSST, 3U A32/D32 2eSST
    VME_XAM_A64_2eSST       = 0x12, // 6U A64/D64 2eSST
    VME_XAM_A40_2eSST       = 0x12, // 3U A40/D32 2eSST
    VME_XAM_A32_2eSST_BCAST = 0x21, // 6U A32/D64 Broadcast 2eSST, 3U A32/D32 Broadcast 2eSST
    VME_XAM_A64_2eSST_BCAST = 0x22, // 6U A64/D64 Broadcast 2eSST
    VME_XAM_A40_2eSST_BCAST = 0x22, // 3U A40/D32 Broadcast 2eSST
};

const int VME_CHAIN_ADDRESS = 0;

// VME Configuration ROM (CR)
const int VME_CR_BASE =  (0x00 | 0x1); // 16-bit addressing, #LWORD=1

#pragma pack(push, 1)
struct vme_cr_t {
  quint8 serial;
  quint8 type;
  quint8 reserved;
  quint8 vendor;
  bool operator==(const struct vme_cr_t &) const;
  bool operator!=(const struct vme_cr_t &r) const { return !(*this == r); }
};
#pragma pack(pop)

// Control/Status Register (CSR)

const int VME_CSR_BASE = (0x04 | 0x1);

#pragma pack(push, 1)
struct vme_csr_t {
  quint16 base;
  bool operator==(const struct vme_csr_t &) const;
  bool operator!=(const struct vme_csr_t &r) const { return !(*this == r); }
};
#pragma pack(pop)

const unsigned int VME_SLOTS = 32;

static int vme_addr_bits(int am)
{
    switch(am) {
    case VME_AM_A24_PRIV_BLT: return 24;
    case VME_AM_A24_PRIV_PROG: return 24;
    case VME_AM_A24_PRIV: return 24;
    case VME_AM_A24_PRIV_MBLT: return 24;
    case VME_AM_A24_BLT: return 24;
    case VME_AM_A24_PROG: return 24;
    case VME_AM_A24: return 24;
    case VME_AM_A24_MBLT: return 24;
//    case VME_AM_A40_BLT: return 40; // BUG: VME_AM_CRCSR_BUG
    case VME_AM_A40_LCK: return 40;
    case VME_AM_A40: return 40;
    case VME_AM_A24_LCK: return 24;
    case VME_AM_CRCSR: return 24;
    case VME_AM_CRCSR_BUG: return 24;
    case VME_AM_A16_PRIV: return 16;
    case VME_AM_A16_LCK: return 16;
    case VME_AM_A16: return 16;
    case VME_AM_3U_2eVME: return 40; // FIXME: decode XAM
    case VME_AM_6U_2eVME: return 64; // FIXME: decode XAM
    case VME_AM_3U_2e2p:  return 40; // FIXME: decode XAM
    case VME_AM_6U_2e2p:  return 64; // FIXME: decode XAM
    case VME_AM_A32_PRIV_BLT: return 32;
    case VME_AM_A32_PRIV_PROG: return 32;
    case VME_AM_A32_PRIV: return 32;
    case VME_AM_A32_PRIV_MBLT: return 32;
    case VME_AM_A32_BLT: return 32;
    case VME_AM_A32_PROG: return 32;
    case VME_AM_A32: return 32;
    case VME_AM_A32_MBLT: return 32;
    case VME_AM_A32_LCK: return 32;
    case VME_AM_A64_LCK: return 64;
    case VME_AM_A64_BLT: return 64;
    case VME_AM_A64: return 64;
    case VME_AM_A64_MBLT: return 64;
    default: return 64;
    }
    return 64;
}

#endif /* VME_TYPES_H */
