//
//    Copyright 2010 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include <QDebug>
#include <QStringList>

#include "daqcommon/rawdata.h"
#include "decoder/dvs.h"
#include "daqcommon/device_ids.h"
#include "decoder_modlist.h"
#include "hptdc/hptdc.h"
#include "decode_bric.h"
#include "modules/bric_types.h"
#include "decode_tqdc.h"
#include "decoder/decode_trig.h"
#include "RawDumper.h"

const int DUMP_MAX_LINES = 20;

RawDumper::RawDumper()
{
    clear();
}

void RawDumper::clear()
{
    dbufState.clear();
}

QString RawDumper::decodeRawDataWord(const unsigned int d)
{
    QString str;
    const unsigned int dt = d & ADLVME_DTYPE_MASK_ALL;
    const unsigned int wt = dt>>28;
    {
        str += daq_str_dtype_mod((unsigned) dt >> 28, dbufState.getModuleId());
        str += "  ";

        if (dt == ADLVME_DTYPE_STAT) {
            int type = (d >> 24) & 0xF;
            switch (type) {
                case 1: str += QString("temp %1").arg((quint16)(d & 0xFFFF) / 256.);
            }
        }
        if (dt == ADLVME_DTYPE_EHDR) {
            dbufState.handleEHDR(d);
            str += QString("ev %1").arg(dbufState.getEventNumber());
        }
        if (dt == ADLVME_DTYPE_ETRL) {
            dbufState.handleETRL(d);
            int rdst = decode_etrl_readout_status(d);
            bool timeout = rdst & ETRL_STATUS_BIT_TIMEOUT;
            str += QString("wc %1 %2").arg(decode_etrl_wc(d)).arg(timeout?"TIMEOUT ":"");
        }
        if (dt == ADLVME_DTYPE_MHDR) {
            dbufState.handleMHDR(d);
            str += QString("ev %1, id %2 (%3), ga %4").arg(dbufState.getEventNumber())
                     .arg(dbufState.getModuleId()).arg(dbufState.getModuleName().c_str())
                      .arg(dbufState.getModuleSlot());
        }
        if (dt == ADLVME_DTYPE_MTRL) {
            dbufState.handleMTRL(d);
            int f = dbufState.getModuleFlags();
            str += QString("wc %1, flags: %2%3%4 %5%6").arg(decode_mtrl_wc(d))
                     .arg(f & 4 ? "F" : "_").arg(f & 1 ? "M" : "_").arg(f & 2 ? "L" : "_")
                     .arg(decode_mtrl_ro(d)?"Overflow ":"")
                     .arg(decode_mtrl_re(d)?"Read_Error ":"");
            //      if (nq > 0) {
            //        printf("*** Q = %f\n", (double) q / nq);
            //      }
        }
        if (d == DATA_SHDR_ENDBURST || d == DATA_STRL_ENDBURST) {
            str += "ENDBURST";
        }
        if (dt == ADLVME_DTYPE_SHDR) {
            dbufState.handleSHDR(d);
        }
        if (dt == ADLVME_DTYPE_STRL) {
            dbufState.handleSTRL(d);
        }
        if ((dt == ADLVME_DTYPE_STRL) && !(d & 0x08000000)) {
            str += QString("wc %1").arg(decode_strl_wc(d));
        }

        if ((d & ADLVME_DTYPE_MASK_CTRL) == 0) {
            //      str += QString("dev %1 ").arg(dn);
            // TQDC decode
            if (DECODE_TQDC) {
                switch (dbufState.getModuleId()) {
                case MODULE_ID_TQDC: {
                        int id = (d >> 24) & 0xF;
                        int sdt = id >> 2; // mode: 0 - TDC, 1..3 - ADC
                        //        int sdn = id & 3;
                        int ch = (d >> 19) & 0x1F;
                        //      ch>>=2;
                        if (wt == 0 || wt == 1) {
                            str += QString("ch=%1 ").arg(ch, 2);
                        }
                        // TDC data
                        if (wt >= 2 && wt <= 6 && sdt == 0) {
                            //          str += QString("%1 ", sdn);
                            //          str.append(tmp);
                            str += "TDC  ";
                            str += hptdc_dataword_str(d);
                        }
                        // ADC data
                        if ((wt == 4 || wt == 5) && sdt != 0) {
                            switch (sdt) {
                            case 0:
                                break;
                            case 1:
                                str += "ADCZ ";
                                break;
                            case 2:
                                str += "ADC  ";
                                break;
                            case 3:
                                str += "QDC  ";
                                break;
                            }
                            str += QString("ch=%1 ").arg(ch, 2);
                        }
                        if (wt == 0) {
                            str += QString("in_cnt_hi: %1").arg(d & 0x0000FFFF);
                        }
                        if (wt == 1) {
                            str += QString("in_cnt_lo: %1").arg(d & 0x0000FFFF);
                        }
                        if ((dt == TQDC_DTYPE_ADC_TS) && (sdt != 0)) {
                            if (((d >> 16) & 7) == 0) {
                                unsigned int tt = d & 0xFFFF;
                                str += QString("TS(31:16) %1 (%2, %3)").arg(tt, 4, 16, QChar('0')).arg(tt).arg(tt%5);
                            } else {
                                if (((d >> 16) & 7) == 1) {
                                    unsigned int tt = d & 0xFFFF;
                                    str += QString("TS(15:0) %1 (%2, %3)").arg(tt, 4, 16, QChar('0')).arg(tt).arg(tt%5);
                                }
                            }
                        } else {
                            if (dt == TQDC_DTYPE_ADC_DATA) {
                                if (sdt == 1 || sdt == 2) {
                                    unsigned int rawval = d & 0x3FFF;
                                    int val = rawval - 0x2000;
                                    str += QString("SAMPL  %1 ").arg(val, 4);
                                    int coarseval = rawval >> 9;
                                    for (int ss = 0; ss < coarseval; ss++)
                                        str += ' ';
                                    str += '*';
                                    for (unsigned int ss = coarseval; ss<32; ss++)
                                        str += ' ';
//                                    for (unsigned int ss = 0; ss < 32; ss++)
//                                        if ((rawval >> 9) == ss)
//                                            str += '*';
//                                        else
//                                            str + ' ';
                                    dbufState.nq++;
                                    dbufState.q += val;
                                }
                                if (sdt == 3) {
                                    int rawval = d & 0x7FFFF;
                                    int val = rawval - 8 * 512;
                                    str += QString("INTEGR  %1").arg(d & 0x7FFFF, 5, 16);
                                    dbufState.nq = 8;
                                    dbufState.q = val;
                                }
                                if (sdt == 0) {
                                    str += QString("DATA  %1").arg(d & 0x7FFFF, 5, 16);
                                }
                            }
                        }
                        break;
                    }
                }
            } // end if(DECODE_TQDC)
            if (DECODE_TDC) {
                if (dbufState.getModuleId() == MODULE_ID_PHTDC
                    || dbufState.getModuleId() == MODULE_ID_TDC32VL
                    || dbufState.getModuleId() == MODULE_ID_TDC64V
                    || dbufState.getModuleId() == MODULE_ID_TDC96
                    || dbufState.getModuleId() == MODULE_ID_MTDC64
                    ) {
                    str += hptdc_dataword_str(d);
                }
            }
            if (DECODE_TRIG && dbufState.getModuleId() == MODULE_ID_TRIG) {
                switch (dt>>28) {
                case 0x0: {
                        str += QString("logic_cnt_read[%1]: %2")
                               .arg(dbufState.mod_word_count_by_datatype[dt], 2).arg(d & 0x0FFFFFFF);
                        break;
                    }
                case 0x1: {
                        str += QString("logic_cnt_all [%1]: %2")
                               .arg(dbufState.mod_word_count_by_datatype[dt], 2).arg(d & 0x0FFFFFFF);
                        break;
                    }
                case 0x3: {
                        str += QString("TLU event ID: %1").arg(d & 0x0FFFFFFF, 7, 16);
                        break;
                    }
                case 0x4: {
                        dbufState.trig_ts_lo = d & 0x0FFFFFFF;
                        str += QString("TS lo: %1").arg(dbufState.trig_ts_lo, 7, 16);
                        break;
                    }
                case 0x5: {
                        dbufState.trig_ts_hi = (d>>20) & 0xFF;
                        double ev_t = dbufState.getTrigTimestamp();
                        str += QString("TS hi: %1, trigbits: %2. Event %3 time: %4 ms")
                                .arg(dbufState.trig_ts_hi, 2, 16)
                                .arg(trig_d5_decode_trigbits(d), 4, 16)
                                .arg(dbufState.getEventNumber())
                                .arg(ev_t*1e3, 0, 'f', 3);
                        break;
                    }
                case 0x7: {
                        str += QString("in_cnt[%1]: %2")
                               .arg(dbufState.mod_word_count_by_datatype[dt], 2)
                               .arg(d & 0x0FFFFFFF);
                        break;
                    }
                }
            }
            if (DECODE_BRIC && dbufState.getModuleId() == MODULE_ID_BRIC) {
                switch (dt>>28) {
                case 0x0: {
                        int w = bric_hdr_decode_window_id(d);
                        int b = bric_hdr_decode_branch_id(d);
                        str += QString("bric %1%2, win %3").arg(b/2).arg((b%2)?"B":"A").arg(w);
                        break;
                    }
                case 0x1: {
                        // branch status
                        bool timeout = BRIC_HEADER_BIT_TIMEOUT & d;
                        bool align_err = BRIC_HEADER_BIT_ALIGN_ERR & d;
                        bool overflow = BRIC_HEADER_BIT_OVERFLOW & d;
                        bool disconnect = BRIC_HEADER_BIT_ABSENT & d;
                        // link status
                        bool link_host_disc = BRIC_HEADER_BIT_LINKHOST_DISCONNECT & d;
                        bool link_host_rxerror = BRIC_HEADER_BIT_LINKHOST_RXERROR & d;
                        bool link_remote_disc = BRIC_HEADER_BIT_LINKREMOTE_DISCONNECT & d;
                        bool link_remote_rxerror = BRIC_HEADER_BIT_LINKREMOTE_RXERROR & d;

                        bool branch_ok = !timeout && !align_err && !overflow && !disconnect;
                        bool link_ok = !link_host_disc && !link_host_rxerror && !link_remote_rxerror; // link_remote_disc normal for last bric in chain
                        bool ok = branch_ok && link_ok;
                        str += QString("%1%2%3%4%5%6%7%8%9")
                               .arg(ok?"Ok ":"")
                               .arg(timeout?"timeout ":"")
                               .arg(align_err?"align_err ":"")
                               .arg(overflow?"overflow ":"")
                               .arg(disconnect?"disconnect ":"")
                               .arg(link_host_disc?"host_disconnect ":"")
                               .arg(link_host_rxerror?"host_rxerror ":"")
                               .arg(link_remote_disc?"end_of_chain ":"")
                               .arg(link_remote_rxerror?"remote_rxerror ":"");
                        break;
                    }
                case 0x4:
                    str += QString("raw bits %1").arg(d&0xFFFF, 4, 16);
                    break;
                case 0x7: {
                        bricbox_lcf_t lcf;
                        lcf.all = (d >> 8) & 0xFFFF;
                        bool timeout[2] = {false, false};
                        bool disconnect[2] = {false, false};
                        bool overflow[2] = {false, false};
                        bool data_error[2] = {false, false};
                        bool ok[2];
                        //		printf("LCF=%04X\n", lcf.all);
                        if (lcf.bit.b1_timeout)
                            timeout[0] = true;
                        if (lcf.bit.b2_timeout)
                            timeout[1] = true;
                        if (lcf.bit.b1_absent)
                            disconnect[0] = true;
                        if (lcf.bit.b2_absent)
                            disconnect[1] = true;
                        if (lcf.bit.b1_overflow)
                            overflow[0] = true;
                        if (lcf.bit.b2_overflow)
                            overflow[1] = true;
                        if (lcf.bit.b1_align_err)
                            data_error[0] = true;
                        if (lcf.bit.b2_align_err)
                            data_error[1] = true;
                        ok[0] = !timeout[0] && !disconnect[0] && !overflow[0] && !data_error[0];
                        ok[1] = !timeout[1] && !disconnect[1] && !overflow[1] && !data_error[1];

                        int st = d & 0x0FF;
                        bool link_LCF_timeout = st & BRIC_STATUS_BIT_LCF_TIMEOUT;
                        bool bric_not_ready   = st & BRIC_STATUS_BIT_TIMEOUT;
                        bool link_disconnect = st & BRIC_STATUS_BIT_DISCONNECT;
                        bool link_data_error = (st & BRIC_STATUS_BIT_DATAERR) | (st & BRIC_STATUS_BIT_ERR_PROP);
                        bool link_crc_error = (st & BRIC_STATUS_BIT_CRC_ERR);
                        bool link_ok = !link_LCF_timeout && !link_disconnect && !link_data_error;
                        str += QString("Link%1%2%3%4%5%6")
                               .arg(link_ok?" OK":"")
                               .arg(link_LCF_timeout?" LCF_TM":"")
                               .arg(bric_not_ready?" NR":"")
                               .arg(link_disconnect?" DC":"")
                               .arg(link_data_error?" ER":"")
                               .arg(link_crc_error?" CE":"");
                        if (!link_LCF_timeout) {
                            str += QString(", br1%1%2%3%4%5")
                                   .arg(ok[0]?" OK":"").arg(timeout[0]?" TM":"")
                                   .arg(disconnect[0]?" DC":"").arg(overflow[0]?" OV":"")
                                   .arg(data_error[0]?" ER":"");
                            str += QString(", br2%1%2%3%4%5")
                                   .arg(ok[1]?" OK":"").arg(timeout[1]?" TM":"")
                                   .arg(disconnect[1]?" DC":"").arg(overflow[1]?" OV":"")
                                   .arg(data_error[1]?" ER":"");
                        }
                        break;
                    }
                }
            } // end if(DECODE_BRIC)
            //      str.append(tmp);
        }
        dbufState.mod_word_count_by_datatype[dt]++;
    }
    //  printdebug(2, str);
    return str;
}

QString RawDumper::dump_raw_str(const unsigned int *buf, size_t start, size_t end, size_t mark, bool compress)
{
    QStringList sl;
    if (!buf)
        return QString();
    bool last_show = true;
    int num_shown = 0;
    for (size_t i = start; i < end; i++) {
        QString s = decodeRawDataWord(buf[i]);
        if (compress) {
            bool show = false;
            if (i == mark) num_shown = 0;
            if (i <= start+2 || i >= end-3) show = true;
            if (i >= mark-3 && i <= mark+3) show = true;
            if (end-start < DUMP_MAX_LINES) show = true;
            if (!show && last_show) sl << QString("  ...");
            last_show = show;
            if (!show) continue;
        }
        num_shown++;
        sl << QString("%1[%2] %3: %4")
                .arg(i == mark ? "> " : "  ")
                .arg(i, 8, 16, QChar('0'))
                .arg(i-start, 5)
                .arg(buf[i], 8, 16, QChar('0'))
                + " "
                + s;
    }
    return sl.join("\n");
}
