
#include <QDebug>
#include <sstream>

#include "daqcommon/rawdata.h"
#include "decoder/decoder.h"
#include "decode_bric.h"
#include "modules/bric_types.h"
#include "decoder/event.h"

void local_dch_decoder::print() const
{
    qDebug() << "*** BRIC decoder state ***";
    Q_ASSERT(ev);
    if (!ev) return;
    qDebug().nospace() << "ev=" << ev->en << ", first_ch=" << first_ch;
    const dch_event_t &bric = ev->bric;
    const dch_branches_t &branches = bric.branches;

    Q_FOREACH(int branch_id, branches.keys()) {
        const dch_branch_t &branch = branches.value(branch_id);
        const dch_windows_t &windows = branch.wins;
        qDebug() << "branch" << branch_id << ":"
                 << (branch.timeout ? "timeout" : "")
                 << (branch.disconnect ? "disconnect" : "")
                 << (branch.data_error ? "data_error" : "")
                 << (branch.isOk() ? "Ok" : "");
        Q_FOREACH(int window_id, windows.keys()) {
            const dch_window_t &win = windows.value(window_id);
            size_t nbits = win.bits.size()*8;
            if (nbits == 0) continue;
            qDebug() << "window" << window_id << ":" << nbits << "bits";
            QString str;
            for (size_t n = 0; n<nbits; n+=8) {
                str += QString("%1 %2").arg((unsigned int)win.bits[n/8], 2, 16, QChar('0'));
                if ((n/8 % 4) == 3) str += "  ";
                if ((n/8 % (4*5)) == (4*5-1)) str += "\n";
            }
            qDebug() << str;
        }
    }
    qDebug() << "**************************";
}

int local_dch_decoder::decode()
{
    Q_ASSERT(ev);
    dch_event_t &bric = ev->bric;
    dch_branches_t &branches = bric.branches;

    unsigned int dwc = 0;
    unsigned int window_id = 0;
    unsigned int branch_id = 0;
    int found_status = 0;
    std::string branch_id_str, bric_id_str;
    //  unsigned int window = 0, branch = 0;

    //  printf("ev %d, event_decode_bric: len=%d\n", ev->en, len);
    for (size_t p = 1; p < len; p++) {
        unsigned int d = buf[p];
        unsigned int dt = d & ADLVME_DTYPE_MASK_ALL;
        if (dt == 0x00000000) { // header
            window_id = bric_hdr_decode_window_id(d);
            branch_id = bric_hdr_decode_branch_id(d);
            //		printf("bric hdr: branch %d, window %d\n", branch_id, window_id);
            dwc = 0;
            {
                std::ostringstream ost;
                ost << "bric_" << branch_id/2 << ((branch_id%2)?"B":"A");
                branch_id_str = ost.str();
            }
            {
                std::ostringstream ost;
                ost << "bric_" << branch_id/2;
                bric_id_str = ost.str();
            }
        }
        if (dt == 0x10000000) { // header 2
            // branch status
            bool timeout = BRIC_HEADER_BIT_TIMEOUT & d;
            bool align_err = BRIC_HEADER_BIT_ALIGN_ERR & d;
            bool overflow = BRIC_HEADER_BIT_OVERFLOW & d;
            bool absent = BRIC_HEADER_BIT_ABSENT & d;
            if (timeout) setStat(branch_id_str+"_timeout");
            if (align_err) setStat(branch_id_str+"_disconnect");
            if (overflow) setStat(branch_id_str+"_oveflow");
            if (absent) setStat(branch_id_str+"_align_err");
            // link status
            bool link_host_disc = BRIC_HEADER_BIT_LINKHOST_DISCONNECT & d;
            bool link_host_rxerror = BRIC_HEADER_BIT_LINKHOST_RXERROR & d;
            bool link_remote_disc = BRIC_HEADER_BIT_LINKREMOTE_DISCONNECT & d;
            bool link_remote_rxerror = BRIC_HEADER_BIT_LINKREMOTE_RXERROR & d;
            if (link_host_disc) setStat(bric_id_str+"_host_disconnect");
            if (link_host_rxerror) setStat(bric_id_str+"_host_rx_error");
            if (link_remote_disc) setStat(bric_id_str+"_end_of_chain");
            if (link_remote_rxerror) setStat(bric_id_str+"_remote_rx_error");
        }
        if (dt == 0x40000000) { // data
            unsigned char b0 = d&0xFF;
            unsigned char b1 = (d>>8)&0xFF;
            //			unsigned int off = ((dwc+1) % 2 ) * 2;
            //			unsigned int ncard = window->num_bits/32;
            dch_branch_t &branch = branches[branch_id];
            dch_window_t &window = branch.wins[window_id];
            window.bits.push_back(b0);
            window.bits.push_back(b1);
            //    		printf("bric bits: %02X%02X\n", b1, b0);
            //    		printf("bric raw bits [%04X]: %02X %02X\n", off, window->bits[ncard*4 + off], window->bits[ncard*4 + off + 1]);
            //		unsigned int bits = d & 0x0FFFFFFF;
            //	    if (ev->en < 10)
            //	  	  printf("bric branch %d window %d, raw bits: [%d] %04X\n", branch_id, window_id, window?window->num_bits:0, d&0xFFFF);
            dwc++;
        }
        if (dt == 0x70000000) { // status
            found_status++;
            //		struct dch_branch_t *br0 = &bev->branch[0];
            //		struct dch_branch_t *br1 = &bev->branch[1];
            // serial link status
            int st = d & 0x0FF;
            bool link_LCF_timeout = st & BRIC_STATUS_BIT_LCF_TIMEOUT;
            bool bric_not_ready   = st & BRIC_STATUS_BIT_TIMEOUT;
            bool link_disconnect = st & BRIC_STATUS_BIT_DISCONNECT;
            bool link_data_error = (st & BRIC_STATUS_BIT_DATAERR) | (st & BRIC_STATUS_BIT_ERR_PROP);
            bool link_crc_error = st & BRIC_STATUS_BIT_CRC_ERR;
            //		bool link_ok = !link_LCF_timeout && !link_disconnect && !link_data_error;

            if (link_disconnect) {
                setStat("bric_link_down");
            } else {
                if (link_data_error) {
                    setStat("bric_link_data_error");
                }
                if (link_crc_error) {
                    setStat("bric_link_crc_error");
                }
                if (link_LCF_timeout) {
                    setStat("bric_LCF_timeout");
                } else {
                    if (bric_not_ready) {
                        setStat("bric_not_ready");
                    }
                    // Link Control Frame (LCF)
                    bricbox_lcf_t lcf;
                    lcf.all = (d >> 8) & 0xFFFF;
                    //		printf("LCF=%04X\n", lcf.all);
                    if (!lcf.bit.bric_ready) {
                        setStat("bricbox_not_ready");
                    }
                    if (lcf.bit.rx_error) {
                        setStat("bricbox_link_data_error");
                    }

                    /*
                if (lcf.bit.b1_timeout) {
//		  br0->timeout = true;
                  setStat("bricbox_b1_timeout");
                }
                if (lcf.bit.b2_timeout) {
//		  br1->timeout = true;
                  setStat("bricbox_b2_timeout");
                }
                if (lcf.bit.b1_absent) {
//		  br0->disconnect = true;
                  setStat("bricbox_b1_disconnect");
                }
                if (lcf.bit.b2_absent) {
//		  br1->disconnect = true;
                  setStat("bricbox_b2_disconnect");
                }
                if (lcf.bit.b1_overflow) {
//		  br0->overflow = true;
                  setStat("bricbox_b1_oveflow");
                }
                if (lcf.bit.b2_overflow) {
//		  br1->overflow = true;
                  setStat("bricbox_b2_oveflow");
                }
                if (lcf.bit.b1_align_err) {
//		  br0->data_error = true;
                  setStat("bricbox_b1_align_err");
                }
                if (lcf.bit.b2_align_err) {
//		  br1->data_error = true;
                  setStat("bricbox_b2_align_err");
                }
*/
                } // end of LCF timeout
            } // end of link_down
            /*
            // decode port 0 to branches 0, 1;
                unsigned int st[3];
                st[0] = d & 0xFF;
                st[1] = (d>>8) & 0xFF;
                st[2] = (d>>16) & 0xFF;
                bool timeout    = st[0] & BRIC_STATUS_BIT_TIMEOUT;
                bool disconnect = st[0] & BRIC_STATUS_BIT_DISCONNECT;
                bool data_error = (st[0] & BRIC_STATUS_BIT_DATAERR) | (st[0] & BRIC_STATUS_BIT_ERR_PROP);
                if (timeout) { br0->timeout = true; br1->timeout = true; setStat("bric_timeout"); }
                if (disconnect) { br0->disconnect = true; br1->disconnect = true; setStat("bric_offline"); }
                if (data_error) { br0->data_error = true; br1->data_error = true; setStat("bric_data_error"); }
//		bool bev->error_prop = st[0] & 0x08;
//		printf("bric status: %s%s%s\n", timeout?" TIMEOUT":"", disconnect?" DISCONNECT":"", data_error?" DATA_ERROR":"");
*/
        }
    }
    //  if (ev->en < 10) print();
    if (found_status == 0)
        setStat("bricbox_missing_status");
    if (found_status > 1)
        setStat("bricbox_toomany_status");
    return 0;
}
