
#include "decode_tdc.h"

#include <QDebug>

#include <algorithm>

#include "daqcommon/rawdata.h"
#include "decoder/decoder.h"
#include "hal/hw_manager.h"
#include "modules/tdc_types.h"
#include "hal/tdc_mod.h"
#include "decoder/decode_base.h"
#include "hptdc/hptdc.h"

namespace {
    const mfam_t mf = MF_TDC;
}

local_tdc_decoder::local_tdc_decoder(const hw_manager &p, event_t * const event)
    : local_base_decoder(p, event),
      tdcresindex(TDC_RES_INDEX_100),
      tdcpresindex(TDC_RES_INDEX_100),
      pair(false),
      edge(0)
{
}

bool local_tdc_decoder::isVeryHighResolution() const
{
    return tdc_res_vh(tdcresindex);
}

void local_tdc_decoder::print() const
{
    Q_ASSERT(ev);
    qDebug() << "*** TDC decoder state ***";
    if (!ev) return;
    qDebug().nospace() << "ev=" << ev->en << ", first_ch=" << first_ch << ", nhits=" << ev->tdc.hits.size();
    qDebug() << "**************************";
}

void local_tdc_decoder::updateDataFormat()
{
    if (hwf.get_nch_byfam(mf) < 1) return;
    tdcresindex = static_cast<TdcResIndex>(hwf.get_int("res", mf));
    tdcpresindex = static_cast<TdcResIndex>(hwf.get_int("pres", mf));
    edge = hwf.get_int("edge", mf);
    pair = hwf.get_int("pair", mf);
//    qDebug() << QString("updateDataFormat(), res=%1, pres=%2, pair=%3")
//            .arg(tdcresindex).arg(tdcpresindex).arg(pair);
}

int local_tdc_decoder::decode_ch(unsigned int d) const
{
    unsigned int id = (d >> 24) & 0xF;
    unsigned int tch = (d >> 19) & 0x1F;
    int ch = map_tdcidch2inp(id, tch);
    return ch;
}

QDebug operator<<(QDebug dbg, const tdc_hit_t &value) {
    dbg.nospace() << "time=" << value.time << ", trail=" << value.trail_edge << ", wid=" << value.wid;
    return dbg;
}

tdc_hits_t local_tdc_decoder::combine_hit_edges(const tdc_hits_t &hits)
{
    tdc_hits_t newhits;
    Q_FOREACH(int ch, hits.keys()) {
        const tdc_hitv_t &chitv = hits.value(ch);
        tdc_hitv_t hitv(chitv);
        std::sort(hitv.begin(), hitv.end());
        const size_t n = hitv.size();
        tdc_hitv_t newhitv;
        int state = 0;
        int tlead = 0;
        if (hitv.size() == 1)
            newhitv = hitv;
        else
            for (size_t i = 0; i<n; i++) {
                const tdc_hit_t &h = hitv[i];
                bool trail = h.trail_edge;
                if (!trail && state == 0) { tlead = h.time; state = 1; continue; } // TL
                if (!trail && state == 1) { newhitv.push_back(tdc_hit_t(tlead, 0)); tlead = h.time; state=1; continue; } // LL, leading without trailing
                if (trail && state == 1) { newhitv.push_back(tdc_hit_t(tlead, h.time-tlead)); state = 0; continue; } // LT
                if (trail && state == 0) { setStat(QString("trailing without leading in ch %1").arg(ch).toStdString()); } // TT, trailing without leading
            }
        newhits[ch] = newhitv;
//        qDebug() << "ch " << ch << ", count " << newhitv.size() << ": " << newhitv;
    }
    return newhits;
}

int local_tdc_decoder::decode()
{
    Q_ASSERT(ev);
    if (!ev) return -1;
    // not count module header and trailer
    //  unsigned int mhdr = buf[0];
    unsigned int mtrl = buf[len];

    tdc_event_t &tdc = ev->tdc;
    tdc_hits_t local_hits;
    //    tdc_hits_t &hits = tdc.hits;
    tdc_cnts_t &cnts = tdc.cnts;

    bool vh = tdc_res_vh(tdcresindex);

    if (decode_mtrl_ro(mtrl))
        tdc.error |= 0x10000;
    if (decode_mtrl_re(mtrl))
        tdc.error |= 0x20000;
    for (size_t p = 1; p < len; p++) {
        unsigned int d = buf[p];
        unsigned int dt = buf[p] & ADLVME_DTYPE_MASK_ALL;

        int raw_ch = decode_ch(vh ? (d & ~(3<<19)): d); // bits 20:19 == time[1:0] in Very High mode
        // TDC data
        if ((dt == 0x40000000) || (dt == 0x50000000)) {
            // hit
            int rawtime = decode_tdc_rawtime(d, tdcresindex);
            unsigned int hit_ch = first_ch + raw_ch;
            int hit_time, hit_wid;
            bool trail_edge = false;
            if (dt == 0x50000000)
                trail_edge = true;
            if (pair) {
                int rawwid = decode_tdc_rawwid(d, tdcpresindex);
                hit_time = rawtime & 0xFFF;
                hit_wid = rawwid;
            } else {
                hit_time = rawtime;
                hit_wid = 0;
            }
            local_hits[hit_ch].push_back(tdc_hit_t(hit_time, hit_wid, trail_edge));
        } // dt 4 or 5
        // counters
        if (dt == 0x00000000 || dt == 0x10000000) {
            int nn = (d>>16)&0x0FFF;
            nn += first_ch;
            unsigned int val = d&0xFFFF;
//            qDebug() << QString("d = 0x%1, nn = %2, val = %3").arg(d, 0, 16).arg(nn).arg(val);
            if (dt & 0x10000000)
            {cnts[nn] = (cnts[nn] & 0xFFFF0000u) | val;}
            else
            {cnts[nn] = (cnts[nn] & 0x0000FFFFu) | (val<<16);}
//            qDebug() << "cnts[" << nn << "]=" << cnts[nn];
            //  		printf(PREFIX "cnt[%d]=%u\n", nn, ev->tdc.cnt[nn]);
        }
        if (dt == 0x60000000) { // tdc error
            ev->tdc.error |= d&0x7FFF;
            handle_hptdc_error(d);
        }
    } // for (len)

    if (edge == 3) {
        local_hits = combine_hit_edges(local_hits);
    }

    Q_FOREACH(int ch, local_hits.keys()) {
        tdc.hits[ch] = local_hits[ch];
    }

    if (error_flag)
        return -1;
    return 0;
}
