
#include <cstddef>
#include <cstdio>

#include <QDebug>
#include <vector>
#include <algorithm>

#include "daqcommon/rawdata.h"
#include "decoder/decoder.h"
#include "hal/base_mod.h"
#include "hal/hw_manager.h"
#include "modules/tqdc_types.h"
#include "decode_tqdc.h"
#include "hptdc/hptdc.h"
#include "decoder/event.h"

using namespace std;

const int ADC_BITS = 10;
const int ADC_BITS_rev2 = 14;
const int ADC_ACC_BITS = 19;

local_tqdc_decoder::local_tqdc_decoder(const hw_manager &p, event_t * const event)
    : local_base_decoder(p, event)
{
}

void local_tqdc_decoder::print() const
{
    Q_ASSERT(ev);
    if (!ev) return;
    tqdc_event_t &tqdc = ev->tqdc;
    const tqdc_hits_t_t &hits_t = tqdc.hits_t;
    const tqdc_hits_q_t &hits_q = tqdc.hits_q;
    const tqdc_wfs_t &wfs = ev->tqdc.wfs;
    const tqdc_cnts_t &cnts = ev->tqdc.cnts;
    qDebug() << "*** TQDC decoder state ***";
    qDebug().nospace() << "ev=" << ev->en << ", first_ch=" << first_ch
            << ", hits_t[" << hits_t.size()
            << "], hits_q[" << hits_q.size()
            << "], wfs[" << wfs.size()
            << "], cnts[" << cnts.size()
            << "]";
    Q_FOREACH(int ch, wfs.keys()) {
        const tqdc_wfv_t &wfv = wfs.value(ch);
        for(tqdc_wfv_t::const_iterator jt=wfv.begin(); jt!=wfv.end(); ++jt) {
            const tqdc_wf_t &wf = (*jt);
            qDebug().nospace() << "wf: ch=" << ch
                    << ", len=" << wf.samp.size()
                    << ", charge=" << wf.getCharge();
        }
    }
    qDebug() << "**************************";
}

unsigned int local_tqdc_decoder::flush_waveforms()
{
    Q_ASSERT(ev);
    if (!ev) return -1;
    tqdc_event_t &tqdc = ev->tqdc;
    tqdc_wfs_t &wfs = tqdc.wfs;
    for(tmp_wfs_t::const_iterator it = tmp_wfs.begin(); it!=tmp_wfs.end(); ++it) {
        int ch = (*it).first;
        wfs[ch].push_back((*it).second);
    }
    tmp_wfs.clear();
    return 0;
}

unsigned int local_tqdc_decoder::flush_integrator()
{
    Q_ASSERT(ev);
    if (!ev) return -1;
    tqdc_event_t &tqdc = ev->tqdc;
    tqdc_hits_q_t &hits_q = tqdc.hits_q;
    for(tmp_hits_q_t::const_iterator it = tmp_hits_q.begin(); it!=tmp_hits_q.end(); ++it) {
        int ch = (*it).first;
        hits_q[ch].push_back((*it).second);
    }
    tmp_hits_q.clear();
    return 0;
}

TdcResIndex local_tqdc_decoder::get_tdcresindex() const
{
    if (hwf.get_nch_byfam(MF_TQDC) < 1) return TDC_RES_INDEX_100;
    return static_cast<TdcResIndex>(hwf.get_int("res", MF_TQDC));
}

bool local_tqdc_decoder::get_tdcpair() const
{
    if (hwf.get_nch_byfam(MF_TQDC) < 1) return false;
    return (hwf.get_int("pair", MF_TQDC));
}

int local_tqdc_decoder::get_samp_len() const
{
    if (hwf.get_nch_byfam(MF_TQDC) < 1) return 1;
    return (hwf.get_int("adc_match", MF_TQDC));
}

int local_tqdc_decoder::decode()
{
    Q_ASSERT(ev);
    Q_ASSERT(hwmod);
    if (!ev) return -1;

    const int module_rev = hwmod->get_rev();
    tqdc_event_t &tqdc = ev->tqdc;
    tqdc_hits_q_t &hits_q = tqdc.hits_q;
    tqdc_hits_t_t &hits_t = tqdc.hits_t;
    tqdc_wfs_t &wfs = tqdc.wfs;
    tqdc_cnts_t &cnts = tqdc.cnts;

    // count hits before decode
    int n0_t = 0;
    Q_FOREACH(tqdc_hitv_t_t hit, hits_t) n0_t += hit.size();
    int n0_q = 0;
    Q_FOREACH(tqdc_hitv_q_t hit, hits_q) n0_q += hit.size();

    TdcResIndex tdcresindex = get_tdcresindex();
    bool vh = tdc_res_vh(tdcresindex);
    int samp_len = std::max(1, get_samp_len());

    int cur_adc_ts = 0;
    int cur_trig_ts = 0;

    for (size_t p = 1; p < len; p++) {
        const unsigned int d = buf[p];
        const unsigned int dt = buf[p] & ADLVME_DTYPE_MASK_ALL;

//        const unsigned int id = (d >> 24) & 0x3;
        const unsigned int mode = (d >> 26) & 0x3;
        const unsigned int rawch = (d >> 19) & 0x1F;
        const unsigned int ch = first_ch+rawch;

        //  printf("tqdc_decode: [%d] id %01X, sdt %d, sdn %d, ch %d\n", p, id, sdt, sdn, ch);
        if (mode == 0) {
            // counters
            if (dt == 0x00000000 || dt == 0x10000000) {
                int nn = first_ch + ((d>>19)&0xF);
                int val = d&0xFFFF;
                if (dt & 0x10000000)
                    cnts[nn] = (cnts[nn] & 0xFFFF0000) | val;
                else
                    cnts[nn] = (cnts[nn] & 0x0000FFFF) | (val<<16);
                //  		printf("tqdc: cnt[%d]=%d\n", nn, ev->tqdc.cnt[nn]);
            }
            // TDC data
            if ((dt == 0x40000000) || (dt == 0x50000000)) {
                // hit
                int rawtime = 0;
                if (vh) {
                    unsigned int clear_bits = (3<<19); // clear bits 19, 20
                    unsigned int rcdata = ((d >> 24) & 0x3) << 19; // move RC data bits 24, 25 to 19, 20
                    unsigned int tdcd = (d & ~clear_bits) | rcdata;
                    rawtime = decode_tdc_rawtime(tdcd, tdcresindex);
                } else {
                    rawtime = decode_tdc_rawtime(d, tdcresindex);
                }
                hits_t[ch].push_back(tqdc_hit_t_t(rawtime, 0));
                //          printf("TQDC decoder: d=%08X, tdcresindex=%d, rawtime=%d\n", d, tdcresindex, rawtime);
            }
            if (dt == 0x60000000) { // tdc error
                tqdc.tdc_error |= d&0x7FFF;
                handle_hptdc_error(d);
            }
        } else {
            // test/adc/qdc data
            if (dt == TQDC_DTYPE_ADC_TS) {
                // software integrator - flush if any
                flush_integrator();
                flush_waveforms();
                // this is timestamp
                {
                    if (((d >> 16) & 7) == 0) {
                        // trigger timestamp
                        //            ts[ch] = (ts[ch] & 0x0000FFFF) | ((d & 0xFFFF) << 16);  // set high 16
                        cur_trig_ts = (d & 0xFFFF);
                        tqdc.ts = cur_trig_ts;
                        //            printf("ch %3d: TRIG TS=%d\n", ch, cur_trig_ts);
                    } else if (((d >> 16) & 7) == 1) {
                        // adc slide timestamp
                        //            ts[ch] = (ts[ch] & 0xFFFF0000) | ((d & 0xFFFF));    // set low 16
                        cur_adc_ts = (d & 0xFFFF);
                        //            tmp_wfs[ch].ts = ts_q;
                        //            printf("ch %3d:  ADC TS=%d\n", ch, cur_adc_ts);
                    }
                }
            } else if (dt == TQDC_DTYPE_ADC_DATA) {
                int adc_ts = cur_adc_ts - cur_trig_ts; // relative timestamp
                if (mode == 1 || mode == 2) {
                    int val;
                    if (module_rev < 2) {
                        val = (d & ((1<<ADC_BITS)-1)) - (1<<(ADC_BITS-1));
                    } else {
                        val = (d & ((1<<ADC_BITS_rev2)-1)) - (1<<(ADC_BITS_rev2-1));
//                        val = qint16(d);
                    }
                    tmp_wfs[ch].samp.push_back(val);
                    tmp_wfs[ch].ts = adc_ts;
                } else if (mode == 3) {
                    double val;
                    if (module_rev < 2) {
                        int rawval = d & ((1<<ADC_ACC_BITS)-1);
                        val = static_cast<double>(rawval) / samp_len - (1<<(ADC_BITS-1));
                    } else {
                        int rawval = d & ((1<<ADC_ACC_BITS)-1);
                        val = static_cast<double>(rawval) / samp_len - (1<<(ADC_BITS_rev2-1));
//                        val = static_cast<double>(qint16(d)) / samp_len;
                    }
                    tqdc_hit_q_t hit;
                    hit.ts = adc_ts;
                    hit.charge = val;
                    hits_q[ch].push_back(hit);
                }
            }
        }
    }
    // software integrator - flush if any
    flush_integrator();
    // save waveform if any and event wf buffer is not full
    flush_waveforms();

    // populate hits_q from wfs
    hits_q.clear();
    Q_FOREACH(int ch, wfs.keys()) {
        const tqdc_wfv_t &wfv = wfs.value(ch);
        for(tqdc_wfv_t::const_iterator jt = wfv.begin(); jt!=wfv.end(); ++jt) {
            const tqdc_wf_t &wf = (*jt);
            tqdc_hit_q_t hit_q;
            hit_q.charge = wf.getCharge();
            hit_q.ts = wf.ts;
            hits_q[ch].push_back(hit_q);
        }
    }

    // count hits after decode
    int n_t = 0;
    Q_FOREACH(tqdc_hitv_t_t hit, hits_t) n_t += hit.size();
    int n_q = 0;
    Q_FOREACH(tqdc_hitv_q_t hit, hits_q) n_q += hit.size();

    // Include this module counts only
    setStat("decoder_tqdc_hits_t", n_t - n0_t);
    setStat("decoder_tqdc_hits_q", n_q - n0_q);

    if (error_flag)
        return -1;
    return 0;
}
