
#include "decoder/decoder.h"

#include <QDebug>

#include <map>

#include "daqcommon/daqexc.h"
#include "daqcommon/device_ids.h"
#include "daqcommon/rawdata.h"
#include "hptdc/hptdc.h"
#include "daqcommon/vme_types.h"
#include "hal/base_mod.h"
#include "hal/hw_manager.h"
//#include "hal/hws.h" // MAX_HW_MODULES
#include "decoder/eventstats.h"
#include "decoder/decoder_modlist_inc.h"
#include "decoder/dvs.h"
#include "modules/bric_types.h"
#include "modules/msc_types.h"
#include "modules/trig_types.h"
#include "decoder/RawDumper.h"

//-----------------------------------------------------------------------//

base_decoder::base_decoder(const hw_manager &p)
    : ev(new event_t),
    hwf(p),
    buf(NULL),
    len(0),
    dbric(new local_dch_decoder(p, ev)),
    dmsc(new local_msc_decoder(p, ev)),
    dtrig(new local_trig_decoder(p, ev)),
    dtqdc(new local_tqdc_decoder(p, ev)),
    dtdc32vl(new local_tdc32vl_decoder(p, ev)),
    dtdc64v(new local_tdc64v_decoder(p, ev)),
    dtdc96(new local_tdc96_decoder(p, ev)),
    dmtdc64(new local_mtdc64_decoder(p, ev)),
    dphtdc(new local_phtdc_decoder(p, ev))
{
    register_decoder(dbric);
    register_decoder(dmsc);
    register_decoder(dtrig);
    register_decoder(dtqdc);
    register_decoder(dtdc32vl);
    register_decoder(dtdc64v);
    register_decoder(dtdc96);
    register_decoder(dmtdc64);
    register_decoder(dphtdc);
    reset();
}

base_decoder::~base_decoder()
{
    Q_FOREACH(local_base_decoder *p, decoders) {
        delete p;
    }
    delete ev;
}

void base_decoder::register_decoder(local_base_decoder *p)
{
    decoders.push_back(p);
}

void base_decoder::reset()
{
    clear();
    setbuf(0, 0);
    map_modules();
    map_channels();
}

void base_decoder::clear()
{
    Q_ASSERT(ev);
    if (ev) ev->clear();
    Q_FOREACH(local_base_decoder *p, decoders) {
        p->clear();
    }
}

void base_decoder::map_channels()
{
    std::map<mfam_t, int> ch_by_mfam;
    //  ch_offset.clear();
    for (unsigned int slot_num = 0; slot_num <= VME_SLOTS; slot_num++) {
        const base_mod *mod = hwf.find_mod_by_slot_const(slot_num);
        if (!mod) continue;
        int nch = mod->get_nch();
        mfam_t fam = mod->get_fam();
        int ofs = ch_by_mfam[fam];
        ch_offset[slot_num] = ofs;
        //    printf("========== Channel offset for slot %d: %d\n", slot_num, ofs);
        ch_by_mfam[fam] += nch;
    }
}

void base_decoder::map_modules()
{
    /*
  key_by_slot.clear();
  if (!hwf)
        throw exc("base_decoder: null hwf");
  for (int slot_num=VME_MIN_GA; slot_num<=VME_MAX_GA; slot_num++) {
        const base_mod *hwfmod = hwf->find_mod_by_slot_const(slot_num);
        ddb_key_t key = hwfmod->get_ddb_key();
        key_by_slot[slot_num] = key;
  }
*/
}

namespace {
    unsigned int data_dtype(unsigned int data)
    {
        return data & ADLVME_DTYPE_MASK_ALL;
    }
}

/*
 * Module data decoders. Each module type has personal decode function.
 */
bool base_decoder::decode_module_data(mtype_t mid, size_t pmh, size_t pmt, const base_mod *hwmod)
{
    //  bool handled = false;
    int slot_number = hwmod?hwmod->get_slot():0;
    int ch_ofs = ch_offset[slot_number];
    local_base_decoder *decoder = 0;

    switch (mid) {
    case MODULE_ID_PHTDC:
        decoder = dphtdc;
        break;
    case MODULE_ID_TDC32VL:
        decoder = dtdc32vl;
        break;
    case MODULE_ID_TDC64V:
        decoder = dtdc64v;
        break;
    case MODULE_ID_TDC96:
        decoder = dtdc96;
        break;
    case MODULE_ID_MSC16V:
        decoder = dmsc;
        break;
    case MODULE_ID_MTDC64:
        decoder = dmtdc64;
        break;
    case MODULE_ID_PHLOM:
        break;
    case MODULE_ID_TQDC:
        decoder = dtqdc;
        break;
    case MODULE_ID_TRIG:
        decoder = dtrig;
        break;
    case MODULE_ID_BRIC:
        decoder = dbric;
        break;
    default:
        qWarning() << QString("Unhandled module data: mid=%1").arg(mid);
    }
    if (!decoder) return false;
    decoder->prepare_decode(&buf[pmh], pmt - pmh, ch_ofs, hwmod);
    decoder->decode_module_header();
    decoder->decode();
    decoder->decode_module_trailer();
    decoder->post_decode();
    return true;
}

/*!
 * Decode one event. First word is expected to be EVENT HEADER, last word EVENT TRAILER.
 * \returns true on success
 *
 */
bool base_decoder::decode_buf(const unsigned int *pbuf, size_t s, unsigned int rawdata_flags)
{
    Q_ASSERT(ev);
    size_t p;
    std::map<int, int>nmod;

    ev->clear();
    if (rawdata_flags & 1)
        ev->evtype = EVT_EOB; // endspill data
    else
        ev->evtype = EVT_NORM;
    setbuf(pbuf, s);

    if (len < 1) {
        qDebug() << "event_decode: empty input buffer";
        return true;                 // empty is Ok
    }
    if (len < 2) {
        qDebug() << "event_decode: at least 2 words required";
        return false;                // require at least 2 words: EHDR, ETRL
    }

    if (data_dtype(buf[0]) != ADLVME_DTYPE_EHDR) {
        qDebug() << QString("event_decode: buf[0]=%1  (1)").arg(buf[0], 8, 16);
        return false;
    }
    if (data_dtype(buf[len - 1]) != ADLVME_DTYPE_ETRL) {
        qDebug() << QString("event_decode: buf[%1]=%2  (2)").arg(len - 1).arg(buf[len - 1], 8, 16);
        return false;
    }
    ev->en = decode_ehdr_en(buf[0]);
    ev->size = len;
    p = 1;
    nmod.clear();
    while (p < len - 1) {
        int mod = 0;
        size_t pmh = 0, pmt = 0;
        // search for MHDR
        while (p < len - 1) {
            if (data_dtype(buf[p]) == ADLVME_DTYPE_MHDR) {
                mod = 1;
                pmh = p;
                break;
            }
            p++;
        }
        // find MTRL only if MHDR was found
        while ((mod == 1) && (p < len - 1)) {
            if (data_dtype(buf[p]) == ADLVME_DTYPE_MTRL) {
                mod = 2;
                pmt = p;
                break;
            }
            /*
// MHDR, EHDR, ETRL are also terminators! (structure broken)
      if ((data_dtype (buf[p]) == ADLVME_DTYPE_MHDR) ||
          (data_dtype (buf[p]) == ADLVME_DTYPE_EHDR) ||
          (data_dtype (buf[p]) == ADLVME_DTYPE_ETRL) ||
          (data_dtype (buf[p]) == ADLVME_DTYPE_SHDR) ||
          (data_dtype (buf[p]) == ADLVME_DTYPE_STRL)
         ) {
        mod = 2;
        pmt = p;
        break;
      }
*/
            p++;
        }
        // process module data if any
        if ((mod == 2) && (pmh < pmt)) {
            int mid = decode_mhdr_id(buf[pmh]);
            int slot_num = decode_mhdr_ga(buf[pmh]);
            // count this module
            {
                nmod[mid]++;
                //        printf("p=%d, nmod[%d]=%d\n", p, mid, nmod[mid]);
                //      	debug(0, "Decoding module data: mid=%d %s", mid, module_id_str((mtype_t)mid));
                const base_mod *hwmod = hwf.find_mod_by_slot_const(slot_num); // FIXME: optimize
                if (!hwmod) {
                    qWarning() << QString("base_decoder::decode_buf: !hwmod for slot %1").arg(slot_num);
                    throw exc("base_decoder::decode_buf: !hwmod");
                }
                decode_module_data((mtype_t)mid, pmh, pmt, hwmod);
            }
        }
        p++;
    }
    //  if (error_flag) {
    //	debug(0, "event_decode: error_flag is set");
    //    return false;
    //  }
    //  debug(1, "event_decode: ok");
    return true;
}

void base_decoder::dataFormatChanged()
{
    qDebug() << "base_decoder::dataFormatChanged()";
    updateDataFormat();
}

void base_decoder::updateDataFormat()
{
    Q_FOREACH(local_base_decoder *p, decoders) {
        p->updateDataFormat();
    }
    map_modules();
    map_channels();
}

void base_decoder::print() const
{
    Q_FOREACH(local_base_decoder *p, decoders) {
        p->print();
    }
}
