//
//    Copyright 2010 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "decoder/dvs.h"

#include <QDebug>
#include <QStringList>

//#include "colorlog/log.h"

#include "hal/hws.h"
#include "daqcommon/device_ids.h"
#include "daqcommon/rawdata.h"
#include "RawDumper.h"

class CrcHarness {
public:
    enum {DI = 0xD5};
    explicit CrcHarness();
//    explicit CrcHarness(const CrcHarness &r);
    quint8 initValue() const { return 0xFF;}
    quint8 next(quint8 crc, quint32 data) const;
private:
    CrcHarness(const CrcHarness &r);
    CrcHarness &operator=(const CrcHarness &r);
    void initialize();
    quint8 crc8_table[256];
};
// 1+x^2+x^4+x^6+x^7+x^8
// x^8 + x^7 + x^6 + x^4 + x^2 + 1
// GP = 1 1101 0101
// DI = 0xD5
CrcHarness::CrcHarness()
{
    initialize();
}

//CrcHarness::CrcHarness(const CrcHarness &r)
//{
//    if (this != &r) *this = r;
//}

void CrcHarness::initialize()
{
    quint8 crc;
    for(int i=0; i<256; i++) {
        crc = i;
        for (int j=0; j<8; j++)
            crc = (crc << 1) ^ ((crc & 0x80) ? DI : 0);
        crc8_table[i] = crc & 0xFF;
//        printf("table[%d] = %3d (0x%02X)\n", i, crc, crc);
    }
}

quint8 CrcHarness::next(quint8 crc, quint32 data) const
{
    quint8 d;
    d = (data>>24) & 0xFF;
    crc = crc8_table[crc ^ d];

    d = (data>>16) & 0xFF;
    crc = crc8_table[crc ^ d];

    d = (data>>8) & 0xFF;
    crc = crc8_table[(crc ^ (d & 0xFF)) & 0xFF];

    d = data & 0xFF;
    crc = crc8_table[(crc ^ (d & 0xFF)) & 0xFF];
    return crc;
}


//#define PRINT_DEBUG(x) {if (!suppressedDebug()) printDebug(x);}
#define SET_WARN(p, x)  {if (!suppressedWarn()) setWarning(p, x);}
#define SET_ERR(p, x)   {if (!suppressedCrit())  setError(p, x);}

namespace {
    const unsigned int N_PRINT = 10;

    const int PRINT_N_WARN = 10;
}

dvs_t::dvs_t()
    : aborted(0), spill_number(0), nv(0), nev(0), pnmod(-1), modnum(0), hwif_spwc(0), sptrcnt(0),
      ec(dvs_errcount_t()),
      crcHarness(new CrcHarness),
      ds(DS_INIT), nerr(0), nwarn(0), pen(-1), en(0), ev_st(0),
      ewc(0), estart(0), eend(0),
      mid(MAX_HW_MODULES), mga(MAX_HW_MODULES), mwc(MAX_HW_MODULES), men(MAX_HW_MODULES), mcrc(MAX_HW_MODULES),
      printCountCrit(0), printCountWarn(0), nprintinf(0), printCountDebug(0)
{
    clear();
}

dvs_t::~dvs_t()
{
    delete crcHarness;
}

bool dvs_t::suppressedDebug() const
{
    return printCountDebug > N_PRINT;
}

bool dvs_t::suppressedWarn() const
{
    return printCountWarn > N_PRINT;
}

bool dvs_t::suppressedCrit() const
{
    return printCountCrit > N_PRINT;
}

void dvs_t::printDebug(int level, const QString &text)
{
//    if (!colorlog::debug_level_ge(level)) return;
    if (printCountDebug < N_PRINT)
        qDebug() << text;
    else
        if (printCountDebug == N_PRINT)
            qDebug() << "...more...";
    printCountDebug++;
}

void dvs_t::printWarn(const QString &text)
{
    if (printCountWarn < N_PRINT)
        qWarning() << text;
    else
        if (printCountWarn == N_PRINT)
            qWarning() << "...more...";
    printCountWarn++;
}

void dvs_t::printCrit(const QString &text)
{
    if (printCountCrit < N_PRINT)
        qCritical() << text;
    else
        if (printCountCrit == N_PRINT)
            qCritical() << "...more...";
    printCountCrit++;
}


void dvs_t::setWarning(size_t pos, const QString &text)
{
    nwarn++;
    ev_st |= EV_ST_WARN;

    if (suppressedWarn()) return;
    if (ds == DS_DATA || ds == DS_MHDR || ds == DS_MTRL)
        printWarn(QString("[%1] ev %2, module %3 (%4, slot %5): %6")
                 .arg(pos, 8, 16, QChar('0')).arg(en).arg(modnum)
                 .arg(module_id_str((mtype_t)mid[modnum]))
                 .arg(mga[modnum]).arg(text));
    else if (ds == DS_EHDR || ds == DS_ETRL)
        printWarn(QString("[%1] ev %2: %6")
                 .arg(pos, 8, 16, QChar('0')).arg(en).arg(text));
    else
        printWarn(QString("[%1]: %6")
                 .arg(pos, 8, 16, QChar('0')).arg(text));
}

void dvs_t::setError(size_t pos, const QString &text)
{
    nerr++;
    ev_st |= EV_ST_ERR;
    if (suppressedCrit()) return;
    if (ds == DS_DATA || ds == DS_MHDR || ds == DS_MTRL)
        printCrit(QString("[%1] ev %2, module %3 (%4, slot %5): %6")
                 .arg(pos, 8, 16, QChar('0')).arg(en).arg(modnum)
                 .arg(module_id_str((mtype_t)mid[modnum]))
                 .arg(mga[modnum]).arg(text));
    else if (ds == DS_EHDR || ds == DS_ETRL)
        printCrit(QString("[%1] ev %2: %6")
                 .arg(pos, 8, 16, QChar('0')).arg(en).arg(text));
    else
        printCrit(QString("[%1]: %6")
                 .arg(pos, 8, 16, QChar('0')).arg(text));
}

void dvs_t::clear()
{
//    *this = dvs_t();
    aborted = 0;
    spill_number = 0;
    nv = 0;
    nev = 0;
    pnmod = -1;
    modnum = 0;
    hwif_spwc = 0;
    sptrcnt = 0;
    ec = dvs_errcount_t();
    ds = DS_INIT;
    nerr = 0;
    nwarn = 0;
    pen = -1;
    en = 0;
    ev_st = 0;
    ewc = 0;
    estart = 0;
    eend = 0;
    mid.clear();  mid.resize(MAX_HW_MODULES);
    mga.clear();  mga.resize(MAX_HW_MODULES);
    mwc.clear();  mwc.resize(MAX_HW_MODULES);
    men.clear();  men.resize(MAX_HW_MODULES);
    mcrc.clear();mcrc.resize(MAX_HW_MODULES);
    printCountCrit = 0;
    printCountWarn = 0;
    nprintinf = 0;
    printCountDebug = 0;
}

dvs_errcount_t::dvs_errcount_t()
    : fatal(0), ehdr_en(0), etrl_wc(0), modnum(0),
      mod_en(MAX_HW_MODULES), mod_wc(MAX_HW_MODULES), mod_crc(MAX_HW_MODULES)
{
}

const char *dvs_t::strds(dv_ds_t ds)
{
    switch (ds) {
    case DS_INIT:
        return "INIT";
    case DS_SHDR:
        return "SHDR";
    case DS_EHDR:
        return "EHDR";
    case DS_MHDR:
        return "MHDR";
    case DS_DATA:
        return "DATA";
    case DS_MTRL:
        return "MTRL";
    case DS_ETRL:
        return "ETRL";
    case DS_STRL:
        return "STRL";
    }
    return "";
}

const char *daq_str_dtype_mod(int dt, int mid)
{
    switch (mid) {
    case MODULE_ID_PHTDC:
    case MODULE_ID_TDC96:
    case MODULE_ID_TDC64V:
    case MODULE_ID_TDC32VL:
        if (dt == 0x0) return "   TGHDR";
        if (dt == 0x1) return "   TGTRL";
        if (dt == 0x2) return "   THDR ";
        if (dt == 0x3) return "   TTRL ";
        if (dt == 0x4) return "    LD  ";
        if (dt == 0x5) return "    TR  ";
        if (dt == 0x6) return "   TERR ";
        if (dt == 0x7) return "   TDEBG";
        goto def;
        break;
    case MODULE_ID_BRIC:
        if (dt == 0x0) return "   HDR1 ";
        if (dt == 0x1) return "   HDR2 ";
        if (dt == 0x4) return "   DATA ";
        if (dt == 0x7) return "   STAT ";
        if (dt > 0 && dt < 8) return "   ???? ";
        goto def;
        break;
    case MODULE_ID_TQDC:
        if (dt == 0x0) return "   CNTL ";
        if (dt == 0x1) return "   CNTH ";
        if (dt == 0x4) return "    D   ";
        if (dt == 0x5) return "    D   ";
        if (dt == 0x7) return "   ???? ";
        goto def;
    case MODULE_ID_TRIG:
        if (dt == 0x0) return "   LC_R ";
        if (dt == 0x1) return "   LC_A ";
        if (dt == 0x2) return "   ???? ";
        if (dt == 0x3) return "   EVID ";
        if (dt == 0x4) return "   TS_L ";
        if (dt == 0x5) return "   TS_H ";
        if (dt == 0x6) return "   ???? ";
        if (dt == 0x7) return "   ICNT ";
        goto def;
    case MODULE_ID_MTDC64:
        if (dt == 0x4) return "    LD  ";
        if (dt == 0x5) return "    TR  ";
        goto def;
    case MODULE_ID_MSC16V:
        if (dt == 0x0) return "   GR 0 ";
        if (dt == 0x1) return "   GR 1 ";
        if (dt == 0x2) return "   GR 2 ";
        if (dt == 0x3) return "   GR 2 ";
        if (dt == 0x4) return "   TS   ";
        goto def;
    default:
        goto def;
    }
    def:
    return daq_str_dtype(dt);
}

void daq_print_dbuf(const unsigned int *buf, size_t p)
{
    QString str = QString("[%1]: %2 %3")
                  .arg(p, 8, 16, QChar('0'))
                  .arg(buf[p], 8, 16, QChar('0'))
                  .arg(daq_dbuf_str(buf, p));
    qDebug() << str;
}

QString daq_dbuf_str(const unsigned int *buf, size_t p)
{
    return QString("[%1]: %2 ")
            .arg(p, 8, 16, QChar('0'))
            .arg(buf[p], 8, 16, QChar('0')) +
            GlobalRawDumper::Instance()->decodeRawDataWord(buf[p]);
}

bool shm_evlist_t::put(unsigned int pstart, unsigned int pend)
{
    Q_ASSERT(magic == SHM_EVLIST_MAGIC);
    if (nev + 1 < SHM_EVLIST_MAXENT) {
        start[nev] = pstart;
        end[nev] = pend;
        ev_flags[nev] |= 0x1;
        nev++;
    } else {
        if (nev + 1 == SHM_EVLIST_MAXENT)
            qWarning("shm_evlist_put: list overflow");
        return false;
    }
    return true;
}

void shm_evlist_t::clear()
{
    *this = shm_evlist_t();
    magic = SHM_EVLIST_MAGIC; // set valid
}

void dvs_t::newspill()
{
    pen = -1;                // reset event number
}

void dvs_t::newevent()
{
    ewc = 0;
    ev_st = 0;
    // reset module counter
    modnum = 0;
    for (int i = 0; i < MAX_HW_MODULES; i++) {
        mwc[i] = 0;
        mcrc[i] = crcHarness->initValue();
    }
}

void dvs_t::newmodule(int en, int id, int ga)
{
    if (modnum < MAX_HW_MODULES) {
        men[modnum] = en;
        mid[modnum] = id;
        mga[modnum] = ga;
        mcrc[modnum] = crcHarness->initValue();
    }
    // reset module[modnum] counters
    if (modnum < MAX_HW_MODULES) {
        mwc[modnum] = 0;
        mcrc[modnum] = crcHarness->initValue();
    }
}

void daq_dump_raw(const unsigned int *buf, size_t start, size_t end, size_t mark, bool compress)
{
    qDebug() << GlobalRawDumper::Instance()->dump_raw_str(buf, start, end, mark, compress);
}

bool dvs_t::check_endspill_data()
{
    bool result = true;
    if (aborted) return false;
    if (nv == 0) return false;
    if (ds != DS_STRL) {
        //      daq_print_dbuf (buf, p);
        printCrit(QString("Decoder end state is %1, not %2").arg(strds(ds)).arg(strds(DS_STRL)));
        //      daq_dump_raw (buf, p - 5, (p + 5) < len ? p + 5 : len, p);
        nerr++;
        //  	  ev_st |= 1;
        result = false;
    }
    if (nv < (hwif_spwc-15)) {
        printCrit(QString("Data lost: hardware wordcount %1, decoded %2 (%3 words lost)")
                 .arg(hwif_spwc).arg(nv).arg(hwif_spwc-15-nv));
        nerr++;
        //    ev_st |= 1;
        result = false;
    }
    return result;
}

void dvs_t::dump_event(const unsigned int *buf)
{
    if (suppressedDebug()) return;
    const size_t dump_pre = 5;
    size_t s, e;

    if (!buf)
        return;
    s = estart;
    e = eend;
    if (s > dump_pre)
        s -= dump_pre;
    else
        s = 0;
    RawDumper localRawDumper;
    QStringList sl;
    int before = estart-s;
    QString str_before;
    if (before == 1) str_before = QString(" %1 word before").arg(before);
    if (before > 1) str_before = QString(" %1 words before").arg(before);
    sl << QString("--- DUMP EVENT %1 BEGIN ---%2").arg(en).arg(str_before);
    sl << localRawDumper.dump_raw_str(buf, s, e, estart);
    sl << QString("--- DUMP EVENT %1 END ---").arg(en);
    printDebug(0, sl.join("\n"));
}

bool dvs_t::verify_raw(shm_evlist_t *shm_evlist, const unsigned int *buf,
                      size_t start, size_t *inend, int endspill)
{
    const int DUMP_PREV = 5;
    const int DUMP_POST = 5;
    dv_ds_t newds;

    if (!inend)
        return false;
    size_t len = *inend;
    if (len < 1)
        return false;

    if (start >= len)
        return true;

    if (aborted)
        return true;

    if (!shm_evlist || shm_evlist->magic != SHM_EVLIST_MAGIC) {
        printCrit("Invalid shm_evlist, raw decoder aborted");
        aborted = 1;
        nerr++;
        return false;
    }
    size_t p;
    for (p = start; p < len; p++) {
        const unsigned int dtype = buf[p] & ADLVME_DTYPE_MASK_ALL;
        if (dtype == ADLVME_DTYPE_SHDR) {
            if (buf[p]  == DATA_SHDR_ENDBURST) { // ENDBURST DATA
                newspill();
                pnmod = -1;
                pen = -1;
            } else {
                spill_number = decode_strl_spill_number(buf[p]);
                qDebug() << QString("dvs_newspill at %1, ds=%2").arg(p).arg(strds(ds));
            }
            if ((ds == DS_INIT) || (ds == DS_STRL)) {
                newds = DS_SHDR;
                goto ds_change;
            }
        }

        if (dtype == ADLVME_DTYPE_STAT) {
            ewc++;
            if (modnum < MAX_HW_MODULES)
                mwc[modnum]++;
            continue;
        }
        if (dtype == ADLVME_DTYPE_EHDR) {
            // TODO: check EHDR STATUS (bits 20-27)
            estart = p;
            en = decode_ehdr_en(buf[p]);
            if (en != static_cast<int>((pen + 1) & DATA_EHDR_EN_MASK)) {
                ec.ehdr_en++;
                SET_WARN(p, QString("event number not incrementing: %1 after %2").arg(en).arg(pen));
                if (nwarn < PRINT_N_WARN) {
                    daq_print_dbuf(buf, p);
                }
            }
            pen = en;
            newevent();

            bool EHDR_change_ok = true;
            QString msg;
            if (ds == DS_INIT) {
                EHDR_change_ok = false;
                msg += "EHDR in state INIT, possibly lost SHDR?";
            }
            if (ds == DS_STRL) {
                EHDR_change_ok = false;
                msg += "EHDR in state STRL, possibly lost SHDR?";
            }
            if (ds == DS_DATA) {
                EHDR_change_ok = false;
                msg += "EHDR in state DATA, lost MTRL, ETRL?";
            }
            if (ds == DS_MHDR) {
                EHDR_change_ok = false;
                msg += "EHDR in state MHDR, possibly lost MHDR, ETRL?";
            }
            if (ds == DS_MTRL) {
                EHDR_change_ok = false;
                msg += "EHDR in state MTRL, possibly lost ETRL?";
            }
            if (!EHDR_change_ok) {
                SET_ERR(p, msg);
                if (nerr < PRINT_N_WARN) {
                    daq_print_dbuf(buf, p);
                }
            }
            //      if ((ds == DS_SHDR) || (ds == DS_ETRL)) {
            if ((ds == DS_INIT) || (ds == DS_SHDR) ||
                (ds == DS_ETRL) || (ds == DS_MHDR) ||
                (ds == DS_MTRL) || (ds == DS_DATA)) {
                newds = DS_EHDR;
                goto ds_change;
            }
        }

        if (dtype == ADLVME_DTYPE_MHDR) {
            ewc++;
            const int mhdr_en = decode_mhdr_en(buf[p]);    // MHDR event number
            const int mhdr_en_masked = mhdr_en & DATA_EHDR_EN_MASK & DATA_MHDR_EN_MASK;
            const int en_masked = en & DATA_EHDR_EN_MASK & DATA_MHDR_EN_MASK;
            const int ga = decode_mhdr_ga(buf[p]);
            if (mhdr_en_masked != en_masked) {
                if (modnum < MAX_HW_MODULES)
                    ec.mod_en[modnum]++;
                SET_WARN(p, QString("module event number bad: %1").arg(mhdr_en));
                if (nwarn < PRINT_N_WARN) {
                    daq_print_dbuf(buf, p);
                }
            }
            newmodule(mhdr_en, decode_mhdr_id(buf[p]), ga);
            mcrc[modnum] = crcHarness->next(mcrc[modnum], buf[p]);
            if ((ds == DS_EHDR) || (ds == DS_MTRL)) {
                newds = DS_MHDR;
                goto ds_change;
            }
        }

        if (ds == DS_MTRL && buf[p] == 0xFFFFFFFF) { // 64-bit padding workaround
            ewc++;
            // do nothing
            newds = ds;
            goto ds_change;
        }

        if ((dtype & ADLVME_DTYPE_MASK_CTRL) == 0) {
            ewc++;
            if (modnum < MAX_HW_MODULES) {
                mwc[modnum]++;
                mcrc[modnum] = crcHarness->next(mcrc[modnum], buf[p]);
            }
            if ((ds == DS_EHDR) ||(ds == DS_MHDR) || (ds == DS_DATA)) {
                newds = DS_DATA;
                goto ds_change;
            }
        }

        if (dtype == ADLVME_DTYPE_MTRL) {
            ewc++;
            int mtrl_wc = decode_mtrl_wc(buf[p]);
            if (modnum < MAX_HW_MODULES) {
                // check for wordcount
                if (mwc[modnum] != mtrl_wc &&  mwc[modnum] != mtrl_wc + 1) { // 64-bit padding workaround
                    ec.mod_wc[modnum]++;
                    SET_WARN(p, QString("wordcount bad: %1, counted %2").arg(mtrl_wc).arg(mwc[modnum]));
                }
            }
            quint32 mtrl_crc = decode_mtrl_crc(buf[p]);
            if ((mcrc[modnum] != (int)mtrl_crc) && !endspill) {
                ec.mod_crc[modnum]++;
                SET_WARN(p, QString("crc mismatch: %1, counted %2")
                         .arg(mtrl_crc, 2, 16, QChar('0'))
                         .arg(mcrc[modnum], 2, 16, QChar('0')));
            }
            if ((ds == DS_MHDR) || (ds == DS_DATA)) {
                // readout error
                if (decode_mtrl_re(buf[p])) {
                    SET_WARN(p, "readout error");
                }
                // readout overflow
                if (decode_mtrl_ro(buf[p])) {
                    SET_WARN(p, "readout overflow");
                }
                // at last, increment module count
                modnum++;
                if (modnum > MAX_HW_MODULES) {
                    ec.modnum++;
                    SET_ERR(p, QString("too much modules: %1 (no EHDR?)").arg(modnum));
                    daq_print_dbuf(buf, p);
                }
                newds = DS_MTRL;
                goto ds_change;
            }
        }

        if (dtype == ADLVME_DTYPE_ETRL) {
            eend = p;
            int etrl_wc = decode_etrl_wc(buf[p]);    // wordcount
            if (etrl_wc != ewc) {
                ec.etrl_wc++;
                SET_WARN(p, QString("event wordcount bad: %1, counted %2").arg(etrl_wc).arg(ewc));
            }
//            int rdst = decode_etrl_readout_status(buf[p]);
//            bool readout_timeout = rdst & ETRL_STATUS_BIT_TIMEOUT;
//            if (readout_timeout) {
//                SET_WARN(p, "readout timeout");
//            }
            if (ds == DS_MHDR) {
                SET_ERR(p, "lost module trailer");
            }
            bool ETRL_change_ok = true;
            QString msg;
            if (ds == DS_DATA) {
                ETRL_change_ok = false;
                msg += "ETRL in state DATA, possibly lost MTRL?";
            }
            if (ds == DS_MHDR) {
                ETRL_change_ok = false;
                msg += "ETRL in state MHDR, possibly lost DATA, MTRL?";
            }
            if (!ETRL_change_ok) {
                SET_ERR(p, msg);
                if (nerr < PRINT_N_WARN) {
                    daq_print_dbuf(buf, p);
                }
            }

            if ((ds == DS_EHDR) || (ds == DS_MHDR) ||
                (ds == DS_MTRL) || (ds == DS_DATA)) {
                nev++;
                if (pnmod != -1) {
                    if (modnum != pnmod) {
                        ec.modnum++;
                        SET_ERR(p, QString("number of modules in event changing: was %1, now %2")
                                     .arg(pnmod).arg(modnum));
                        if (nerr < PRINT_N_WARN) {
                            daq_print_dbuf(buf, p);
                        }
                    }
                }
                pnmod = modnum;
                // dump in case of errors
                if (ev_st & (EV_ST_WARN | EV_ST_ERR)) {
                    //          if (nwarn < PRINT_N_WARN)
                    dump_event(buf);
                }
                //        if (nev < 2) daq_dump_event(dvs, buf);
                bool ok = shm_evlist->put(estart, eend);
                if (!ok)
                    SET_ERR(p, "shm_evlist_put error");
                newds = DS_ETRL;
                goto ds_change;
            }
        }

        if (dtype == ADLVME_DTYPE_STRL) {
            if (buf[p] == DATA_STRL_ENDBURST) {
                // ENDBURST DATA
                //		pnmod = -1;
                //		pen = -1;
                newds = DS_SHDR;
                goto ds_change;
            } else {
                if (ds == DS_INIT) {
                    SET_ERR(p, "unexpected STRL %08X");
                    //		ev_st |= 1;
                    daq_dump_raw(buf, p-DUMP_PREV, p+/*DUMP_POST*/20, p, true);
                    continue;
                } else {
                    sptrcnt = decode_strl_wc(buf[p]);
                    newds = DS_STRL;
                    goto ds_change;
                }
            }
        }

        ec.fatal++;
        SET_ERR(p, QString("unexpected %1 in state %2")
                     .arg(daq_str_dtype((unsigned) dtype >> 28)).arg(strds(ds)));
        if (nerr < PRINT_N_WARN) {
            daq_dump_raw(buf, p-DUMP_PREV, p+DUMP_POST, p, true);
        }

        continue;

        ds_change:
        ds = newds;
        nv++;
        if (ds == DS_STRL) {
            p++;
            break;
        }
    }
    /*
  if (nv > 0) {
    if (endspill && !aborted && (ds != DS_STRL)) {
      daq_print_dbuf (buf, p);
      printerr("Last word is not STRL");
      daq_dump_raw (buf, p - 5, (p + 5) < len ? p + 5 : len, p);
      nerr++;
    }
  }
  if (endspill) {
    printnorm("Verify: %d words, %d events, %d modules", nv, nev,
               pnmod >= 0 ? pnmod : modnum);
  }
*/
    //  *inend = nv;
    *inend = p;
    // printnorm("daq_verify_raw: end 0x%08X", p);


    if (endspill) {
        if (nerr)
            qWarning() << QString("Raw data: %1 errors, %2 warnings").arg(nerr).arg(nwarn);
        else
            if (nwarn)
                qWarning() << QString("Raw data: %1 warnings").arg(nwarn);
        else
            qWarning() << "Raw data:: Ok";
    }
    bool ok = !nwarn && !nerr;
    return ok;
}

void dvs_t::print_verify_summary() const
{
    //  dump_raw(buf, estart, eend, -1);
    qDebug() << " ---------- Summary ---------- ";
    if (aborted)
        qWarning() << "Data verification aborted";
    qDebug().nospace() << "Modules: " << pnmod << ". Errors: " << nerr << ", warnings: " << nwarn;
    qDebug() << "Total triggers (cnt in STRL):  " << sptrcnt;
    qDebug() << "Received events (ETRL count):  " << nev;
    if (ec.fatal)
        qWarning() << "Fatal data error";
    if (ec.ehdr_en)
        qWarning() << "Event header event number errors: " << ec.ehdr_en;
    if (ec.etrl_wc)
        qWarning() << "Event trailer wordcount errors: " << ec.etrl_wc;
    if (ec.modnum)
        qWarning() << "Bad module count errors: " << ec.modnum;
    for (int m = 0; m < pnmod; m++) {
        int n_en = ec.mod_en[m];
        int n_wc = ec.mod_wc[m];
        int n_crc = ec.mod_crc[m];
        if (n_en || n_wc || n_crc) {
            qWarning().nospace() << "*** Module " << m << " ***";
            if (n_en)
                qWarning() << "  Module header event number errors: " << ec.mod_en[m];
            if (n_wc)
                qWarning() << "  Module trailer wordcount errors: " << ec.mod_wc[m];
            if (n_crc)
                qWarning() << "  Module data checksum errors: " << ec.mod_crc[m];
        }
    }
}
