//
//    Copyright 2010 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#ifndef DVS_H
#define DVS_H

#include <cstddef>
#include <vector>
#include <string>

const unsigned int SHM_EVLIST_MAGIC = 0x544E5645;
const unsigned int SHM_EVLIST_MAXENT = 1000000;

struct shm_evlist_t
{
    void clear();
    bool put(unsigned int pstart, unsigned int pend);

    unsigned int magic;
    unsigned int nev;           // number of events
    unsigned int start[SHM_EVLIST_MAXENT]; // event start offset, words
    unsigned int end[SHM_EVLIST_MAXENT];   // event end-?+1 offset, words
    unsigned int ev_flags[SHM_EVLIST_MAXENT];
};

// Error Counters
class dvs_errcount_t {
public:
    dvs_errcount_t();
    unsigned int fatal;         // uncorrectable error in data format
    unsigned int ehdr_en;       // bad event number in EHDR
    unsigned int etrl_wc;       // bad wordcount in ETRL
    unsigned int modnum;        // bad number of modules in event
    std::vector<unsigned int> mod_en;  // bad event number in MHDR
    std::vector<unsigned int> mod_wc;  // bad wordcount in MTRL
    std::vector<unsigned int> mod_crc; // bad checksum
};

class QString;
class CrcHarness;

// Data Verify Status
class dvs_t
{
private:
    enum dv_ds_t {
        DS_INIT,
        DS_SHDR,
        DS_EHDR,
        DS_MHDR,
        DS_DATA,
        DS_MTRL,
        DS_ETRL,
        DS_STRL,
    };
    enum ev_st_t{
        EV_ST_ERR = 1,
        EV_ST_WARN = 2,
    };

public:
    explicit dvs_t();
    ~dvs_t();
    void clear();
    void print_verify_summary()  const;
    bool check_endspill_data();
    bool verify_raw(shm_evlist_t *shm_evlist, const unsigned int *buf, size_t start, size_t *inend, int endspill);

    int aborted;                // verify aborted - fatal data error
    unsigned int spill_number;  // SHDR spill number
    int nv;                     // number of words verified
    int nev;                    // number of events done
    int pnmod;                  // number of modules in previous event
    int modnum;                 // number of modules in current event
    int hwif_spwc;              // spill wordcount read from hardware
    int sptrcnt;                // counter in spill trailer (total triggers)

    dvs_errcount_t ec;          // total counters

private:
    dvs_t(const dvs_t &);
    dvs_t &operator=(const dvs_t);
    void dump_event (const unsigned int *buf);
    void newspill();
    void newevent();
    void newmodule(int en, int id, int ga);
    const char *strds(dv_ds_t ds);
    bool suppressedDebug() const;
    bool suppressedWarn() const;
    bool suppressedCrit() const;
    void printDebug(int level, const QString &text);
//    void printnorm(const QString &text);
    void printWarn(const QString &text);
    void printCrit(const QString &text);
    void setWarning(size_t pos, const QString &text);
    void setError(size_t pos, const QString &text);

    CrcHarness const * crcHarness;
    dv_ds_t ds;                 // dvs state
    int nerr;                   // number of errors
    int nwarn;                  // number of warnings
    int pen;                    // previous event's number
    int en;                     // this event's number
    int ev_st;                  // this event's status
    int ewc;                    // event wordcount
    int estart, eend;           // event start and end marker positions (len=end-start+1)
    std::vector<int> mid;       // module ID
    std::vector<int> mga;       // module GA
    std::vector<int> mwc;       // module wordcount
    std::vector<int> men;       // module event number
    std::vector<int> mcrc;      // module data checksum

    unsigned int printCountCrit;
    unsigned int printCountWarn;
    unsigned int nprintinf;
    unsigned int printCountDebug;
};

const char *daq_str_dtype(int dtype);
//! print one data word: buf[p], including trailing '\n'
void daq_print_dbuf (const unsigned int *buf, size_t p);
QString daq_dbuf_str (const unsigned int *buf, size_t p);

void daq_dump_raw (const unsigned int *buf, size_t start, size_t end, size_t mark, bool compress);
const char *daq_str_dtype_mod (int dt, int mid);


#endif // DVS_H
