
#include <QDebug>
#include <map>
#include <string>
#include <sstream>

#include "decoder/eventstats.h"
#include "hal/base_mod.h"
#include "hal/hw_manager.h"

using namespace std;

void printEventStats(const EventStats &eventStats)
{
    qDebug() << "Event statistics by module";
    typedef std::map<std::string, qint64> total_t;
    total_t total;
    for(EventStats::const_iterator it = eventStats.begin(); it!=eventStats.end(); ++it) {
        const std::string &name = (*it).first;
        const ModuleStats &m = (*it).second;
        for(ModuleStats::const_iterator jt = m.begin(); jt!=m.end(); ++jt) {
            const ddb_key_t &key = (*jt).first;
            qint64 count = (*jt).second;
            total[name] += count;
            qDebug().nospace() << QString::fromStdString(name) << "[" << key << "]: " << count;
        }
    }
    qDebug() << "Event statistics summary";
    for(total_t::const_iterator it = total.begin(); it!=total.end(); ++it) {
        const std::string &name = (*it).first;
        qint64 count = (*it).second;
        qDebug().nospace() << QString::fromStdString(name) << ": " << count;
    }
}

void EventStatsSum::add(const EventStats &eventStats)
{
    for(EventStats::const_iterator it = eventStats.begin(); it!=eventStats.end(); ++it) {
        const std::string &name = (*it).first;
        const ModuleStats &m = (*it).second;
        for(ModuleStats::const_iterator jt = m.begin(); jt!=m.end(); ++jt) {
            const ddb_key_t &key = (*jt).first;
            qint64 count = (*jt).second;
            data[name][key] += count;
        }
    }
    //  printEventStats(data);
}

void EventStatsSum::clear()
{
    data.clear();
}

std::string EventStatsSum::getHtml() const
{
//    std::string html;
    std::ostringstream os;

    // Find all modules
    typedef std::map<ddb_key_t, bool> mods_t;
    mods_t mods;
    for(EventStats::const_iterator it = data.begin(); it!=data.end(); ++it) {
        const ModuleStats &m = (*it).second;
        for(ModuleStats::const_iterator jt = m.begin(); jt!=m.end(); ++jt) {
            const ddb_key_t key = (*jt).first;
            mods[key] = true;
        }
    }

    // Find all variables
    typedef std::map<std::string, bool> vars_t;
    vars_t vars;
    for(EventStats::const_iterator it = data.begin(); it!=data.end(); ++it) {
        const std::string &name = (*it).first;
        vars[name] = true;
    }

    // Map module keys to table column
    std::map<ddb_key_t, int> mod_col;
    int ncols = 2;
    for(mods_t::const_iterator it = mods.begin(); it!=mods.end(); ++it) {
        mod_col[(*it).first] = ncols;
        ncols++;
    }

    // Map variable name to table row, totals is 1
    std::map<std::string, int> var_row;
    int nrows = 1;
    for(vars_t::const_iterator it = vars.begin(); it!=vars.end(); ++it) {
        var_row[(*it).first] = nrows;
        nrows++;
    }


    // make table, count totals
    std::map<int, std::map<int, std::string> > table;
    table[0][0] = "Name";
    table[0][1] = "Total";
    // vertical header
    for(vars_t::const_iterator it = vars.begin(); it!=vars.end(); ++it) {
        const std::string &name = (*it).first;
        table[var_row[name]][0] = name;
    }
    // horizontal header
    for(mods_t::const_iterator it = mods.begin(); it!=mods.end(); ++it) {
        const ddb_key_t &key = (*it).first;
        //    ostringstream o; o << hex << key;
        const base_mod *mod = hwf.find_mod_by_key_const(key);
        if (!mod) continue;
        ostringstream o; o << mod->get_slot() << ": " << mod->get_name();
        table[0][mod_col[key]] = o.str();
    }

    std::map<std::string, qint64> total;
    for(EventStats::const_iterator it = data.begin(); it!=data.end(); ++it) {
        const std::string &name = (*it).first;
        const ModuleStats &m = (*it).second;
        int row = var_row[name];
        for(ModuleStats::const_iterator jt = m.begin(); jt!=m.end(); ++jt) {
            const ddb_key_t key = (*jt).first;
            const qint64 count = (*jt).second;
            total[name] += count;
            int col = mod_col[key];
            ostringstream o; o << count;
            table[row][col] = o.str();
        }
        ostringstream o; o << total[name];
        table[row][1] = o.str();
    }

    // make html
    // title
    os << "<html>\n";
    os << "<h1>Event statistics by module</h1>\n";
    os << "<table border=1>\n";

    for (int row = 0; row < nrows; row++) {
        os << "<tr>";
        for (int col = 0; col<ncols; col++) {
            std::string text = table[row][col];
            if (text.empty())
                text = "&nbsp;";
            os << "<td>" << text << "</td>";
        }
        os << "</tr>";
    }
    /*
  os << "<th>";
  os << "<td>Title</td>";
  os << "<td>Total</td>";
  foreach(ddb_keys, it)
    os << "<td>" << hex << (*it).first << "</td>";
  os << "</th>\n";
  foreach_const(data, it) {
        const std::string &name = (*it).first;
        const ModuleStats &m = (*it).second;
        os << "<tr>";
        os << "<td>" << name << "</td>";
        foreach_const(m, jt) {
          qint64 count = (*jt).second;
          total[name] += count;
        }
        foreach_const(m, jt) {
          const ddb_key_t &key = (*jt).first;
          qint64 count = (*jt).second;
          os << "<td>"  "</td><td>" << key << "</td><td>" << count << "</td></tr>\n";
        }
        os << "<td>" << total[name] << "</td>";
        os << "</tr>\n";
  }
*/
    os << "</table>\n";
    os << "</html>\n";
    return os.str();
}
