
#include "hwif_vme_mlink.h"

#include <QDebug>
#include <QCoreApplication>

#include <iostream>
#include <iomanip>
#include <sstream>
#include <cerrno>
#include <cstdio>
#include <cstring>
#include <cassert>

#include "colorlog/log.h"
#include "mlink/frame_types.h"
#include "mlink/mlink_frame_header.h"
#include "mlink/frame_hw.h"

#include "daqcommon/daqexc.h"
#include "hal/hwif_event.h"
#include "daqcommon/vme_types.h"
#include "fvme_reg.h"
#include "daqcommon/sysutil.h"
//#include "vmedaq_jtag_impl_fvme.h"
#include "daqcommon/QDebugFormatter.h"
#include "daqcommon/TextFormatter.h"
#include "daqcommon/UnixTimeStamp.h"
#include "daqcommon/vmedaq_gettimeofday.h"

#define PREFIX "vme_mlink: "

using namespace std;
using colorlog::debug_level_ge;
using namespace mlink;
namespace {
enum {
    FVME_DEVICE_ID = 0xD100
};
enum {
    REG16_DEVICE_ID    = 0x0042,
    REG64_SERIAL_ID    = 0x0050,
    REG16_TEMPERATURE  = 0x0054,
    REG16_FW_VER       = 0x0056,
    REG16_FW_REV       = 0x0057,
    REG32_STATS_BASE   = 0x0080
};

//enum {FVME_STATS_NUM = 16 };
enum {FVME_STATS_NUM = 4 }; // others unused

    const char *fvme_reg_name(unsigned int reg_addr)
    {

        const char *fvme_reg_name_str[16] = {
            "control_0",
            "control_1",
            "am_0",
            "am_1",
            "addr_lo",
            "addr_hi",
            "w_data_lo",
            "w_data_hi",
            "r_data_lo",
            "r_data_hi",
            "status_1",
            "status_2",
            "link_setup",
            "test",
            "spwc_lo",
            "spwc_hi"
        };
        if (reg_addr < 16)
            return fvme_reg_name_str[reg_addr];
        return "unknown";
    }

    const char *spill_src_str[8] = {
        "TTC", "IRQ1", "IRQ2", "IRQ3", "IRQ4", "IRQ5", "IRQ6", "IRQ7"
            };

    const char *trig_src_str[8] = {
        "TTC", "IRQ1", "IRQ2", "IRQ3", "IRQ4", "IRQ5", "IRQ6", "IRQ7"
            };

    std::ostream &operator<<(std::ostream &stream, const fvme_reg_status_0_t &st)
    {
        int irqs = st.bit.irqs;
        bool ok = !st.bit.tmout && st.bit.dtack && !st.bit.berr;
        stream  << (ok?" Ok":"")
                << (st.bit.tmout?" TIMEOUT":"")
                << (st.bit.dtack?"":" NO_DTACK")
                << (st.bit.berr?" BERR":"")
                << ((irqs & 0x01)?" IRQ0":"")
                << ((irqs & 0x02)?" IRQ1":"")
                << ((irqs & 0x04)?" IRQ2_SPILL":"")
                << ((irqs & 0x08)?" IRQ3":"")
                << ((irqs & 0x10)?" IRQ4":"")
                << ((irqs & 0x20)?" IRQ5_TRIG":"")
                << ((irqs & 0x40)?" IRQ6":"")
                << ((irqs & 0x80)?" IRQ7":"");
        return stream;
    }

    //void snprintf_vme_status(char *str1, size_t size, const fvme_reg_status_0_t &st)
    //{
    //    std::ostringstream ost;
    //    ost << st;
    //    snprintf(str1, size, "%s", ost.str().c_str());
    //}

    const unsigned int max_io_req_size = 9; // maximum number of command words for one i/o

    std::ostream &operator<<(std::ostream &stream, const vme_io_t &io)
    {
        fvme_reg_status_0_t st; st.all = io.st;
        int addr_bits = vme_addr_bits(io.am);
        int ga_base = io.addr >> (addr_bits - 5);
        stream << "VME " << (io.write?"WRITE":"READ ") << ": AM=" << std::hex << std::setw(2) << io.am
                << "h, addr=" << std::setw(addr_bits/4) << io.addr
                << "h, slot=" << std::dec << std::setw(2) << ga_base
                << ", " << (io.write?"w":"r") << "data=" << std::hex << std::setw(8);
        if (io.write) stream << io.wdata;
        else if (st.bit.dtack) stream << io.rdata;
        else stream << "xxxxxxxx";
        stream << "h, "<< st << " (" << st.all << ")";
        return stream;
    }
}

void hwif_vme_mlink::driverVersionCheck()
{
#ifdef Q_OS_WIN
    qDebug() << "driverVersionCheck: not implemented";
#else
    unsigned long driverVersion = 0;
    unsigned long driverRevision = 0;
    mlink::ml_get_driver_version(&driverVersion);
    mlink::ml_get_driver_revision(&driverRevision);
    //  log_info("Driver version:%08lX, revision %08lX", driverVersion, driverRevision);
    if (driverVersion != 0 && driverVersion != 0x00010020) {
        log_err("Incompatible driver version:%lu.%lu, required 1.32", driverVersion>>16, driverVersion&0xFFFF);
        throw exc_hw("Incompatible driver version, required 1.32");
    }
#endif
}


hwif_vme_mlink::hwif_vme_mlink()
    : hwif_vme()
{
    //  Jtag_Impl *p = new Jtag_Impl_Fvme();
    //  setJtagImpl(p);

    spill_started_after_run_enable = false;
    fvme_run_mode = false;
    fvme_req_seq = 0;
    fvme_spwc = 0;
    fvme_ctrl_0.all = 0;
    fvme_ctrl_1.all = 0;
    isOpen = false;
#if defined (Q_OS_UNIX)
    if (mlink::ml_open("/dev/mlink0")) {
        throw exc_hwif("ml_open failed");
    }
    isOpen = true;
#endif
    driverVersionCheck();
    //  sysreset();
    //  while(true) {
    //  if (!selftest())
    //	throw exc_hwif("FVME selftest failed");
    //  }
}

hwif_vme_mlink::~hwif_vme_mlink()
{
#if defined (Q_OS_UNIX)
    mlink::ml_close();
#endif
}

bool hwif_vme_mlink::check_link_state()
{
#ifdef Q_OS_WIN
    qDebug() << "check_link_state: not implemented";
    return true;
#else
    mlink_status_t st;
    int ret = mlink::ml_get_driver_status(&st);
    if (ret!=0) {
        qDebug() << QString("read status failed: %1").arg(mlink::ml_strerror(ret));
        throw exc_hwif("read status failed");
    }
    if (st.bit.link_down)
        throw exc_hwif("Link down");
    if (st.bit.data_error)
        throw exc_hwif("Data error");
    if (st.bit.remote_error)
        throw exc_hwif("Remote error");
    if (st.bit.no_tx_clock || st.bit.no_rx_clock || st.bit.pci_error)
        throw exc_hwif("Interface failure");
//    qDebug() << "Link OK";
    return true;
#endif
}

void hwif_vme_mlink::w_a16_d16(unsigned int addr, int data)
{
    vme_maw16(VME_AM_A16, addr, data);
}

void hwif_vme_mlink::r_a16_d16(unsigned int addr, int *data)
{
    vme_mar16(VME_AM_A16, addr, data);
}

void hwif_vme_mlink::w_a16_d32(unsigned int addr, int data)
{
    vme_maw32(VME_AM_A16, addr, data);
}

void hwif_vme_mlink::r_a16_d32(unsigned int addr, int *data)
{
    vme_mar32(VME_AM_A16, addr, data);
}

void hwif_vme_mlink::w_a16_d64(unsigned int addr, qint64 data)
{
    //  int am = VME_AM_A16;
    int datalo = data & 0xFFFFFFFF;
    int datahi = data >> 32;
    w_a16_d32(addr, datalo);
    w_a16_d32(addr+4, datahi);
}

void hwif_vme_mlink::r_a16_d64(unsigned int addr, qint64 *data)
{
    //  int am = VME_AM_A16;
    int datahi = 0;
    int datalo = 0;
    r_a16_d32(addr, &datalo);
    r_a16_d32(addr+4, &datahi);
    quint64 data64 = (((quint64)datahi) << 32) | ((quint32)datalo);
    *data = data64;
}

void hwif_vme_mlink::w_r_a16_d32(unsigned int addr, int wdata, int *rdata)
{
    vme_maw32(VME_AM_A16, addr, wdata);
    vme_mar32(VME_AM_A16, addr, rdata);
}

void hwif_vme_mlink::w_a24_d32(unsigned int addr, int data)
{
    vme_maw32(VME_AM_A24, addr, data);
}

void hwif_vme_mlink::r_a24_d32(unsigned int addr, int *data)
{
    vme_mar32(VME_AM_A24, addr, data);
}

void hwif_vme_mlink::w_crcsr_d16(unsigned int addr, int data)
{
    vme_maw16(VME_AM_CRCSR_BUG, addr, data);
}

void hwif_vme_mlink::r_crcsr_d16(unsigned int addr, int *data)
{
    vme_mar16(VME_AM_CRCSR_BUG, addr, data);
}

void hwif_vme_mlink::w_blk_a24_d32(unsigned int start_addr, size_t n32, int *data)
{
    assert(data);
    VME_AM am = VME_AM_A24;
    for (size_t i = 0; i<n32;i++) {
        unsigned int addr = start_addr + i*4;
        vme_maw32(am, addr, data[i]);
    }
}

void hwif_vme_mlink::r_blk_a24_d32(unsigned int start_addr, size_t n32, int *data)
{
    assert(data);
    VME_AM am = VME_AM_A24;
    for (size_t i = 0; i<n32;i++) {
        unsigned int addr = start_addr + i*4;
        vme_mar32(am, addr, &data[i]);
    }
}

void hwif_vme_mlink::w_a16_d32(const std::vector<int> &addr, const std::vector<int> &data)
{
    assert(data.size() == addr.size());
    if (addr.empty())
        return;

    vme_maw32_blk(addr.size(), VME_AM_A16, &addr[0], &data[0]);
}

std::vector<int> hwif_vme_mlink::r_a16_d32(const std::vector<int> &addr)
{
    if (addr.empty())
        return std::vector<int>();
    std::vector<int> rdata(addr.size());
    vme_mar32_blk(addr.size(), VME_AM_A16, &addr[0], &rdata[0]);
    return rdata;
}

/**
 * Test 32-bit register available at VME address <addr>
 */
bool hwif_vme_mlink::testreg32(int addr, unsigned int loops)
{
    quint32 p = 0;
    unsigned int nerr = 0, nok = 0;
    const VME_AM am = VME_AM_A16;
    int badbits = 0;

    int save_value = 0;
    vme_mar32(am, addr, &save_value);
    double tv1 = UnixTimeStamp::sec();
    for(unsigned int i=0; i<loops; i++) {
        int t = rand();
        quint32 w = t;
        vme_maw32(am, addr, w);
        quint32 r = 0;
        int tmp_r = 0;
        vme_mar32(am, addr, &tmp_r);
        r = tmp_r;
        if (w != r) {
            nerr++;
            badbits |= w^r;
            if (r == p)
                qDebug() << QString("vme_testreg32 (%1/0x%2) W: 0x%3, R:0x%4, stuck data")
                .arg(am, 2, 16).arg(addr, 8, 16).arg(w, 8, 16).arg((quint32)r, 8, 16);
            else
                qDebug() << QString("vme_testreg32 (%1/0x%2) W: 0x%3, R:0x%4, xor 0x%5")
                        .arg(am, 2, 16).arg(addr, 8, 16).arg(w, 8, 16).arg((quint32)r, 8, 16).arg(w ^ r, 8, 16);
            break;
        } else {
            nok++;
        }
        p = r;
    }
    double tv2 = UnixTimeStamp::sec();
    vme_maw32(am, addr, save_value);
    if (nerr == 0) {
        qDebug() << QString("vme_testreg32 (%1/0x%2): Ok after %3 loops, %4 ms")
                .arg(am, 2, 16).arg(addr, 8, 16)
                .arg(nok).arg(1e3 * (tv2-tv1), 0, 'f', 3);
        return true;
    } else {
        qDebug() << QString("vme_testreg32 (%1/0x%2): error after %3 Ok loops")
            .arg(am, 2, 16).arg(addr, 8, 16)
            .arg(nok);
        qDebug() << QString("vme_testreg32 (%1/0x%2) failed, bad bits: %3")
            .arg(am, 2, 16).arg(addr, 8, 16)
            .arg(badbits, 8, 16, QChar('0'));

        return false;
    }
}

///*
// * Test 16-bit register available at VME address <addr>
// */
//bool hwif_vme_mlink::testreg16(int addr, unsigned int loops)
//{
//    int ret = 0;
//    unsigned int nerr = 0, nok = 0;
//    struct timeval tv1, tv2;
//    const VME_AM am = VME_AM_A16;

//    gettimeofday(&tv1, NULL);

//    addr |= 1; // LWORD=1 in 16-bit mode

//    //  t = 0x0000;
//    //  w = t & 0xFFFF;
//    //  if (vme_w16(w)) goto err;

//    // save
//    int save_value = 0;
//    vme_mar32(am, addr, &save_value);
//    while (1) {
//        int t = rand();
//        //    t = 0x0001;
//        int w = t & 0xFFFF;
//        vme_maw16(am, addr, w);
//        int r = 0;
//        int p = 0;
//        int pw = 0;
//        {
//            int tmpr;
//            vme_mar16(am, addr, &tmpr);
//            for (int i = 0; i < 20; i++) {
//                vme_mar16(am, addr, &r);
//                if (r != tmpr) {
//                    log_debug(1, "vme_testreg16 (%02X/0x%08X): W: 0x%04X, R#0:0x%04X, R#%d:0x%04X read error", am, addr, w, tmpr, i + 1, r);
//                    ret = -ETESTFAIL;
//                    goto err;
//                }
//            }
//        }
//        if (w != r) {
//            nerr++;
//            if ((r == p) && (nok > 0))
//                log_debug(1, "vme_testreg16 (%02X/0x%08X): W: 0x%04X, R:0x%04X,  stuck data on read", am, addr, w, r);
//            else if ((r == pw) && (nok > 0))
//                log_debug(1, "vme_testreg16 (%02X/0x%08X): W: 0x%04X, R:0x%04X,  stuck data on write", am, addr, w, r);
//            else
//                log_debug(1, "vme_testreg16 (%02X/0x%08X): W: 0x%04X, R:0x%04X,  xor 0x%04X", am, addr, w, r, w ^ r);
//            ret = -ETESTFAIL;
//            goto err;
//        } else {
//            nok++;
//        }
//        if (nok >= loops)
//            break;
//        //    t <<= 1; if (t==0) t=1;
//        p = r;
//        pw = w;
//    }

//    // restore
//    vme_maw32(am, addr, save_value);
//    gettimeofday(&tv2, NULL);

//    log_debug(3, "vme_testreg16 (%02X/0x%08X): Ok after %d loops, %.3f ms", am, addr, nok, 1e3 * tv_dt(tv1, tv2));
//    return true;
//    err:
//    if (ret == -ETESTFAIL)
//        log_debug(1, "vme_testreg16 (%02X/0x%08X): error after %d Ok loops (Data error)", am, addr, nok);
//    else
//        log_debug(1, "vme_testreg16 (%02X/0x%08X): error after %d Ok loops (%s)", am, addr, nok, ml_strerror(-ret));
//    throw exc("hwif_vme_mlink::testreg16 failed");

//    return false;
//}

/*
 * Test RAM available at VME address <addr> with size 2^<radix> longwords
 */
bool hwif_vme_mlink::testram32(int addr, unsigned int radix, unsigned int count)
{
    unsigned int nok = 0;
    struct timeval tv1, tv2;
    const VME_AM am = VME_AM_A24;

    gettimeofday(&tv1, NULL);

    if (radix > 24) { // 16 Mwords maximum
        qDebug() << QString("Invalid radix %1").arg(radix);
        return false;
    }
    size_t memsz = 1u << radix;
    std::vector<int> aa(memsz);
    std::vector<int> test(memsz);

    //  if ((ret = vme_setam (am))) goto err;

    for (size_t i = 0; i < memsz; i++)
        aa[i] = addr | (i << 2);

    for (unsigned int l = 0; l < count; l++) {
        for (size_t i = 0; i < memsz; i++) {
            int j = (rand() % memsz);
            int x = aa[j];
            aa[j] = aa[i];
            aa[i] = x;
        }                           // disorder

        // W-R-W-R quick test
        for (size_t i = 0; i < memsz; i++)
            test[i] = rand();

        vme_maw32_blk(memsz, am, &aa[0], &test[0]);

        for (size_t i = 0; i < memsz; i++) {
            vme_maw32(am, aa[i], test[i]);
            //    if (vme_ar32 (aa[i], &r)) return -1;
            int r = 0;
            vme_mar32(am, aa[i], &r);
            if (r != test[i]) {
                qDebug() << QString("RAM W-R-W-R test failed (step %1): addr=0x%2: wrote 0x%3, read 0x%4, xor 0x%5")
                .arg(i)
                .arg((quint32)aa[i], 2, 16)
                .arg((quint32)test[i], 8, 16)
                .arg((quint32)r, 8, 16)
                .arg((quint32)(test[i] ^ r), 8, 16);
                return false;
            }
        }

        // W-W-R-R test
        for (size_t i = 0; i < memsz; i++)
            test[i] = rand();

        for (size_t i = 0; i < memsz; i++) {
            vme_maw32(am, aa[i], test[i]);
        }
        int pd = 0;
        for (size_t i = 0; i < memsz; i++) {
            int r = 0;
            vme_mar32(am, aa[i], &r);
            if (r != test[i]) {
                qDebug() << QString("RAM WWW-RRR test failed (step %1): addr=0x%2: wrote 0x%3, read 0x%4, xor 0x%5")
                .arg(i)
                .arg((quint32)aa[i], 2, 16)
                .arg((quint32)test[i], 8, 16)
                .arg((quint32)r, 8, 16)
                .arg((quint32)(test[i] ^ r), 8, 16);
                //        usleep(100000);
                if ((i > 0) && (pd == test[i]))
                    qDebug() << "Stuck data";
                return false;
            }
            pd = r;
        }
        nok++;
    }

    gettimeofday(&tv2, NULL);
    double dt = tv2.tv_sec - tv2.tv_sec + 1e-6*(tv2.tv_usec - tv1.tv_usec);

    qDebug() << QString("vme_testram: %1 words Ok after %2 loops in %3 ms")
            .arg(memsz)
            .arg(nok)
            .arg(1e3 * dt, 0, 'f', 3);
    return true;
    //err:
    //  log_debug(1, "vme_testram: error after %d Ok loops", nok);
    //  throw exc("hwif_vme_mlink::testram32 failed");
    //  return false;
}

///*
// * Test RAM available at VME address <addr> with size 2^<radix> 16-bit words
// */
//bool hwif_vme_mlink::testram16(int addr, unsigned int radix, unsigned int count)
//{
//    unsigned int nok = 0;
//    VME_AM am = VME_AM_A24;

//    if (radix > 24) { // 16 Mwords maximum
//        log_debug(1, "Invalid radix %d", radix);
//        return (-EINVAL);
//    }
//    size_t memsz = 1 << radix;
//    std::vector<int> aa(memsz);
//    std::vector<int> test(memsz);

//    addr |= 1; // LWORD=1 in 16-bit mode

//    for (size_t i = 0; i < memsz; i++)
//        aa[i] = addr | (i << 1);

//    for (unsigned int l = 0; l < count; l++) {

//        for (size_t i = 0; i < memsz; i++) {
//            int j = (rand() % memsz);
//            int x = aa[j];
//            aa[j] = aa[i];
//            aa[i] = x;
//        }                           // disorder

//        // W-R-W-R quick test
//        for (size_t i = 0; i < memsz; i++)
//            test[i] = rand() & 0xFFFF;

//        for (size_t i = 0; i < memsz; i++) {
//            vme_maw16(am, aa[i], test[i]);
//            //    if (vme_ar16 (aa[i], &r)) return -1;
//            int r = 0;
//            vme_mar16(am, aa[i], &r);
//            if (r != test[i]) {
//                log_debug(1, "RAM W-R-W-R test failed (step %d): addr=0x%04X: wrote 0x%04X, read 0x%04X, xor 0x%04X",
//                          i, aa[i], test[i], r, test[i] ^ r);
//                return -ETESTFAIL;
//            }
//        }

//        // W-W-R-R test
//        for (size_t i = 0; i < memsz; i++)
//            test[i] = rand() & 0xFFFF;

//        for (size_t i = 0; i < memsz; i++) {
//            vme_maw16(am, aa[i], test[i]);
//        }
//        int pd = 0;
//        for (size_t i = 0; i < memsz; i++) {
//            int r = 0;
//            vme_mar16(am, aa[i], &r);
//            if (r != test[i]) {
//                log_debug(1, "RAM WWW-RRR test failed (step %d): addr=0x%04X: wrote 0x%04X, read 0x%04X, xor 0x%04X",
//                          i, aa[i], test[i], r, test[i] ^ r);
//                //        usleep(100000);
//                if ((i > 0) && (pd == test[i]))
//                    log_debug(1, "Stuck data");
//                return -ETESTFAIL;
//            }
//            pd = r;
//        }
//        nok++;
//    }
//    log_debug(2, "vme_testram16: %u words Ok after %d loops", memsz, nok);
//    return true;
//    //err:
//    //  log_debug(1, "vme_testram16: error after %d Ok loops", nok);
//    //  throw exc("hwif_vme_mlink::testram16 failed");
//    //  return false;
//}

void hwif_vme_mlink::set_cblt_address(unsigned int chain_address)
{
    fvme_reg_am_1_t fvme_am_1; fvme_am_1.all = 0;
//    fvme_am_1.bit.am = VME_AM_6U_2e2p;
//    fvme_am_1.bit.am = VME_AM_A64_MBLT;
    fvme_am_1.bit.am = VME_AM_A32_BLT;
    fvme_am_1.bit.read = 1;
    fvme_am_1.bit.iack_n = 1;
    fvme_am_1.bit.cblt = 1;

    fvme_req_buf_t req;
    req.push_back(FVME_REQ_WR | (FVME_REQ_REG_AM_1<<16) | fvme_am_1.all); // write am_1
    req.push_back(FVME_REQ_WR | (FVME_REQ_REG_AD_0<<16) | (chain_address & 0xFFFF));
    req.push_back(FVME_REQ_WR | (FVME_REQ_REG_AD_1<<16) | (chain_address>>16));
    qDebug() << QString("set_cblt_address: am_1=%1, chain_address=%2")
            .arg(fvme_am_1.all, 4, 16).arg(chain_address, 4, 16);
    fvme_request(req);

    // set SPILL and TRIG sources
    fvme_ctrl_0.bit.spill_src = 2;
    fvme_ctrl_0.bit.trig_src = 5;
    qDebug() << QString("set_cblt_address: spill_src=%1, trig_src=%2")
            .arg(fvme_ctrl_0.bit.spill_src, 4, 16).arg(fvme_ctrl_0.bit.trig_src, 4, 16);
    set_ctrl_0();
    /*
#ifndef TESTCNT
  if (vme_setaddr (0))
    throw exc("vme_setaddr failed");
  if (vme_setam (VME_AM_A32_BLT))
    throw exc("vme_set_am failed");
#endif
*/
}

quint16 hwif_vme_mlink::fvme_get_reg16(int reg)
{
    fvme_req_buf_t req;
    req.push_back(FVME_REQ_RD | (reg<<16)); // read bits 15:0
    const fvme_req_buf_t &ack = fvme_request(req);
    if (ack.size() != req.size()) {
        log_err("bad reply for fvme_get_reg16: size mismatch: expected %d, got %d", req.size(), ack.size());
        return 0;
    }
    return ack[0] & 0xFFFF;
}

void hwif_vme_mlink::fvme_set_reg16(int reg, quint16 data)
{

}

quint32 hwif_vme_mlink::fvme_get_reg32(int reg)
{
    fvme_req_buf_t req;
    req.push_back(FVME_REQ_RD | (reg<<16)); // read bits 15:0
    req.push_back(FVME_REQ_RD | ((reg+1)<<16)); // read bits 31:16
    const fvme_req_buf_t &ack = fvme_request(req);
    if (ack.size() != req.size()) {
        log_err("bad reply for block command: size mismatch: expected %d, got %d", req.size(), ack.size());
        return 0;
    }
    quint32 data = ((ack[1] & 0xFFFF) << 16) | (ack[0] & 0xFFFF);
    return data;
}

void hwif_vme_mlink::fvme_set_reg32(int reg, quint32 data)
{
    fvme_req_buf_t req;
    req.push_back(FVME_REQ_WR | (reg<<16) | (data & 0xFFFF)); // write bits 15:0
    req.push_back(FVME_REQ_WR | (reg+1<<16) | ((data >> 16) & 0xFFFF)); // write bits 31:16
    fvme_request(req);
}

quint64 hwif_vme_mlink::fvme_get_reg64(int reg)
{
    fvme_req_buf_t req;
    req.push_back(FVME_REQ_RD | (reg<<16)); // read bits 15:0
    req.push_back(FVME_REQ_RD | ((reg+1)<<16)); // read bits 31:16
    req.push_back(FVME_REQ_RD | ((reg+2)<<16)); // read bits 31:16
    req.push_back(FVME_REQ_RD | ((reg+3)<<16)); // read bits 31:16
    const fvme_req_buf_t &ack = fvme_request(req);
    if (ack.size() != req.size()) {
        log_err("bad reply for block command: size mismatch: expected %d, got %d", req.size(), ack.size());
        return 0;
    }
    quint32 datalo = ((ack[1] & 0xFFFF) << 16) | (ack[0] & 0xFFFF);
    quint32 datahi = ((ack[3] & 0xFFFF) << 16) | (ack[2] & 0xFFFF);
    quint64 data = ((quint64)(datahi) << 32) | ((quint64)(datalo));
    return data;
}

void hwif_vme_mlink::fvme_set_reg64(int reg, quint64 data)
{

}

bool hwif_vme_mlink::fvme_get_device_info()
{
    quint16 device_id = fvme_get_reg16(REG16_DEVICE_ID);
    if (device_id != FVME_DEVICE_ID) {
        qCritical() << QString("Device ID not supported: %1").arg(device_id, 4, 16, QChar('0'));
        return false;
    }
    quint64 serial_id_code = fvme_get_reg64(REG64_SERIAL_ID);
    quint32 serial_id = serial_id_code >> 8;
    quint16 temp = fvme_get_reg16(REG16_TEMPERATURE);
    quint16 fw_ver = fvme_get_reg16(REG16_FW_VER);
    quint16 fw_rev = fvme_get_reg16(REG16_FW_REV);
    QString fw_str = QString("FVME %1.%2 rev. %3, s/n %4")
            .arg((fw_ver>>8) & 0xFF)
            .arg((fw_ver) & 0xFF)
            .arg(fw_rev)
            .arg(serial_id, 8, 16, QChar('0'))
            ;
    qWarning() << fw_str;
    return true;
}

QVector<quint32> hwif_vme_mlink::fvme_get_stats()
{
    QVector<quint32> stats;
    fvme_req_buf_t req;
    for(int i=0; i<FVME_STATS_NUM; i++) {
        int reg = REG32_STATS_BASE + 2 * i;
        req.push_back(FVME_REQ_RD | (reg<<16)); // read bits 15:0
        req.push_back(FVME_REQ_RD | ((reg+1)<<16)); // read bits 31:16
    }
    const fvme_req_buf_t &ack = fvme_request(req);
    for(int i=0; i<ack.size() / 2; i++) {
        quint32 datalo = (ack[2*i+0] & 0xFFFF);
        quint32 datahi = (ack[2*i+1] & 0xFFFF);
        stats.push_back((datahi << 16) | datalo);
    }
    return stats;
}

void hwif_vme_mlink::sysreset()
{
    qDebug() << "System Reset...";
#ifdef Q_OS_WIN
    qDebug() << "sysreset: not implemented";
    return;
#else
    if (mlink::ml_remote_reset()) {
        throw exc_hwif("ml_remote_reset failed");
    }
#endif
    softclear();
    struct timeval tv;
    gettimeofday(&tv, NULL);
    fvme_req_seq = 0;
    while (true) {
        struct timeval tv2;
        gettimeofday(&tv2, NULL);
        double dt = tv2.tv_sec - tv.tv_sec + 1e-6*(tv2.tv_usec - tv.tv_usec);
        if (dt > getSysresetPostDelay()) break;
        qApp->processEvents();
    }
    qDebug() << "System Reset Ok";
}

void hwif_vme_mlink::start_communication()
{
    // reset driver
#ifdef Q_OS_WIN
    qDebug() << "start_communication: not implemented";
    return;
#else
    if (mlink::ml_reset()) {
        qCritical() << QString("ml_reset failed");
        throw exc_hwif("ml_reset failed");
    }
#endif
    fvme_get_device_info();
    fvme_run_mode = false;
    fvme_ctrl_0.all = 0;
    fvme_ctrl_1.all = 0;
    set_ctrl_0();
    set_ctrl_1();
    softclear();
    if (!selftest()) {
        qCritical() << "VME controller selftest failed";
        throw exc_hw("VME controller selftest failed");
    }
    sysreset();
}

void hwif_vme_mlink::reset()
{
    set_run_mode(0);
    softclear();
}

void hwif_vme_mlink::post_start_run()
{
    set_cblt_address(VME_CHAIN_ADDRESS);
    set_run_mode(1);
}

void hwif_vme_mlink::pre_stop_run()
{
    set_run_mode(0);
    //  drain_data();
    check_readout_chain_status();
    selftest();
}



/* VME bus operation */
void hwif_vme_mlink::vme_maw16(VME_AM am, int addr, int data)
{
    vme_maw32(am, addr | 1, data);
}

fvme_req_buf_t hwif_vme_mlink::fvme_request(const fvme_req_buf_t &req)
{
#ifdef Q_OS_WIN
    qDebug() << "fvme_request: not implemented";
    return fvme_req_buf_t();
#else
    for (int i=0; i<3; i++) {
        try {
            return fvme_request_impl(req);
        }
        catch (exc_hwif &e) {
            qWarning() << "Resetting interface";
            mlink::ml_reset();
            qWarning() << "Retrying VME request";
        }
    }
    // retry ugly and propagate exception if any
    return fvme_request_impl(req);
#endif
}

fvme_req_buf_t hwif_vme_mlink::fvme_request_impl(const fvme_req_buf_t &req)
{
#ifdef Q_OS_WIN
    qDebug() << "fvme_request_impl: not implemented";
    return fvme_req_buf_t();
#else
    fvme_req_buf_t ack;
    if (req.empty()) {
        return ack;
    }
    int timeout = 1000;

    const int L = 4; // log_debug level

    unsigned int vme_am = 0;
    unsigned int vme_addr = 0;
    unsigned int vme_wdata = 0;
    unsigned int vme_write = 0;
    unsigned int vme_iack = 0;
    unsigned int vme_cblt = 0;

    if (debug_level_ge(L)) {
        for (size_t i = 0; i<req.size(); i++) {
            unsigned int reg_addr = (req[i]>>16) & 0x3FFF;
            unsigned int reg_data = req[i] & 0xFFFF;
            int write = (req[i] & 0x80000000)?0:1;
            int start = req[i] & 0x40000000;

            switch (reg_addr) {
            case 2:
            case 3:
                vme_am = reg_data&0x3F;
                vme_write = reg_data&0x2000;
                vme_iack = reg_data&0x4000;
                vme_cblt = reg_data&0x8000;
                break;
            case 4:
                vme_addr &= ~0x0000FFFF;
                vme_addr |= reg_data;
                break;
            case 5:
                vme_addr &= ~0xFFFF0000;
                vme_addr |= reg_data;
                break;
            case 6:
                vme_wdata &= ~0x0000FFFF;
                vme_wdata |= reg_data;
                break;
            case 7:
                vme_wdata &= ~0xFFFF0000;
                vme_wdata |= reg_data;
                break;
            }


            if (start) {
                if (debug_level_ge(L))
                    qDebug() << QString("[%1] *** VME %2 AM %3, ADDR %4 %5%6")
                    .arg(i, 3).arg(vme_write?"write":"read").arg(vme_am, 2, 16)
                    .arg(vme_addr, 8, 16)
                    .arg(vme_iack?"IACK ":"").arg(vme_cblt?"CBLT ":"");
            } else {
                QString strdata;
                if (write)
                    strdata = QString("=%1").arg(reg_data, 4, 16, QChar('0'));
                else
                    strdata = " read";
                const char *reg_name = fvme_reg_name(reg_addr);
                QString txt;
                switch (reg_addr) {
                case 0:
                case 1:
                    txt = QString("%1 SPILL=%2, TRIG=%3").arg(reg_addr==1?"RUN_":"PROG_")
                          .arg(spill_src_str[reg_data&0x7])
                          .arg(trig_src_str[(reg_data>>3)&0x7]);
                    break;
                case 2:
                case 3:
                    txt = QString("%1VME: %2%3%4").arg(reg_addr==3?"RUN_":"PROG_")
                          .arg(vme_cblt?"CBLT ":"").arg(vme_iack?"IACK ":"")
                          .arg(vme_write?"WRITE":"READ");
                    break;
                }
                if (debug_level_ge(L))
                    qDebug() << QString("[%1] %2%3 %4").arg(i, 3).arg(reg_name).arg(strdata).arg(txt);
            }
        }
    }


    int ret = 0;
    if ((ret = mlink::ml_send_data_frame(static_cast<const int *>(static_cast<const void *>(&req[0])),
                                         req.size(), FRAME_TYPE_FVME_REQ, 1, 2, fvme_req_seq++, timeout))) {
        qDebug() << QString("vme_maw32: ml_send_data_frame returned %1 (%2)").arg(ret).arg(mlink::ml_strerror(-ret));
        check_link_state();
        throw exc_hwif("fvme_request: ml_send_data_frame failed");
    }
    int exprxlen = 0;
    for (size_t i=0; i<req.size(); i++) if (req[i]&0x80000000) exprxlen++;
    //  if (exprxlen==0) return; // WTF???

    //  log_debug(0, "fvme_request: trying to receive frame");

    int rawbuf[ML_RX_BUFFER_WORDS];
    memset(rawbuf, 0, sizeof(rawbuf));
    bool eoj = false;
    while (!eoj) {
        int rawbytes = sizeof(rawbuf);
        struct mlink::frameset_t frameset;
        mlink::ml_recv_data_frame_multi(rawbuf, rawbytes, &frameset, timeout);
        if (frameset.count == 0) {
            check_link_state();
            qDebug() << "Req: " << req;
            throw exc_hwif("FVME request timeout");
            eoj = true;
        }
        for (size_t fn = 0; fn<frameset.count; fn++) {
            const mlink::framedesc_t *fdesc = &frameset.fdesc[fn];
            const size_t datawords = fdesc->len - (ML_FRAME_HEADER_WORDS+ML_FRAME_TRAILER_WORDS);
            const int *frame_data = &fdesc->rawdata[ML_FRAME_HEADER_WORDS];
            const frame_hdr_t *fh = (const frame_hdr_t *) fdesc->rawdata;
            const int type = fh->type;
            if (type == FRAME_TYPE_FVME_ACK) {
                if (datawords > 0) {
                    ack.resize(datawords);
                    memcpy(&ack[0], frame_data, datawords*4);
                }
                eoj = true;
            } else {
                qDebug() << QString("Dropped frame type 0x%1").arg(type, 4, 16, QChar('0'));
            }
        }
    }
    return ack;
#endif
}

size_t hwif_vme_mlink::max_blk_io_size() const {
    return ML_MAX_FRAME_DATA_WORDS / max_io_req_size;
}

int hwif_vme_mlink::blk_io(vme_block_io_t &multi_io)
{
    if (multi_io.empty()) return 0; // nothing to do
    const size_t count = multi_io.size();

    // build commands
    fvme_req_buf_t req;
    // poison cache
    int p_am = ~multi_io[0].am;
    unsigned int p_addr_hi = ~((multi_io[0].addr>>16)&0xFFFF);
    unsigned int p_addr_lo = ~((multi_io[0].addr)&0xFFFF);
    unsigned int p_wdata_hi = ~((multi_io[0].wdata>>16)&0xFFFF);
    unsigned int p_wdata_lo = ~((multi_io[0].wdata)&0xFFFF);
    //  unsigned int p_addr = ~multi_io[0].addr;
    //  unsigned int p_wdata = ~multi_io[0].wdata;
    bool p_write = !multi_io[0].write;
    for (size_t i = 0; i<count; i++) {
        if (req.size() + max_io_req_size > ML_MAX_FRAME_DATA_WORDS) {
            qWarning() << QString("block size too large (does not fit in one frame), requested %1, available %2")
                    .arg(count).arg(max_blk_io_size());
            return(-1);
        }
        const vme_io_t &io = multi_io[i];
        const VME_AM am = io.am;
        unsigned int addr_hi = (io.addr>>16)&0xFFFF;
        unsigned int addr_lo = (io.addr)&0xFFFF;
        unsigned int wdata_hi = (io.wdata>>16)&0xFFFF;
        unsigned int wdata_lo = (io.wdata)&0xFFFF;
        //    unsigned int addr = io.addr;
        //    unsigned int wdata = io.wdata;
        bool write = io.write;
        if (p_am != am || p_write != write) {
            req.push_back(FVME_REQ_WR | (FVME_REQ_REG_AM_0<<16) | (am&0x3F) | FVME_REQ_REG_AM_BIT_IACK_N | (write?FVME_REQ_REG_AM_BIT_WRITE:FVME_REQ_REG_AM_BIT_READ));
            p_am = am;
            p_write = write;
        }
        if (p_addr_hi != addr_hi) {
            req.push_back(FVME_REQ_WR | (FVME_REQ_REG_AD_1<<16) | addr_hi); // write VME address 16-31
            p_addr_hi = addr_hi;
        }
        if (p_addr_lo != addr_lo) {
            req.push_back(FVME_REQ_WR | (FVME_REQ_REG_AD_0<<16) | addr_lo); // write VME address 0-15
            p_addr_lo = addr_lo;
        }
        if (write && p_wdata_hi != wdata_hi) {
            req.push_back(FVME_REQ_WR | (FVME_REQ_REG_WR_1<<16) | wdata_hi); // VME data 16-31
            p_wdata_hi = wdata_hi;
        }
        if (write && p_wdata_lo != wdata_lo) {
            req.push_back(FVME_REQ_WR | (FVME_REQ_REG_WR_0<<16) | wdata_lo); // VME data 0-15
            p_wdata_lo = wdata_lo;
        }
        req.push_back(FVME_REQ_WR | 0x40000000); // start bus operation
        if (!write) {
            req.push_back(FVME_REQ_RD | (FVME_REQ_REG_RD_0<<16)); // VME data 0-15
            req.push_back(FVME_REQ_RD | (FVME_REQ_REG_RD_1<<16)); // VME data 16-31
        }
        req.push_back(FVME_REQ_RD | (FVME_REQ_REG_ST_0<<16)); // VME status
    }

    // count required ack size
    size_t expect_ack_size = 0;
    for (size_t i = 0;i<count; i++) {
        const vme_io_t &io = multi_io[i];
        expect_ack_size++; // status
        if (!io.write)
            expect_ack_size+=2; // read data
    }
    const fvme_req_buf_t &ack = fvme_request(req);
    if (ack.size() != expect_ack_size) {
        log_err("bad reply for block command: size mismatch: expected %d, got %d", expect_ack_size, ack.size());
        //	return(-1);
    }
    // analyze reply
    const unsigned int cmd_mask = 0xFFFF0000;
    const unsigned int cmd_rdata_0 = FVME_REQ_RD | (FVME_REQ_REG_RD_0<<16);
    const unsigned int cmd_rdata_1 = FVME_REQ_RD | (FVME_REQ_REG_RD_1<<16);
    const unsigned int cmd_status  = FVME_REQ_RD | (FVME_REQ_REG_ST_0<<16);
    size_t ack_p = 0;
    for (size_t i = 0; i<count; i++) {
        vme_io_t *io = &multi_io[i];
        bool write = io->write;
        if (ack_p >= ack.size()) {
            log_err("ack is truncated, terminating reply analysis, command is not acknowledged");
            return -1;
        }
        //	unsigned int ack = ack[ack_p];
        size_t status_pos = write?ack_p:(ack_p+2);
        bool cmd_ok = true;
        if (!write) {
            if ((ack[ack_p] & cmd_mask) != cmd_rdata_0) {
                log_err("bad reply for block command: address mismatch: expected %08X, got %08X", cmd_rdata_0, ack[ack_p]);
                cmd_ok = false;
            }
            if ((ack[ack_p+1] & cmd_mask) != cmd_rdata_1) {
                log_err("bad reply for block command: address mismatch: expected %08X, got %08X", cmd_rdata_1, ack[ack_p+1]);
                cmd_ok = false;
            }
        }
        if ((ack[status_pos] & cmd_mask) != cmd_status) {
            log_err("bad reply for block command: address mismatch: expected %08X, got %08X", cmd_status, ack[status_pos]);
            cmd_ok = false;
        }
        if (!cmd_ok) {
            //	  return(-1);
        }
        // read data, status
        if (!write) {
            io->rdata = (ack[ack_p] & 0xFFFF) | ((ack[ack_p+1] & 0xFFFF)<<16);
        }
        io->st = ack[status_pos];
        // move pointer
        if (!write)
            ack_p += 2;
        ack_p++;
    }
    return 0;
}

void hwif_vme_mlink::vme_maw32_blk(size_t count, VME_AM am, const int *addr, const int *data)
{
    if (count == 0)
        return;
    std::vector<vme_io_t> io(count);
    for (size_t i = 0; i<count; i++) {
        io[i].am = am;
        io[i].addr = addr[i];
        io[i].write = true;
        io[i].wdata = data[i];
    }
    if (blk_io(io))
        throw exc("vme_blk_io failed in vme_maw32_blk");

    //  bool ok = true;
    for (size_t i = 0; i<count; i++) {
        fvme_reg_status_0_t st;
        st.all = io[i].st;
        bool tmout = st.bit.tmout;
        bool dtack = st.bit.dtack;
        bool berr = st.bit.berr;
        bool this_ok = !tmout && dtack && !berr;
        //    if (!this_ok) ok = false;
        //    char str1[1024];
        //        snprintf_vme_status(str1, 1023, st);
        //    log_debug(this_ok?3:1, str1);

#ifdef DEBUG_VME
        if (this_ok) {
            std::ostringstream os;
            os << io;
            std::string str = os.str();
            qDebug() << QString::fromStdString(str);
        }
#endif
        if (!this_ok) {
            std::ostringstream os;
            os << io;
            std::string str = os.str();
            qWarning() << QString::fromStdString(str);

            if (tmout)
                throw exc_vme_tmout(str);
            if (!dtack)
                throw exc_vme_nack(str);
            if (berr)
                throw exc_vme_berr(str);
        }
    }
}

void hwif_vme_mlink::vme_mar32_blk(size_t count, VME_AM am, const int *addr, int *data)
{
    if (count == 0)
        return;
    std::vector<vme_io_t> io(count);
    for (size_t i = 0; i<count; i++) {
        io[i].am = am;
        io[i].addr = addr[i];
        io[i].write = false;
    }
    if (blk_io(io))
        throw exc("vme_blk_io failed in vme_maw32_blk");

    //  bool ok = true;
    for (size_t i = 0; i<count; i++) {
        fvme_reg_status_0_t st;
        st.all = io[i].st;
        bool tmout = st.bit.tmout;
        bool dtack = st.bit.dtack;
        bool berr = st.bit.berr;
        bool this_ok = !tmout && dtack && !berr;
        //    if (!this_ok) ok = false;
        //    char str1[1024];
        //        snprintf_vme_status(str1, 1023, st);
        //    log_debug(this_ok?3:1, "%s", str1);
        data[i] = io[i].rdata;

#ifdef DEBUG_VME
        if (this_ok) {
            std::ostringstream os;
            os << io;
            std::string str = os.str();
            qDebug() << QString::fromStdString(str);
        }
#endif
        if (!this_ok) {
            std::ostringstream os;
            os << io;
            std::string str = os.str();
            qWarning() << QString::fromStdString(str);

            if (tmout)
                throw exc_vme_tmout(str);
            if (!dtack)
                throw exc_vme_nack(str);
            if (berr)
                throw exc_vme_berr(str);
        }
    }
}

void hwif_vme_mlink::vme_maw32(VME_AM am, int addr, int data)
{
    vme_block_io_t io(1);

    io[0].am = am;
    io[0].addr = addr;
    io[0].write = true;
    io[0].wdata = data;
    io[0].rdata = 0;
    io[0].st = 0;

    if (blk_io(io))
        throw exc("vme_blk_io failed in vme_maw32");

    fvme_reg_status_0_t st;
    st.all = io[0].st;
    bool tmout = st.bit.tmout;
    bool dtack = st.bit.dtack;
    bool berr = st.bit.berr;
    bool this_ok = !tmout && dtack && !berr;

    //  char str1[1024];
    //  snprintf_vme_status(str1, 1023, st);

    //  log_debug(ok?3:1, "%s", str1);

#ifdef DEBUG_VME
    if (this_ok) {
        std::ostringstream os;
        os << io;
        std::string str = os.str();
        qDebug() << QString::fromStdString(str);
    }
#endif
    if (!this_ok) {
        std::ostringstream os;
        os << io;
        std::string str = os.str();
        qWarning() << QString::fromStdString(str);

        if (tmout)
            throw exc_vme_tmout(str);
        if (!dtack)
            throw exc_vme_nack(str);
        if (berr)
            throw exc_vme_berr(str);
    }
}

void hwif_vme_mlink::vme_mar16(VME_AM am, int addr, int *data)
{
    vme_mar32(am, addr | 1, data);
    if (data) *data &= 0xFFFF;
}

void hwif_vme_mlink::vme_mar32(VME_AM am, int addr, int *data)
{
    vme_block_io_t io(1);

    io[0].am = am;
    io[0].addr = addr;
    io[0].write = false;
    io[0].wdata = 0;
    io[0].rdata = 0;
    io[0].st = 0;

    if (blk_io(io))
        throw exc("vme_blk_io failed in vme_maw32");

    if (data) *data = io[0].rdata;
    fvme_reg_status_0_t st;
    st.all = io[0].st;
    bool tmout = st.bit.tmout;
    bool dtack = st.bit.dtack;
    bool berr = st.bit.berr;
    bool this_ok = !tmout && dtack && !berr;

    //  char str1[1024];
    //  snprintf_vme_status(str1, 1023, st);

    //  log_debug(ok?3:1, "%s", str1);

#ifdef DEBUG_VME
    if (this_ok) {
        std::ostringstream os;
        os << io;
        std::string str = os.str();
        qDebug() << QString::fromStdString(str);
    }
#endif
    if (!this_ok) {
        std::ostringstream os;
        os << io;
        std::string str = os.str();
        qWarning() << QString::fromStdString(str);

        if (tmout)
            throw exc_vme_tmout(str);
        if (!dtack)
            throw exc_vme_nack(str);
        if (berr)
            throw exc_vme_berr(str);
    }
}

void hwif_vme_mlink::set_ctrl_0()
{
    fvme_req_buf_t req(1, FVME_REQ_WR | (FVME_REQ_REG_CTRL_0<<16) | fvme_ctrl_0.all); // write ctrl_0
    fvme_request(req);
}

void hwif_vme_mlink::set_ctrl_1()
{
    fvme_req_buf_t req(1, FVME_REQ_WR | (FVME_REQ_REG_CTRL_1<<16) | fvme_ctrl_1.all); // write ctrl_1
    fvme_request(req);
}

void hwif_vme_mlink::softclear()
{
    spill_started_after_run_enable = false;
    fvme_ctrl_0.bit.softclear = 1;
    try {
        set_ctrl_0();
    }
    catch (exc_hwif &) {
    }
    //  drain_data();
    clear_events();
    fvme_ctrl_0.bit.softclear = 0;
    try {
        set_ctrl_0();
    }
    catch (exc_hwif &) {
    }
}

void hwif_vme_mlink::stat_clr()
{
    fvme_ctrl_0.bit.stat_clr = 1;
    set_ctrl_0();
    fvme_ctrl_0.bit.stat_clr = 0;
    set_ctrl_0();
}

void hwif_vme_mlink::start_of_burst_signal()
{
    spill_started_after_run_enable = true;
}

void hwif_vme_mlink::end_of_burst_signal()
{
    const QVector<quint32> &stats = fvme_get_stats();
    stat_clr();
    for(int i=0; i<stats.size(); i++) {
        qDebug() << QString("stats[%1]: %2").arg(i).arg(stats[i]);
    }
}

void hwif_vme_mlink::set_run_mode(bool new_run_mode)
{
    if (fvme_run_mode == new_run_mode) return;
    if (new_run_mode) {
        fvme_ctrl_1.bit.run_mode = 1;
        fvme_spwc = 0;
    } else {
        fvme_ctrl_1.bit.run_mode = 0;
    }

//    qDebug() << QString("%1_RUN_MODE request").arg(fvme_ctrl_1.bit.run_mode?"START":"STOP");
    set_ctrl_1();

    fvme_run_mode = new_run_mode;
//    qDebug() << QString("fvme run mode is %1").arg(fvme_run_mode);
}

bool hwif_vme_mlink::check_readout_chain_status()
{
    bool status = false;
    try {
        fvme_reg_status_0_t cblt_st;
        cblt_st.all = 0;
        const fvme_req_buf_t req(1, FVME_REQ_RD | (FVME_REQ_REG_ST_0<<16)); // VME status
//        qDebug() << "READ_STATUS request";
        const fvme_req_buf_t &ack = fvme_request(req);
        if ((ack.size() != 1) || (((ack[0]>>16)& 0x3FFF) != FVME_REQ_REG_ST_0)) {
            qWarning() << "failed to read CBLT status";
        } else {
            cblt_st.all = ack[0] & 0xFFFF;
        }
        // begin read spill wordcount
        {
            fvme_req_buf_t reqbuf;
            reqbuf.push_back(FVME_REQ_RD | (FVME_REQ_REG_SPWC_0<<16)); // spill wordcount lo
            reqbuf.push_back(FVME_REQ_RD | (FVME_REQ_REG_SPWC_1<<16)); // spill wordcount hi
            const fvme_req_buf_t &ackbuf = fvme_request(reqbuf);
            if (
                    (ackbuf.size() != 2) ||
                    (((ackbuf[0]>>16) & 0x3FFF) != FVME_REQ_REG_SPWC_0) ||
                    (((ackbuf[1]>>16) & 0x3FFF) != FVME_REQ_REG_SPWC_1)
                    ) {
                qWarning() << "failed to read FVME SPWC registers";
            } else {
                unsigned int spwc = (ackbuf[0]&0xFFFF) | ((ackbuf[1]&0xFFFF)<<16);
                fvme_spwc = spwc;
                qDebug() << QString("Read SPWC=%1").arg(spwc);
            }
        }
        // end read spill wordcount
        bool cblt_status = (!cblt_st.bit.tmout && cblt_st.bit.dtack);
        bool spill_any_data = fvme_spwc > 1+16; // SHDR+16*STRL // FIXME
        if (spill_any_data) {
            if (!cblt_st.bit.berr)
                cblt_status = false;

            if (!cblt_status)
                qWarning() << QString("CBLT failed:%1%2%3")
                .arg(cblt_st.bit.berr?"":" NO_BERR")
                .arg(cblt_st.bit.tmout?" TIMEOUT":"")
                .arg(cblt_st.bit.dtack?"":" NO_DTACK");
            qDebug() << QString("VME CBLT STATUS: %1%2%3  %4%5%6%7%8%9%10%11")
                    .arg((cblt_st.bit.tmout)?" TIMEOUT":"")
                    .arg((cblt_st.bit.dtack)?"":" NO_DTACK")
                    .arg((cblt_st.bit.berr)?" BERR":"")
                    .arg((cblt_st.bit.irqs & 0x01)?" IRQ0":"")
                    .arg((cblt_st.bit.irqs & 0x02)?" IRQ1":"")
                    .arg((cblt_st.bit.irqs & 0x04)?" IRQ2_SPILL":"")
                    .arg((cblt_st.bit.irqs & 0x08)?" IRQ3":"")
                    .arg((cblt_st.bit.irqs & 0x10)?" IRQ4":"")
                    .arg((cblt_st.bit.irqs & 0x20)?" IRQ5_TRIG":"")
                    .arg((cblt_st.bit.irqs & 0x40)?" IRQ6":"")
                    .arg((cblt_st.bit.irqs & 0x80)?" IRQ7":"");
            status = cblt_status;
        }
    }
    catch (exc_vme &e) {
        qWarning() << e.what();
    }
    return status;
}

bool hwif_vme_mlink::interspill_test()
{
//    qDebug() << "hwif_vme_mlink::interspill_test";
    return selftest();
}

bool hwif_vme_mlink::selftest()
{
    bool test_ok = true;

    const int nloops = FVME_SELFTEST_LOOPS;
    int l = 0;
    while (test_ok && l<nloops) {
        const int nset = 10;
        //  int buf[nset*2];
        std::vector<int> data(nset*2);
        std::vector<int> addr(nset*2);
        //  size_t s = 0;

        /*
// Two register test
  for (int i=0;i<nset;i++) {
        data[i] = rand() % 0xFFFF;
        addr[i] = (i%2)?FVME_REQ_REG_TEST_1:FVME_REQ_REG_TEST_0;
  }
  for (int i=0;i<nset;i+=2) {
        buf[s++] = FVME_REQ_WR | (addr[i]<<16) | data[i]; // test resister 0
        buf[s++] = FVME_REQ_WR | (addr[i+1]<<16) | data[i+1]; // test resister 1
        buf[s++] = FVME_REQ_RD | (addr[i]<<16); // test resister 0
        buf[s++] = FVME_REQ_RD | (addr[i+1]<<16); // test resister 1
  }
*/
        // One register test
        for (int i = 0; i<nset; i++) {
            data[i] = rand() % 0xFFFF;
            addr[i] = FVME_REQ_REG_TEST_0;
        }

        fvme_req_buf_t buf;
        for (int i = 0; i<nset; i++) {
            buf.push_back(FVME_REQ_WR | (addr[i]<<16) | data[i]); // test resister 0
            buf.push_back(FVME_REQ_RD | (addr[i]<<16)); // test resister 0
        }
        const fvme_req_buf_t &ack = fvme_request(buf);

        // check answer
        for (size_t i = 0; i<ack.size(); i++) {
            quint32 word = ack[i];
            quint32 expect = FVME_REQ_RD | (addr[i]<<16) | data[i];
            if (word != expect) {
                qWarning() << QString("fvme selftest failed at step %1: got %2, expected %3")
                        .arg(i)
                        .arg(word, 8, 16)
                        .arg(expect, 8, 16);
                test_ok = false;
                break;
            }
        }
        l++;
    }
//    if (test_ok) qDebug() << "fvme selftest Ok";
    return test_ok;
}

void hwif_vme_mlink::get_events(hwev_queue_t *q, int timeout_ms)
{
#ifdef Q_OS_WIN
    qDebug() << "get_events: not implemented";
    return;
#else
    if (!isOpen) return;
    mlink::mlink_event_t ml_ev;

    int ret = 0;
    if ((ret = mlink::ml_get_event(&ml_ev, timeout_ms))) {
        qWarning() << QString("ml_get_event failed: %1").arg(strerror(-ret));
        return;
    }
    //  log_debug(0, "ml_get_event: mask %08X", ml_ev);
    //  if (ml_ev & (ML_EV_BIT_SPILL_ON | ML_EV_BIT_SPILL_OFF))
    //	ml_ev &= ~ML_EV_BIT_SPILL_ON;

    if (ml_ev & ML_EV_BIT_SPILL_ON)
        q->push(hwifEvent(1, hwifEvent::SPILL_ON));
    if (ml_ev & ML_EV_BIT_SPILL_OFF)
        q->push(hwifEvent(1, hwifEvent::SPILL_OFF));

    if (ml_ev & ML_EV_BIT_DATA)
        q->push(hwifEvent(1, hwifEvent::DATA));

    if (ml_ev & ML_EV_BIT_RX_PKT)
        q->push(hwifEvent(1, hwifEvent::RXPKT));
#endif
}

void hwif_vme_mlink::clear_events()
{
#ifdef Q_OS_WIN
    qDebug() << "clear_events: not implemented";
    return;
#else
    mlink::mlink_event_t ml_ev;

    int ret = 0;
    if ((ret = mlink::ml_get_event(&ml_ev, 0))) {
        qWarning() << QString("ml_get_event failed: %1").arg(strerror(-ret));
        return;
    }
#endif
}
