
#include <QDebug>

#include "vmedaq_jtag_impl_fvme.h"
#include "hal/base_hwif.h"
#include "daqcommon/vme_types.h"
#include "colorlog/log.h"

#define PREFIX "jtag: "

using namespace colorlog;

Jtag_Impl_Options_Vme::Jtag_Impl_Options_Vme(base_hwif *p, unsigned int addr, unsigned int addr_b32)
    : hwif(p),
    io_addr(addr),
    io_b32_addr(addr_b32),
    enable_32bit_jtag((io_b32_addr != 0) && (io_b32_addr != io_addr))
{
}

Jtag_Impl_Fvme::Jtag_Impl_Fvme(const Jtag_Impl_Options_Vme &p)
    : Jtag_Impl(), options(p)
{
    clearStats();
}

Jtag_Impl_Fvme::~Jtag_Impl_Fvme()
{
    printStats();
}

base_hwif *Jtag_Impl_Fvme::getHwif()
{
    Q_ASSERT(options.hwif);
    return options.hwif;
}

int Jtag_Impl_Fvme::jtag_io (int tms, int tdi)
{
    if (tms)
        tms = 1;
    if (tdi)
        tdi = 1;

    int w = tdi | (tms << 1);

    vme_block_io_t io(2);
    io[0].am = VME_AM_A16;
    io[0].addr = get_io_addr();
    io[0].write = 1;
    io[0].wdata = w;
    io[1].am = io[0].am;
    io[1].addr = io[0].addr;
    io[1].write = 0;
    getHwif()->blk_io(io);
    int r = io[1].rdata;
    int tdo = r & 1;

    log_debug(JTAG_LOG_LEVEL, "[%08X]: single TMS %d TDI %d TDO %d", get_io_addr(), tms, tdi, tdo);
    return tdo;
}

#ifdef USE_JTAG32
int Jtag_Impl_Fvme::jtag_io_b32 (int tdi)
{
    if (!options.enable_32bit_jtag)
        return Jtag_Impl::jtag_io_b32(tdi);

    vme_block_io_t io(2);
    io[0].am = VME_AM_A16;
    io[0].addr = get_io_b32_addr();
    io[0].write = 1;
    io[0].wdata = tdi;
    io[1].am = io[0].am;
    io[1].addr = io[0].addr;
    io[1].write = 0;
    getHwif()->blk_io(io);
    int tdo = io[1].rdata;

    log_debug(JTAG_LOG_LEVEL, "[%08X]: single 32  TDI %08X TDO %08X", get_io_b32_addr(), tdi, tdo);
    return tdo;
}

void Jtag_Impl_Fvme::jtag_io_b32_blk (unsigned int count, const unsigned int *tdi32Data, unsigned int *tdo32Data)
{
    if (!options.enable_32bit_jtag) {
        Jtag_Impl::jtag_io_b32_blk(count, tdi32Data, tdo32Data);
        return;
    }

    size_t off = 0;
    size_t blocksize = getHwif()->max_blk_io_size() / 2; // write and read
    vme_block_io_t iobuf;
    while (off < count) {
        iobuf.clear();
        size_t sz = (count-off);
        if (sz > blocksize) sz = blocksize;
        for (size_t n=0; n<sz; n++) {
            vme_io_t io;
            io.am = VME_AM_A16;
            io.addr = get_io_b32_addr();
            io.write = 1;
            io.wdata = tdi32Data?tdi32Data[off+n]:0;
            iobuf.push_back(io);
            io.write = 0;
            iobuf.push_back(io);
        }
        getHwif()->blk_io(iobuf);
        if (tdo32Data) {
            for (unsigned int n=0;n<sz;n++) {
                tdo32Data[off+n] = iobuf[n*2+1].rdata;
            }
        }

        for (unsigned int n=0;n<sz;n++) {
            log_debug(JTAG_LOG_LEVEL, "[%08X]: block 32 [% 4d]  TDI %08X TDO %08X", get_io_b32_addr(), n, iobuf[n*2].wdata, iobuf[n*2+1].rdata);
        }
        off += sz;
    }
}
#endif

void Jtag_Impl_Fvme::jtag_io_b (size_t count, const bool *btms, const bool *btdi, bool *btdo)
{
    size_t off = 0;
    base_hwif * const hwif = getHwif();
    size_t blocksize = hwif->max_blk_io_size() / 2; // write and read
    vme_block_io_t iobuf;
    while (off < count) {
        iobuf.clear();
        size_t sz = (count-off);
        if (sz > blocksize) sz = blocksize;
        for (size_t n=0; n<sz; n++) {
            vme_io_t io;
            io.am = VME_AM_A16;
            io.addr = get_io_addr();
            io.write = 1;
            io.wdata = 0;
            if (btms && btms[off+n])
                io.wdata |= 2;
            if (btdi && btdi[off+n])
                io.wdata |= 1;
            iobuf.push_back(io);
            io.write = 0;
            iobuf.push_back(io);
        }
        getHwif()->blk_io(iobuf);
        if (btdo) {
            for (unsigned int n=0;n<sz;n++) {
                btdo[off+n] = iobuf[n*2+1].rdata&1;
            }
        }
        for (unsigned int n=0;n<sz;n++) {
            log_debug(JTAG_LOG_LEVEL, "[%08X]: block  1 [% 4d]  TMS %d TDI %d TDO %d", get_io_b32_addr(), n, (iobuf[n*2].wdata&2)!=0, iobuf[n*2].wdata&1, iobuf[n*2+1].rdata&1);
        }
        off += sz;
    }
}

void Jtag_Impl_Fvme::clearStats()
{
    count_pkt = 0;
    count_io = 0;
    count_io_b = 0;
    count_io_b32 = 0;
}
void Jtag_Impl_Fvme::printStats() const
{
    qDebug() << "JTAG packets: " << count_pkt;
    qDebug() << "JTAG io:      " << count_io;
    qDebug() << "JTAG io_b:    " << count_io_b;
    qDebug() << "JTAG io_b32:  " << count_io_b32;
}
