
#include "base_bmod.h"

#include <QDebug>
#include <QString>
#include <QMutex>
#include <QWaitCondition>

#include <stdexcept>
#include <vector>

#include "hal/base_hwif.h"
#include "daqcommon/daq_ipc.h"
#include "daqcommon/hal_types.h"

#include "colorlog/log.h"

using colorlog::debug_level_ge;

base_bmod_hardware::base_bmod_hardware(const bmod_props &props)
    : hwif(props.hwif), reg_base(props.reg_base)
{
    Q_ASSERT(hwif);
}

base_bmod::base_bmod(const bmod_props &props)
    : base_bmod_hardware(props),
    //    shm_status(0),
    shm_pio_rawdata(props.shm_pio_rawdata),
    config_rom_cache_valid(false),
    messages_enabled(true)
{
    status_reg.all = MODULE_STATUS_DEFAULT;
    reg_ctrl.all = MODULE_CTRL_DEFAULT;
}

void base_bmod_hardware::w_a24_d32(unsigned int addr, int data)
{  hwif->w_a24_d32(addr, data);}

void base_bmod_hardware::r_a24_d32(unsigned int addr, int *data)
{  hwif->r_a24_d32(addr, data);}

void base_bmod_hardware::w_a16_d64(unsigned int addr, qint64 data)
{  hwif->w_a16_d64(addr, data);}

void base_bmod_hardware::r_a16_d64(unsigned int addr, qint64 *data)
{  hwif->r_a16_d64(addr, data);}

void base_bmod_hardware::w_a16_d32(unsigned int addr, int data)
{  hwif->w_a16_d32(addr, data);}

void base_bmod_hardware::r_a16_d32(unsigned int addr, int *data)
{  hwif->r_a16_d32(addr, data);}

void base_bmod_hardware::w_a16_d16(unsigned int addr, int data)
{  hwif->w_a16_d16(addr, data);}

void base_bmod_hardware::r_a16_d16(unsigned int addr, int *data)
{  hwif->r_a16_d16(addr, data);}

void base_bmod_hardware::w_blk_a24_d32(unsigned int start, size_t n32, int *data)
{  hwif->w_blk_a24_d32((get_reg_base()<<8)+start, n32, data);}

void base_bmod_hardware::r_blk_a24_d32(unsigned int start, size_t n32, int *data)
{  hwif->r_blk_a24_d32((get_reg_base()<<8)+start, n32, data);}

bool base_bmod_hardware::testreg32(int addr, unsigned int loops)
{  return hwif->testreg32(addr, loops);}

//bool base_bmod_hardware::testreg16(int addr, unsigned int loops)
//{  return hwif->testreg16(addr, loops);}

bool base_bmod_hardware::testram32(int addr, unsigned int radix, unsigned int count)
{  return hwif->testram32(addr, radix, count);}

//bool base_bmod_hardware::testram16(int addr, unsigned int radix, unsigned int count)
//{  return hwif->testram16(addr, radix, count);}

void base_bmod_hardware::set_reg(unsigned int reg, int data)
{
    w_a16_d32(reg*4 | get_reg_base(), data);
}

void base_bmod_hardware::get_reg(unsigned int reg, int *data)
{
    r_a16_d32(reg*4 | get_reg_base(), data);
}

void base_bmod_hardware::set_reg_v(const std::vector<int> &regs, const std::vector<int> &data)
{
    std::vector<int> addr;
    for (size_t i = 0; i<regs.size(); i++) {
        addr.push_back(regs[i]*4 | get_reg_base());
    }
    hwif->w_a16_d32(addr, data);
}

std::vector<int> base_bmod_hardware::get_reg_v(const std::vector<int> &regs)
{
    const size_t count = regs.size();
    std::vector<int> addr;
    for (size_t i = 0; i<count; i++) {
        addr.push_back(regs[i]*4 | get_reg_base());
    }
    return hwif->r_a16_d32(addr);
}

void base_bmod::reset()
{
    clear_error_flag();
    clear_overflow_flag();
    //  get_reg(MODULE_REG_CTRL, &ctrl);
    reg_ctrl.all = MODULE_CTRL_DEFAULT;
    reg_ctrl.bit.reset = 1;
    set_reg(MODULE_REG_CTRL, reg_ctrl.all);
    reg_ctrl.bit.reset = 0;
    set_reg(MODULE_REG_CTRL, reg_ctrl.all);
    reg_ctrl.bit.dummy_blt_data = 0; // FIXME
    set_reg(MODULE_REG_CTRL, reg_ctrl.all);
    read_module_status(messagesEnabled());
}

void base_bmod::read_configuration()
{
    qint64 data = 0;
    qint32 datalo = 0;
    qint32 datahi = 0;
    qint32 tmp = 0;
    get_reg(MODULE_REG_CTRL, &tmp);
    debug_mod(2, QString("Read CTRL = %1").arg((quint32)tmp, 8, 16, QChar('0')));
    get_reg(MODULE_REG_STATUS, &tmp);
    debug_mod(2, QString("Read STATUS = %1").arg((quint32)tmp, 8, 16, QChar('0')));
    // serial id is optional
    try {
        get_reg(MODULE_REG_SERIAL_lo, &datalo);
        debug_mod(2, QString("Read SERIAL_lo = %1").arg((quint32)datalo, 8, 16, QChar('0')));
        get_reg(MODULE_REG_SERIAL_hi, &datahi);
        debug_mod(2, QString("Read SERIAL_hi = %1").arg((quint32)datahi, 8, 16, QChar('0')));
    }
    catch (std::runtime_error &e) {
        info_mod(QString("Error reading Serial ID: %1").arg(e.what()));
    }
    data = (qint64)datahi << 32 | (qint64)datalo;
    if (data == 0 || data == -1) { // ID not present
        warn_mod("ID not present");
    } else {
        quint32 serial_id = (data >> 8) & 0xFFFFFFFF;
        set_serial(serial_id);
    }
    qint32 fw_ver = 0;
    qint32 fw_rev = 0;
    try {
        get_reg(MODULE_REG_FW_VER, &fw_ver);
        get_reg(MODULE_REG_FW_REV, &fw_rev);
    }
    catch (std::runtime_error &e) {
        info_mod(QString("Error reading Firmware ID: %1").arg(e.what()));
    }
    set_fw_ver(fw_ver);
    set_fw_rev(fw_rev);

    print_info();
}

bool base_bmod::detect()
{
    return true;
}

bool base_bmod::selftest()
{
    bool test_ok = true;

    debug_mod(1, "selftest...");

    test_ok = testreg32(get_reg_base() + MODULE_REG_TEST * 4, MOD_SELFTEST_LOOPS);
    read_module_status();

    if (test_ok)
        debug_mod(1, "selftest Ok");
    else
        debug_mod(0, "selftest failed");
    return (test_ok && status_ok());
}

void base_bmod::update_status()
{
    read_module_status();
    double temp = read_temp();
    if (temp != 0.) {
        info_mod(QString("t %1 \u00b0C").arg(temp, 0, 'f', 1));
//        setCalibrationCondition("current_temperature", temp);
    }
}

void base_bmod::preparerun()
{
    update_status();
}

void base_bmod::calibration()
{
    update_status();
}

void base_bmod::start_of_burst()
{
}

void base_bmod::end_of_burst()
{
    read_module_status();
    end_of_burst_readout();
}

void base_bmod::end_of_burst_readout ()
{
    append_rawdata_mhdr();
    append_rawdata_status();
    append_rawdata_mtrl(1);
}

void base_bmod::clear_statistics()
{
    memset(&setup->stats, 0, sizeof(setup->stats));
}

void base_bmod::update_statistics()
{
    setup->stats.total++;
    //  if (readout_enabled()) print_flags();
    if (get_error_flag()) setup->stats.err++;
    if (get_warning_flag()) setup->stats.warn++;
    if (get_failure_flag()) setup->stats.fatal++;
    if (get_overflow_flag()) setup->stats.over++;
}

bool base_bmod::interspill_test()
{
    read_module_status();
    return (status_ok());
}

void base_bmod::print_info() const
{
    base_mod::print_info();
}

bool base_bmod::isSpillMaster() const
{
    return get_fam() == MF_TRIG;
}

void base_bmod::start_run()
{
    setup_readout_chain();

    if (!isSpillMaster()) {
        reg_ctrl.bit.run = 1;
        set_reg(MODULE_REG_CTRL, reg_ctrl.all);
    }
}

void base_bmod::post_start_run()
{
    if (isSpillMaster()) {
        reg_ctrl.bit.run = 1;
        set_reg(MODULE_REG_CTRL, reg_ctrl.all);
    }
}

void base_bmod::setup_readout_chain()
{
    int modflags = get_flags();

    reg_ctrl.bit.cblt = (modflags & STDMOD_FLAG_CBLT)?1:0;
    if (reg_ctrl.bit.cblt) {
        reg_ctrl.bit.first = (modflags & STDMOD_FLAG_FIRST)?1:0;
        reg_ctrl.bit.mid = (modflags & STDMOD_FLAG_MID)?1:0;
        reg_ctrl.bit.last = (modflags & STDMOD_FLAG_LAST)?1:0;
    } else {
        reg_ctrl.bit.first = 0;
        reg_ctrl.bit.mid = 0;
        reg_ctrl.bit.last = 0;
    }
    set_reg(MODULE_REG_CTRL, reg_ctrl.all);
}

void base_bmod::pre_stop_run()
{
    if (isSpillMaster()) {
        reg_ctrl.bit.run = 0;
        set_reg(MODULE_REG_CTRL, reg_ctrl.all);
    }
}

void base_bmod::stop_run()
{
    if (!isSpillMaster()) {
        reg_ctrl.bit.run = 0;
        set_reg(MODULE_REG_CTRL, reg_ctrl.all);
    }
    reg_ctrl.bit.cblt = 0;
    set_reg(MODULE_REG_CTRL, reg_ctrl.all);
}

void base_bmod::start_spill()
{
}

void base_bmod::stop_spill()
{
}

void base_bmod::initialize_setup_entry(common_setup_t *s)
{
    Q_ASSERT(s);
    s->magic = BASE_SETUP_MAGIC;
    s->size = get_setup_size();
}

module_status_reg_t base_bmod::get_status_reg()
{
    int tmp;
    get_reg(MODULE_REG_STATUS, &tmp);
    status_reg.all = tmp;
    return status_reg;
}

module_status_reg_t base_bmod::read_module_status(bool messages)
{
    get_status_reg();
    // overflow bit has active low
    if (status_reg.bit.overflow==0) {
        set_overflow_flag();
    }
    //  else clear_overflow_flag();
    if (status_ok())
        clear_error_flag();
    else
        set_error_flag();
    if (messages)
        print_status_reg();

    if (setup) setup->mon.temperature = read_temp();
    return status_reg;
}

bool base_bmod::status_ok() const
{
    module_status_reg_t mask;
    mask.all = 0x0000FFFF;
    mask.bit.overflow = 0;
    return ((status_reg.all & mask.all) == mask.all);
}

bool base_bmod::ttc_clock_ok() const
{
    return status_reg.bit.ttc_lock;
}

void base_bmod::print_status_reg() const
{
    bool ok = status_ok();
    bool err = !status_reg.bit.ttc_lock || !status_reg.bit.device_ok || !status_reg.bit.readerror;
    QString str = QString("%1 %2%3%4%5%6%7%8%9%10%11")
                  .arg(status_reg.all, 8, 16, QChar('0'))
                  .arg(!status_reg.bit.ttc_lock?"No_TTC_clock ":"")
                  .arg(!status_reg.bit.device_ok?"General_error ":"")
                  .arg(!status_reg.bit.subdev_0_ok?"Sub_0_err ":"")
                  .arg(!status_reg.bit.subdev_1_ok?"Sub_1_err ":"")
                  .arg(!status_reg.bit.subdev_2_ok?"Sub_2_err ":"")
                  .arg(!status_reg.bit.subdev_3_ok?"Sub_3_err ":"")
                  .arg(!status_reg.bit.subdev_4_ok?"Sub_4_err ":"")
                  .arg(!status_reg.bit.subdev_5_ok?"Sub_5_err ":"")
                  .arg(!status_reg.bit.readerror?"Readout_Error ":"")
                  .arg(!status_reg.bit.overflow?"Overflow ":"");

    if (ok) {
        debug_mod(1, "STATUS: Ok");
    } else {
        if (err)
            err_mod(str);
        else
            warn_mod(str);
    }
}

void base_bmod::start_of_burst_signal()
{
    clear_error_flag(); // Clear any errors
}

void base_bmod::end_of_burst_signal()
{
}

void base_bmod::clear_error_flag()
{
    clr_flags_bit(STDMOD_FLAG_WARN);
    clr_flags_bit(STDMOD_FLAG_ERR);
    clr_flags_bit(STDMOD_FLAG_FAIL);
}

void base_bmod::set_failure_flag()
{
    set_flags_bit(STDMOD_FLAG_FAIL);
}

void base_bmod::set_error_flag()
{
    set_flags_bit(STDMOD_FLAG_ERR);
}

void base_bmod::set_warning_flag()
{
    set_flags_bit(STDMOD_FLAG_WARN);
}

void base_bmod::set_overflow_flag()
{
    set_flags_bit(STDMOD_FLAG_OVER);
}

void base_bmod::clear_overflow_flag()
{
    clr_flags_bit(STDMOD_FLAG_OVER);
}

void base_bmod::append_rawdata(int d)
{
    put_pio_rawdata_word(d, shm_pio_rawdata);
}

void base_bmod::append_rawdata_mhdr()
{
    const unsigned int ga = get_slot();
    const unsigned int id = get_id();
    const unsigned int ev = 0;

    unsigned int mhdr = (8<<28)|(ga<<23)|(id<<16)|ev;
    append_rawdata(mhdr);
}

void base_bmod::append_rawdata_status()
{
    quint16 temp_value = get_temperature() * 256;
    unsigned int status = (0xE<<28)|(1<<24)|(temp_value&0xFFFF);
    append_rawdata(status);
}

void base_bmod::append_rawdata_mtrl(unsigned int wc)
{
    const unsigned int ga = get_slot();
    unsigned int mtrl = (9<<28)|(ga<<23)|(0<<20)|0xF0000|wc;
    append_rawdata(mtrl);
}

double base_bmod::read_temp()
{
    int data = 0;
    try {
        get_reg(MODULE_REG_TEMPERATURE_1, &data);
    }
    catch (std::runtime_error &) {
    }
    if (data == -1) return 0;
    double temp = 0.0625 * (data & 0x7FF);
    return temp;
}

void base_bmod::run_task(bmod_task_id_t task_id)
{
    debug_mod(3, task_id_str(task_id));
    switch (task_id) {
    case BMOD_TASK_READ_CONFIG: read_configuration(); break;
    case BMOD_TASK_SELFTEST: selftest(); break;
    case BMOD_TASK_RESET: reset(); break;
    case BMOD_TASK_UPDATE_STATUS: update_status(); break;
    case BMOD_TASK_PREPARERUN: preparerun(); break;
    case BMOD_TASK_CALIBRATION: calibration(); break;
    case BMOD_TASK_START_RUN: start_run(); break;
    case BMOD_TASK_POST_START_RUN: post_start_run(); break;
    case BMOD_TASK_PRE_STOP_RUN: pre_stop_run(); break;
    case BMOD_TASK_STOP_RUN: stop_run(); break;
    case BMOD_TASK_START_SPILL: start_spill(); break;
    case BMOD_TASK_STOP_SPILL: stop_spill(); break;
    case BMOD_TASK_START_OF_BURST: start_of_burst(); break;
    case BMOD_TASK_END_OF_BURST: end_of_burst(); break;
    case BMOD_TASK_START_OF_BURST_SIGNAL: start_of_burst_signal(); break;
    case BMOD_TASK_END_OF_BURST_SIGNAL: end_of_burst_signal(); break;
    case BMOD_TASK_CLEAR_STATISTICS: clear_statistics(); break;
    case BMOD_TASK_UPDATE_STATISTICS: update_statistics(); break;
    case BMOD_TASK_INTERSPILL_TEST: interspill_test(); break;
    }
}

#ifdef DEBUG
void base_bmod::debug_mod(int level, const QString &text) const { log_pri_mod(LOG_DEBUG, level, text); }
#else
void base_bmod::debug_mod(int level, const QString &text) const { Q_UNUSED(level); Q_UNUSED(text); }
#endif
void base_bmod::info_mod(const QString &text) const { log_pri_mod(LOG_INFO, 0, text); }
void base_bmod::notice_mod(const QString &text) const { log_pri_mod(LOG_NOTICE, 0, text); }
void base_bmod::warn_mod(const QString &text) const { log_pri_mod(LOG_WARNING, 0, text); }
void base_bmod::err_mod(const QString &text) const { log_pri_mod(LOG_ERR, 0, text); }

void base_bmod::log_pri_mod(int priority, int level, const QString &str) const
{
    QString s = QString("[debug %1] ").arg(level)+QString("%1 (serial %2, slot %3): %4").arg(get_name()).arg(get_serial(), 8, 16, QChar('0')).arg(get_slot(), 2).arg(str);
    switch (priority) {
    case LOG_ERR:
        qCritical() << s;
        break;
    case LOG_WARNING:
        qWarning() << s;
        break;
    default:
        if (priority != LOG_DEBUG || debug_level_ge(level))
            qDebug() << s;
    }
}

void base_bmod::waitMs(unsigned long ms) const
{
    QMutex dummy;
    dummy.lock();
    QWaitCondition wc;
    wc.wait(&dummy, ms);
}
