#ifndef BASE_BMOD_H
#define BASE_BMOD_H

#include <QtCore/QtGlobal>
#include <vector>

#include "hal/base_mod.h"
#include "hal/mod_regs.h"

class base_hwif;
struct shm_status_t;
struct shm_pio_rawdata_t;

class bmod_props
{
public:
    bmod_props(base_hwif *_hwif, int _reg_base, shm_pio_rawdata_t *_shm_pio_rawdata)
        : hwif(_hwif), reg_base(_reg_base), shm_pio_rawdata(_shm_pio_rawdata) {}
    base_hwif * const hwif;
    int reg_base;
    shm_pio_rawdata_t * const shm_pio_rawdata;
};

class base_bmod_hardware {
public:
    base_bmod_hardware(const bmod_props &props);
    // proxy calls to hwif
    void w_a24_d32 (unsigned int addr, int data);
    void r_a24_d32 (unsigned int addr, int *data);
    void w_a16_d64 (unsigned int addr, qint64 data);
    void r_a16_d64 (unsigned int addr, qint64 *data);
    void w_a16_d32 (unsigned int addr, int data);
    void r_a16_d32 (unsigned int addr, int *data);
    void w_a16_d16 (unsigned int addr, int data);
    void r_a16_d16 (unsigned int addr, int *data);
    void w_blk_a24_d32 (unsigned int start, size_t n32, int *data);
    void r_blk_a24_d32 (unsigned int start, size_t n32, int *data);
    bool testreg32 (int addr, unsigned int loops);
//    bool testreg16 (int addr, unsigned int loops);
    bool testram32 (int addr, unsigned int radix, unsigned int count);
//    bool testram16 (int addr, unsigned int radix, unsigned int count);

    void set_reg (unsigned int reg, int data);
    void get_reg (unsigned int reg, int *data);

    void set_reg_v (const std::vector<int> &regs, const std::vector<int> &data);
    std::vector<int> get_reg_v (const std::vector<int> &regs);

protected:
    int get_reg_base() const {return reg_base;}

private:
    base_hwif * const hwif;
    const int reg_base;
};

class base_bmod : protected base_bmod_hardware, public virtual base_mod
{
public:
    base_bmod(const bmod_props &props);
    virtual ~base_bmod() {}

    virtual void print_info() const;

    virtual void run_task(bmod_task_id_t task_id);
    virtual bool detect ();

    // module actions - FIXME: move to protected
    virtual void read_configuration();
    virtual bool selftest ();
    virtual void reset ();
    virtual void update_status();
    virtual void preparerun();
    virtual void calibration();
    virtual void start_run();
    virtual void post_start_run();
    virtual void pre_stop_run();
    virtual void stop_run();
    virtual void start_spill();
    virtual void stop_spill();
    virtual void start_of_burst();
    virtual void end_of_burst();
    virtual void end_of_burst_readout(); // called from end_of_burst()
    virtual void clear_statistics();
    virtual void update_statistics();
    virtual bool interspill_test();
    virtual void start_of_burst_signal();
    virtual void end_of_burst_signal();

    virtual void initialize_setup_entry(common_setup_t *s);

protected:
    virtual int reg_index_jtag()  const {return 3;}
    virtual int reg_index_jtagb() const {return 8;}
    virtual module_status_reg_t get_status_reg();
    virtual module_status_reg_t read_module_status(bool messages = true);
    virtual bool status_ok() const;
    virtual bool ttc_clock_ok() const;
    virtual void print_status_reg() const;

    virtual void clear_error_flag(); // also clears warning, overflow
    virtual void set_failure_flag();
    virtual void set_error_flag();
    virtual void set_warning_flag();
    virtual void set_overflow_flag();
    virtual void clear_overflow_flag();

    void append_rawdata(int d);
    void append_rawdata_mhdr();
    void append_rawdata_status();
    void append_rawdata_mtrl(unsigned int wc);

    virtual double read_temp();
    bool isSpillMaster() const;

    void log_pri_mod(int priority, int level, const QString &str) const;
    void debug_mod(int level, const QString &text) const;
    void info_mod(const QString &text) const;
    void notice_mod(const QString &text) const;
    void warn_mod(const QString &text) const;
    void err_mod(const QString &text) const;

    void waitMs(unsigned long ms) const;

private:
    bool messagesEnabled() const { return messages_enabled; }
    void setup_readout_chain();

    shm_pio_rawdata_t * const shm_pio_rawdata;
    bool config_rom_cache_valid;
    mutable bool messages_enabled;
    module_ctrl_reg_t reg_ctrl;
    module_status_reg_t status_reg;
};

#endif /* BASE_BMOD_H */
