/*
 * abstract hardware interface
 */

#ifndef BASE_HWIF_H
#define BASE_HWIF_H

#include <QtCore/QtGlobal>
#include <list>
#include <queue>
#include "daqcommon/vme_types.h"

class hwifEvent;

typedef std::priority_queue<hwifEvent> hwev_queue_t;

// block operation
struct vme_io_t {
    VME_AM am;
    unsigned int addr;
    bool write;
    int wdata;
    int rdata;
    unsigned int st;
};

typedef std::vector<vme_io_t> vme_block_io_t;

// Configuration ROM
#pragma pack(push, 1)
typedef struct {
    quint8 serial;
    quint8 type;
    quint8 reserved;
    quint8 vendor;
} base_cr_t;
#pragma pack(pop)

struct vme_cr_t;
struct vme_csr_t;

struct hwif_discover_t {
    hwif_discover_t(const vme_cr_t&, const vme_csr_t&, int ga, int serial_id);
    base_cr_t cr;
    quint32 serial_id;
    quint32 reg_base;
    quint8 slot;
};

class base_hwif {
public:
    virtual ~base_hwif() {}

    // single bus operations
    virtual void w_a16_d16 (unsigned int addr, int data) = 0;
    virtual void r_a16_d16 (unsigned int addr, int *data) = 0;
    virtual void w_a16_d32 (unsigned int addr, int data) = 0;
    virtual void r_a16_d32 (unsigned int addr, int *data) = 0;
    virtual void w_r_a16_d32 (unsigned int addr, int wdata, int *rdata);
    virtual void w_a16_d64 (unsigned int addr, qint64 data) = 0;
    virtual void r_a16_d64 (unsigned int addr, qint64 *data) = 0;
    virtual void w_a24_d32 (unsigned int addr, int data) = 0;
    virtual void r_a24_d32 (unsigned int addr, int *data) = 0;

    // block bus operation
    virtual int blk_io (vme_block_io_t &multi_io) = 0;
    virtual size_t max_blk_io_size() const = 0;

    // grouped operations (not BLT, really) with address autoincrement
    virtual void w_blk_a24_d32 (unsigned int start_addr, size_t n32, int *data) = 0;
    virtual void r_blk_a24_d32 (unsigned int start_addr, size_t n32, int *data) = 0;

    virtual void w_a16_d32(const std::vector<int> &addr, const std::vector<int> &data) = 0;
    virtual std::vector<int> r_a16_d32(const std::vector<int> &addr) = 0;

    // Test 32-bit register available at address <addr>
    virtual bool testreg32 (int addr, unsigned int loops) = 0;
    // Test 16-bit register available at address <addr>
    //  virtual bool testreg16 (int addr, unsigned int loops);
    // Test RAM available at address <addr> with size 2^<radix> longwords
    virtual bool testram32 (int addr, unsigned int radix, unsigned int count) = 0;
    // Test RAM available at address <addr> with size 2^<radix> words
    //  virtual bool testram16 (int addr, unsigned int radix, unsigned int count) = 0;

    virtual std::list<hwif_discover_t> discover() = 0;
    virtual void set_run_mode(bool mode) = 0;

    virtual void start_communication() = 0;
    virtual void sysreset() = 0;
    virtual void reset() = 0;
    virtual bool selftest() = 0;
    virtual void post_start_run() = 0;
    virtual void pre_stop_run() = 0;
    virtual bool interspill_test() = 0;
    virtual void start_of_burst_signal() = 0;
    virtual void end_of_burst_signal() = 0;

    virtual void get_events(hwev_queue_t *q, int timeout_ms) = 0;
    virtual unsigned int get_spwc() const = 0;

protected:
    base_hwif() {}
};

#endif /* BASE_HWIF_H */
