
/*
 * BRIC backend (hardware) implementation
 */

#include "bric_bmod.h"

#include <QString>

#include "modules/bric_types.h"
#include "debug_hal.h"

namespace {
#pragma pack(push, 1)
    union bric_bricset_t {
        unsigned int all;
        struct {
            unsigned int win_enable:16; // 16
            unsigned int read32:1; // 1 = read 32 windows, 0 = read 16 windows
            unsigned int pol:1; //
            unsigned int overflow_threshold:4; // limit branch length by 32*N cards
            unsigned int reserved:10;
        } bit;
    };
#pragma pack(pop)

    enum BRIC_REG {
        BRIC_REG_CONTROL = 0,
        BRIC_REG_BRICSET = 1,
        BRIC_REG_TEST = BRIC_REG_BRICSET
    };
    enum BRICBOX_REG {
        BRIC2_REG_CTRL      = 0x40,
        BRIC2_REG_STATUS    = 0x41,
        BRIC2_REG_DEVICE_ID = 0x42,
        BRIC2_REG_TEST      = 0x43,
        BRIC2_REG_WINMASK   = 0x44,
        BRIC2_REG_LIMITS    = 0x45,
        BRIC2_REG_DELAYS    = 0x46,
        BRIC2_REG_TEMP      = 0x54,
        BRIC2_REG_FW_VER    = 0x56,
        BRIC2_REG_FW_REV    = 0x57
    };
    const quint16 BRIC2_REG_BASE = 0x0100;
}

void bric_bmod::write_remote_reg(int reg, quint16 data)
{
    Q_ASSERT(reg < 0x0100);
    // TODO: implement retransmit
    try {
        data &= 0xFFFF;
        debug_mod(0, QString("write remote reg %1 = %2").arg(reg, 4, 16, QChar('0')).arg(data, 4, 16, QChar('0')));
        set_reg(reg + BRIC2_REG_BASE, data);
        int check = 0;
        get_reg(reg + BRIC2_REG_BASE, &check);
        if ((check & 0xFFFF) != (data & 0xFFFF)) {
            err_mod(QString("write remote reg %1 failed: data error").arg(reg, 4, 16, QChar('0')));
        }
    }
    catch (std::exception &e) {
        err_mod(QString("write remote reg %1 failed: communication error").arg(reg, 4, 16, QChar('0')));
    }
}

quint16 bric_bmod::read_remote_reg(int reg)
{
    Q_ASSERT(reg < 0x0100);
    int data = 0;
    // TODO: implement retransmit
    try {
        get_reg(reg + BRIC2_REG_BASE, &data);
        data &= 0xFFFF;
        debug_mod(0, QString("read remote reg %1 = %2").arg(reg, 4, 16, QChar('0')).arg(data, 4, 16, QChar('0')));
    }
    catch (std::exception &e) {
        err_mod(QString("read remote reg %1 failed").arg(reg, 4, 16, QChar('0')));
    }
    return data;
}

bool bric_bmod::detect()
{
    /*
    int data = 0;
    // read remote registers
    get_reg(BRIC_REG_WINMASK, &data);
    debug_mod(0, QString("read remote reg 44h = %1").arg(data, 8, 16, QChar('0')));
    data = 0;
    get_reg(BRIC_REG_LIMITS, &data);
    debug_mod(0, QString("read remote reg 45h = %1").arg(data, 8, 16, QChar('0')));
    // write remote registers
    data = 0x12345678;
    set_reg(BRIC_REG_WINMASK, data);
    debug_mod(0, QString("write remote reg 44h = %1").arg(data, 8, 16, QChar('0')));
    data = 0;
    get_reg(BRIC_REG_WINMASK, &data);
    debug_mod(0, QString("read remote reg 44h = %1").arg(data, 8, 16, QChar('0')));
*/
    // BRIC2 - write remote registers
    quint16 device_id = read_remote_reg(BRIC2_REG_DEVICE_ID);
    quint16 fw_ver = read_remote_reg(BRIC2_REG_FW_VER);
    quint16 fw_rev = read_remote_reg(BRIC2_REG_FW_REV);
    info_mod(QString("device_id %1, fw %2.%3.%4").arg(device_id).arg((fw_ver>>8)&0xFF).arg(fw_ver&0xFF).arg(fw_rev));
    setup->ndev = 1;
    setup->nch = BRIC_NCH;
    return true;
}

void bric_bmod::reset()
{
    base_bmod::reset();
    waitMs(1); // 100 us delay required
    load_settings(setup);
    read_module_status();
    debug_mod(BRIC_DEBUG_LEVEL+2, "Reset Ok");
}

void bric_bmod::load_settings(const common_setup_t *p)
{
    Q_ASSERT(p);

    const bric_setup_t *s = static_cast<const bric_setup_t *> (p);
    //  set_mode_vme ();

    debug_mod(3, QString("window enable mask %1").arg(s->win_enable, 8, 16, QChar('0')));

    bric_bricset_t bricset;
    bricset.all = 0;
    bricset.bit.win_enable = s->win_enable;
    //  bricset.bit.read32 = 1;
    bricset.bit.pol = 1;
    // Set overflow threshold to less than BRIC readout fifo size (4092 cards, or 16 win x 255 cards)
    // N=overflow_threshold size in cards is calculated as: cards = 31+32*N
    // 0 - 31
    // 1 - 63 (31+32)
    // 2 - 95 (31+64)
    // 3 - 127 (31+96)
    // 4 - 159 (31+128)
    // 5 - 191 (31+160)
    // 6 - 223 (31+192)
    // 7 - 255 (31+224)
    // ...
    // F - 511 (31+480)

    bricset.bit.overflow_threshold = 0; // hardcoded
    //  bricset.bit.overflow_threshold = s->branch_size_limit_index[0];
    if (bricset.bit.overflow_threshold > 15) bricset.bit.overflow_threshold = 15;

    set_reg(BRIC_REG_BRICSET, bricset.all);
    debug_mod(0/*BRIC_DEBUG_LEVEL*/, QString("BRICSET    = %1, win_enable=%2, ov_thr=%3")
              .arg(bricset.all, 8, 16, QChar('0'))
              .arg(bricset.bit.win_enable)
              .arg(bricset.bit.overflow_threshold));

    if (s->fw_ver == 0x0100) {
    // BRIC2 - write remote registers
        quint16 device_id = read_remote_reg(BRIC2_REG_DEVICE_ID);
        quint16 fw_ver = read_remote_reg(BRIC2_REG_FW_VER);
        quint16 fw_rev = read_remote_reg(BRIC2_REG_FW_REV);
        info_mod(QString("device_id %1, fw %2.%3.%4").arg(device_id).arg((fw_ver>>8)&0xFF).arg(fw_ver&0xFF).arg(fw_rev));
        write_remote_reg(BRIC2_REG_WINMASK, bricset.bit.win_enable);
        write_remote_reg(BRIC2_REG_LIMITS, 30);
        write_remote_reg(BRIC2_REG_DELAYS, 0x140F); // period=20, delay=15
    }

    debug_mod(BRIC_DEBUG_LEVEL+1, "load_settings Ok");
}

bool bric_bmod::interspill_test()
{
    bool test_ok = true;

    test_ok = testreg32(get_reg_base() + BRIC_REG_TEST * 4, 10);
    read_module_status();
    if (test_ok)
        debug_mod(BRIC_DEBUG_LEVEL, "interspill test Ok");
    else
        debug_mod(0, "interspill_test failed");
    return test_ok;
}

bool bric_bmod::selftest()
{
    bool test_ok = true;

    debug_mod(BRIC_DEBUG_LEVEL, "selftest...");

    test_ok = testreg32(get_reg_base() + BRIC_REG_TEST * 4, 100);
    read_module_status();

    if (test_ok)
        debug_mod(BRIC_DEBUG_LEVEL, "selftest Ok");
    else
        debug_mod(0, "selftest failed");
    return test_ok;
}

void bric_bmod::preparerun()
{
    base_bmod::preparerun();
    load_settings(setup);
    debug_mod(BRIC_DEBUG_LEVEL, "preparerun Ok");
}

void bric_bmod::start_run()
{
    reset();
    base_bmod::start_run();
    //  load_settings(setup);
    debug_mod(BRIC_DEBUG_LEVEL, "start_run Ok");
}
