//
//    Copyright 2011 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "hal/device_db.h"

#include <QDebug>
#include <QDir>
#include <QFile>
#include <QDataStream>

#include "QtSingleApplication/QtLockedFile"

namespace {
    const quint32 dbfile_hdr_magic = 0x42445F44;
}

ddb_ent::ddb_ent(const void *data, size_t size)
    : QByteArray(static_cast<const char *>(data), size)
{
}

ddb_ent::ddb_ent(const QByteArray &p)
    : QByteArray(p)
{
}

QString device_db::defaultFileName()
{
    return "dev_db.dat";
}

ddb_ent device_db::find(ddb_key_t key) const
{
    ddbmap_t::const_iterator it = devs.find(key);
    if (it == devs.end())
        return ddb_ent();
    else
        return *it;
}

void device_db::put(ddb_key_t key, const void *data, size_t size)
{
    Q_ASSERT(data);
    put(key, ddb_ent(static_cast<const char *>(data), size));
}

void device_db::put(ddb_key_t key, const ddb_ent &data)
{
    devs.insert(key, data);
}


device_db::device_db(const QString &filename)
    : fn(filename)
{
}

void device_db::clear()
{
    devs.clear();
}

bool device_db::load()
{
    devs.clear();

    QtLockedFile f(fn);
    if (!f.open(QIODevice::ReadOnly)) {
        qDebug() << "Can't open file" << fn << ":" << f.errorString();
        return false;
    }
    if (!f.lock(QtLockedFile::ReadLock)) {
        qWarning() << "Can't lock file" << fn << ":" << f.errorString();
        return false;
    }
    QDataStream ds(&f);
    quint32 fileHeader;
    ds >> fileHeader;
    if (fileHeader != dbfile_hdr_magic) {
        ds.setStatus(QDataStream::ReadCorruptData);
    }
    ddbmap_t fileData;
    ds >> fileData;
    f.unlock();
    f.close();
    if (f.error() != QFile::NoError) {
        qWarning() << "Can't read file" << fn << " :" << f.errorString();
        return false;
    }
    if (ds.status() != QDataStream::Ok) {
        qWarning() << "Can't parse file" << fn;
        return false;
    }
    devs = fileData;
    return true;
}

bool device_db::save() const
{
    return saveAs(fn);
}

bool device_db::saveAs(const QString &fileName) const
{

    QtLockedFile f(fileName);
    if (!f.open(QIODevice::WriteOnly)) {
        qWarning() << "Can't open file" << fileName << ":" << f.errorString();
        return false;
    }
    if (!f.lock(QtLockedFile::WriteLock)) {
        qWarning() << "Can't lock file" << fileName << ":" << f.errorString();
        return false;
    }
    QDataStream ds(&f);
    ds << dbfile_hdr_magic;
    ds << devs;

    f.unlock();
    f.close();
    return true;
}

QList<ddb_key_t> device_db::keys() const
{
    return devs.keys();
}
