
#include <QtCore/QtGlobal>

#include "hptdc/hptdc_types.h"
#include "hptdc/hptdc.h"
#include "hptdc_settings.h"
#include "mf_types.h"

hptdc_settings::hptdc_settings(hptdc_setup_t *ps, int *pmask)
    : s(ps),
    channel_mask(pmask)
{
    Q_ASSERT(s);
    Q_ASSERT(channel_mask);
}

void hptdc_settings::set_defaults()
{
    hptdc_set_setup_defaults(s);
    hptdc_setup_set_trigmode(s);  /* some working mode */
}

void hptdc_settings::get_id(int *id) const
{
    if (id)
        *id = s->bit.tdc_id;
}

void hptdc_settings::set_id(int id)
{
    s->bit.tdc_id = id;
}

//void hptdc_settings::set_param(bool ENABLE_TDC_HEADERS, bool ENABLE_TDC_TRAILERS, int HPTDC_CORE_CLOCK)
//{
//    s->bit.enable_local_header = ENABLE_TDC_HEADERS?1:0;
//    s->bit.enable_local_trailer = ENABLE_TDC_TRAILERS?1:0;
//    s->bit.core_clock_source = HPTDC_CORE_CLOCK;   // 0->40MHz, 1->80MHz, 2->160MHz
//}

void hptdc_settings::get_window(int *matchwin, int *latency) const
{
    if (matchwin)
        *matchwin = s->bit.match_window;
    if (latency)
        *latency = 0x1000 - s->bit.trigger_count_offset;
}

void hptdc_settings::set_window(int matchwin, int latency)
{
    if (matchwin > 4087)
        matchwin = 4087;
    latency &= 0xFFF;

    s->bit.enable_matching = 1;
    s->bit.enable_automatic_reject = 1;
    s->bit.roll_over = 0xFFF;  // no roll-over
    s->bit.coarse_count_offset = 0;
    s->bit.enable_relative = 1;
    s->bit.enable_fixed_pattern = 0;
    s->bit.match_window = matchwin;
    s->bit.search_window = matchwin + 8;
    s->bit.trigger_count_offset = (0xFFF - latency + 1) & 0xFFF;
    s->bit.reject_count_offset = (s->bit.trigger_count_offset - 4) & 0xFFF;
}

void hptdc_settings::get_pair(bool *pair, int *pres) const
{
    if (pair)
        *pair = s->bit.enable_pair;
    if (pres)
        *pres = s->bit.width_select;
}

void hptdc_settings::set_pair(bool pair, int pres)
{
    if (pair)
        pair = 1;
    s->bit.enable_pair = pair;
    s->bit.width_select = pres;
}

void hptdc_settings::get_resolution(TdcResIndex *res) const
{
    hptdc_setup_get_resindex(*s, res);
}

void hptdc_settings::set_resolution(TdcResIndex resindex)
{
    // 0=100ps, 1=200ps, 2=400ps, 3=800ps, 4=1.6ns
    hptdc_setup_set_resindex(s, resindex);
}

void hptdc_settings::get_edge(int *edge) const
{
    if (!edge) return;

    int d = 0;
    if (s->bit.enable_leading)
        d |= 1;
    if (s->bit.enable_trailing)
        d |= 2;
    *edge = d;
}

void hptdc_settings::set_edge(int edge)
{
    hptdc_setup_set_edge(s, edge);   /* 1 = /   2 = \  3 = /\  */
}

void hptdc_settings::get_hdr(bool *hdr) const
{
    int h;
    h = s->bit.enable_local_header;
    if (hdr)
        *hdr = h;
}

void hptdc_settings::get_trl(bool *trl) const
{
    int t;
    t = s->bit.enable_local_trailer;
    if (trl)
        *trl = t;
}

void hptdc_settings::set_hdr(bool hdr)
{
    s->bit.enable_local_header = hdr?1:0;
}

void hptdc_settings::set_trl(bool trl)
{
    s->bit.enable_local_trailer = trl?1:0;
}

void hptdc_settings::get_rcadj(int rcadj[4]) const
{
    if (rcadj) {
        rcadj[0] = s->bit.rc_adjust_0;
        rcadj[1] = s->bit.rc_adjust_1;
        rcadj[2] = s->bit.rc_adjust_2;
        rcadj[3] = s->bit.rc_adjust_3;
    }
}

void hptdc_settings::set_rcadj(int rcadj[4])
{
    s->bit.rc_adjust_0 = rcadj[0];
    s->bit.rc_adjust_1 = rcadj[1];
    s->bit.rc_adjust_2 = rcadj[2];
    s->bit.rc_adjust_3 = rcadj[3];
}

void hptdc_settings::get_ch_enable(int ch, bool *en) const
{
    if (en)
        *en = (*channel_mask & (1<<ch));
}

void hptdc_settings::set_ch_enable(int ch, bool en)
{
    unsigned int mask;
    if (ch == SET_ALL_CH || ch == SET_MOD_CH) mask=0xFFFFFFFF;
    else
        mask = (1<<ch);
    if (en)
        *channel_mask |= mask;
    else
        *channel_mask &= ~mask;
}

int hptdc_settings::get_min_win() { return 1; }
int hptdc_settings::get_max_win() { return 4095; }

int hptdc_settings::get_min_lat() { return 1; }
int hptdc_settings::get_max_lat() { return 4095; }

bool hptdc_settings::get_int(int *value, const std::string &tag, unsigned int mch) const
{
    Q_ASSERT(value);

    if (tag == "hptdc_min_win") {*value = get_min_win(); return true;}
    if (tag == "hptdc_max_win") {*value = get_max_win(); return true;}
    if (tag == "hptdc_min_lat") {*value = get_min_lat(); return true;}
    if (tag == "hptdc_max_lat") {*value = get_max_lat(); return true;}
    if (tag == "en") {
        bool en;
        get_ch_enable(mch, &en);
        *value = en;
        return true;
    }
    if (tag == "win") {
        int mw, lat;
        get_window(&mw, &lat);
        *value = mw;
        return true;
    }
    if (tag == "lat") {
        int mw, lat;
        get_window(&mw, &lat);
        *value = lat;
        return true;
    }
    if (tag == "res") {
        TdcResIndex resindex;
        get_resolution(&resindex);
        *value = resindex;
        return true;
    }
    if (tag == "edge") {
        get_edge(value);
        return true;
    }
    if (tag == "pair") {
        bool pair;
        int pres;
        get_pair(&pair, &pres);
        *value = pair;
        return true;
    }
    if (tag == "pres") {
        bool pair;
        int pres;
        get_pair(&pair, &pres);
        *value = pres;
        return true;
    }
    if (tag == "hdr") {
        bool h;
        //    get_hdr_trl(&h, &t);
        get_hdr(&h);
        *value = h;
        return true;
    }
    if (tag == "trl") {
        bool t;
        //    get_hdr_trl(&h, &t);
        get_trl(&t);
        *value = t;
        return true;
    }
    if (tag == "rcadj") {
        int rcadj[4] = {0, 0, 0, 0};
        get_rcadj(rcadj);
        *value= hptdc_pack_rcadj(rcadj);
        return true;
    }

    return false;
}

bool hptdc_settings::set_int(int value, const std::string &tag, int mch)
{
    if (tag == "en") {
        set_ch_enable(mch, value);
        return true;
    }
    if (tag == "win") {
        int mw, lat;
        get_window(&mw, &lat);
        mw = value;
        set_window(mw, lat);
        return true;
    }
    if (tag == "lat") {
        int mw, lat;
        get_window(&mw, &lat);
        lat = value;
        set_window(mw, lat);
        return true;
    }
    if (tag == "res") {
        set_resolution(static_cast<TdcResIndex>(value));
        return true;
    }
    if (tag == "edge") {
        set_edge(value);
        return true;
    }
    if (tag == "pair") {
        bool pair;
        int pres;
        get_pair(&pair, &pres);
        pair = value;
        set_pair(pair, pres);
        return true;
    }
    if (tag == "pres") {
        bool pair;
        int pres;
        get_pair(&pair, &pres);
        pres = value;
        set_pair(pair, pres);
        return true;
    }
    if (tag == "hdr") {
        //    bool h, t;
        //    get_hdr_trl(&h, &t);
        //    h = value;
        //    set_hdr_trl(h, t);
        set_hdr(value);
        return true;
    }
    if (tag == "trl") {
        //    bool h, t;
        //    get_hdr_trl(&h, &t);
        //    t=value;
        //    set_hdr_trl(h, t);
        set_trl(value);
        return true;
    }
    if (tag == "rcadj") {
        int rcadj[4] = {0, 0, 0, 0};
        hptdc_unpack_rcadj(value, rcadj);
        set_rcadj(rcadj);
        return true;
    }
    return false;
}
