
#include "hal/hw_backend.h"

#include <QDebug>
#include <QDir>

#include "daqcommon/daqexc.h"
#include "base_bmod.h"
#include "hwif_vme.h"
#include "daqcommon/daq_ipc.h"
#include "hal/mod_regs.h"
#include "fvme/hwif_vme_mlink.h"
#include "hw_ipc.h"
#include "hal/hwif_event.h"
#include "base_bmod_factory.h"
#include "daqcommon/vme_types.h"
#include "daqcommon/bitop.h"

#include "hal/device_db.h"

hw_backend::hw_backend(shm_status_t *st, shm_pio_rawdata_t *rd)
    : ddb(new device_db(defaultDbFile())),
      sharedSetup(new SharedSetupBackend()),
#ifdef Q_OS_WIN
      hwif(new hwif_vme_mlink()),
#else
      hwif(new hwif_vme_mlink()),
#endif
      shm_status(st),
      shm_pio_rawdata(rd)
{
    open_ddb();
}

hw_backend::~hw_backend()
{
    delete hwif;
    delete sharedSetup;
    delete ddb;
}

QString hw_backend::defaultDbFile()
{
    const QString &home = QDir::homePath();
    return home + "/" + device_db::defaultFileName();
}

void hw_backend::initialize_database()
{
    //  hws->clear();
    delete_modules();
    ddb->clear();
    ddb->save();
}

void hw_backend::open_ddb()
{
    ddb->load();
}

bool hw_backend::ddb_save_as(const QString &fn) const
{
    return ddb->saveAs(fn);
}

void hw_backend::autodetect_modules(base_hwif *ifc)
{
    delete_modules();
    sharedSetup->clear();
    typedef std::list<hwif_discover_t> crs_t;
    const crs_t &crs = ifc->discover();

    int mod_key = 0;
    for(crs_t::const_iterator it = crs.begin(); it != crs.end(); ++it) {
        const hwif_discover_t &cr = (*it);

        mtype_t id = (mtype_t)cr.cr.type;
        bmod_props props(ifc, cr.reg_base, shm_pio_rawdata);
        base_bmod *p = base_bmod_factory::create(id, props);
        Q_ASSERT(p);

        // set module config
        // add entry to shm
        int mod_shm_index = sharedSetup->alloc(p->get_setup_size()); // allocated and cleared
        common_setup_t *setup = static_cast<common_setup_t *>(sharedSetup->get_data_ptr(mod_shm_index));
        Q_ASSERT(setup);
        // setup is now all zeroes
        p->initialize_setup_entry(setup);
        // setup is formally valid, but empty
        p->set_setup(setup, p->get_setup_size());
        p->set_vend(cr.cr.vendor);
        p->set_id(static_cast<mtype_t>(cr.cr.type));
        p->set_rev(cr.cr.serial);
        p->set_serial(cr.serial_id);
        p->set_slot(cr.slot); // slot number is variable
        p->set_defaults();
        add_module(mod_key, p);

        // set again - add_module overwrites everything with DDB data
        p->set_vend(cr.cr.vendor);
        p->set_id(static_cast<mtype_t>(cr.cr.type));
        p->set_rev(cr.cr.serial);
        p->set_serial(cr.serial_id);
        p->set_slot(cr.slot); // slot number is variable

        p->detect(); // add module even if detect failed
        p->set_slot(cr.slot); // slot number is run-time data, must be re-set after ddb is read in add_module()

//        p->print_setup();

        mod_key++;
    }
}
/*
void hw_backend::open ()
{
  autodetect_modules();
//  hws->print();
//  print_modules();
//  hw_manager::open();
}

void hw_backend::close ()
{
//  hw_manager::close();
  delete_hwifs();
}
*/
void hw_backend::print() const
{
    hw_manager::print();
    qDebug() << ddb;
}

void hw_backend::sync_to_db()
{
    Q_FOREACH(const base_mod *pb, mods) {
        const base_bmod *p = dynamic_cast<const base_bmod *>(pb);
        Q_ASSERT(p);
        ddb_key_t k = p->get_ddb_key();
        // create local object
        //            ddb_ent *e = new ddb_ent();
        //            e->set_data(p->get_setup(), p->get_setup_size());
        ddb->put(k, p->get_setup(), p->get_setup_size());
    }
    ddb->save();
}

void hw_backend::add_module(const int key, base_mod * p)
{
    ddb_key_t k = p->get_ddb_key();
    // create local object
    const QByteArray &data = ddb->find(k);
    bool found = !data.isNull();
    if (found) {
        bool st = p->copy_setup(static_cast<const common_setup_t *>(static_cast<const void *>(data.constData())), data.size()); // try to copy data from DDB to shm
        if (!st) found=0;
    }
    if (!found) {
        // not found in DDB, create new entry
        qDebug() << QString("hw_manager::add_module: creating db entry with key %1").arg(k, 8, 16);
        ddb->put(k, p->get_setup(), p->get_setup_size());
    }
    mods[key] = p;
}

/**************************************************************************/
/*   TASKS                                                                */
/**************************************************************************/

namespace {
    void local_info(const QString &msg)
    {
        qDebug() << QString("hw_backend: %1").arg(msg);
    }

    // initialized map class
    class TaskEnableReadoutRequiredFor: public std::map<bmod_task_id_t, bool>
    {
    public:
        TaskEnableReadoutRequiredFor();
        bool readoutRequired(const bmod_task_id_t &task_id) const;
    };

    TaskEnableReadoutRequiredFor::TaskEnableReadoutRequiredFor()
    {
        (*this)[BMOD_TASK_READ_CONFIG] = false;
        (*this)[BMOD_TASK_SELFTEST] = false;
        (*this)[BMOD_TASK_RESET] = false;
        (*this)[BMOD_TASK_UPDATE_STATUS] = false;
        (*this)[BMOD_TASK_PREPARERUN] = false;
        (*this)[BMOD_TASK_CALIBRATION] = false;
        (*this)[BMOD_TASK_START_RUN] = false;
        (*this)[BMOD_TASK_POST_START_RUN] = false;
        (*this)[BMOD_TASK_PRE_STOP_RUN] = false;
        (*this)[BMOD_TASK_STOP_RUN] = false;
        (*this)[BMOD_TASK_START_OF_BURST] = false;
        (*this)[BMOD_TASK_END_OF_BURST] = false;
        (*this)[BMOD_TASK_START_OF_BURST_SIGNAL] = false;
        (*this)[BMOD_TASK_END_OF_BURST_SIGNAL] = false;
        (*this)[BMOD_TASK_CLEAR_STATISTICS] = false;
        (*this)[BMOD_TASK_UPDATE_STATISTICS] = false;
        (*this)[BMOD_TASK_INTERSPILL_TEST] = false;
    }

    bool TaskEnableReadoutRequiredFor::readoutRequired(const bmod_task_id_t &task_id) const
    {
        return find(task_id) != end();
    }

    const TaskEnableReadoutRequiredFor taskEnableReadoutRequiredFor;
}

void hw_backend::start_communication()
{
    hwif->start_communication();
}

void hw_backend::autodetect()
{
    local_info("autodetect");
    sharedSetup->clear();
    //  hws->print();
    autodetect_modules(hwif);
    bmod_task(BMOD_TASK_READ_CONFIG);
}

void hw_backend::endburst()
{
    local_info("endburst");
    bmod_task(BMOD_TASK_END_OF_BURST);
    bmod_task(BMOD_TASK_UPDATE_STATISTICS);
}

void hw_backend::bmod_task(bmod_task_id_t task_id)
{
    local_info(base_mod::task_id_str(task_id));

    Q_FOREACH(base_mod *pb, mods) {
        base_bmod *p = dynamic_cast<base_bmod *> (pb);
        Q_ASSERT(p);
        //	if (!taskEnableReadoutRequiredFor.readoutRequired(task_id) || p->readout_enabled())
        p->run_task(task_id);
    }
}

void hw_backend::interspill_test()
{
    local_info("interspill_test");
    hwif->interspill_test();
    bmod_task(BMOD_TASK_INTERSPILL_TEST);
}

void hw_backend::preparerun()
{
    local_info("preparerun");
    build_cblt_chain();
    bmod_task(BMOD_TASK_PREPARERUN);
    sync_to_db();
}

void hw_backend::update_status()
{
    local_info("update_status");
    bmod_task(BMOD_TASK_UPDATE_STATUS);
}

void hw_backend::calibration()
{
    local_info("calibration");
    bmod_task(BMOD_TASK_CALIBRATION);
}

void hw_backend::reset()
{
    local_info("reset");
    hwif->reset();
    build_cblt_chain();
    bmod_task(BMOD_TASK_CLEAR_STATISTICS);
    bmod_task(BMOD_TASK_RESET);
    bmod_task(BMOD_TASK_READ_CONFIG);
}

void hw_backend::selftest()
{
    local_info("selftest");
    hwif->selftest();
    bmod_task(BMOD_TASK_SELFTEST);
}

void hw_backend::start_run()
{
    local_info("start_run");
    bmod_task(BMOD_TASK_START_RUN);
    bmod_task(BMOD_TASK_POST_START_RUN);
    hwif->post_start_run();
}

void hw_backend::stop_run()
{
    local_info("stop_run");
    hwif->pre_stop_run();
    bmod_task(BMOD_TASK_PRE_STOP_RUN);
    bmod_task(BMOD_TASK_STOP_RUN);
}

void hw_backend::start_spill()
{
    local_info("start_spill");
    //  hwif->start_spill();
    bmod_task(BMOD_TASK_START_SPILL);
}

void hw_backend::stop_spill()
{
    local_info("stop_spill");
    bmod_task(BMOD_TASK_STOP_SPILL);
    //  hwif->stop_spill();
}

void hw_backend::start_of_burst_signal()
{
    local_info("start_of_burst_signal");
    hwif->start_of_burst_signal();
    bmod_task(BMOD_TASK_START_OF_BURST_SIGNAL);
}

void hw_backend::end_of_burst_signal()
{
    local_info("end_of_burst_signal");
    hwif->end_of_burst_signal();
    bmod_task(BMOD_TASK_END_OF_BURST_SIGNAL);
}

// mods container has items sorted by detect() order.
void hw_backend::build_cblt_chain()
{
    base_bmod *pfirst = 0, *plast = 0;
    //  int en_chain;
    int found_first = 0, found_last = 0;

    // find first and last module with CBLT bit set
    Q_FOREACH(base_mod *pb, mods) {
        base_bmod *p = dynamic_cast<base_bmod *>(pb);
        Q_ASSERT(p);
        bool cblt_enable = p->get_cblt_enable();
        if (cblt_enable && !found_first) {
            pfirst = p;
            found_first = 1;
        }
        if (cblt_enable) {
            plast = p;
            found_last = 1;
        }
    }

    // set flags
    Q_FOREACH(base_mod *pb, mods) {
        base_bmod *p = dynamic_cast<base_bmod *>(pb);
        Q_ASSERT(p);
        base_mod_flags modflags = p->get_flags();
        bool cblt_enable = p->get_cblt_enable();

        if (cblt_enable) {
            bool is_first = (found_first && (p == pfirst));
            bool is_last = (found_last && (p == plast));
            bool is_mid = (!is_first && !is_last);
            modflags |= STDMOD_FLAG_CBLT;
            set_flag_value(&modflags, STDMOD_FLAG_FIRST, is_first);
            set_flag_value(&modflags, STDMOD_FLAG_LAST, is_last);
            set_flag_value(&modflags, STDMOD_FLAG_MID, is_mid);
        } else {
            set_flag_value(&modflags, STDMOD_FLAG_FIRST, false);
            set_flag_value(&modflags, STDMOD_FLAG_LAST, false);
            set_flag_value(&modflags, STDMOD_FLAG_MID, false);
            set_flag_value(&modflags, STDMOD_FLAG_CBLT, false);
        }

//        p->debug_mod(1, QString("hw_build_cblt_chain: %1, flags=%2 %3 %4 %5 %6")
//                  .arg(module_id_str(p->get_id()))
//                  .arg(static_cast<unsigned int>(modflags), 8, 16)
//                  .arg(modflags & STDMOD_FLAG_CBLT ? "CBLT" : "")
//                  .arg(modflags & STDMOD_FLAG_FIRST ? "FIRST" : "")
//                  .arg(modflags & STDMOD_FLAG_MID ? "MID" : "")
//                  .arg(modflags & STDMOD_FLAG_LAST ? "LAST" : ""));
        p->set_flags(modflags);
    }
}

void hw_backend::get_events(hwev_queue_t *q, int timeout_ms)
{
    hwif->get_events(q, timeout_ms);
}

unsigned int hw_backend::get_spwc() const
{
    return hwif->get_spwc();
}
