//
//    Copyright 2010 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "hw_ipc.h"

#include <QDebug>
#include <QSharedMemory>
#include <QSharedPointer>

#include <sstream>
#include <cerrno>
#include <cstring>

#include "daqcommon/daqexc.h"
#include "hal/hws.h"

//const unsigned int SHM_HWS_HDR_KEY = 0x1399;
//const unsigned int SHM_HWS_START_KEY = 0x1300;
class hw_ipcPrivate
{
public:
    hw_ipcPrivate();
    QSharedMemory shmHdr;
    QVector<QSharedPointer<QSharedMemory> > shmModules;
};
hw_ipcPrivate::hw_ipcPrivate()
    : shmHdr("SHM_HWS_HDR_KEY")
{
}


hw_ipc::hw_ipc(AccessMode mode)
    : d(new hw_ipcPrivate()),
    hdrAccessMode(mode),
    dataAccessMode(ReadWrite),
    hws(new hws_t(attach_hws_hdr(mode)))
{
}

hw_ipc::~hw_ipc()
{
    delete hws;
    delete d;
}

hws_hdr_t *hw_ipc::attach_hws_hdr(AccessMode mode)
{
    Q_ASSERT(d);
    if (mode == ReadOnly) {
        if (!d->shmHdr.attach(QSharedMemory::ReadOnly)) {
            QString str = QString("hw_ipc: failed to attach memory segment %1").arg(d->shmHdr.key());
            qDebug() << str;
            throw exc(str.toStdString());
        }
        return static_cast<hws_hdr_t *>(d->shmHdr.data());
    } else {
        size_t hws_hdr_sz = sizeof(hws_hdr_t);
        if (!d->shmHdr.attach()) {
            if (!d->shmHdr.create(hws_hdr_sz)) {
                QString str = QString("failed to create memory segment %1 with size %2: %3")
                              .arg(d->shmHdr.key())
                              .arg(hws_hdr_sz)
                              .arg(d->shmHdr.errorString());
                qDebug() << str;
                throw exc(str.toStdString());
            }
        }
        hws_hdr_t *hdr = static_cast<hws_hdr_t *>(d->shmHdr.data());
        Q_ASSERT(hdr);
        hws_initialize_hdr(hdr);
        return hdr;
    }
}

common_setup_t *hw_ipc::get_data_ptr(int n)
{
    return hws->get_data_ptr(n);
}

size_t hw_ipc::get_data_size(int n) const
{
    return hws->get_data_size(n);
}

int hw_ipc::alloc(size_t s)
{
    if (!hws->hasSpace()) {
        QString str = QString("hw_ipc::alloc - no free entries");
        qDebug() << str;
        throw exc(str.toStdString());
    }
    QSharedPointer<QSharedMemory> shmEntry(new QSharedMemory(QString("SHM_HWS_ENTRY_%1").arg(hws->count())));
    if (!shmEntry->attach()) {
        if (!shmEntry->create(s)) {
            QString str = QString("failed to create memory segment %1 with size %2: %3")
                          .arg(shmEntry->key())
                          .arg(s)
                          .arg(shmEntry->errorString());
            qDebug() << str;
            throw exc(str.toStdString());
        }
    }

    void *p = shmEntry->data();
    Q_ASSERT(p);
    memset(p, 0, s);
    d->shmModules.append(shmEntry);
    return hws->addEntry(p, s);
}

SharedSetupBackend::SharedSetupBackend()
    : hw_ipc(ReadWrite)
{
}

SharedSetupBackend::~SharedSetupBackend()
{
    clear();
}

SharedSetupFrontend::SharedSetupFrontend()
    : hw_ipc(ReadOnly)
{
}

SharedSetupFrontend::~SharedSetupFrontend()
{
}

void hw_ipc::attach()
{
    bool any_readonly = false;
    const hws_hdr_t *hdr = hws->hdr;
    qDebug() << QString("hw_ipc::attach - %1 entries").arg(hws->count());
    for (int i=0; i<hws->count(); i++) {
        const hws_ent_t *ent = &hdr->ent[i];
        if (!hws_ent_valid(ent)) {
            QString str = QString("hw_ipc::attach - invalid entry #%1").arg(i);
            qDebug() << str;
            throw exc(str.toStdString());
        }
        QSharedPointer<QSharedMemory> shmEntry(new QSharedMemory(QString("SHM_HWS_ENTRY_%1").arg(i)));
        bool code = shmEntry->attach();
        if (shmEntry->error() == QSharedMemory::PermissionDenied) {
            any_readonly = true;
            code = shmEntry->attach(QSharedMemory::ReadOnly);
        }
        if (!code) {
            QString str = QString("failed to attach memory segment %1: %3")
                          .arg(shmEntry->key())
                          .arg(shmEntry->errorString());
            qDebug() << str;
            throw exc(str.toStdString());
        }
        d->shmModules.push_back(shmEntry);
        void *p = shmEntry->data();
        Q_ASSERT(p);
        hws->attachData(i, static_cast<common_setup_t *>(p));
    }
    if (any_readonly) {
        dataAccessMode = ReadOnly;
        qWarning() << "hws readonly";
    }
}

void hw_ipc::detach()
{
    clear();
}

void hw_ipc::clear()
{
    if (hdrAccessMode == ReadOnly)
        return;
    Q_ASSERT(hws);
    hws_hdr_t *hdr = hws->hdr;
    Q_ASSERT(hdr);
    //    const int nent = hws->count();
    hdr->nent=0;
    //    for (int i = 0; i < nent; i++) {
    //        hws->data[i] = NULL; // invalidate data ptr
    //        memset(&hdr->ent[i], 0, sizeof(hws_ent_t)); // invalidate entry
    //    }
    d->shmModules.clear(); // automatically detach
}

int hw_ipc::get_nent() const
{
    return hws->count();
}

//void hw_ipc::print() const
//{
//    hws->print();
//}

