
#include "hal/hw_manager.h"

#include <QDebug>

#include "daqcommon/daqexc.h"
#include "daqcommon/daq.h"
#include "daqcommon/daq_ipc.h"
#include "daqcommon/hal_types.h"
#include "hal/base_mod.h"
#include "base_mod_factory.h"

hw_manager::hw_manager()
    : daq_hal_initialized(false)
{
}

void hw_manager::print_summary() const
{
    qDebug() << "+---------- hw_manager summary ----------";
    qDebug() << "+   Number of modules: "<< mods.size();

    Q_FOREACH(const base_mod *p, mods) {
        p->print_info();
    }
    qDebug() << "+---------- end --------------------------";
    //  print_modules();// print brief module setup (too much text...)
    int nc_adc = get_nch_byfam(MF_ADC);
    int nc_tdc = get_nch_byfam(MF_TDC);
    int nc_tqdc = get_nch_byfam(MF_TQDC);
    int nc_trig = get_nch_byfam(MF_TRIG);
    int nc_dch = get_nch_byfam(MF_DCH);
    int nc_msc = get_nch_byfam(MF_MSC);
    qDebug() << QString("HAL: %1 ADC, %2 TDC, %3 TQDC, %4 TRIG, %5 DCH, %6 MSC channels")
                .arg(nc_adc).arg(nc_tdc).arg(nc_tqdc).arg(nc_trig).arg(nc_dch).arg(nc_msc);
}

void hw_manager::delete_modules()
{
    Q_FOREACH(base_mod *p, mods) {
        delete p;
    }
    mods.clear();
}

void hw_manager::print_modules() const
{
    qDebug() << "+--------- Modules ---------";
    Q_FOREACH(const base_mod *p, mods) {
        p->print_info();
        //  	  p.print_setup();
    }
    qDebug() << "+----------- end -----------";
}

int hw_manager::hal_initialized()
{
    return daq_hal_initialized;
}

int hw_manager::get_mod_nch_by_fam_nmod(mfam_t mf, int find_nmod) const
{
    int nch = 0;
    int nmod = 0;
    Q_FOREACH(const base_mod *p, mods) {
        Q_ASSERT(p);
        mfam_t cfam = p->get_fam();
        if (match_mfam(cfam, mf)) {
            if (nmod == find_nmod) {
                nch = p->get_nch();
                break;
            }
            nmod++;
        }
    }
    return nch;
}

bool hw_manager::match_mfam(mfam_t cfam, mfam_t mf) const
{
    return cfam == mf;
}

int hw_manager::get_nch_byfam(mfam_t mf) const
{
    int fch = 0;
    Q_FOREACH(const base_mod *p, mods) {
        Q_ASSERT(p);
        mfam_t cfam = p->get_fam();
        if (match_mfam(cfam, mf)) {
            fch += p->get_nch();
        }
    }
    return fch;
}

int hw_manager::get_nmod_byfam(mfam_t mf) const
{
    int nmod = 0;
    Q_FOREACH(const base_mod *p, mods) {
        Q_ASSERT(p);
        mfam_t cfam = p->get_fam();
        if (match_mfam(cfam, mf)) {
            nmod++;
        }
    }
    return nmod;
}

int hw_manager::get_nch(mtype_t mt) const
{
    int fch = 0;
    Q_FOREACH(const base_mod *p, mods) {
        Q_ASSERT(p);
        if (p->get_id() == mt) {
            fch += p->get_nch();
        }
    }
    return fch;
}

const base_mod *hw_manager::find_mod_by_slot_const(int slot) const
{
    Q_FOREACH(const base_mod *p, mods) {
        Q_ASSERT(p);
        if (p->get_slot() == slot)
            return p;
    }
    return 0;
}

const base_mod *hw_manager::find_mod_by_key_const(ddb_key_t key) const
{
    Q_FOREACH(const base_mod *p, mods) {
        Q_ASSERT(p);
        if (p->get_ddb_key() == key)
            return p;
    }
    return 0;
}

base_mod *hw_manager::find_mod_by_slot(int slot)
{
    Q_FOREACH(base_mod *p, mods) {
        Q_ASSERT(p);
        if (p->get_slot() == slot)
            return p;
    }
    return 0;
}

modlist_t hw_manager::find_mod_by_fam(mfam_t mf)
{
    modlist_t l;
    Q_FOREACH(base_mod *p, mods) {
        Q_ASSERT(p);
        if (match_mfam (p->get_fam(), mf))
            l.push_back(p);
    }
    return l;
}

bool hw_manager::mf_present(mfam_t mf) const
{
    Q_FOREACH(const base_mod *p, mods) {
        Q_ASSERT(p);
        if (p->get_nch() == 0) continue;
        if (match_mfam (p->get_fam(), mf))
           return true;
    }
    return false;
}

bool hw_manager::mt_present(mtype_t mt) const
{
    Q_FOREACH(const base_mod *p, mods) {
        Q_ASSERT(p);
        if (p->get_id() == mt)
           return true;
    }
    return false;
}

/*
modlist_t hw_manager::find_mod_by_fam_const(mfam_t mf) const
{
    modlist_t l;
    Q_FOREACH(const base_mod *p, mods) {
        Q_ASSERT(p);
        if (match_mfam (p->get_fam(), mf))
            l.push_back(p);
    }
    return l;
}
*/
/*
modlist_t hw_manager::find_mod_by_type(mtype_t mt)
{
    modlist_t l;
    Q_FOREACH(base_mod *p, mods) {
        if (p->get_id() == mt)
            l.push_back(p);
    }
    return l;
}
*/
base_mod *hw_manager::find_mod_by_fam_ch (mfam_t mf, int ch, int *mch=0)
{
    int tmch = -1, found = 0, fch = 0;
    base_mod *mod = 0;
    Q_ASSERT(ch >= 0);

    Q_FOREACH(base_mod *p, mods) {
        mod = p;
        Q_ASSERT(p);
        if (match_mfam (p->get_fam(), mf)) {
            int mnch = p->get_nch();
            fch += mnch;
            if (fch > ch) {
                tmch = mnch - (fch - ch);
                found=1;
                break;
            }
        }
    }

    if (!found) {
        return 0;
    }
    if (mch) *mch = tmch;

    return mod;
}

const base_mod *hw_manager::find_mod_by_fam_ch_const (mfam_t mf, int ch, int *mch=0) const
{
    int tmch = -1, found = 0, fch = 0;
    const base_mod *mod = 0;
    Q_ASSERT(ch >= 0);

    Q_FOREACH(const base_mod *p, mods) {
        mod = p;
        Q_ASSERT(p);
        if (match_mfam (p->get_fam(), mf)) {
            int mnch = p->get_nch();
            fch += mnch;
            if (fch > ch) {
                tmch = mnch - (fch - ch);
                found=1;
                break;
            }
        }
    }

    if (!found) {
        return 0;
    }
    if (mch) *mch = tmch;

    return mod;
}
/*
int hw_manager::get_ch_info (mfam_t mf, int ch, mtype_t *mt, int *key) const
{
    int ret = -1, tkey = -1, tmch = -1, found = 0, fch = 0;
    base_mod *mod=0;
    Q_ASSERT(ch >= 0);

    for (modmap_t::const_iterator it = mods.begin(); it != mods.end(); ++it) {
        tkey = (*it).first;
        mod = (*it).second;
        if (match_mfam (mod->get_fam(), mf)) {
            int mnch = mod->get_nch();
            fch += mnch;
            if (fch > ch) {
                tmch = mnch - (fch - ch);
                found=1;
                break;
            }
        }
    }

    if (!found) {
        QString s = QString ("hw_manager::get_ch_info: not found (mf %1, ch %2)")
                    .arg(mf).arg(ch);
        qWarning() << s;
        throw exc(s.toStdString());
    }
    mtype_t tmt = mod->get_id();
    if (mt) *mt = tmt;
    if (key) *key = tkey;

    qDebug() << "Ch info: mf " << mf << ", ch " << ch << ": mt " <<
            module_id_str(tmt) << ", key " << tkey;
    return ret;
}
*/
/*
int hw_manager::get_mid_by_mfch (mfam_t mf, int ch) const
{
    // Get module ID by global channel number
    int mid;
    get_ch_info (mf, ch, 0, &mid);
    return (mid);
}
*/
// check that all channels have equal value
bool hw_manager::is_all_channels_same(const std::string &tag, mfam_t mf) const
{
    std::vector<int> v;
    for (int i=0; i<get_nch_byfam(mf); i++) {
        v.push_back(get_int(tag, mf, i));
    }
    if (v.size() > 1) {
        for (size_t i=1; i<v.size(); i++)
            if (v[i] != v[0]) return false;
    }
    return true;
}

QVariant  hw_manager::get_var(const std::string &tag, mfam_t mf, int ch) const
{
    // if ch < -1, set ch to first channel in module
    int mch = 0;
    if (ch < 0) {
        //	ch = GET_FIRST_CH;
        int nmod = -ch-2;
        ch = get_mod_nch_by_fam_nmod(mf, nmod) * nmod;
    }
    const base_mod *p = find_mod_by_fam_ch_const(mf, ch, &mch);
    if (!p) {
        QString str = QString("hw_manager::get_var: mf not found (tag %1, mf %2, ch %3)")
                      .arg(QString::fromStdString(tag)).arg(mf).arg(ch);
        qWarning() << str;
        throw exc(str.toStdString());
    }
    return p->get_var(tag, mch);
}

int hw_manager::get_int(const std::string &tag, mfam_t mf, int ch) const
{
    // if ch < -1, set ch to first channel in module
    int mch = 0;
    if (ch < 0) {
        //	ch = GET_FIRST_CH;
        int nmod = -ch-2;
        ch = get_mod_nch_by_fam_nmod(mf, nmod) * nmod;
    }
    const base_mod *p = find_mod_by_fam_ch_const(mf, ch, &mch);
    if (!p) {
        QString str = QString("hw_manager::get_int: mf not found (tag %1, mf %2, ch %3)")
                      .arg(QString::fromStdString(tag)).arg(mf).arg(ch);
        qWarning() << str;
        throw exc(str.toStdString());
    }
    return p->get_int(tag, mch);
}

//! ch >=  0 - regular channel
//! ch == -1 - ALL channels
//! ch == -2 - All channels in module #0
//! ch == -3 - All channels in module #1
//! ...
void  hw_manager::set_var(const QVariant &value, const std::string &tag, mfam_t mf, int ch)
{

    if (isReadonly()) {
        QString str = QString("hw_manager::set_var: read-only (tag %1, mf %2, ch %3)")
                .arg(QString::fromStdString(tag)).arg(mf).arg(ch);
        qWarning() << str;
        throw exc(str.toStdString());
    }
    if (ch == SET_ALL_CH) {
        modlist_t l = find_mod_by_fam(mf);
        for (modlist_t::iterator it = l.begin(); it != l.end(); ++it) {
            base_mod *p = (*it);
            p->set_var(value, tag, SET_ALL_CH);
        }
        return;
    } else if (ch < -1) {
        int nmod = -ch-2;
        modlist_t l = find_mod_by_fam(mf);
        int this_mod = 0;
        for (modlist_t::iterator it = l.begin(); it != l.end(); ++it) {
            if (this_mod == nmod) {
                base_mod *p = (*it);
                p->set_var(value, tag, SET_ALL_CH);
                break;
            }
            this_mod++;
        }
    } else {
        int mch = SET_ALL_CH;
        base_mod *p = find_mod_by_fam_ch(mf, ch, &mch);
        if (!p) return;
        if (ch == SET_MOD_CH)
            mch = SET_ALL_CH;
        p->set_var(value, tag, mch);
    }
}

void hw_manager::set_int(int value, const std::string &tag, mfam_t mf, int ch)
{
    if (isReadonly()) {
        QString str = QString("hw_manager::set_int: read-only (tag %1, mf %2, ch %3)")
                      .arg(QString::fromStdString(tag)).arg(mf).arg(ch);
        qWarning() << str;
        throw exc(str.toStdString());
    }
    if (ch == SET_ALL_CH) {
        modlist_t l = find_mod_by_fam(mf);
        for (modlist_t::iterator it = l.begin(); it != l.end(); ++it) {
            base_mod *p = (*it);
            p->set_int(value, tag, SET_ALL_CH);
        }
        return;
    } else if (ch < -1) {
        int nmod = -ch-2;
        modlist_t l = find_mod_by_fam(mf);
        int this_mod = 0;
        for (modlist_t::iterator it = l.begin(); it != l.end(); ++it) {
            if (this_mod == nmod) {
                base_mod *p = (*it);
                p->set_int(value, tag, SET_ALL_CH);
                break;
            }
            this_mod++;
        }
    } else {
        int mch = SET_ALL_CH;
        base_mod *p = find_mod_by_fam_ch(mf, ch, &mch);
        if (!p) return;
        if (ch == SET_MOD_CH)
            mch = SET_ALL_CH;
        p->set_int(value, tag, mch);
    }
}

std::string hw_manager::get_string(const std::string &tag, mfam_t mf, int ch)
{
    const base_mod *p = find_mod_by_fam_ch(mf, ch);
    if (!p) return std::string();
    return p->get_string(tag, ch);
}

void hw_manager::set_string(const std::string &str, const std::string &tag, mfam_t mf, int ch)
{
    if (isReadonly()) {
        QString s = QString ("hw_manager::set_string: read-only (tag %1, mf %2, ch %3)")
                    .arg(QString::fromStdString(tag)).arg(mf).arg(ch);
        qWarning() << s;
        throw exc(s.toStdString());
    }
    if (ch == SET_ALL_CH) {
        modlist_t l = find_mod_by_fam(mf);
        for (modlist_t::iterator it = l.begin(); it != l.end(); ++it) {
            (*it)->set_string(str, tag, ch);
        }
        return;
    }
    base_mod *p = find_mod_by_fam_ch(mf, ch);
    if (!p) return;
    p->set_string(str, tag, ch);
}

void hw_manager::insert_module(int key, common_setup_t *setup, size_t size)
{
    Q_ASSERT(setup);
    Q_ASSERT(size >= sizeof(common_setup_t));
    Q_ASSERT(setup->magic == BASE_SETUP_MAGIC);

    mods.remove(key);
    try {
        mtype_t id = setup->device_id;
        base_mod *p = base_mod_factory::create(id);
        p->set_setup(setup, size);
        mods[key] = p;
    }
    catch(exc &e) {
        qWarning() << e.what();
    }
}
