
#include "hwif_vme.h"

#include <QDebug>

#include <iostream>
#include <sstream>
#include <cerrno>
#include <stdexcept>
#include <map>

#include "daqcommon/daqexc.h"
#include "daqcommon/device_ids.h"
#include "daqcommon/vme_types.h"
#include "daqcommon/vme_util.h"
#include "hal/mod_regs.h"
#include "colorlog/log.h"

#undef log_debug

using colorlog::log_debug;


double hwif_vme::getSysresetPostDelay() {
//    return (0.015+0.030)*2; // *2 for buggy TQDC
    return 2 * 0.080; // ID read delay 80ms, increase by 2 for maximum allowable clock ratio (i.e. set 40, actual 20 MHz)
}

hwif_vme::hwif_vme()
    : vme_nst(VME_SLOTS)
{
}

void hwif_vme::validate_ga(int ga)
{
    Q_ASSERT(ga >= 0);
    Q_ASSERT(ga < vme_nst);
}

void hwif_vme::read_cr(unsigned int ga, struct vme_cr_t *cr)
{
    validate_ga(ga);
    Q_ASSERT(cr);
    quint16 *tmpbuf = static_cast<quint16 *>(static_cast<void *>(cr));

    const int baseaddr = (ga<<19) | VME_CR_BASE;
    log_debug(5, QString("read_cr (ga=%1, address %2)...").arg(ga).arg(baseaddr, 8, 16, QChar('0')).toStdString());

    std::vector<int> addrv;
    for (size_t i = 0; i < (sizeof(struct vme_cr_t) + 1) / 2; i++) {
        addrv.push_back((i << 1) | baseaddr);
    }

    for (size_t i = 0; i < addrv.size(); i++) {
        int data = 0;
        r_crcsr_d16(addrv[i], &data);
        tmpbuf[i] = data;
    }
    log_debug(5, "read_cr Ok for address %08X", baseaddr);
}

void hwif_vme::read_csr(unsigned int ga, struct vme_csr_t *csr)
{
    validate_ga(ga);
    Q_ASSERT(csr);
    quint16 *tmpbuf = static_cast<quint16 *>(static_cast<void *>(csr));

    const int baseaddr = (ga<<19) | VME_CSR_BASE;
    log_debug(5, "read_csr (ga=%d, address %08X)...", ga, baseaddr);

    std::vector<int> addrv;
    for (size_t i = 0; i < (sizeof(struct vme_csr_t) + 1) / 2; i++) {
        addrv.push_back((i << 1) | baseaddr);
    }

    for (size_t i = 0; i < addrv.size(); i++) {
        int data = 0;
        r_crcsr_d16(addrv[i], &data);
        tmpbuf[i] = data;
    }
    log_debug(5, "read_csr Ok for address %08X", baseaddr);
}

void hwif_vme::write_csr(unsigned int ga, const struct vme_csr_t * csr)
{
    validate_ga(ga);
    Q_ASSERT(csr);
    const quint16 *tmpbuf = static_cast<const quint16 *>(static_cast<const void *>(csr));

    const int baseaddr = (ga<<19) | VME_CSR_BASE;

    std::vector<std::pair<int, int> > addrdata;

    for (size_t i = 0; i < (sizeof(struct vme_csr_t) + 1) / 2; i++) {
        addrdata.push_back(std::pair<int, int>((i << 1) | baseaddr, tmpbuf[i]));
    }
    for (size_t i = 0; i < addrdata.size(); i++) {
        const int addr = addrdata[i].first;
        const int data = addrdata[i].second;
        w_crcsr_d16((i << 1) | addr, data);
    }
    log_debug(5, "write_csr Ok for address %08X", baseaddr);
}


//----------------------------------------------------------------------------------

std::map<int, struct vme_cr_t> hwif_vme::find_all_cr()
{
    std::map<int, struct vme_cr_t> found_cr;

    for (unsigned int ga = 0; ga < VME_SLOTS; ga++) {
        struct vme_cr_t cr;

        log_debug(2, "Searching position %d", ga);
        try {
            read_cr(ga, &cr);
        }
        catch (exc_vme_tmout &e) {
            continue;
        }
        if (cr.vendor != VME_CR_VENDOR_ID) {
            log_warn("Position %d, unknown module: vendor ID %02X", ga, cr.vendor);
            continue;
        }
        found_cr[ga] = cr;
    }
    return found_cr;
}

namespace {
    unsigned char crc8(unsigned char val, unsigned char crc)
    {
        unsigned char i = 8;

        crc ^= val;
        do {
            if (crc & 0x01) {
                crc >>= 1;
                crc ^= 0x8C;
            } else {
                crc >>= 1;
            }
        } while (--i);

        return crc;
    }
}

std::list<hwif_discover_t> hwif_vme::discover()
{
    typedef std::map<int, struct vme_cr_t> found_cr_t;
    const found_cr_t &found_cr = find_all_cr();

    // Read CSR, assign base address, write CSR
    std::map<int, struct vme_csr_t> save_csr;

    for(found_cr_t::const_iterator it = found_cr.begin(); it!=found_cr.end(); ++it) {
        const int ga = (*it).first;
        //	const struct vme_cr_t &cr = (*it).second;
        const unsigned int base_a16 = ga<<11;
        struct vme_csr_t csr;
        read_csr(ga, &csr);
        // Write Base Address in CSR
        csr.base = base_a16;
        save_csr[ga] = csr;
        write_csr(ga, &csr);
    }

    // Read serial ID, if available
    std::map<int, int> save_serial_id;
    for(found_cr_t::const_iterator it = found_cr.begin(); it!=found_cr.end(); ++it) {
        const int ga = (*it).first;
        const struct vme_cr_t &cr = (*it).second;
        const struct vme_csr_t &csr = save_csr[ga];
        save_serial_id[ga] = cr.serial; // failover to hw rev. for backward compatibility
        try {
            const int addr = MODULE_REG_SERIAL * 4 + csr.base;
            qint64 data64 = 0;
            unsigned char *d = (unsigned char *)&data64;
            r_a16_d64(addr, &data64);
//            qDebug() << QString("read ID data from %1: %2").arg(addr, 8, 16, QChar('0')).arg(data64, 16, 16, QChar('0'));
            unsigned char crccode = 0;
            for (int i = 0; i<7; i++)
                crccode = crc8(d[i], crccode);

            if (d[0] == 0x00 || d[0] == 0xFF) {
                // no device id
                continue;
            }
            if (d[0] != 0x28) {
                log_err("read serial ID from %08X failed: bad device code", addr);
                continue;
            }
            if (crccode != d[7]) {
                log_err("read serial ID from %08X failed: bad crc (ID=%02X%02X%02X%02X%02X%02X%02X%02X)", addr, d[0], d[1], d[2], d[3], d[4], d[5], d[6], d[7]);
                continue;
            }
            quint64 serial_id = (data64>>8)&0xFFFFFFFF;
//            log_debug(1, "read serial ID from %08X: %08llX", addr, serial_id);
            save_serial_id[ga] = serial_id;
        }
        catch(exc_vme &e) {
        }
    }

    std::list<hwif_discover_t> m;

    for(found_cr_t::const_iterator it = found_cr.begin(); it!=found_cr.end(); ++it) {
        const int ga = (*it).first;
        const struct vme_cr_t &cr = (*it).second;
        const struct vme_csr_t &csr = save_csr[ga];
        const int serial_id = save_serial_id[ga];
//        const char *id_str = module_id_str((mtype_t)cr.type);
        hwif_discover_t pcr(cr, csr, ga, serial_id);
        QString s();
//        log_info("Position %2d, vendor ID %02X, type %02X, rev %02X: ID %02X, serial %08X, A16 base %04X",
//                 pcr.slot, pcr.cr.vendor, pcr.cr.type, pcr.cr.serial, cr.type, pcr.serial_id, pcr.reg_base);
        m.push_back(pcr);
    }
    /*

  while(true) {
        sysreset();
  m.clear();
  foreach_const (found_cr, jt) {
        const int ga = (*jt).first;
        const struct vme_cr_t &cr = (*jt).second;
    const struct vme_csr_t &csr = save_csr[ga];
    const int serial_id_test = save_serial_id[ga];

        bool test_performed = false;
        bool test_ok = true;
        double temp = 0;

    try {
// Test for serial read
          if (true) {
          if (cr.type == MODULE_ID_TQDC && serial_id_test < 0x16000) {
                test_ok = false;
                log_err("slot %d: Bad serial ID %08X", ga, serial_id_test);
//  		  throw std::runtime_error("Bad serial ID");
          }
          for (int i=0; i<50; i++) {
//  		printf(".");
//  		fflush(stdout);
        test_performed = true;
                int data2 = 0;
                start_communication(); // SYSRESET
// Check CR
                struct vme_cr_t cr2;
                read_cr (ga, &cr2);
                if (cr != cr2) {
                  test_ok = false;
                  log_err("slot %d CR read test FAILED: first read %s, now read %s", ga, cr_str(cr).c_str(), cr_str(cr2).c_str());
//  		  throw std::runtime_error("CR read test FAILED");
                  break;
                }
// Recover all CSR after SYSRESET
                foreach_const(save_csr, it) {
                  const int cur_ga = (*it).first;
                  const struct vme_csr_t &cur_csr = (*it).second;
                  write_csr (cur_ga, &cur_csr);
                }

// Check all CSR
                foreach_const(save_csr, it) {
                  const int cur_ga = (*it).first;
                  const struct vme_csr_t &cur_csr = (*it).second;
                  struct vme_csr_t test_csr;
                  read_csr (cur_ga, &test_csr);
                if (cur_csr != test_csr) {
                  test_ok = false;
                  log_err("slot %d CSR read test FAILED: first read %s, now read %s", cur_ga, csr_str(test_csr).c_str(), csr_str(cur_csr).c_str());
                  throw std::runtime_error("CSR read test FAILED");
                  break;
                }
                } // end foreach

                if (cr.type == MODULE_ID_TQDC) {
        const int addr_id = MODULE_REG_SERIAL_1 * 4 + csr.base;
                r_a16_d32(addr_id, &data2);
        log_debug(1, "slot %d, read serial ID from %08X: %08X", ga, addr_id, data2);
                if (data2 != serial_id_test) {
                  test_ok = false;
                  log_err("slot %d: ID changed on step %d: %08X to %08X", ga, i, serial_id_test, data2);
                  serial_id_test = data2;
//  		  throw std::runtime_error("ID changed");
//  		  break;
                }
                const int addrt = MODULE_REG_TEMPERATURE_1 * 4 + csr.base;
                r_a16_d32(addrt, &data2);
        int raw_temp = data2 & 0x7FF;
        temp = 0.0625 * raw_temp;
        log_debug(1, "slot %d: read temp from %08X: %.1fC, %08X", ga, addrt, temp, data2);
                if (temp < 10 || temp > 70) {
                  test_ok = false;
                  log_err("slot %d: invalid temp on step %d: %.1fC, %08X", ga, i, temp, data2);
//  		  throw std::runtime_error("invalid temp");
//  		  break;
                }

          }
          }

//	  if (serial_id_test != -1 && serial_id_test != 0) {
//  		serial_id = serial_id_test;
//  	  }
    }
    }
    catch (exc_vme_tmout &e) { }
    if (test_performed) {
    if (test_ok) {
          log_info("slot %d - test Ok, temp %.1fC", ga, temp);
    } else {
          log_err("slot %d - test FAILED", ga);
    }
    }

    }

  }
*/
    return m;
}
