
/*
 * MSC backend (hardware) implementation
 */

#include "msc_bmod.h"

#include <QDebug>
#include <QString>
#include <QFile>
#include <QTextStream>
#include <QDateTime>
#include <cmath>

#include "daqcommon/bitop.h"
#include "daqcommon/daq_ipc.h"
#include "daqcommon/daqexc.h"
#include "modules/msc_types.h"
#include "debug_hal.h"

//#include "msc_calibration.h"
#include "decoder/CalibrationData.h"

namespace {
    enum MSC_REG {
        MSC_REG_DACSET    = 5,
        MSC_REG_STEPTIMER = 6,
        MSC_REG_TEST      = 7,
        MSC_REG_COMPSTAT  = 0xA,
        MSC_REG_RAMSIZE   = 0xB,
        MSC_REG_ADC_RAWSAMPLE = 0x20
    };
    const int MSC_N_COUNTERS = 16; // FIXME
}

void msc_bmod::dac_exec(const std::vector<int> &cmd)
{
    std::vector<int> regs(cmd.size(), MSC_REG_DACSET);
//    for(int i=0; i<cmd.size(); i++)
//        info_mod(QString("DAC write: %1").arg(cmd[i], 8, 16, QChar('0')));
    set_reg_v(regs, cmd);
}

//--------------------------------------------------------------------------
// DAC commands
//--------------------------------------------------------------------------

//! command: set DAC input register for channel ch to value val
std::vector<int> msc_bmod::dac_set_cmd (int ch, int val)
{
    int n, a;
    n = (ch / 8) & 0x01; // only one bit allowed
    a = ch % 8;
    val &= 0x0FFF;
    std::vector<int> cmd;
    cmd.push_back((n<<16) | (a << 12) | val);
    //  cmd.push_back(tmpcmd);
//    printf("DAC: ch %2d, val %4d, cmd=%08X\n", ch, val, cmd[0]);
    return cmd;
}

//! LDAC# low - This sets LDAC permanently low, thus
//! allowing the DAC registers to be updated continuously.
const int LDAC_Low = 0xA000;

//! LDAC# high - The DAC registers are latched,
//! and the input registers may change without
//! affecting the contents of the DAC registers.
const int LDAC_High = 0xA001;

//! LDAC# signle update - This causes a single pulse on LDAC#,
//! thus updating the DAC registers once.
const int LDAC_Update = 0xA002;

std::vector<int> msc_bmod::dac_reset_cmd ()
{
    std::vector<int> cmd;
    for (int i = 0; i < 2; i++) {
        cmd.push_back((i<<16) | 0xF000); //!< Data and Control Reset
        cmd.push_back((i<<16) | 0x8000 | ((0) << 4) | ((0) << 2) | 0); //!< gain, buf, vdd
//        cmd.push_back((i<<16) | LDAC_Low);
        cmd.push_back((i<<16) | LDAC_High);
    }
    return cmd;
}

std::vector<int> msc_bmod::dac_load_cmd ()
{
    std::vector<int> cmd;
//    for (int i = 0; i < 2; i++) {
//        cmd.push_back((i<<16) | LDAC_Update);
//    }
    cmd.push_back((0<<16) | LDAC_Update);
    cmd.push_back((1<<16) | LDAC_Update);
    cmd.push_back((0<<16) | LDAC_Update);
    cmd.push_back((1<<16) | LDAC_Update);
    return cmd;
}

std::vector<int> msc_bmod::dac_set_all_cmd(int dac)
{
    std::vector<int> cmd;
    for (int c=0; c<MSC_NCH; c++) {
        const std::vector<int> &tmp = dac_set_cmd(c, dac);
        cmd.insert(cmd.end(), tmp.begin(), tmp.end());
    }
    return cmd;
}

//--------------------------------------------------------------------------
// DAC executables
//--------------------------------------------------------------------------

//! reset all DACs to power-on state
void msc_bmod::dac_reset ()
{
    dac_exec(dac_reset_cmd());
}

//! load DAC output registers from input registers
void msc_bmod::dac_load ()
{
    dac_exec(dac_load_cmd());
}

//! set DAC channel ch to val
void msc_bmod::dac_set (int ch, int val)
{
    dac_exec(dac_set_cmd(ch, val));
}

void msc_bmod::dac_set_all(int dac)
{
    dac_exec(dac_set_all_cmd(dac));
}

void msc_bmod::dac_set_load(int ch, int dac)
{
    std::vector<int> cmd = dac_set_cmd(ch, dac);
    const std::vector<int> tmp = dac_load_cmd();
    cmd.insert(cmd.end(), tmp.begin(), tmp.end());
    dac_exec(cmd);
}
void msc_bmod::dac_set_all_load(int dac)
{
    std::vector<int> cmd = dac_set_all_cmd(dac);
    const std::vector<int> tmp = dac_load_cmd();
    cmd.insert(cmd.end(), tmp.begin(), tmp.end());
    dac_exec(cmd);
}

void msc_bmod::set_thr (int force_zero)
{
    QString sf, sm;

    for (int c = 0; c < MSC_NCH; c++) {
        const msc_setup_t *s = static_cast<const msc_setup_t *>(setup);
//        bool neg = (c<8)?s->neg_a:s->neg_b;
//        debug_mod(MSC_DEBUG_LEVEL, QString("ch %1, neg=%2").arg(c, 2).arg(neg));
        double thr = s->thr[c];
//        if (neg) thr *= -1.;
        int dac = dac_value ((force_zero?0:thr) + thr_offset(c));
        if (dac < 0)
            dac = 0;
        if (dac > dac_max())
            dac = dac_max();
        thr = dac_mv(dac);
        dac_set(c, dac);

        sf += QString("%1 ").arg(dac, 4);
        sm += QString("%1 ").arg(thr, 6, 'f', 1);
    }
    dac_load();
    debug_mod (MSC_DEBUG_LEVEL, QString("Thr raw %1: %2").arg(force_zero?"force_zero":"").arg(sf));
    debug_mod (MSC_DEBUG_LEVEL, QString("Thr mV  %1: %2").arg(force_zero?"force_zero":"").arg(sm));
}

void msc_bmod::load_registers(int force_zero)
{
    dac_reset ();
    set_thr (force_zero); // dac_load inside
    set_reg(MSC_REG_STEPTIMER, 800);
}

bool msc_bmod::detect ()
{
    info_mod("detect");
    msc_setup_t *s = static_cast<msc_setup_t *>(setup);

    // test DACs
    if (0) {
        int d=0;
        while (true) {
            dac_set_all(d);
            d += 256;
            d &= 0xFFF;
            int rd = 0;
            get_reg(MSC_REG_COMPSTAT, &rd);
            debug_mod(0, QString("COMPSTAT = %1").arg(rd, 16, 2, QChar('0')));
        }
    }

    if ((true != testreg32(get_reg_base() + MSC_REG_TEST*4, MOD_SELFTEST_LOOPS))) {
        err_mod("Register test failed");
        return false;
    }

    quint32 ch_count = 0;
    quint32 ram_size = 0;
/*
//    int patt = 0x5a5a5a5a;
    int patt = 1;
    for(int i=0; i<1;i++) {
        int addr = 0;
        int wd = patt;
//        patt = ~patt;
        if (patt != 0) patt = patt * 2; else patt = 1;
        w_a24_d32((get_reg_base()<<8)+addr, wd);
        const int read_try_count = 100;
        std::vector<int> v;
        for(int i=0; i<read_try_count; i++) {
            int rd = 0;
            r_a24_d32((get_reg_base()<<8)+addr, &rd);
            v.push_back(rd);
        }
        std::map<int, int> m;
        for(int i=0; i<v.size(); i++) {
            m[v[i]]++;
        }
        if (m.size() > 1) {
//            warn_mod(QString("MemIO test: unstable read"));
            for(std::map<int, int>::const_iterator it=m.begin(); it != m.end(); ++it) {
                int value = it->first;
                int count = it->second;
                warn_mod(QString("MemIO test: unstable read: %1  - %2 times")
                         .arg((quint32)value, 8, 16, QChar('0'))
                         .arg(count)
                         );
            }
        }
        int rd = v[0];
        warn_mod(QString("MemIO test: wrote %1, read %2, xor %3")
                 .arg((quint32)wd, 8, 16, QChar('0'))
                 .arg((quint32)rd, 8, 16, QChar('0'))
                 .arg((quint32)(rd^wd), 8, 16, QChar('0'))
                 );
        usleep(100000);
    }
*/

    // test data bus
    try {
        if ((true != testreg32(get_reg_base() + MSC_REG_TEST*4, 5))) {
            err_mod("Register test failed");
            goto nx;
        }
        int d = 0;
        get_reg(MSC_REG_RAMSIZE, &d);
        int ch_radix = ((unsigned)d >> 8) & 0xFF;
        int ram_radix = (unsigned)d & 0xFF;
        err_mod(QString("MSC_REG_RAMSIZE = %1, ch_radix = %2, ram_radix = %3")
                .arg(d, 8, 16, QChar('0'))
                .arg(ch_radix)
                .arg(ram_radix)
                );
        ch_count = 1 << ch_radix;
        ram_size = 1 << ram_radix;
        if (ch_count == 0 || ch_count > MSC_NCH) {
            err_mod(QString("Detect: Invalid channel count %1").arg(ch_count));
            ch_count = 2; // safe default
        }
        if (ram_size == 0 || ram_size >= 16*1024*1024) {
            err_mod(QString("Detect: Invalid memory size %1").arg(ram_size));
            ram_size = 1024; // safe default
        }
    }
    catch (exc &e) {
        err_mod(e.what());
        goto nx;
    }

    debug_mod (MSC_DEBUG_LEVEL, QString("Detect: %1 channels x %2 samples").arg(ch_count).arg(ram_size));
    s->nch = ch_count;
    s->ndev = MSC_NDEV;
    return true;
    nx:
    err_mod ("Detect: not found");
    return false;
}

void msc_bmod::reset ()
{
    base_bmod::reset();
//    debug_mod(MSC_DEBUG_LEVEL, "msc_bmod reset");
//    debug_mod(MSC_DEBUG_LEVEL, "Reset Ok");
}

/**************************************************************************/

bool msc_bmod::selftest ()
{
    return testreg32(get_reg_base() + MSC_REG_TEST * 4, MOD_SELFTEST_LOOPS);
}

void msc_bmod::preparerun ()
{
    load_registers();
    debug_mod(MSC_DEBUG_LEVEL, "preparerun");
}

void msc_bmod::end_of_burst_readout ()
{

    // encode
    unsigned int wc = 0;

    append_rawdata_mhdr();
/*
    const msc_counters_t &cnt = get_counters();
    for (size_t n=0; n<cnt.size(); n++) {
        unsigned int cnthi, cntlo;
        cnthi = cnt[n] >> 16;
        cntlo = cnt[n] & 0xFFFF;
        append_rawdata( (0<<28)|(n<<19)|cnthi); wc++;
        append_rawdata( (1<<28)|(n<<19)|cntlo); wc++;
    }
*/
    append_rawdata_status(); wc++;
    append_rawdata_mtrl(wc);
}

msc_counters_t msc_bmod::get_counters ()
{
    msc_counters_t cnt(MSC_N_COUNTERS);
    r_blk_a24_d32(0x800, cnt.size(), static_cast<int *>(static_cast<void *>(&cnt[0])));

    bool countersAvailable = false;
    for (size_t n=0; n<cnt.size(); n++) {
        if (cnt[n] != 0xFFFFFFFFu)
            countersAvailable = true;
    }
    if (countersAvailable)
        return cnt;
    else
        return msc_counters_t();
}

namespace {
    const int cal_dac_delta_mv = 1000;
    const int cal_dac_min = msc_mod::dac_value(-cal_dac_delta_mv);
    const int cal_dac_max = msc_mod::dac_value(cal_dac_delta_mv);
    const size_t cal_steps = cal_dac_max-cal_dac_min;

    int cal_dac_value(int step) {
        return cal_dac_min+step;
    }
}

void msc_bmod::calibration()
{
    //  load_registers(FORCE_ZERO);
    //  base_bmod::reset();
    //  usleep(100000);

    update_status();
//    if (!ttc_clock_ok()) {
//        warn_mod("Clock problem, calibration not performed");
//        return;
//    }
    double currentTemperature = get_temperature();
//    const QVariant &currentTemperature = getCalibrationCondition("current_temperature");
//    setCalibrationCondition("temperature", currentTemperature);

    std::vector<std::vector<double> > comp_by_ind(MSC_NCH, std::vector<double>(cal_steps, 0));
    unsigned int cmpbits_min = 0xFFFFFFFF;
    unsigned int cmpbits_max = 0;

    for (size_t i=0; i<cal_steps; i++) {
        const int dac = cal_dac_value(i);
        //    const double thr = dac_mv (dac);
        //	dac_set_all_load(0);
        //    dac_reset ();
        dac_set_all_load(dac);
        std::vector<double> sample_sum(MSC_NCH, 0.);
        size_t num_samples = 10;
        // new style
        //	std::vector<int> regs(num_samples, MSC_REG_COMPSTAT);
        //	std::vector<int> cmpbits = get_reg_blk(regs);
        const std::vector<int> regs(num_samples, MSC_REG_COMPSTAT);
        const std::vector<int> regdata = get_reg_v (regs);
//        qDebug() << "dac " << dac << ": " << QVector<int>::fromStdVector(regdata);
        if (regdata.size() != num_samples) {
            err_mod(QString("get_reg_v failed in calibration: requested %1 samples, got %2").arg(num_samples).arg(regdata.size()));
            return;
        }
//        debug_mod(0, QString("COMP=%1").arg(regdata[0], 8, 16, QChar('0')));
        size_t got_samples = regdata.size();
        for (size_t s=0; s<got_samples; s++) {
            unsigned int cmpbits = regdata[s];
            cmpbits_min &= cmpbits;
            cmpbits_max |= cmpbits;
            for (int c=0; c<MSC_NCH; c++) {
                if (cmpbits & (1<<c))
                    sample_sum[c]++;
            }
        }
        /*
// old style
        size_t got_samples = num_samples;
        for (size_t s=0; s<num_samples; s++) {
          int cmpbits = 0;
          get_reg(MSC_REG_COMPSTAT, &cmpbits);
          for (int c=0; c<MSC_NCH; c++) {
                if (cmpbits & (1<<c))
                  sample_sum[c]++;
          }
        }
*/
        for (int c=0; c<MSC_NCH; c++) {
            sample_sum[c] /= got_samples;
            comp_by_ind[c][i] = sample_sum[c];
        }
    }
    bool calibration_present = true;
    if (cmpbits_min == cmpbits_max) {
        calibration_present = false;
        err_mod(QString("threshold calibration failed, search range %1 .. %2 mV, cmpbits=%3")
                 .arg(dac_mv(cal_dac_min), 5, 'f', 1)
                 .arg(dac_mv(cal_dac_max), 5, 'f', 1)
                 .arg(cmpbits_min, 8, 26, QChar('0')));
    }

    if (calibration_present) {
        std::vector<double> thr_cal(MSC_NCH, 0);
        double syst_err = 0;
        for (int c=0; c<MSC_NCH; c++) {
            bool found = false;
            for (size_t i=1; i<cal_steps; i++) {
                double samp = comp_by_ind[c][i];
                double prev_samp = comp_by_ind[c][i-1];
                if (prev_samp >= 0.5 && samp < 0.5) {
                    int dac = cal_dac_value(i-1);
                    double thr = dac_mv(dac);
                    qDebug() << "***" << c << dac << thr;
                    double manual_thr = thr_offset(c);
                    double err = thr-manual_thr;
                    syst_err += err;
                    thr_cal[c] = thr;
                    found = true;
                    break;
                }
            }
            if (!found) {
                err_mod(QString("ch %1 threshold calibration failed, search range %2 .. %3 mV")
                         .arg(c, 2)
                         .arg(dac_mv(cal_dac_min), 5, 'f', 1)
                         .arg(dac_mv(cal_dac_max), 5, 'f', 1));
            }
        }
        syst_err /= MSC_NCH;

        for (int c=0; c<MSC_NCH; c++) {
            double err = thr_cal[c]-thr_offset(c);
            if (fabs(err) >= 0.5)
                notice_mod(QString("ch %1, threshold calibration %2 mV, factory %3 mV, error %4 mV")
                           .arg(c, 2)
                           .arg(thr_cal[c], 5, 'f', 1)
                           .arg(thr_offset(c), 5, 'f', 1)
                           .arg(thr_cal[c]-thr_offset(c), 5, 'f', 1));
        }
        if (syst_err > 5.0)
            err_mod(QString("threshold systematic error %1 mV").arg(syst_err, 0, 'f', 2));
        else if (syst_err > 2.0)
            warn_mod(QString("threshold systematic error %1 mV").arg(syst_err, 0, 'f', 2));
        else if (syst_err > 0.5)
            notice_mod(QString("threshold systematic error %1 mV").arg(syst_err, 0, 'f', 2));

        setCalibrationValue("thr_offset/temp", currentTemperature);
        setCalibrationValue("thr_offset/date", QDateTime::currentDateTime().toUTC().toString());
        for (int c=0; c<MSC_NCH; c++) {
            setCalibrationValue(QString("thr_offset/%1").arg(c), thr_cal[c]);
        }
    }
    calibrationData->sync();
}

double msc_bmod::thr_offset(int c)
{
    return getCalibrationValue(QString("thr_offset/%1").arg(c)).toDouble();
}

int msc_bmod::get_regnumber_test() const
{
    return MODULE_REG_TEST;
}
