
/*
 * TDC backend (hardware) implementation
 */

#include "tdc_bmod.h"

#include <QString>

#include "hptdc/hptdc.h"
#include "daqcommon/bitop.h"
#include "modules/tdc_types.h"
#include "vmedaq_jtag_engine_hptdc.h"
#include "vmedaq_jtag_impl.h"
#include "fvme/vmedaq_jtag_impl_fvme.h"
#include "hal/base_hwif.h"
#include "debug_hal.h"

namespace tdc {
    // custom registers (32-bit)
    const unsigned int MODULE_REG_TEST = 1;
    const unsigned int MODULE_REG_MASK_0 = 10;
    const unsigned int MODULE_REG_MASK_1 = 11;
}

//--------------------------------------------------------------------------

tdc_bmod::tdc_bmod(const bmod_props &props)
    : base_bmod(props),
    jtagImpl(new Jtag_Impl_Fvme(
            Jtag_Impl_Options_Vme(props.hwif,
                                  get_reg_base() | 4*reg_index_jtag(),
                                  get_reg_base() | 4*reg_index_jtagb()))),
    jtag(new JTAG_Engine_HPTDC(*jtagImpl))
{
    hptdc_lock = false;
    hptdc_idcode_test = false;
    hptdc_bypass_test = false;
    hptdc_err = false;
    hptdc_warn = false;
    hptdc_initialized = false;
}

tdc_bmod::~tdc_bmod()
{
    delete jtag;
    delete jtagImpl;
}

//--------------------------------------------------------------------------

bool tdc_bmod::check_dll (int *status)
{
    const tdc_setup_t *s = static_cast<const tdc_setup_t *>(setup);

    int locked = 1;
    const std::vector<hptdc_status_t> &st = jtag->hptdc_get_status (s->ndev);
    for (int i = 0; i < s->ndev; i++)
        if (!st[i].bit.dll_lock)
            locked = 0;

    if (status) {
        *status = 0;
        for (int i = 0; i < s->ndev; i++)
            if (st[i].bit.dll_lock)
                *status |= (1 << i);
        else
            *status &= ~(1 << i);
    }
    if (!locked) {
        for (int i = 0; i < s->ndev; i++) {
            if (st[i].bit.dll_lock)
                debug_mod(5, QString("HPTDC #%1: DLL LOCKED").arg(i));
            else
                debug_mod(1, QString("HPTDC #%1: DLL unlocked").arg(i));
        }
        debug_mod (1, QString("Some of %1 HPTDC chips have DLL unlocked, possible setup or clock problem").arg(s->ndev));
        hptdc_lock = false;
        return false;
    }
    hptdc_lock = true;
    return true;
}

bool tdc_bmod::reset_hptdc_chips_pll ()
{
    const tdc_setup_t *s = static_cast<const tdc_setup_t *>(setup);
    hptdc_control_t default_ctrl;
    default_ctrl.bit.enable_pattern = 0x4;
    default_ctrl.bit.Global_reset = 0;
    default_ctrl.bit.enable_channel = 0;
    default_ctrl.bit.dll_reset = 1;
    default_ctrl.bit.pll_reset = 1;

    std::vector<hptdc_control_t> ctrl(s->ndev, default_ctrl);
    jtag->hptdc_write_control(ctrl);
    waitMs(1);
    for (int i = 0; i < s->ndev; i++)
        ctrl[i].bit.pll_reset = 0;
    jtag->hptdc_write_control(ctrl);
    waitMs(1);
    for (int i = 0; i < s->ndev; i++)
        ctrl[i].bit.dll_reset = 0;
    jtag->hptdc_write_control(ctrl);
    waitMs(1);
    return check_dll (NULL);
}

void tdc_bmod::reset_hptdc_chips ()
{
    const tdc_setup_t *s = static_cast<const tdc_setup_t *>(setup);

    hptdc_control_t default_ctrl;
    default_ctrl.bit.enable_pattern = 0x4;
    default_ctrl.bit.Global_reset = 1;
    default_ctrl.bit.enable_channel = 0;
    default_ctrl.bit.dll_reset = 0;
    default_ctrl.bit.pll_reset = 0;
    std::vector<hptdc_control_t> ctrl(s->ndev, default_ctrl);

    //  clear_error_flag();

    jtag->hptdc_write_control(ctrl);
    for (int i = 0; i < s->ndev; i++)
        ctrl[i].bit.Global_reset = 0;
    jtag->hptdc_write_control (ctrl);

    TdcResIndex resindex = TDC_RES_INDEX_100;
    hptdc_setup_get_resindex(s->tdcsetup[0], &resindex);
    bool hires = tdc_res_vh(resindex);

    //  int nch_by_dev = (s->ndev > 0) ? (s->nch / s->ndev):0;
    for (int hptdc_num = 0; hptdc_num < s->ndev; hptdc_num++) {
        ctrl[hptdc_num].bit.enable_pattern = 0x5; // turn ON
        int chmask = 0;
        for (int hptdc_ch = 0; hptdc_ch < 32; hptdc_ch++) { // hptdc has 32 channnels
            int ch = map_chtdc2inp(hptdc_num, hptdc_ch);
            //   	  printf("mapped hptdc_num=%d, hptdc_ch=%d to ch=%d, en=%d\n", hptdc_num, hptdc_ch, ch, s->ch_enable[ch]);
            if (ch != -1) {
                if (s->ch_enable[ch]) {
                    if (hires)
                        chmask |= ((0xF << hptdc_ch)); // need 0xF to enable hi-res mode
                    else
                        chmask |= ((0x1 << hptdc_ch));
                }
                //        printf("tdc %d, subch %d (tdc_ch %d): enable: %d, ch_mask=%08X\n", i, j, tdc_ch, s->ch_enable[tdc_ch], chmask);
            }
        }
        //    printf("========== %d: mask=%08X\n", hptdc_num, chmask);
        //	log_debug(TDC_DEBUG_LEVEL+2, "MASK=%08X", chmask);
        ctrl[hptdc_num].bit.enable_channel = chmask;
        //    ctrl[hptdc_num].bit.enable_channel = 0xFFFFFFFF;
    }
    jtag->hptdc_write_control(ctrl);
}

std::vector<quint32> tdc_bmod::jtag_readout ()
{
    std::vector<quint32> data;
    // TODO: hptdc jtag readout
    /*
  const tdc_setup_t *s = static_cast<const tdc_setup_t *>(setup);
  hptdc_readout_t readout[TDC_MAXDEV];
  int i, any_ready;

  while (*count < limit) {
    hptdc_get_jtag_readout (readout, s->ndev);
    any_ready = 0;
    for (i = 0; i < s->ndev; i++)
      if (readout[i].bit.ready) {
        data[(*count)++] = readout[i].bit.data;
        if (*count >= limit)
          return;
        any_ready = 1;
      }
    if (!any_ready)
      return;
  }
*/
    return data;
}

/*
void tdc_print_status (int d)
{

  printf("TQDC status %08X %s%s%s%s  wordcount %d\n",
          d,
          d & TQDC_STAT_BIT_SPILL ? "SPILL " : "NO_SPILL ",
          d & TQDC_STAT_BIT_EVBUSY ? "EV_BUSY " : "",
          d & TQDC_STAT_BIT_CLKLOCK ? "CLK_LOCK " : "CLK_UNLOCKED ",
          d & TQDC_STAT_BIT_FIFOEMPTY ? "FIFO_EMPTY " : "FIFO_HAS_DATA ", d & 0x00000FFF);
}

int tqdc_get_status (const tdc_setup_t * s, int *st)
{

  int d, ret;

  if (!s) {
    log_debug(0, "NULL setup");
    return (-EINVAL);
  }

  if ((ret = vme_setam (vme_reg_am())))
    return (ret);
  if ((ret = vme_ar32 (vme_reg_base() + TQDC_STAT_ADDR_OFFSET, &d)))
    return (ret);               // read status
  if (st)
    *st = d;
//  tqdc_print_status(d);
  return 0;
}

int tqdc_check_runstatus (const tdc_setup_t * s)
{

  int d, ret;

  if (!s) {
    log_debug(0, "NULL setup");
    return (-EINVAL);
  }
  if ((ret = tqdc_get_status (s, &d)))
    return (ret);               // read status
  if (!(d & TQDC_STAT_BIT_CLKLOCK) ||
//    (d&TQDC_STAT_BIT_TDCERR) ||
      (d & TQDC_STAT_BIT_EVBUSY)
//  ||  (d&TQDC_STAT_BIT_EVREADY)
    ) {
    log_debug(0, "check_runstatus: Bad state\n");
    tqdc_print_status (d);
    return (-EREMOTEIO);
  }
  return 0;
}
*/
bool tdc_bmod::detect ()
{
    tdc_setup_t *s = static_cast<tdc_setup_t *>(setup);
    s->nch = tdc_nch();
    s->ndev = tdc_ndev();

    //  update_jtag_address();
    jtag->jtag_reset ();
    const std::vector<quint32> &idcodes = jtag->jchain_detect_ndev();
    for (int i = 0; i < 10; i++) {
        const std::vector<quint32> &idcodes_tmp = jtag->jchain_detect_ndev();

        if (idcodes != idcodes_tmp) {
            err_mod ("JTAG detect: unstable");
//            return false;
        }
    }
    debug_mod(2, QString("Detect: board with %1 HPTDC devices").arg(idcodes.size()));
    if (idcodes.size() > TDC_MAXDEV) {
        warn_mod (QString("Detect: too much devices: %1").arg(idcodes.size()));
//        goto nx;
    }
    s->ndev = std::min(static_cast<int>(idcodes.size()), static_cast<int>(TDC_MAXDEV));
    for (int i=0; i<s->ndev; i++)
        s->hptdc_idcode[i] = idcodes[i];
    //  log_debug(0, "********** tdc_nch() = %d", tdc_nch());
    return true;
//    nx:
//    warn_mod ("Detect: not found");
//    return false;
}

void tdc_bmod::reset ()
{
    hptdc_initialized = false;
    base_bmod::reset();
}

/**************************************************************************/
/* Test functions */
bool tdc_bmod::test_idcode (int loops)
{
    bool test_1_ok = true, test_2_ok = true;
    int n = loops/2;
    int n_1_ok=0, n_2_ok=0;

    //  update_jtag_address();

    if (n<1) n=1;
    const tdc_setup_t *s = static_cast<const tdc_setup_t *>(setup);
    jtag->jtag_reset ();

    //printf("s->ndev=%d\n",s->ndev);
    for (int i = 0; i < n; i++) {
        if (jtag->jchain_test_idcode (0, s->ndev)) {
            test_1_ok = false;
            continue;
        }
        n_1_ok++;
    }
    if (test_1_ok)
        debug_mod(5, QString("IDCODE (TLR) OK"));
    else
        debug_mod(0, QString("IDCODE (TLR) FAILED %1 of %2 runs").arg(n - n_1_ok).arg(n));

    for (int i = 0; i < n; i++) {
        if (jtag->jchain_test_idcode (1, s->ndev)) {
            test_2_ok = false;
            continue;
        }
        n_2_ok++;
    }
    if (test_2_ok)
        debug_mod(5, QString("IDCODE (IRSCAN) OK"));
    else
        debug_mod(0, QString("IDCODE (IRSCAN) FAILED %1 of %2 runs").arg(n - n_2_ok).arg(n));

    hptdc_idcode_test = (test_1_ok && test_2_ok);

    return test_1_ok && test_2_ok;
}

bool tdc_bmod::test_bypass (int loops)
{
    const tdc_setup_t *s = static_cast<const tdc_setup_t *>(setup);
    for (int i = 0; i < loops; i++) {
        if (jtag->jchain_test_bypass (s->ndev)) {
            debug_mod(0, QString("BYPASS failed after %1 Ok loops").arg(i));
            hptdc_bypass_test = false;
            return false;
        }
    }
    hptdc_bypass_test = true;
    debug_mod (5, "BYPASS Ok");
    return true;
}

bool tdc_bmod::test_bist (int loops)
{
    Q_UNUSED(loops);
    // TODO: hptdc BIST
    bool bist_ok = true;
    /*
  const tdc_setup_t *s = static_cast<const tdc_setup_t *>(setup);
  hptdc_setup_t ts[TQDC_NDEV];
  hptdc_bist_t bist[TQDC_NDEV];

  for (int l = 0; l < loops; l++) {
    update_jtag_address();
    jtag->reset ();

// Initialize setup bits
    for (int i = 0; i < s->ndev; i++) {
      hptdc_set_setup_defaults (&setup[i]);
      setup[i].bit.enable_ttl_clock = 0;
      setup[i].bit.core_clock_source = 0;   // 0->40MHz, 1->80MHz, 2->160MHz
    }

// First, do setup and internal memory test
    hptdc_write_setup (setup, s->ndev);

    hptdc_get_bist (bist, s->ndev);
    for (int i = 0; i < s->ndev; i++) {
      if (!hptdc_bist_ok (&bist[i])) {
        printf("\n************ TDC %d BIST FAILED ************\n", i);
        hptdc_print_bist (&bist[i]);
        bist_ok = false;
      }
    }
    if (!bist_ok)
      goto err;
  }

  log_debug(5, "HPTDC BIST %s after %d loops", bist_ok?"Ok":"FAILED", l);
*/
    return bist_ok;
}

bool tdc_bmod::interspill_test ()
{
    bool test_ok = true;
    //  set_mode_vme ();
    setup_with_retries(5);
    if (!check_status ()) test_ok = false;
    if (!test_idcode (10)) test_ok = false;
    read_module_status();
    if (!status_ok()) {
        test_ok = false;
    }
    return test_ok;
}

bool tdc_bmod::selftest ()
{
    bool test_ok = true;
    int fast = 1;
    if (get_error_flag() && hptdc_initialized) {
        warn_mod ("selftest: ERROR flag set before test");
    }
    //  set_mode_vme ();
    tdc_bmod::reset();
    if (true!=testreg32(get_reg_base()+4*get_regnumber_test(), MOD_SELFTEST_LOOPS)) test_ok = false;
    //  if (true!=testram32(vme_reg_base()<<8, 4, 1)) test_ok = false; // BUG
    //  update_jtag_address();
    if (!test_idcode (fast?2:200)) test_ok = false;
    if (!test_bypass (fast?2:200)) test_ok = false;

    /*  if (!fast) */{
        // BIST require clock40
        if (!test_bist (fast?1:5)) test_ok = false;
        //  if (!test_patterns(5)) test_ok = false;
        tdc_bmod::reset();
        // Ok. now load settings.
        preparerun();
    }

    read_module_status();
    if (!status_ok()) {
        test_ok = false;
    }

    debug_mod(test_ok?5:0, QString("selftest %1").arg(test_ok?"Ok":"FAILED"));
    return test_ok;
}

//! write hptdc setup, retry on errors
bool tdc_bmod::setup_with_retries(size_t num_tries)
{
    const tdc_setup_t *s = static_cast<const tdc_setup_t *>(setup);
    Q_ASSERT(s);

    //  tdc_bmod::reset();

    // softclear

    module_ctrl_reg_t ctrl;
    ctrl.all=0;
    int tmp=0;
    get_reg(MODULE_REG_CTRL, &tmp);
    ctrl.all = tmp;
    ctrl.bit.reset = 1;
    set_reg(MODULE_REG_CTRL, ctrl.all);
    ctrl.bit.reset = 0;
    set_reg(MODULE_REG_CTRL, ctrl.all);

    //  log_log_debug(2, "stop_run Ok");


    //  update_jtag_address();
    size_t attempts = 0;
    bool prepare_ok = false;
    while (attempts <= num_tries) {
        attempts++;
        std::vector<hptdc_setup_t> hptdc_setup(s->ndev);
        for (int i=0; i<s->ndev; i++)
            hptdc_setup[i] = s->tdcsetup[i];
        jtag->hptdc_write_setup(hptdc_setup);
        if (!reset_hptdc_chips_pll ()) {
            debug_mod(1, QString("pll reset failed, retry %1").arg(attempts));
            continue;
        }
        reset_hptdc_chips ();
        if (!check_status ()) {
            debug_mod(0, QString("status read failed, retry %1").arg(attempts));
            continue;
        }
        hptdc_initialized = true;

        write_channel_mask();

        read_module_status();
        prepare_ok = status_ok();
        if (!prepare_ok) {
            debug_mod(TDC_DEBUG_LEVEL, QString("status ERROR, retry %1").arg(attempts));
            continue;
        }

        if (prepare_ok) break;
    }
    if (attempts>1 || !prepare_ok) {
        if (attempts >= num_tries && !prepare_ok) {
            err_mod(QString("setup %1 after %2 attempts").arg(prepare_ok?"Ok":"FAILED").arg(attempts));
        } else {
            warn_mod(QString("setup %1 after %2 attempts").arg(prepare_ok?"Ok":"FAILED").arg(attempts));
        }
    } else {
        debug_mod(TDC_DEBUG_LEVEL, QString("setup %1").arg(prepare_ok?"Ok":"FAILED"));
    }
    return prepare_ok;
}

void tdc_bmod::preparerun ()
{
//    // debug
//    const tdc_setup_t *s = static_cast<const tdc_setup_t *>(setup);
//    Q_ASSERT(s);
//    for (int i=0; i<s->ndev; i++)
//        hptdc_print_setup(s->tdcsetup[i]);
    setup_with_retries(5);
}

bool tdc_bmod::check_status ()
{
    const tdc_setup_t *s = static_cast<const tdc_setup_t *>(setup);
    int locked = 1, err = 0, warn = 0;
    const std::vector<hptdc_status_t> &st = jtag->hptdc_get_status(s->ndev);
    for (int i = 0; i < s->ndev; i++) {
        if (st[i].bit.error & HPTDC_STATUS_ERROR_MASK) {
            warn_mod(QString("HPTDC #%1: %2").arg(i)
                     .arg(hptdc_status_error_toString(st[i].bit.error)));
            //      if (st[i].bit.error & HPTDC_STATUS_ERROR_MASK) {
            set_warning_flag();
            warn = 1;
            //      }
        }
        if (!st[i].bit.dll_lock) {
            //      warn_mod ("DLL unlocked in dev #%d", i);
            set_error_flag(); // critical!
            locked = 0;
        }
    }
    if (!locked) {
        QString str;
        for (int i = 0; i < s->ndev; i++)
            if (!st[i].bit.dll_lock)
                str += QString::number(i);
        warn_mod(QString("HPTDC %1: DLL unlocked").arg(str));
    }

    //  if (err) {
    //    log_warn ("check_status: error");
    //  }
    //  if (!locked) {
    //    log_warn ("check_status: DLL unlocked");
    //  }
    debug_mod(TDC_DEBUG_LEVEL, QString("check_status: %1%2%3")
              .arg(!err && locked ? "Ok" : "")
              .arg(err ? "error" : "")
              .arg(locked ? "" : "DLL unlocked"));
    hptdc_warn = warn;
    return !err && locked;
}

void tdc_bmod::end_of_burst ()
{
    read_module_status();
    check_status();
    end_of_burst_readout();
}

module_status_reg_t tdc_bmod::get_status_reg()
{
    module_status_reg_t tmp = base_bmod::get_status_reg();
    // mask tdc errors
    if (!hptdc_initialized) {
        tmp.bit.subdev_0_ok = 1;
        tmp.bit.subdev_1_ok = 1;
        tmp.bit.subdev_2_ok = 1;
        tmp.bit.subdev_3_ok = 1;
        tmp.bit.subdev_4_ok = 1;
        tmp.bit.subdev_5_ok = 1;
    }
    return tmp;
}

bool tdc_bmod::status_ok() const
{
    return base_bmod::status_ok() && hptdc_lock && !hptdc_err;
}

void tdc_bmod::clear_error_flag()
{
    base_bmod::clear_error_flag();
    hptdc_warn = false;
    hptdc_err = false;
}

void tdc_bmod::write_channel_mask()
{
    if (get_id() == MODULE_ID_MTDC64) {
        const tdc_setup_t *s = static_cast<const tdc_setup_t *>(setup);
        unsigned int mask0=0, mask1=0;
        for (int i=0; i<32; i++) {
            mask0 |= s->ch_enable[i]?(1<<i):0;
            mask1 |= s->ch_enable[32+i]?(1<<i):0;
        }
        set_reg(tdc::MODULE_REG_MASK_0, mask0);
        set_reg(tdc::MODULE_REG_MASK_1, mask1);
    }
}

int tdc_bmod::get_regnumber_test() const
{
    return tdc::MODULE_REG_TEST; // hardware BUG in TDC96, should be MODULE_REG_TEST
}

// end
