
/*
 * TDC generic module
 */


#include <QDebug>
#include <QString>

#include "modules/tdc_types.h"
#include "hal/tdc_mod.h"
#include "daqcommon/daqexc.h"
#include "daqcommon/bitop.h"
#include "hptdc/hptdc.h"

void hptdc_settings_tdc::set_defaults()
{
    Q_ASSERT(s);
    hptdc_settings::set_defaults();
    set_core_clk(HPTDC_CORE_CLOCK_80);
    set_edge (1); /* 1 = /   2 = \  3 = /\  */
    set_window (40, 40);
    set_resolution (TDC_RES_INDEX_100);  /* 0=100ps, 1=200ps, 2=400ps, 3=800ps, 4=1.6ns */
}

size_t tdc_mod::get_setup_size() const {return sizeof(tdc_setup_t);}

void tdc_mod::set_defaults()
{
    base_mod::set_defaults();

    unsigned int tdc_id = 0;
    for(int i=0; i<hptdc.size(); i++) {
        QSharedPointer<hptdc_settings_tdc> p = hptdc[i];
        Q_ASSERT(p);
        p->set_defaults();
        p->set_id (tdc_id++);
    }

    tdc_setup_t *s = static_cast<tdc_setup_t *>(setup);

    for (unsigned int i = 0; i < tdc_nch(); i++) {
        s->ch_enable[i] = 1;
    }
}

void tdc_mod::set_setup(common_setup_t *f_setup, size_t f_size)
{
    base_mod::set_setup(f_setup, f_size);
    tdc_setup_t *s = static_cast<tdc_setup_t *>(setup);
    hptdc.clear();
    for (unsigned int i=0;i<tdc_ndev();i++)
        hptdc.push_back(QSharedPointer<hptdc_settings_tdc>(new hptdc_settings_tdc(&s->tdcsetup[i], &s->ch_enable[i])));
//    qDebug() << "Created TDC setup for "<< tdc_ndev() << " hptdc devices" <<endl;
}

void tdc_mod::print_private_setup() const
{
    Q_ASSERT(setup_valid());

    const tdc_setup_t *s = static_cast<const tdc_setup_t *>(setup);

    qDebug() << "*** TDC private setup" ;

    QString str = "Channels: ";
    for (int i=0;i<s->nch;i++) {
        if (i>0) str += ", ";
        str += QString::number(i) + ": " + (s->ch_enable[i]?"ON ":"off");
    }
    qDebug() << str;

    for (unsigned int i=0;i<tdc_ndev();i++) {
        hptdc_print_brief_setup(s->tdcsetup[i]);
    }
}

int tdc_mod::get_int(const std::string &tag, unsigned int mch) const
{
    const tdc_setup_t *s = static_cast<const tdc_setup_t *>(setup);

    if (!validate_mch(mch)) {
        QString str = QString("%1::get_int: bad channel number %2").arg(get_name()).arg(mch);
        throw exc(str.toStdString());
    }
    if (mch >= TDC_MAXNCH)
        mch = GET_FIRST_CH;

    if (tag == "tdc_matchwin_min") {return 1;}
    if (tag == "tdc_matchwin_max") {return 200;}
    if (tag == "tdc_matchwin") {return s->matchwin;}
    if (tag == "tdc_latency_min") {return 0;}
    if (tag == "tdc_latency_max") {return 200;}
    if (tag == "tdc_latency") {return s->latency;}
    if (tag == "en") {return s->ch_enable[mch];}

    int value;
    if (hptdc[hptdc_dev_num(mch)]->get_int(&value, tag, mch)) return value;

    // unhandled
    QString str = QString("%1::get_int: Unknown tag '%2'").arg(get_name()).arg(QString::fromStdString(tag));
    qWarning() << str;
    throw exc_mod_unknown_tag(str.toStdString());
}

namespace {
    const int tdc_hptdc_match_scale = 2;
    const int tdc_hptdc_match_offset = 0;
    const int tdc_hptdc_lat_scale = 2;
    const int tdc_hptdc_lat_offset = 21;
}

void tdc_mod::set_int(int value, const std::string &tag, int mch)
{
    tdc_setup_t *s = static_cast<tdc_setup_t *>(setup);

    //  if (tag == "trigset") {s->trigset.all=value;return;}
    if (mch == SET_ALL_CH) {
        bool found=false;
        for (int i=0;i<hptdc.size();i++)
            if (hptdc[i]->set_int(value, tag, mch)) found=true;
        if (found) return;
    }

    if (tag == "tdc_matchwin") {
        s->matchwin = value;
        for (int i=0;i<hptdc.size();i++)
            hptdc[i]->set_int(tdc_hptdc_match_offset + tdc_hptdc_match_scale*value, "win", SET_ALL_CH);
        return;
    }
    if (tag == "tdc_latency") {
        s->latency = value;
        for (int i=0;i<hptdc.size();i++)
            hptdc[i]->set_int(tdc_hptdc_lat_offset + tdc_hptdc_lat_scale*value, "lat", SET_ALL_CH);
        return;
    }

    if (tag == "en") {
        if (mch == SET_ALL_CH)
            for (int c=0;c<get_nch();c++)
                s->ch_enable[c]=value;
        else
            s->ch_enable[mch]=value;
        return;
    }

    if (mch < 0)
        mch = SET_ALL_CH;

    if (hptdc[hptdc_dev_num(mch)]->set_int(value, tag, mch)) return;

    // unhandled
    QString str = QString("%1::set_int: Unknown tag '%2'").arg(get_name()).arg(QString::fromStdString(tag));
    qWarning() << str;
    throw exc_mod_unknown_tag(str.toStdString());
}
