
/*
 * TQDC backend (hardware) implementation
 */

#include "tqdc_bmod.h"

#include <QDebug>
#include <QString>
#include <QFile>
#include <QTextStream>
#include <QDateTime>
#include <cmath>

#include "daqcommon/bitop.h"
#include "daqcommon/daq_ipc.h"
#include "daqcommon/daqexc.h"
//#include "modules/tqdc_types.h"
#include "debug_hal.h"

//#include "tqdc_calibration.h"
#include "decoder/CalibrationData.h"

namespace {
    enum TQDC_REG {
        TQDC_REG_ADCSET0  = 1,
        TQDC_REG_ADCSET1  = 2,
        //	TQDC_REG_JTAG    = 3,
        TQDC_REG_TRIGSET  = 4,
        TQDC_REG_DACSET   = 5,
        TQDC_REG_TRLAT    = 6,
        //	TQDC_REG_JTAGB   = 8,
        TQDC_REG_COMPSTAT = 0xA,
        TQDC_REG_ADC_RAWSAMPLE = 0x20,
    };

    union tqdc_adcset_reg_t {
        unsigned int all;
        struct {
            unsigned int search:10;
            unsigned int match:5;
            unsigned int neg:1;
            unsigned int mask:8;
            unsigned int reserved2:4;
            unsigned int id:2;
            unsigned int mode:2;
        } bit;
    } ;

}

/******************************************************************************/
#if 0
void tqdc_bmod::dac_shift (int n, int d)
{
    d &= 0xFFFF;
    d |= n << 16;
    set_reg(TQDC_REG_DACSET, d);

    /*
{
  int i, x;
  char p[200], q[100];

  sprintf(p, "DAC %d shifted %04x: ", n, d);
  if (d & 0x8000) {
    sprintf(q, "CTRL ");
    strcat (p, q);
    if ((d & 0xE000) == 0x8000)
      sprintf(q, "Gain %d, BUF %d, Vdd %d", (d >> 4) & 3, (d >> 2) & 3, d & 3);
    if ((d & 0xE000) == 0xA000)
      sprintf(q, "LDAC: %d", d & 3);
    if ((d & 0xE000) == 0xC000)
      sprintf(q, "Power-Down: %02x", d & 0xFF);
    if ((d & 0xF000) == 0xE000)
      sprintf(q, "Reset Data");
    if ((d & 0xF000) == 0xF000)
      sprintf(q, "Reset Data and Control");
    strcat (p, q);
  } else {
    sprintf(q, "DATA ");
    strcat (p, q);
    sprintf(q, "A%d, %03x (%d)", d >> 12, d & 0xFFF, d & 0xFFF);
    strcat (p, q);
  }
  log_debug(1, "%s", p);
}
*/
}
#endif

void tqdc_bmod::dac_exec(const std::vector<int> &cmd)
{
    std::vector<int> regs(cmd.size(), TQDC_REG_DACSET);
//    for(int i=0; i<cmd.size(); i++)
//        info_mod(QString("DAC write: %1").arg(cmd[i], 8, 16, QChar('0')));
    set_reg_v(regs, cmd);
}

//--------------------------------------------------------------------------
// DAC commands
//--------------------------------------------------------------------------

//! command: set DAC input register for channel ch to value val
std::vector<int> tqdc_bmod::dac_set_cmd (int ch, int val)
{
    int n, a;
    n = (ch / 8) & 0x01; // only one bit allowed
    a = ch % 8;
    a = (a ^ 3) & 7;                    // map TQDC->DAC ch
    val &= 0x0FFF;
    std::vector<int> cmd;
    cmd.push_back((n<<16) | (a << 12) | val);
    //  cmd.push_back(tmpcmd);
//    printf("DAC: ch %2d, val %4d, cmd=%08X\n", ch, val, cmd[0]);
    return cmd;
}

//! LDAC# low - This sets LDAC permanently low, thus
//! allowing the DAC registers to be updated continuously.
const int LDAC_Low = 0xA000;

//! LDAC# high - The DAC registers are latched,
//! and the input registers may change without
//! affecting the contents of the DAC registers.
const int LDAC_High = 0xA001;

//! LDAC# signle update - This causes a single pulse on LDAC#,
//! thus updating the DAC registers once.
const int LDAC_Update = 0xA002;

std::vector<int> tqdc_bmod::dac_reset_cmd ()
{
    std::vector<int> cmd;
    for (int i = 0; i < 2; i++) {
        cmd.push_back((i<<16) | 0xF000); //!< Data and Control Reset
        cmd.push_back((i<<16) | 0x8000 | ((0) << 4) | ((0) << 2) | 0); //!< gain, buf, vdd
//        cmd.push_back((i<<16) | LDAC_Low);
        cmd.push_back((i<<16) | LDAC_High);
    }
    return cmd;
}

std::vector<int> tqdc_bmod::dac_load_cmd ()
{
    std::vector<int> cmd;
//    for (int i = 0; i < 2; i++) {
//        cmd.push_back((i<<16) | LDAC_Update);
//    }
    cmd.push_back((0<<16) | LDAC_Update);
    cmd.push_back((1<<16) | LDAC_Update);
    cmd.push_back((0<<16) | LDAC_Update);
    cmd.push_back((1<<16) | LDAC_Update);
    return cmd;
}

std::vector<int> tqdc_bmod::dac_set_all_cmd(int dac)
{
    std::vector<int> cmd;
    for (int c=0; c<TQDC_NCH; c++) {
        const std::vector<int> &tmp = dac_set_cmd(c, dac);
        cmd.insert(cmd.end(), tmp.begin(), tmp.end());
    }
    return cmd;
}

//--------------------------------------------------------------------------
// DAC executables
//--------------------------------------------------------------------------

//! reset all DACs to power-on state
void tqdc_bmod::dac_reset ()
{
    dac_exec(dac_reset_cmd());
}

//! load DAC output registers from input registers
void tqdc_bmod::dac_load ()
{
    dac_exec(dac_load_cmd());
}

//! set DAC channel ch to val
void tqdc_bmod::dac_set (int ch, int val)
{
    dac_exec(dac_set_cmd(ch, val));
}

void tqdc_bmod::dac_set_all(int dac)
{
    dac_exec(dac_set_all_cmd(dac));
}

void tqdc_bmod::dac_set_load(int ch, int dac)
{
    std::vector<int> cmd = dac_set_cmd(ch, dac);
    const std::vector<int> tmp = dac_load_cmd();
    cmd.insert(cmd.end(), tmp.begin(), tmp.end());
    dac_exec(cmd);
}
void tqdc_bmod::dac_set_all_load(int dac)
{
    std::vector<int> cmd = dac_set_all_cmd(dac);
    const std::vector<int> tmp = dac_load_cmd();
    cmd.insert(cmd.end(), tmp.begin(), tmp.end());
    dac_exec(cmd);
}
//--------------------------------------------------------------------------
void tqdc_bmod::set_adc (const tqdc_setup_t &s)
{
//    const tqdc_setup_t *s = static_cast<const tqdc_setup_t *>(setup);

    tqdc_adcset_reg_t adcset[2];
    for (int dev=0;dev<2;dev++) {
        adcset[dev].all=0; // reset
        // set channel enable
        unsigned int mask=0;
        for (int i = 0; i < 8; i++)
            if (s.ch_enable[i+8*dev])
                mask |= (1 << i);
        adcset[dev].bit.mask = mask & 0xFF;
        adcset[dev].bit.search = s.adc_search & 0x3FF;
        adcset[dev].bit.match = (s.adc_match-1) & 0x1F;
        adcset[dev].bit.id = dev & 0x3;
        adcset[dev].bit.mode = s.adc_mode & 0x3;
        adcset[dev].bit.neg = (dev==0)?s.neg_a:s.neg_b;
    }

    set_reg(TQDC_REG_ADCSET0, adcset[0].all);
    set_reg(TQDC_REG_ADCSET1, adcset[1].all);

    for (int dev=0;dev<2;dev++)
        debug_mod(TQDC_DEBUG_LEVEL, QString("ADCSET%1 = %2  id:%3 mode:%4 srch:%5 mtch:%6 enable:%7 neg:%8")
                  .arg(dev).arg(adcset[dev].all, 8, 16, QChar('0'))
                  .arg(adcset[dev].bit.id).arg(adcset[dev].bit.mode)
                  .arg(adcset[dev].bit.search).arg(adcset[dev].bit.match).arg(adcset[dev].bit.mask).arg(adcset[dev].bit.neg));
}

void tqdc_bmod::normalize_trigset()
{
    tqdc_setup_t *s = static_cast<tqdc_setup_t *>(setup);
    s->trigset.bit.shape &= 0xF;
    s->trigset.bit.alt_ch_1 &= 0xF;
    s->trigset.bit.alt_ch_2 &= 0xF;
    s->trigset.bit.alt_en_1 &= 0x1;
    s->trigset.bit.alt_en_2 &= 0x1;
    s->trigset.bit.match8 &= 0x1;
    s->trigset.bit.enable &= 0xFFFF;

    // Set 'hit' trigger mask
//    s->trigset.bit.enable = 0xFFFF; // enable all
    s->trigset.bit.enable = 0xFFFF; // disable all
    for (int c = 0; c < TQDC_NCH; c++) {
        s->trigset.bit.enable |= (s->ch_enable[c] & 1) << c;
    }
    // disable 'hit' trigger line if 'st' is used
//    if (s->trigset.bit.alt_en_1)
//        s->trigset.bit.enable &= ~(1<<s->trigset.bit.alt_ch_1);
//    if (s->trigset.bit.alt_en_2)
//        s->trigset.bit.enable &= ~(1<<s->trigset.bit.alt_ch_2);
}

void tqdc_bmod::set_trigset (int force_zero)
{
    normalize_trigset();
    const tqdc_setup_t *tqdc_setup = static_cast<const tqdc_setup_t *>(setup);
    const tqdc_trigset_reg_t &trigset = tqdc_setup->trigset;
    set_reg(TQDC_REG_TRIGSET, trigset.all);
    int maj = tqdc_setup->maj;
    int min = tqdc_setup->min;

    debug_mod (TQDC_DEBUG_LEVEL, QString("TRIGSET = %1  enable:%2, shp:%3 alt1(%4):%5 alt2(%6):%7  match8:%8")
               .arg(trigset.all, 8, 16, QChar('0'))
               .arg(trigset.bit.enable, 4, 16, QChar('0'))
               .arg(trigset.bit.shape)
               .arg(trigset.bit.alt_en_1?"ON":"off").arg(trigset.bit.alt_ch_1)
               .arg(trigset.bit.alt_en_2?"ON":"off").arg(trigset.bit.alt_ch_2)
               .arg(trigset.bit.match8?"ON":"off"));

    // setup trigger RAM (DPM)

    debug_mod(TQDC_DEBUG_LEVEL, QString("min=%1, maj=%2").arg(min).arg(maj));
    for (int n=0;n<2;n++) {
        int tram[8];
        // clear
        for (int i=0;i<8;i++)
            tram[i]=0;
        //
        if (!force_zero) {
            for (int i=0;i<256;i++) {
                int nb = 0;
                for (int b=0;b<8;b++)
                    if (i&(1<<b)) nb++;
                if (nb>=min && nb<=maj)
                    set_flag_value(&tram[i/32], 1<<(i%32), 1);
            }
        }
        for (int i=0;i<8;i++)
            debug_mod(TQDC_DEBUG_LEVEL, QString("TRMASK[%1, %2] = %3").arg(n).arg(i).arg(tram[i], 8, 16, QChar('0')));
        tram[0] &= ~1;
        w_blk_a24_d32(n*32, 8, tram);
    }
}

void tqdc_bmod::set_thr (int force_zero)
{
    QString sf, sm;

    for (int c = 0; c < TQDC_NCH; c++) {
        const tqdc_setup_t *s = static_cast<const tqdc_setup_t *>(setup);
        bool neg = (c<8)?s->neg_a:s->neg_b;
        debug_mod(TQDC_DEBUG_LEVEL, QString("ch %1, neg=%2").arg(c, 2).arg(neg));
        double thr = s->thr[c];
        if (neg) thr *= -1.;
        int dac = dac_value ((force_zero?0:thr) + thr_offset(c));
        if (dac < 0)
            dac = 0;
        if (dac > dac_max())
            dac = dac_max();
        thr = dac_mv(dac);
        dac_set(c, dac);

        sf += QString("%1 ").arg(dac, 4);
        sm += QString("%1 ").arg(thr, 6, 'f', 1);
    }
    dac_load();
    debug_mod (TQDC_DEBUG_LEVEL, QString("Thr raw %1: %2").arg(force_zero?"force_zero":"").arg(sf));
    debug_mod (TQDC_DEBUG_LEVEL, QString("Thr mV  %1: %2").arg(force_zero?"force_zero":"").arg(sm));
}

void tqdc_bmod::set_trlat ()
{
    int trlat = (static_cast<const tqdc_setup_t *>(setup))->trlat;
    quint32 d = - trlat;
    set_reg(TQDC_REG_TRLAT, d);
    debug_mod(TQDC_DEBUG_LEVEL, QString("TrLat   = %1  %2").arg(d, 8, 16, QChar('0')).arg(-d));
}

void tqdc_bmod::load_registers(int force_zero)
{
    const tqdc_setup_t *s = static_cast<const tqdc_setup_t *>(setup);
    dac_reset ();
    set_thr (force_zero); // dac_load inside
    set_adc (*s);
    set_trlat ();
    set_trigset(force_zero);
}

//--------------------------------------------------------------------------
/*
void tqdc_print_status (int d)
{

  printf("TQDC status %08X %s%s%s%s  wordcount %d\n",
          d,
          d & TQDC_STAT_BIT_SPILL ? "SPILL " : "NO_SPILL ",
          d & TQDC_STAT_BIT_EVBUSY ? "EV_BUSY " : "",
          d & TQDC_STAT_BIT_CLKLOCK ? "CLK_LOCK " : "CLK_UNLOCKED ",
          d & TQDC_STAT_BIT_FIFOEMPTY ? "FIFO_EMPTY " : "FIFO_HAS_DATA ", d & 0x00000FFF);
}

int tqdc_get_status (const tqdc_setup_t * s, int *st)
{

  int d, ret;

  if (!s) {
    log_debug(0, "NULL setup");
    return (-EINVAL);
  }

  if ((ret = vme_setam (vme_reg_am())))
    return (ret);
  if ((ret = vme_ar32 (vme_reg_base() + TQDC_STAT_ADDR_OFFSET, &d)))
    return (ret);               // read status
  if (st)
    *st = d;
//  tqdc_print_status(d);
  return 0;
}

int tqdc_check_runstatus (const tqdc_setup_t * s)
{

  int d, ret;

  if (!s) {
    log_debug(0, "NULL setup");
    return (-EINVAL);
  }
  if ((ret = tqdc_get_status (s, &d)))
    return (ret);               // read status
  if (!(d & TQDC_STAT_BIT_CLKLOCK) ||
//    (d&TQDC_STAT_BIT_TDCERR) ||
      (d & TQDC_STAT_BIT_EVBUSY)
//  ||  (d&TQDC_STAT_BIT_EVREADY)
    ) {
    log_debug(0, "check_runstatus: Bad state\n");
    tqdc_print_status (d);
    return (-EREMOTEIO);
  }
  return 0;
}
*/
bool tqdc_bmod::detect ()
{
    info_mod("detect");

    // test data bus
    try {
        if ((true != testreg32(get_reg_base() + TQDC_REG_ADCSET0*4, 5))) {
            err_mod("ADCSet0 test failed");
            goto nx;
        }
        if ((true != testreg32(get_reg_base() + TQDC_REG_ADCSET1*4, 5))) {
            err_mod("ADCSet1 test failed");
            goto nx;
        }
    }
    catch (exc &e) {
        err_mod(e.what());
        goto nx;
    }
    if (! tdc_bmod::detect()) {
        goto nx;
    }
    setup->nch = TQDC_NCH;
    setup->ndev = TQDC_NDEV;
    return true;
    nx:
    err_mod ("Detect: not found");
    return false;
}

void tqdc_bmod::reset ()
{
    debug_mod(TQDC_DEBUG_LEVEL, "tqdc_bmod reset");
    tdc_bmod::reset();
    //  load_registers(FORCE_ZERO);
    //  calibration();

#if 0 /* DAC test */
    {
        /*
  for (int c=0;c<TQDC_NCH;c++) {
      dac_set (s, c, 0xFFF);
      dac_load(s);
  }
*/
        int step = 256;
        while (1) {
            for (int i = 0; i < dac_max()+1 - step; i += step) {
                for (int c = 0; c < TQDC_NCH; c++) {
                    dac_set (c, i/*0x000*/);
                    dac_load ();
                }
            }
            for (int i = dac_max()+1 - step; i > 0; i -= step) {
                for (int c = 0; c < TQDC_NCH; c++) {
                    dac_set (, c, i/*0xFFF*/);
                    dac_load ();
                }
            }
        }
        //  for (c=0;c<TQDC_NCH;c++)
        if (0) {
            int ch = 0;

            //  printf("DAC ch %d\n", ch);
            for (int i = 0; i < dac_max()+1 - step; i += step) {
                dac_set (ch, i);
                //      dac_load();
            }
            for (int i = dac_max()+1 - step; i > 0; i -= step) {
                dac_set (ch, i);
                //      dac_load();
            }
            /*
        adcw_set_hvadj(0, 0);
        adcw_set_hvadj(0, 2047);
        adcw_set_hvadj(0, 4095);
        adcw_set_hvadj(0, 2047);
//	adcw_set_hvadj(0, 0);
*/
        }
    }
#endif
    debug_mod(TQDC_DEBUG_LEVEL, "Reset Ok");
}

/**************************************************************************/

bool tqdc_bmod::selftest ()
{
    debug_mod(TQDC_DEBUG_LEVEL, "selftest...");
    bool test_ok = true;
    /*
//  int fast = 1;
//  if (get_error_flag()) {
//    log_warn ("selftest: ERROR flag set before test");
//  }
//  set_mode_vme ();
  tdc_bmod::reset();
  if (true!=testram32(vme_reg_base()<<8, 4, 1)) test_ok = false;
*/
    test_ok &= tdc_bmod::selftest();
    if (true!=testram32(get_reg_base()<<8, 4, 1)) test_ok = false;
//  while (true) testreg32(get_reg_base()+4*get_regnumber_test(), 1000);

/*
  debug_mod (test_ok?5:0, "selftest %s", test_ok?"Ok":"FAILED");
*/
    if (test_ok)
        debug_mod(TQDC_DEBUG_LEVEL, "selftest Ok");
        else
        err_mod(QString("selftest failed"));
    return test_ok;
}

void tqdc_bmod::preparerun ()
{
    tdc_bmod::preparerun();

    //  log_debug(TQDC_DEBUG_LEVEL, "Module serial %d - preparerun", get_serial());
    //  calibrate_thresholds();
    load_registers();

    // Setup TDC part
    debug_mod(TQDC_DEBUG_LEVEL, "preparerun");
    //  log_debug(2, "preparerun %s", ok?"Ok":"FAILED");
    return;
}

void tqdc_bmod::end_of_burst_readout ()
{
    const tqdc_counters_t &cnt = get_counters();

    // encode
    unsigned int wc = 0;

    append_rawdata_mhdr();

    for (size_t n=0; n<cnt.size(); n++) {
        unsigned int cnthi, cntlo;
        cnthi = cnt[n] >> 16;
        cntlo = cnt[n] & 0xFFFF;
        append_rawdata( (0<<28)|(n<<19)|cnthi); wc++;
        append_rawdata( (1<<28)|(n<<19)|cntlo); wc++;
    }

    append_rawdata_status(); wc++;
    append_rawdata_mtrl(wc);
}

tqdc_counters_t tqdc_bmod::get_counters ()
{
    tqdc_counters_t cnt(TQDC_N_COUNTERS);
    r_blk_a24_d32(0x800, cnt.size(), static_cast<int *>(static_cast<void *>(&cnt[0])));

    bool countersAvailable = false;
    for (size_t n=0; n<cnt.size(); n++) {
        if (cnt[n] != 0xFFFFFFFFu)
            countersAvailable = true;
    }
    if (countersAvailable)
        return cnt;
    else
        return tqdc_counters_t();
}

//! input: tdc_dev (0..N), tdc_ch(0..31)
//! output: module channel number
int tqdc_bmod::map_chtdc2inp(int tdc_dev, int tdc_ch)
{
    if (tdc_dev < 0 || tdc_dev >= 2) return -1;
    if (tdc_ch < 0 || tdc_ch >= 32) return -1;

    const int tqdc_map_chtdc2inp[2][32] = {
        { 8,-1,-1,-1, 12,-1,-1,-1, 9,-1,-1,-1, 13,-1,-1,-1, 10,-1,-1,-1, 14,-1,-1,-1, 11,-1,-1,-1, 15,-1,-1,-1 },
        { 0,-1,-1,-1, 4,-1,-1,-1, 1,-1,-1,-1, 5,-1,-1,-1, 2,-1,-1,-1, 6,-1,-1,-1, 3,-1,-1,-1, 7,-1,-1,-1 }
    };

    return tqdc_map_chtdc2inp[tdc_dev][tdc_ch];
}

//! input: module channel number
//! output: tdc_ch
//int tqdc_bmod::map_inp2chtdc(int ch)
//{
//  if (ch < 0 || ch >= TQDC_NCH) return 0;
//  int tqdc_map_inp2chtdc[TQDC_NCH] = {
//  };
//  return tqdc_map_inp2chtdc[ch];
//}
//! input: module channel number
//! output: tdc number
//int tqdc_bmod::map_inp2ntdc(int ch)
//{
//  if (ch < 0 || ch >= TQDC_NCH) return 0;
//  return ch/8;
//}

namespace {
    const int cal_dac_delta_mv = 100;
    const int cal_dac_min = tqdc_mod::dac_value(-cal_dac_delta_mv);
    const int cal_dac_max = tqdc_mod::dac_value(cal_dac_delta_mv);
    const size_t cal_steps = cal_dac_max-cal_dac_min;

    int cal_dac_value(int step) {
        return cal_dac_min+step;
    }
}

void tqdc_bmod::calibration()
{
    //  load_registers(FORCE_ZERO);
    //  base_bmod::reset();
    //  usleep(100000);

    update_status();
    if (!ttc_clock_ok()) {
        warn_mod("Clock problem, calibration not performed");
        return;
    }

    // set adc mode for calibration
    set_adc (tqdc_setup_t()); // default settings

    double currentTemperature = get_temperature();
//    const QVariant &currentTemperature = getCalibrationCondition("current_temperature");
//    setCalibrationCondition("temperature", currentTemperature);

    std::vector<std::vector<double> > comp_by_ind(TQDC_NCH, std::vector<double>(cal_steps, 0));
    unsigned int cmpbits_min = 0xFFFFFFFF;
    unsigned int cmpbits_max = 0;

    for (size_t i=0; i<cal_steps; i++) {
        const int dac = cal_dac_value(i);
        //    const double thr = dac_mv (dac);
        //	dac_set_all_load(0);
        //    dac_reset ();
        dac_set_all_load(dac);
        std::vector<double> sample_sum(TQDC_NCH, 0.);
        size_t num_samples = 10;
        // new style
        //	std::vector<int> regs(num_samples, TQDC_REG_COMPSTAT);
        //	std::vector<int> cmpbits = get_reg_blk(regs);
        const std::vector<int> regs(num_samples, TQDC_REG_COMPSTAT);
        const std::vector<int> regdata = get_reg_v (regs);
        if (regdata.size() != num_samples) {
            err_mod(QString("get_reg_v failed in calibration: requested %1 samples, got %2").arg(num_samples).arg(regdata.size()));
            return;
        }
//        debug_mod(0, QString("COMP=%1").arg(regdata[0], 8, 16, QChar('0')));
        size_t got_samples = regdata.size();
        for (size_t s=0; s<got_samples; s++) {
            unsigned int cmpbits = regdata[s];
            cmpbits_min &= cmpbits;
            cmpbits_max |= cmpbits;
            for (int c=0; c<TQDC_NCH; c++) {
                if (cmpbits & (1<<c))
                    sample_sum[c]++;
            }
        }
        /*
// old style
        size_t got_samples = num_samples;
        for (size_t s=0; s<num_samples; s++) {
          int cmpbits = 0;
          get_reg(TQDC_REG_COMPSTAT, &cmpbits);
          for (int c=0; c<TQDC_NCH; c++) {
                if (cmpbits & (1<<c))
                  sample_sum[c]++;
          }
        }
*/
        for (int c=0; c<TQDC_NCH; c++) {
            sample_sum[c] /= got_samples;
            comp_by_ind[c][i] = sample_sum[c];
        }
    }
    bool calibration_present = true;
    if (cmpbits_min == cmpbits_max) {
        calibration_present = false;
        err_mod(QString("threshold calibration failed, search range %1 .. %2 mV, cmpbits=%3")
                 .arg(dac_mv(cal_dac_min), 5, 'f', 1)
                 .arg(dac_mv(cal_dac_max), 5, 'f', 1)
                 .arg(cmpbits_min, 8, 26, QChar('0')));
    }

    if (calibration_present) {
        std::vector<double> thr_cal(TQDC_NCH, 0);
        double syst_err = 0;
        for (int c=0; c<TQDC_NCH; c++) {
            bool found = false;
            for (size_t i=1; i<cal_steps; i++) {
                double samp = comp_by_ind[c][i];
                double prev_samp = comp_by_ind[c][i-1];
                if (prev_samp >= 0.5 && samp < 0.5) {
                    int dac = cal_dac_value(i-1);
                    double thr = dac_mv(dac);
                    double manual_thr = thr_offset(c);
                    double err = thr-manual_thr;
                    syst_err += err;
                    thr_cal[c] = thr;
                    found = true;
                    break;
                }
            }
            if (!found) {
                err_mod(QString("ch %1 threshold calibration failed, search range %2 .. %3 mV")
                         .arg(c, 2)
                         .arg(dac_mv(cal_dac_min), 5, 'f', 1)
                         .arg(dac_mv(cal_dac_max), 5, 'f', 1));
            }
        }
        syst_err /= TQDC_NCH;

        for (int c=0; c<TQDC_NCH; c++) {
            double err = thr_cal[c]-thr_offset(c);
            if (fabs(err) >= 0.5)
                notice_mod(QString("ch %1, threshold calibration %2 mV, factory %3 mV, error %4 mV")
                           .arg(c, 2)
                           .arg(thr_cal[c], 5, 'f', 1)
                           .arg(thr_offset(c), 5, 'f', 1)
                           .arg(thr_cal[c]-thr_offset(c), 5, 'f', 1));
        }
        if (syst_err > 5.0)
            err_mod(QString("threshold systematic error %1 mV").arg(syst_err, 0, 'f', 2));
        else if (syst_err > 2.0)
            warn_mod(QString("threshold systematic error %1 mV").arg(syst_err, 0, 'f', 2));
        else if (syst_err > 0.5)
            notice_mod(QString("threshold systematic error %1 mV").arg(syst_err, 0, 'f', 2));

        setCalibrationValue("thr_offset/temp", currentTemperature);
        setCalibrationValue("thr_offset/date", QDateTime::currentDateTime().toUTC().toString());
        for (int c=0; c<TQDC_NCH; c++) {
            setCalibrationValue(QString("thr_offset/%1").arg(c), thr_cal[c]);
        }
    }
    // Baseline calibration
    std::vector<double> adc_baseline_cal(TQDC_NCH, 0);
    for (int c=0; c<TQDC_NCH; c++) {
        size_t num_samples = 100;
        const std::vector<int> regs(num_samples, TQDC_REG_ADC_RAWSAMPLE+c);
        const std::vector<int> data = get_reg_v (regs);
        double sum = 0;
        const size_t count = data.size();
        for (size_t i=0; i<count; i++) {
            int adc_sample = ((data[i]>>6)&0x3FF) - 512;
            sum += adc_sample;
            //  	  fprintf(stderr, "ch %2zu ADC sample: %08X  % 4d\n", c, data[i], adc_sample);
        }
        adc_baseline_cal[c] = sum / count;
    }
    for (int c=0; c<TQDC_NCH; c++) {
        if (fabs(adc_baseline_cal[c]) > 20.)
            notice_mod(QString("ch %1 adc baseline %2")
                       .arg(c, 2).arg(adc_baseline_cal[c], 5, 'f', 1));
    }

    setCalibrationValue("adc_baseline/temp", currentTemperature);
    setCalibrationValue("adc_baseline/date", QDateTime::currentDateTime().toUTC().toString());
    for (int c=0; c<TQDC_NCH; c++) {
        setCalibrationValue(QString("adc_baseline/%1").arg(c), adc_baseline_cal[c]);
    }
    calibrationData->sync();
}

double tqdc_bmod::thr_offset(int c)
{
    return getCalibrationValue(QString("thr_offset/%1").arg(c)).toDouble();
}

double tqdc_bmod::adc_baseline(int c)
{
    return getCalibrationValue(QString("adc_baseline/%1").arg(c)).toDouble();
}

int tqdc_bmod::get_regnumber_test() const
{
    return MODULE_REG_TEST;
}
