
/*
 * TQDC module
 */


#include <QDebug>
#include <QString>

#include "modules/tqdc_types.h"
#include "tqdc_mod.h"
#include "daqcommon/daqexc.h"
#include "daqcommon/bitop.h"
#include "hptdc_settings.h"

namespace {
    double tqdc_zero_dac_offset = 42.;

    // hardware constants that depend on time offset between adc and tdc
    const int tqdc_adc_match_scale = 4; // 80 MHz ADC / 20 MHz trigger
    const int tqdc_adc_match_offset = 0;
    const int tqdc_hptdc_match_scale = 2;
    const int tqdc_hptdc_match_offset = 0;

    const int tqdc_adc_lat_scale = 4;
    const int tqdc_adc_lat_offset = 44;
    const int tqdc_hptdc_lat_scale = 2;
    const int tqdc_hptdc_lat_offset = 21;
}

class hptdc_settings_tqdc : public hptdc_settings_tdc
{
public:
    hptdc_settings_tqdc(hptdc_setup_t *ps, int *pmask) : hptdc_settings_tdc(ps, pmask) {}
    virtual void set_defaults();
    virtual void get_ch_enable (int ch, bool *en) const;
    virtual void set_ch_enable (int ch, bool en);
};

void hptdc_settings_tqdc::set_defaults()
{
    if (!s) return;
    hptdc_settings_tdc::set_defaults();
    set_lvds_clk();
    set_lvds_hit();
    //  set_core_clk(HPTDC_CORE_CLOCK_160);
    //  set_edge (1); /* 1 = /   2 = \  3 = /\  */
    //  set_window (3, 23);
    //  set_resolution (0);  /* 0=100ps, 1=200ps, 2=400ps, 3=800ps, 4=1.6ns */
}

namespace {
    //! maps subdevice channel(0..8) to hptdc channel
    unsigned int map_dch_to_hptdcch(unsigned int dch)
    {
        int internal_map_dch_to_hptdcch[8] = {0,4,1,5,2,6,3,7};
        if (dch >= 8)
            throw exc("map_dch_to_hptdcch: invalid subdevice channel number");
        return internal_map_dch_to_hptdcch[dch];
    }
}

void hptdc_settings_tqdc::get_ch_enable (int ch, bool *en) const
{
    //  unsigned int tdcnum = ch/8;
    unsigned int tdcch = map_dch_to_hptdcch(ch%8);
    if (en)
        *en = (*channel_mask & (1<<tdcch));
}

void hptdc_settings_tqdc::set_ch_enable (int ch, bool en)
{
    //  unsigned int tdcnum = ch/8;
    unsigned int tdcch = map_dch_to_hptdcch(ch%8);
    unsigned int mask;
    if (ch == SET_ALL_CH) mask=0xFFFFFFFF;
    else
        mask = (1<<tdcch);
    if (en)
        *channel_mask |= mask;
    else
        *channel_mask &= ~mask;
}

size_t tqdc_mod::get_setup_size() const {return sizeof(tqdc_setup_t);}
unsigned int tqdc_mod::tdc_nch() const {return TQDC_NCH;} // number of tdc channels altogether
unsigned int tqdc_mod::tdc_ndev() const {return TQDC_NDEV;} // number of hptdc chips
int tqdc_mod::adc_dev_num(int mch) const {if (mch<0 || mch>=TQDC_NCH) return 0; return mch/8;}
int tqdc_mod::hptdc_dev_num(int mch) const {if (mch<0 || mch>=TQDC_NCH) return 0; return mch/8;}

void tqdc_mod::set_defaults()
{
    tqdc_setup_t *s = static_cast<tqdc_setup_t *>(setup);
    s->set_defaults();

    tdc_mod::set_defaults();

    for(int i=0; i<hptdc.size(); i++) {
        QSharedPointer<hptdc_settings_tdc> x = hptdc[i];
        QSharedPointer<hptdc_settings_tqdc>p = qSharedPointerDynamicCast<hptdc_settings_tqdc>(x);
        Q_ASSERT(p);
        p->set_defaults(); // different defaults for tqdc!
        p->set_id (i);
    }
}
/*
hptdc_settings_tqdc *tqdc_mod::create_hptdc_settings(void *f_setup, size_t f_size) const
{
  tqdc_setup_t *s = static_cast<tqdc_setup_t *>(setup);
  hptdc_settings_tqdc *p;
  p = new hptdc_settings_tqdc(&s->tdcsetup[i], &s->ch_enable[i]);
  return p;
}
*/

void tqdc_mod::set_setup(common_setup_t *f_setup, size_t f_size)
{
    base_mod::set_setup(f_setup, f_size);
    tqdc_setup_t *s = static_cast<tqdc_setup_t *>(setup);
    hptdc.clear();
    for (unsigned int i=0;i<tdc_ndev();i++) {
        hptdc.push_back(QSharedPointer<hptdc_settings_tqdc>(new hptdc_settings_tqdc(&s->tdcsetup[i], &s->ch_enable[i])));
    }
}

void tqdc_mod::print_private_setup() const
{
    Q_ASSERT(setup_valid());
    const tqdc_setup_t *s = static_cast<const tqdc_setup_t *>(setup);
    Q_ASSERT(s);

    qDebug()  << "*** TQDC private setup";

    tdc_mod::print_private_setup();

    qDebug() << QString("TQDC search: %1, match: %2, trlat: %3").arg(s->adc_search).arg(s->adc_match).arg(s->trlat);
    qDebug() << QString("TRIGSET = %1  enable:%2, shp:%3, alt1(%4): %5, alt2(%6): %7, match8:%8, maj:%9, min:%10").arg(s->trigset.all, 8, 16, QChar('0'))
            .arg(s->trigset.bit.enable, 4, 16, QChar('0'))
            .arg(s->trigset.bit.shape)
            .arg(s->trigset.bit.alt_en_1?"ON":"off")
            .arg(s->trigset.bit.alt_ch_1)
            .arg(s->trigset.bit.alt_en_2?"ON":"off")
            .arg(s->trigset.bit.alt_ch_2)
            .arg(s->trigset.bit.match8?"ON":"off")
            .arg(s->maj).arg(s->min);
}

double tqdc_mod::dac_mv_min () const
{
    return (dac_mv (dac_min()));
}

double tqdc_mod::dac_mv_max () const
{
    return (dac_mv (dac_max()));
}

double tqdc_mod::dac_mv (int value)
{
    return (1. * value / 3.2 - 600. - tqdc_zero_dac_offset);
}

int tqdc_mod::dac_value (double mv)
{
    int v = static_cast<int>((mv + 600.+tqdc_zero_dac_offset) * 3.2);
    if (v < dac_min()) v = dac_min();
    if (v > dac_max()) v = dac_max();
    return v;
}


int tqdc_mod::get_int(const std::string &tag, unsigned int mch) const
{
    const tqdc_setup_t *s = static_cast<const tqdc_setup_t *>(setup);
    Q_ASSERT(s);

    if (!validate_mch(mch)) {
        QString str = QString("%1::get_int: bad channel number %1").arg(get_name()).arg(mch);
        throw exc(str.toStdString());
    }

    if (tag == "adc_mode") {return s->adc_mode;}
    if (tag == "adc_match") {return s->adc_match;}
    if (tag == "adc_search") {return s->adc_search;}
    if (tag == "trlat") {return s->trlat;}
    if (tag == "tqdc_matchwin_min") {return 1;}
    if (tag == "tqdc_matchwin_max") {return 200;}
    if (tag == "tqdc_matchwin") {return s->matchwin;}
    if (tag == "tqdc_latency_min") {return 0;}
    if (tag == "tqdc_latency_max") {return 200;}
    if (tag == "tqdc_latency") {return s->latency;}
    if (tag == "thr") {return s->thr[mch];}
    if (tag == "en") {return s->ch_enable[mch];}
    if (tag == "neg") {return mch<8?s->neg_a:s->neg_b;}
    if (tag == "neg_a") {return s->neg_a;}
    if (tag == "neg_b") {return s->neg_b;}
    if (tag == "twoedge") {return s->twoedge;}

    if (tag == "min_thr") {return static_cast<int>(dac_mv_min());}
    if (tag == "max_thr") {return static_cast<int>(dac_mv_max());}
    if (tag == "min_adc_match") {return 2;}
    if (tag == "max_adc_match") {return 29;}
    if (tag == "min_adc_search") {return 0;}
    if (tag == "max_adc_search") {return 0x3FF;}
    if (tag == "min_trlat") {return 0;}
    if (tag == "max_trlat") {return 0x3FF;}
    // trigger part

    //  if (tag == "trigset") {return s->trigset.all;}
    if (tag == "tr_match8") {return s->trigset.bit.match8;}
    if (tag == "tr_maj") {return s->maj;}
    if (tag == "tr_min") {return s->min;}
    if (tag == "tr_en1") {return s->trigset.bit.alt_en_1;}
    if (tag == "tr_en2") {return s->trigset.bit.alt_en_2;}
    if (tag == "tr_alt1") {return s->trigset.bit.alt_ch_1;}
    if (tag == "tr_alt2") {return s->trigset.bit.alt_ch_2;}
    if (tag == "min_tr_shape") {return 0x1;}
    if (tag == "max_tr_shape") {return 0xF;}
    if (tag == "tr_shape") {return s->trigset.bit.shape;}
//    if (tag == "en_trig") {return get_flag_value(s->trigset.bit.enable, 1u<<mch);}

    {
        QSharedPointer<const hptdc_settings_tdc> x = hptdc[hptdc_dev_num(mch)];
        QSharedPointer<const hptdc_settings_tqdc> hs = qSharedPointerDynamicCast<const hptdc_settings_tqdc>(x);
        Q_ASSERT(hs);
        int value;
        if (hs->get_int(&value, tag, mch)) return value;
    }

    // unhandled
    QString str = QString("%1::get_int: Unknown tag '%2'").arg(get_name()).arg(QString::fromStdString(tag));
    throw exc_mod_unknown_tag(str.toStdString());
}

void tqdc_mod::set_int(int value, const std::string &tag, int mch)
{
    tqdc_setup_t *s = static_cast<tqdc_setup_t *>(setup);
    Q_ASSERT(s);

    //  if (tag == "trigset") {s->trigset.all=value;return;}
    if (mch == SET_ALL_CH || mch == SET_MOD_CH) {
        bool found=false;
        for (int i=0;i<hptdc.size();i++)
            if (hptdc[i]->set_int(value, tag, SET_ALL_CH)) found=true;
        if (found) return;
    }

    if (tag == "tqdc_matchwin") {
        s->matchwin = value;
        s->adc_search = tqdc_adc_match_offset + tqdc_adc_match_scale*value;
        for (int i=0;i<hptdc.size();i++)
            hptdc[i]->set_int(tqdc_hptdc_match_offset + tqdc_hptdc_match_scale*value, "win", SET_ALL_CH);
        return;
    }
    if (tag == "tqdc_latency") {
        s->latency = value;
        s->trlat = tqdc_adc_lat_offset + tqdc_adc_lat_scale*value;
        for (int i=0;i<hptdc.size();i++)
            hptdc[i]->set_int(tqdc_hptdc_lat_offset + tqdc_hptdc_lat_scale*value, "lat", SET_ALL_CH);
        return;
    }

    if (tag == "adc_mode") {s->adc_mode = value; return;}
    if (tag == "adc_match") {s->adc_match=value; return;}
    if (tag == "adc_search") {s->adc_search=value; return;}
    if (tag == "trlat") {s->trlat=value;return;}
    if (tag == "thr") {
        if (mch == SET_ALL_CH)
            for (int c=0;c<get_nch();c++)
                s->thr[c]=value;
        else
            s->thr[mch]=value;
        return;
    }
    if (tag == "en") {
        if (mch == SET_ALL_CH)
            for (int c=0;c<get_nch();c++)
                s->ch_enable[c]=value;
        else
            s->ch_enable[mch]=value;return;
    }

    if (tag == "neg_a") {s->neg_a = value; update_hptdc_edge(); return;}
    if (tag == "neg_b") {s->neg_b = value; update_hptdc_edge();return;}
    if (tag == "twoedge") {s->twoedge = value; update_hptdc_edge(); return;}
    //  if (tag == "trigset") {s->trigset.all=value;return;}
    if (tag == "tr_match8") {s->trigset.bit.match8 = value?1:0;return;}
    if (tag == "tr_maj") {s->maj = value;return;}
    if (tag == "tr_min") {s->min = value;return;}
    if (tag == "tr_en1") {s->trigset.bit.alt_en_1 = value?1:0;return;}
    if (tag == "tr_en2") {s->trigset.bit.alt_en_2 = value?1:0;return;}
    if (tag == "tr_alt1") {s->trigset.bit.alt_ch_1 = value & 0xF;return;}
    if (tag == "tr_alt2") {s->trigset.bit.alt_ch_2 = value & 0xF;return;}
    if (tag == "tr_shape") {s->trigset.bit.shape = value & 0xF;return;}
//    if (tag == "en_trig") {unsigned int mask = 1<<mch; if (value) s->trigset.bit.enable |= mask; else s->trigset.bit.enable &= ~mask; return;}

    if (hptdc[hptdc_dev_num(mch)]->set_int(value, tag, mch)) return;

    // unhandled
    QString str = QString("%1::set_int: Unknown tag '%2'").arg(get_name()).arg(QString::fromStdString(tag));
    throw exc_mod_unknown_tag(str.toStdString());
}

void tqdc_mod::update_hptdc_edge()
{
    tqdc_setup_t *s = static_cast<tqdc_setup_t *>(setup);
    Q_ASSERT(s);
    int edge_0 = s->twoedge?3:(s->neg_a?2:1);
    int edge_1 = s->twoedge?3:(s->neg_b?2:1);
    hptdc[0]->set_int(edge_0, "edge", 0);
    hptdc[1]->set_int(edge_1, "edge", 1);
    hptdc[0]->set_int(0, "pair", 0);
    hptdc[1]->set_int(0, "pair", 1);
}

int tqdc_mod::dac_bits()
{
    return 12;
}

int tqdc_mod::dac_min()
{
    return 0;
}

int tqdc_mod::dac_max()
{
    return (1<<dac_bits())-1;
}
