
/*
 * TRIG backend (hardware) implementation
 */

#include "trig_bmod.h"

#include <QString>

#include "daqcommon/bitop.h"
#include "daqcommon/device_ids.h"
#include "daqcommon/daq_ipc.h"
#include "modules/trig_types.h"
#include "daqcommon/rawdata.h"
#include "debug_hal.h"

namespace {
    const int REGSIZE = 4;
}

class TrigRegs
{
public:
    TrigRegs() { trigset.all = 0; timers.all = 0; spillwidth.all = 0; }

    QString str_trigset() const;

    trig_trigset_t trigset;
    trig_timers_t timers;
    spilwid_reg_t spillwidth;
};

QString TrigRegs::str_trigset() const
{
    const unsigned int nm = trigset.bit.nim_enable;
    const unsigned int lm = trigset.bit.lvds_enable;
    QString s;
    QString nmstr;
    for (int i=7; i>0; i--)
        nmstr += (nm&(1<<i))?"1":"0";
    QString lmstr;
    for (int i=8; i>0; i--)
        lmstr += (lm&(1<<i))?"1":"0";
    s = QString("TRIGSET    = %1 sp_ext=%2, NIM=%3(mask %4), spill=%5 (%6%7), noise=%8, pulser=%9, LVD=(mask %10) %11")
             .arg(trigset.all, 8, 16, QChar('0'))
             .arg(trigset.bit.spill_ext_enable)
             .arg(trigset.bit.nim1_enable)
             .arg(nmstr)
             .arg(trigset.bit.spill_select)
             .arg(trigset.bit.spill_select==TRIG_SPILL_EXT?"External":"")
             .arg(trigset.bit.spill_select==TRIG_SPILL_TIMER?"Timer":"")
             .arg(trigset.bit.noise_enable)
             .arg(trigset.bit.pulser_enable)
             .arg(lmstr)
             .arg(trigset.bit.en_xoff_in?", XOFF_ENABLED":"");
    return s;
}

trig_bmod::trig_bmod(const bmod_props &props)
    : base_bmod(props),
    regs(new TrigRegs)
{
}

trig_bmod::~trig_bmod()
{
    delete regs;
}

module_status_reg_t trig_bmod::get_status_reg()
{
    module_status_reg_t r; r.all = 0xFFFFFFFFu;
    return r; // TODO
}

void trig_bmod::load_settings(const common_setup_t *p)
{
    Q_ASSERT(p);

    const trig_setup_t *s = static_cast<const trig_setup_t *> (p);
    //  set_mode_vme ();

    // Normalize
    double spilltimer_step;
    if (s->fw_ver >= 0x0101) {
        spilltimer_step = TRIG_v1_1_SPILLTIMER_STEP_MS;
    } else {
        spilltimer_step = TRIG_SPILLTIMER_STEP_MS;
    }
    unsigned int sp_dly = static_cast<unsigned int>((s->spill_delay-1) * 1. / spilltimer_step);
    unsigned int sp_dur = static_cast<unsigned int>((s->spill_duration-1) * 1. / spilltimer_step);
    unsigned int sp_rep = static_cast<unsigned int>((s->spill_repeat-1) * 1. / spilltimer_step);
    //  log_debug(1, "s->spill_delay=%u, sp_dly=%u", s->spill_delay, sp_dly);
    //  log_debug(1, "s->spill_duration=%u, sp_dur=%u", s->spill_duration, sp_dur);

    for (int i=0;i<TRIG_NCH;i++)
        debug_mod(3, QString("ch %1: en %2, delay %3").arg(i).arg(s->ch_enable[i]).arg(s->ch_delay[i]));
    if (sp_dly < 1) sp_dly = 1;
    if (sp_dur < 1) sp_dur = 1;
    if (sp_rep < 2) sp_rep = 2;
    if (sp_dly > 0xFFFF) sp_dly = 0xFFFF;
    if (sp_dur > 0xFFFF) sp_dur = 0xFFFF;
    if (sp_rep > 0xFFFF) sp_rep = 0xFFFF;
    //  for (int i=0;i<TRIG_NCH;i++)
    //	s->ch_delay[i] &= 0xF;
    //  s->ttc_delay &= 0xF;
    //  s->read_delay &= 0xFFFF;

    //  regs->trigset.all=0;
    int spill_ext_enable = (s->spill_select==TRIG_SPILL_EXT || s->spill_select == TRIG_SPILL_GATE);
    //  if (s->spill_select) s->spill_ext_enable = 0; else s->spill_ext_enable = 1;

    //  regs->trigset.bit.lvds_enable = 0x11; // enable LVD0,4 inputs by default
//    regs->trigset.bit.nim_enable = 0x7F;
    regs->trigset.bit.nim_enable = 0;
    for(int i=1; i<=7; i++)
        if (s->input_enable[i]) regs->trigset.bit.nim_enable |= (1<<(i-1));
    quint32 lvds_en = 0;
    for(int i=0; i<=32; i++)
        if (s->input_enable[8+i]) lvds_en |= (1<<(i-1));
    set_reg(TRIG_REG_LVDSENABLE, lvds_en);
    debug_mod(TRIG_DEBUG_LEVEL, QString("LVDS EN = %1").arg(lvds_en, 8, 16, QChar('0')));

    quint32 nimlogic = ((s->nim_maj_max & 0x7) << 3) | (s->nim_maj_min & 0x7);
    set_reg(TRIG_REG_NIMLOGIC, nimlogic);
    debug_mod(TRIG_DEBUG_LEVEL, QString("NIM maj = %1-%2")
              .arg(nimlogic&7)
              .arg((nimlogic>>3)&7));

    regs->trigset.bit.spill_ext_enable = spill_ext_enable?1:0;
    regs->trigset.bit.spill_gate = (s->spill_select == TRIG_SPILL_GATE)?1:0;
    regs->trigset.bit.lvds_enable = 0;
    if (s->ch_enable[0])     regs->trigset.bit.lvds_enable |= 0x1; // ST
    if (s->ch_enable[1])     regs->trigset.bit.lvds_enable |= 0x0E; // TQDC1
    if (s->ch_enable[2])     regs->trigset.bit.lvds_enable |= 0xF0; // TQDC2
    regs->trigset.bit.nim1_enable = s->ch_enable[3]?1:0; // NIM1
    regs->trigset.bit.spill_select = s->spill_select?1:0;
    regs->trigset.bit.noise_enable = s->noise_trigger?1:0;
    regs->trigset.bit.pulser_enable = s->pulser_trigger?1:0;
    regs->trigset.bit.en_xoff_in = s->en_xoff_in?1:0;

    for (int i=0; i<TRIG_NCH; i++)
        debug_mod(TRIG_DEBUG_LEVEL, QString("ch_enable[%1]=%2").arg(i).arg(s->ch_enable[i]));
    //  for (int i=0; i<8; i++)
    //	log_debug(0, "lvds_enable[%d]=%d", i, s->lvds_enable[i]);

    load_trigset();

    double pulser_period = 80e6 / s->pulser_freq;
    quint32 raw_pulser_period = pulser_period - 1;
    set_reg(TRIG_REG_PULSERPERIOD, raw_pulser_period);
    debug_mod(TRIG_DEBUG_LEVEL, QString("Pulser ferquency = %1 Hz, raw %2")
              .arg(s->pulser_freq)
              .arg(raw_pulser_period, 8, 16, QChar('0')));

    unsigned int d;
    d = sp_dly | (sp_rep<<16);
    set_reg(TRIG_REG_SPILLDEL, d);
    debug_mod(TRIG_DEBUG_LEVEL, QString("SPILLDEL   = %1").arg(d, 8, 16, QChar('0')));


    spilwid_reg_t spw;
    spw.all = 0;
    spw.bit.spillwidth = sp_dur;
    if (s->fw_ver >= 0x0101) {
        spw.bit.en_spill_end_wait = 1;
    } else {
        spw.bit.en_spill_end_wait = 1; //s->en_spill_end_wait;
    }
    //  d = sp_dur;
    set_reg(TRIG_REG_SPILLWIDTH, spw.all);
    debug_mod(TRIG_DEBUG_LEVEL, QString("SPILLWIDTH = %1, en_spill_end_wait=%2")
              .arg(spw.bit.spillwidth, 8, 26, QChar('0')).arg(spw.bit.en_spill_end_wait));

    // TRIGDEL
    d = ((s->ch_delay[3]&0xF)<<12) |
        ((s->ch_delay[2]&0xF)<<8) |
        ((s->ch_delay[1]&0xF)<<4) |
        (s->ch_delay[0]&0xF);

    unsigned int wr_del = 0;
    d |= (wr_del & 0xF) << 16;

    d |= ((s->dch_clock_adj & 0x1F) << 20);

    // invert out CLK10
    //  d |= (1<<25);

    d |= (s->in_shape & 0xF) << 28; // input pulse shaper time

    set_reg(TRIG_REG_TRIGDEL, d);
    //  log_err("TRIGDEL    = %08X", d);
    debug_mod(TRIG_DEBUG_LEVEL, QString("TRIGDEL    = %1").arg(d, 8, 16, QChar('0')));

    /*
const unsigned int read_delay_mask = (1<<TRIG_READ_DELAY_BITS)-1;
const unsigned int ttc_delay_mask = (1<<TRIG_TTC_DELAY_BITS)-1;
const unsigned int xoff_wait_mask = (1<<TRIG_XOFF_WAIT_BITS)-1;

  d = ( (s->read_delay&read_delay_mask)<<16) |
   ((s->xoff_wait & xoff_wait_mask) << TRIG_TTC_DELAY_BITS)
    (s->ttc_delay & ttc_delay_mask);
  set_reg(TRIG_REG_TIMERS, d);
  log_debug(TRIG_DEBUG_LEVEL, "TIMERS     = %08X  TTC_DELAY=%.2f us, READ_DELAY=%.2f us", d,
        (s->ttc_delay&ttc_delay_mask)*TRIG_TTC_DELAY_STEP_US,
        (s->read_delay&read_delay_mask)*TRIG_READ_DELAY_STEP_US);
*/
    set_reg(TRIG_REG_TIMERS, s->timers.all);
    debug_mod(TRIG_DEBUG_LEVEL, QString("TIMERS     = %1  TTC_DELAY=%2 us, READ_DELAY=%3 us, XOFF_WAIT=%4 us")
              .arg(s->timers.all, 8, 16, QChar('0'))
              .arg(s->timers.bit.ttc_delay * TRIG_TTC_DELAY_STEP_US, 0, 'f', 2)
              .arg(s->timers.bit.read_delay * TRIG_READ_DELAY_STEP_US, 0, 'f', 2)
              .arg(s->timers.bit.xoff_wait * TRIG_XOFF_WAIT_STEP_US, 0, 'f', 2));

    d = s->logic & 0xFFFE; // bit #0 always 0
    set_reg(TRIG_REG_LOGIC, d);
    debug_mod(TRIG_DEBUG_LEVEL, QString("LOGIC      = %1").arg(d, 8, 16, QChar('0')));

    d = ((s->out_sel_2 & 0xF) << 4) | (s->out_sel_1 & 0xF);
    set_reg(TRIG_REG_NIMOUTPUT, d);
    debug_mod(TRIG_DEBUG_LEVEL, QString("NIMOutput  = %1").arg(d, 8, 16, QChar('0')));

    debug_mod(TRIG_DEBUG_LEVEL+1, "load_settings Ok");
}


bool trig_bmod::detect ()
{
    trig_setup_t *s = static_cast<trig_setup_t *>(setup);

    s->ndev = 1;
    s->nch = TRIG_NCH;
    return true;
}

void trig_bmod::reset ()
{
    base_bmod::reset();
//    set_reg(TRIG_REG_CONTROL, 0);
    load_settings(setup);
    read_module_status();
    debug_mod(TRIG_DEBUG_LEVEL+2, "Reset Ok");
}

bool trig_bmod::interspill_test ()
{
    int ret = -1;
    //  set_mode_vme ();

    if ((true != testreg32(get_reg_base() + TRIG_REG_TEST * REGSIZE, MOD_SELFTEST_LOOPS))) goto err;

    /*
  trig_setup_t tmp;
  memcpy(&tmp, s, sizeof(trig_setup_t));
  for (int i=0;i<TRIG_NCH;i++)
    tmp.ch_enable[i]=0;
  tmp.spill_select = TRIG_SPILL_EXT;
  tmp.spill_ext_enable = 0;
  if ((ret=trig_load_settings(&tmp))) goto err;
  if ((true!=testreg32(vme_reg_base()|(TRIG_REG_SPILLDEL<<1)|1, 10))) goto err;
  if ((true!=testreg32(vme_reg_base()|(TRIG_REG_SPILLWIDTH<<1)|1, 10))) goto err;
  if ((true!=testreg32(vme_reg_base()|(TRIG_REG_TRIGDEL<<1)|1, 10))) goto err;
  if ((true!=testreg32(vme_reg_base()|(TRIG_REG_TIMERS<<1)|1, 10))) goto err;
  if ((true!=testreg32(vme_reg_base()|(TRIG_REG_LOGIC<<1)|1, 10))) goto err;
  if ((ret=trig_load_settings(s))) goto err;
*/

    read_module_status();
    debug_mod(TRIG_DEBUG_LEVEL, "interspill test Ok");

    return true;
    err:
    read_module_status();
    debug_mod(0, QString("interspill_test failed: %1").arg(strerror(-ret)));
    return false;
}

bool trig_bmod::selftest ()
{
    int ret = -1;
    trig_setup_t tmp;

    //  set_mode_vme ();
    debug_mod(TRIG_DEBUG_LEVEL, "selftest...");

    memcpy(&tmp, setup, sizeof(trig_setup_t));
    for (int i=0;i<TRIG_NCH;i++)
        tmp.ch_enable[i]=0;
    tmp.spill_select = TRIG_SPILL_EXT;
    //  tmp.spill_ext_enable = 0;
    load_settings(&tmp);

    if ((true!=testreg32(get_reg_base() + TRIG_REG_TEST * REGSIZE, MOD_SELFTEST_LOOPS))) goto err;
    /*
  if ((true!=testreg32(vme_reg_base()|(TRIG_REG_SPILLDEL<<1)|1, 100))) goto err;
  if ((true!=testreg32(vme_reg_base()|(TRIG_REG_SPILLWIDTH<<1)|1, 100))) goto err;
  if ((true!=testreg32(vme_reg_base()|(TRIG_REG_TRIGDEL<<1)|1, 100))) goto err;
  if ((true!=testreg32(vme_reg_base()|(TRIG_REG_TIMERS<<1)|1, 100))) goto err;
  if ((true!=testreg32(vme_reg_base()|(TRIG_REG_LOGIC<<1)|1, 100))) goto err;
*/

    load_settings(setup);
    read_module_status();

    debug_mod(TRIG_DEBUG_LEVEL, "selftest Ok");

    return true;
    err:
    debug_mod(0, QString("selftest failed: %1").arg(strerror(-ret)));
    return false;
}

/*
int trig_loadsetup (trig_setup_t * s)
{

  int ret = -1, data;

  if (!s) {
    log_debug(0, "loadsetup: NULL argument");
    return (-EINVAL);
  }

  data = (s->shape & 0xF) | ((s->trigwid & 0xF) << 4);
  if ((ret = trig_set_dreg (s, data)))
    goto err;
  if ((ret = trig_write_ram (s, s->bitmap)))
    goto err;

  log_debug(TRIG_DEBUG_LEVEL, "loadsetup Ok");
  return 0;
err:
  log_debug(0, "loadsetup failed");
  return (ret);
}
*/

void trig_bmod::preparerun()
{
    load_settings(setup);
    read_module_status();
    debug_mod(TRIG_DEBUG_LEVEL, "preparerun Ok");
}

void trig_bmod::end_of_burst_readout()
{
    // revision 0-3: PhLOM version
    // revision 4-7: TTCM version
    // revision 8-11: MTDC-64 built-in, no logic state counters (deprecated)
    // revision 16: MTDC-64 built-in, no logic state counters, interspill timer starts at end of spill, not start of spill,
    //   spill length is now 31 bits
    const int rev = get_rev();
    typedef enum { DF_PHLOM, DF_TTCM, DF_MTDCA, DF_MTDCB } data_format_t;
    data_format_t data_format = DF_MTDCB; // default
    if (rev >= 0 && rev <= 3)
        data_format = DF_PHLOM;
    if (rev >= 4 && rev <= 7)
        data_format = DF_TTCM;
    if (rev >= 8 && rev <= 11)
        data_format = DF_MTDCA;
    if (rev == 16)
        data_format = DF_MTDCB;

    // encode end-of-burst data
    unsigned int wc = 0;

    append_rawdata_mhdr();

    append_rawdata( (5<<28)|(1<<19)); wc++; // Timestamp=0, TriggerBits=(1<<19)

    // read logic state counters
    if (data_format == DF_TTCM) {
        std::vector<unsigned int> count_all, count_read;
        const int num_count = 32;

        for (int i=0;i<num_count;i++) {
            int rtmp;
            int reg = 32 + i;
            r_a16_d32 (get_reg_base() + reg*REGSIZE, &rtmp);
            debug_mod(TRIG_DEBUG_LEVEL, QString("REG[%1] = 0x%2 (%3)").arg(reg, 2).arg((quint32)rtmp, 8, 16, QChar('0')).arg(rtmp));
            if (i%2 == 0)
                count_read.push_back(rtmp);
            else
                count_all.push_back(rtmp);
        }
        bool countersAvailableRead = false;
        for (unsigned int n=0; n<count_read.size(); n++) {
            if (count_read[n] != 0xFFFFFFFFu)
                countersAvailableRead = true;
        }
        if (!countersAvailableRead) count_read.clear();

        bool countersAvailableAll = false;
        for (unsigned int n=0; n<count_all.size(); n++) {
            if (count_all[n] != 0xFFFFFFFFu)
                countersAvailableAll = true;
        }
        if (!countersAvailableAll) count_all.clear();

        for (unsigned int n=0; n<count_read.size(); n++) {
            append_rawdata( (0<<28)|(count_read[n]  & 0x0FFFFFFF)); // triggered logic state count
            wc++;
        }
        for (unsigned int n=0; n<count_all.size(); n++) {
            append_rawdata( (1<<28)|(count_all[n] & 0x0FFFFFFF)); // unblocked logic state count
            wc++;
        }
    }

    append_rawdata_status(); wc++;
    append_rawdata_mtrl(wc);
}

/*
int trig_write_ram (trig_setup_t * s, int data[])
{

  int ret = -1, i;

  if (!s) {
    log_debug(0, "NULL setup");
    return (-EINVAL);
  }

  if (vme_setam (TRIG_BITMAP_AM))
    goto err;
  for (i = 0; i < TRIG_BITMAP_SIZE; i++)
    if (vme_aw32 ((vme_reg_base() << 8) + (i << 2), s->bitmap[i]))
      goto err;

  return 0;
err:
  log_debug(0, "write_ram failed");
  return (ret);
}
*/

trig_incnt_t trig_bmod::get_incnt ()
{
    // TODO: input counters not implemented in hardware
    /*
    trig_incnt_t cnt(TRIG_N_INPUTS);
    // FIXME: use block mode
    //    r_blk_a24_d32(vme_reg_base()<<8), cnt.size(), static_cast<int *>(static_cast<void *>(&cnt[0])));
    for (unsigned int i=0; i<cnt.size(); i++)
    {
        const unsigned int addr = (vme_reg_base()<<8) | (i<<2);
        int d=0;
        r_a24_d32(addr, &d);
        cnt[i] = d;
        log_debug(0, "COUNT[%02d] = 0x%08X", i, d);
    }

    bool countersAvailable = false;

    for (size_t n=0; n<cnt.size(); n++) {
        if (cnt[n] != 0xFFFFFFFFu)
            countersAvailable = true;
    }
    if (countersAvailable)
        return cnt;
    else
    */
    return trig_incnt_t();
}

void trig_bmod::start_spill()
{
    regs->trigset.bit.soft_spill = 0;
    load_trigset();
    regs->trigset.bit.soft_spill = 1;
    load_trigset();
}

void trig_bmod::stop_spill()
{
    regs->trigset.bit.soft_spill = 0;
    load_trigset();
}

void trig_bmod::load_trigset()
{
    set_reg(TRIG_REG_TRIGSET, regs->trigset.all);
    debug_mod(TRIG_DEBUG_LEVEL, regs->str_trigset());
}
