
#include <QtCore/QtGlobal>
#include <cstdio>
#include <cerrno>
#include <cstdarg>
#include <vector>
#include <deque>
#include <stdexcept>

#include "colorlog/log.h"
#undef log_debug

#include "vmedaq_jtag_engine.h"
#include "vmedaq_jtag_impl.h"
#include "daqcommon/bitop.h"
//#include "timer.h"

//#define PREFIX "jtag: "

//using namespace std;
using namespace colorlog;

JTAG_Engine::JTAG_Engine(Jtag_Impl &p)
    : impl(p)
{
    jtagState = JS_TEST_LOGIC_RESET;
    irStopState = JS_RUN_TEST_IDLE;
}

void JTAG_Engine::_internal_jtag_reset () // reset to the Test-Logic-Reset state
{
    // set TMS high, TDI doesn't matter, clock them out five times
    bool btms[5] = {1, 1, 1, 1, 1};
    impl.jtag_io_b(5, btms, 0, 0);
    jtagState = JS_TEST_LOGIC_RESET;

    log_debug(8, "JTAG TEST LOGIC RESET");
}

bool JTAG_Engine::_perform_scan(size_t count, const char *tdiData, char *tdoData)
{
    bool sendBit;
    bool tmsState;

    if (count==0) return true;

    if (tdoData)
        for (unsigned int n=0; n < count>>3; n++)
            tdoData[n >> 3] =0;

    const unsigned int *tdi32Data = static_cast<const unsigned int *>(static_cast<const void *>(tdiData));
    unsigned int *tdo32Data = static_cast<unsigned int *>(static_cast<void *>(tdoData));

    // non-negative, count-1, don't send last bit in b32 instruction: TMS=1
    size_t blen = std::max(static_cast<qint64>(0), (static_cast<qint64>(count) - 1) / 32);

    // optimize: large counts
    enum { J32_BLOCK_XFER_MAX_SIZE = 100 }; // 3200 bits

    if (blen > J32_BLOCK_XFER_MAX_SIZE) {
        log_debug(0, "FIXME: bit count exceeds J32_BLOCK_XFER_MAX_SIZE=%d", J32_BLOCK_XFER_MAX_SIZE);
        return false;
    }

    std::vector<unsigned int> tdibuf(blen, 0);
    if (tdi32Data)
        for (size_t n=0; n<blen; n++)
            tdibuf[n] = tdi32Data[n];

    std::vector<unsigned int> tdobuf(blen, 0);
    impl.jtag_io_b32_blk(blen, &tdibuf[0], &tdobuf[0]);

    for (size_t n=0; n<blen; n++) {
        if (tdo32Data) {
            tdo32Data[n] = tdobuf[n];
        }
    }

    const size_t start_count = blen*32;
    const size_t remain_count = count-start_count; // okay
    btms.resize(remain_count);
    btdi.resize(remain_count);
    btdo.resize(remain_count);
//    bool btms[remain_count], btdi[remain_count], btdo[remain_count];

    for (size_t i=start_count; i<count; i++)  // shift through the data register
    {
        tmsState = (i+1 == count); // TMS = 1 if this is the last bit to be sent for now, otherwise 0
        if (tdiData)                /* not NULL */
            sendBit = tdiData[i >> 3] & (1 << (i & 7)) ? true : false;  // read a bit from the buffer
        else
            sendBit = false;
        btms[i-start_count] = tmsState;
        btdi[i-start_count] = sendBit;
    }                             // when done we will be in JS_EXIT1_DR

    impl.jtag_io_b(remain_count, &btms[0], &btdi[0], &btdo[0]);

    if (tdoData)
        for (size_t p=0; p < remain_count; p++) {
        size_t tdo_p = start_count + p;
        if (btdo[p])
            tdoData[tdo_p >> 3] |= (1 << (p & 7));
        else
            tdoData[tdo_p >> 3] &= ~(1 << (p & 7));
    }
    return true;
}

// count is number of BYTES available to transfer (scanCount is number of bits to actually transfer)
// data is shifted lsb first
//
void JTAG_Engine::_internal_jtag_drscan (const unsigned int count, const char *tdiData, char *tdoData)
{
    switch (jtagState)            // move to JS_SHIFT_DR state
    {
    case JS_TEST_LOGIC_RESET:    // start in JS_TEST_LOGIC_RESET
        impl.jtag_io (0, 0);             // move to IDLE, fall into the idle case

    case JS_RUN_TEST_IDLE: {       // start in JS_RUN_TEST_IDLE
            bool btms[3] = {
                1,             // move to JS_SELECT_DR_SCAN
                0,             // move to JS_CAPTURE_DR
                0              // move to JS_SHIFT_DR
            };
            impl.jtag_io_b(3, btms, 0, 0);
            break;
        }
    case JS_PAUSE_DR: {           // start in JS_PAUSE_DR
            bool btms[2] = {
                1,             // move to JS_EXIT2_DR
                0              // move to JS_SHIFT_DR
            };
            impl.jtag_io_b(2, btms, 0, 0);
            break;
        }
    case JS_PAUSE_IR: {            // start in JS_PAUSE_IR
            bool btms[5] = {
                1,             // move to JS_EXIT2_IR
                1,             // move to JS_UPDATE_IR
                1,             // move to JS_SELECT_DR_SCAN
                0,             // move to JS_CAPTURE_DR
                0              // move to JS_SHIFT_DR
            };
            impl.jtag_io_b(5, btms, 0, 0);
            break;
        }
    default:
        _internal_jtag_reset ();
        throw std::runtime_error("JTAG DR scan: invalid TAP state");
    }

    _perform_scan(count, tdiData, tdoData);

    {
        bool btms[2] = {
            1,           // move to JS_UPDATE_DR
            0            // move to JS_RUN_TEST_IDLE
        };
        impl.jtag_io_b(2, btms, 0, 0);
    }

    {
        if (debug_level_ge(8)) {
            if (count > 32) {
                printf("DR scan TDI[%d]: \n", count);
                bitprint_groups (tdiData, count, 32);
                printf("DR scan TDO[%d]: \n", count);
                bitprint_groups (tdoData, count, 32);
            } else {
                printf("DR scan TDI[%d]: ", count);
                bitprint (tdiData, count);
                printf("DR scan TDO[%d]: ", count);
                bitprint (tdoData, count);
            }
        }
    }
}


void JTAG_Engine::_internal_jtag_irscan (unsigned int count, const char *tdiData, char *tdoData)
{
    switch (jtagState)            // move to JS_SHIFT_DR state
    {
    case JS_TEST_LOGIC_RESET:    // start in JS_TEST_LOGIC_RESET
        impl.jtag_io (0, 0);             // move to IDLE, fall into the idle case

    case JS_RUN_TEST_IDLE:       // start in JS_RUN_TEST_IDLE
        impl.jtag_io (1, 0);             // move to JS_SELECT_DR_SCAN
        impl.jtag_io (1, 0);             // move to JS_SELECT_IR_SCAN
        impl.jtag_io (0, 0);             // move to JS_CAPTURE_IR (instruction register is parallel loaded into the shift register, 01 into lsb's)
        impl.jtag_io (0, 0);             // move to JS_SHIFT_IR
        break;

    case JS_PAUSE_DR:            // start in JS_PAUSE_DR
        impl.jtag_io (1, 0);             // move to JS_EXIT2_DR
        impl.jtag_io (1, 0);             // move to JS_UPDATE_DR
        impl.jtag_io (1, 0);             // move to JS_SELECT_DR_SCAN
        impl.jtag_io (1, 0);             // move to JS_SELECT_IR_SCAN
        impl.jtag_io (0, 0);             // move to JS_CAPTURE_IR
        impl.jtag_io (0, 0);             // move to JS_SHIFT_IR
        break;

    case JS_PAUSE_IR:            // start in JS_PAUSE_IR
        impl.jtag_io (1, 0);             // move to JS_EXIT2_IR
        impl.jtag_io (0, 0);             // move to JS_SHIFT_IR
        break;

    default:
        _internal_jtag_reset ();
        throw std::runtime_error("JTAG IR scan: invalid TAP state");
    }

    _perform_scan(count, tdiData, tdoData);

    switch (irStopState) {
    case JS_RUN_TEST_IDLE:
        impl.jtag_io (1, 0);           // move to JS_UPDATE_IR
        impl.jtag_io (0, 0);           // move to JS_RUN_TEST_IDLE
        break;

    case JS_PAUSE_IR:
        impl.jtag_io (0, 0);           // move to JS_PAUSE_IR
        break;

    case JS_TEST_LOGIC_RESET:
    default:
        impl.jtag_io (1, 0);           // move to JS_UPDATE_IR
        impl.jtag_io (1, 0);           // move to JS_SELECT_DR_SCAN
        impl.jtag_io (1, 0);           // move to JS_SELECT_IR_SCAN
        impl.jtag_io (1, 0);           // move to JS_TEST_LOGIC_RESET
        break;
    }

    {
        if (debug_level_ge(8)) {
            printf("IR scan TDI: ");
            bitprint (tdiData, count);
            printf("IR scan TDO: ");
            bitprint (tdoData, count);
        }
    }
}

void JTAG_Engine::jtag_reset ()               // resets TAP controller, clears pending scan counters
{
    _internal_jtag_reset ();      // reset the interface
    //  irStopState = drStopState = JS_RUN_TEST_IDLE; // default end state is idle
    irStopState = JS_RUN_TEST_IDLE; // default end state is idle
}

void JTAG_Engine::jtag_pause(unsigned int count)
{
    for (unsigned int i = 0; i < count; i++)
        impl.jtag_io (0, 0);
}

void JTAG_Engine::jtag_drscan (unsigned int count, const void *tdiData, void *tdoData)
{
    _internal_jtag_drscan(count, static_cast<const char *>(tdiData), static_cast<char *>(tdoData));
}

//bool JTAG_Engine::jtag_drscan_chk (unsigned int count, const void *tdiData, void *tdoData)
//{
//    // shift and update DR
//    jtag_drscan (count, tdiData, tdoData);

//    // check if DR has changed
//    if (bitcmp (tdiData, tdoData, count)) {
//        // shift and update DR again
//        jtag_drscan (count, tdiData, tdoData);

//        // if DR has changed again, it cannot be loaded, fail
//        if (bitcmp (tdiData, tdoData, count)) {
//            log_debug(0, "jtag_drscan_chk: data still changing");
//            if (debug_level_ge(1)) {
//                bitprintdiff(tdiData, tdoData, count);
//            }
//            return true;
//        } else {
//            return 1;
//        }
//    } else {
//        return false;
//    }
//    return false;
//}

void JTAG_Engine::jtag_irscan (unsigned int count, const void *tdiData, void *tdoData)
{
    //  irscanCount = count;
    _internal_jtag_irscan (count, static_cast<const char *>(tdiData), static_cast<char *>(tdoData));
}

void JTAG_Engine::debug(int level, const char *fmt, ...) const  // used only with -Dlog_debug
{
    const size_t LOGBUFFER_SIZE = 1024;
    static char buf[LOGBUFFER_SIZE+1];
    va_list msg;

    va_start (msg, fmt);
    vsnprintf(buf, LOGBUFFER_SIZE, fmt, msg);
    va_end (msg);
    buf[LOGBUFFER_SIZE] = 0;

    log_debug(level, "%s%s", "JTAG ", buf);
    //  mod->log_debug_mod (level, "%s%s", "JTAG ", buf);
}
