
#include <cstdio>
#include <cstring>
#include <cerrno>
#include <vector>

#include "colorlog/log.h"
#undef log_debug

#include "daqcommon/bitop.h"
#include "vmedaq_jtag_engine.h"
#include "vmedaq_jtag_engine_hptdc.h"
#include "hptdc/hptdc.h"
#include "hptdc/htpdc_chip.h"

using namespace colorlog;

namespace {
    const unsigned int TDI_TEST_PATT = 0xa5a5a5a5;
    const unsigned int TDI_TEST_PATT_SHL = 0x4B4B4B4B;
}

void JTAG_Engine_HPTDC::jchain_irscan (const std::vector<hptdc_ir_t> &ir)
{
    if (ir.empty())
        return;
    const size_t ndev = ir.size();

    const size_t scanbits = TDC_IR_BITS * ndev;
    const size_t irsz = BYTES_FOR_BITS (scanbits);
    std::vector<unsigned char> ir_in(irsz);
    std::vector<unsigned char> ir_out(irsz);
    for (unsigned int i = 0; i < ndev; i++)
        bitcopy (&ir_in[0], i * TDC_IR_BITS, (unsigned char *) &ir[i], 0, TDC_IR_BITS);

    //  printf("irscan tdi: "); bitprint(ir_in, scanbits);
    jtag_irscan (scanbits, &ir_in[0], &ir_out[0]);
    //  printf("irscan tdo: "); bitprint(ir_out, scanbits);
    if (bitcmp (&ir_in[0], &ir_out[0], scanbits)) {
        //        jtag_irscan (scanbits, &ir_in[0], &ir_out[0]);
        //        if (bitcmp (&ir_in[0], &ir_out[0], scanbits)) {
        //            log_debug(0, "Failed to load IR");
        //            if (debug_level_ge(2)) {
        //                printf("IR: ");
        //                bitprintdiff_groups (&ir_in[0], &ir_out[0], scanbits, 5);
        //            }
        //            //      bitprintdiff(ir_in, ir_out, scanbits);
        //            ret = -EREMOTEIO;
        //            goto err;
        //        }
        log_debug(3, "IR changed");
    } else {
        log_debug(3, "IR not changed");
    }
}

void JTAG_Engine_HPTDC::jchain_irscan_all (unsigned int ir_in, unsigned int ndev)
{
    jchain_irscan(std::vector<hptdc_ir_t>(ndev, ir_in));
}

//! \return IDCODE for devices found
std::vector<quint32> JTAG_Engine_HPTDC::jchain_detect_ndev()
{
    std::vector<quint32> idcodes;

    const unsigned int maxdevs = 16;
    const unsigned int scandevs = maxdevs+1; // reserve one for pattern pass test
    const size_t scanbits = scandevs * 32;
    //  const size_t idsz = BYTES_FOR_BITS (scanbits);
    std::vector<unsigned int> id_in(scandevs, TDI_TEST_PATT);
    std::vector<unsigned int> id_out(scandevs, 0);

    jtag_reset();

    for (unsigned int n = 0; n < 2; n++)
        jtag_drscan(scanbits, &id_in[0], &id_out[0]);
    bool endsrch = false;
    for (unsigned int i = 0; i < scandevs; i++) {
        const unsigned int id = id_out[i];
        switch (id) {
        case 0x0:
        case 0xFFFFFFFF:
            if (i == 0)
                log_warn("Broken JTAG chain: TDO %08X", id);
            else
                log_warn("Broken JTAG chain: TDO #%d - TDI #%d: %08X", i, i - 1, id);
            return idcodes;
        case TDI_TEST_PATT:
            log_debug(3, "Detected %d devices", i);
            endsrch = true;
            break;
        case TDI_TEST_PATT >> 1:
        case (TDI_TEST_PATT << 1)|0:
        case (TDI_TEST_PATT << 1)|1:
            log_warn("Broken JTAG controller: bad shift detected: data %08X, expected %08X",
                     id, TDI_TEST_PATT);
            endsrch = true;
            break;
        case HPTDC_IDCODE_1_2:
        case (HPTDC_IDCODE_1_2>>1):
        case (HPTDC_IDCODE_1_2<<1)|0:
        case (HPTDC_IDCODE_1_2<<1)|1:
            log_debug(3, "Device #%d is HPTDC 1.2", i);
            idcodes.push_back(id);
            break;
        case HPTDC_IDCODE_1_3:
        case (HPTDC_IDCODE_1_3>>1):
        case (HPTDC_IDCODE_1_3<<1)|0:
        case (HPTDC_IDCODE_1_3<<1)|1:
            log_debug(3, "Device #%d is HPTDC 1.3", i);
            idcodes.push_back(id);
            break;
        default:
            log_warn("Unknown device #%d with ID %08X", i, id);
            return idcodes;
//            idcodes.push_back(id);
        }
        if (endsrch)
            break;
    }
    if (id_out[maxdevs] != TDI_TEST_PATT) {
        log_warn("Broken JTAG chain: TDI not passed, TDO is %08X", id_out[maxdevs]);
        return idcodes;
    }
    return idcodes;
}

int JTAG_Engine_HPTDC::jchain_test_idcode (unsigned int load_ir, unsigned int ndev)
{
    if (ndev == 0)
        return 0;
    int ret = 0;

    const unsigned int scandevs = ndev+1;
    const unsigned int scanbits = scandevs * 32;
    //  const size_t idsz = BYTES_FOR_BITS (scanbits);
    std::vector<unsigned int> id_in(scandevs);
    std::vector<unsigned int> id_out(scandevs);

    if (load_ir) {
        jchain_irscan_all (HPTDC_I_IDCODE, ndev);
    } else {
        jtag_reset();
    }

    for (unsigned int i = 0; i < scandevs; i++)
        id_in[i] = TDI_TEST_PATT;
    for (unsigned int n = 0; n < 2; n++)
        jtag_drscan (scanbits, &id_in[0], &id_out[0]);

    if ((id_out[0] == 0xFFFFFFFF) || (id_out[0] == 0)) {
        log_debug(1, "ID[0] = 0x%08x", id_out[0]);
        ret = -ENODEV;
        goto err;
    }
    if (id_out[0] == TDI_TEST_PATT) {
        log_debug(1, "TDI shorted to TDO");
        ret = -ENODEV;
        goto err;
    }
    for (unsigned int i = 0; i < scandevs; i++) {
        if (i == ndev) {
            if (id_out[i] == TDI_TEST_PATT) {
                log_debug(3,  "JTAG chain is Ok");
                break;
            } else {
                log_debug(1, "TDI problem: pattern not passed");
                if (debug_level_ge(2)) {
                    printf("TDI: ");
                    bitprint_groups (&id_in[0], scanbits, 32);
                    printf("TDO: ");
                    bitprint_groups (&id_out[0], scanbits, 32);
                }
            }
        }
        if ((id_out[i] != HPTDC_IDCODE_1_2) && (id_out[i] != HPTDC_IDCODE_1_3)) {
            log_debug(0, "Not an HPTDC device #%d: %08x", i, id_out[i]);
            ret = -ENODEV;
            goto err;
        } else {
            if (id_out[i] == HPTDC_IDCODE_1_2)
                log_debug(3, "HPTDC 1.2 device #%d", i);
            if (id_out[i] == HPTDC_IDCODE_1_3)
                log_debug(3, "HPTDC 1.3 device #%d", i);
        }
    }

    log_debug(3, "IDCODE Ok");
    return 0;
    err:
    log_debug(0, "jchain_irscan_all() failed: %s", strerror(-ret));
    return ret;
}

int JTAG_Engine_HPTDC::jchain_test_bypass(unsigned int ndev)
{
    if (ndev == 0)
        return 0;
    //    int ret = 0;
    const unsigned char testpattern[] = { 0x5A, 0x29, 0x3E, 0x91, 0xEA, 0x83, 0xB1, 0xA5 };

    const unsigned int pattbits = 70;
    const unsigned int scanbits = pattbits + ndev;
    const size_t sz = BYTES_FOR_BITS (scanbits);

    std::vector<unsigned char> in(sz);
    std::vector<unsigned char> out(sz);
    std::vector<unsigned char> tmp(sz);

    for (unsigned int i=0; i<sz; i++)
        if (i < sizeof(testpattern))
            in[i] = testpattern[i];
    else
        in[i] = 0x1E;
    /* all "1" is BYPASS opcode */
    jchain_irscan_all (0xFF, ndev);
    jtag_drscan (scanbits, &in[0], &out[0]);
    /*
  printf("TDI: "); bitprint(in, scanbits);
  printf("TDO: "); bitprint(out, scanbits);
*/
    bitcopy (&tmp[0], 0, &out[0], ndev, pattbits);
    if (bitcmp (&in[0], &tmp[0], pattbits)) {
        log_debug(0, "Bypass failed");
        if (debug_level_ge(2)) {
            printf("TDI: ");
            bitprint (&in[0], pattbits);
            printf("TDO: ");
            bitprint (&tmp[0], pattbits);
        }
        //    goto err;
    } else {
        log_debug(3, "Bypass Ok");
    }

    return 0;
    //    err:
    //    log_debug(0, "jchain_test_bypass() failed: %s", strerror(-ret));
    //    return ret;
}

std::vector<hptdc_status_t> JTAG_Engine_HPTDC::hptdc_get_status(size_t ndev)
{
    std::vector<hptdc_status_t> status(ndev);
    if (ndev == 0)
        return status;

    const size_t scanbits = TDC_STATUS_BITS * ndev;
    const size_t sz = BYTES_FOR_BITS (scanbits);
    std::vector<unsigned char> buf(sz);

    jchain_irscan_all (HPTDC_I_Status, ndev);
    jtag_drscan (scanbits, NULL, &buf[0]);

    for (unsigned int i = 0; i < ndev; i++)
        bitcopy (&status[i], 0, &buf[0], i * TDC_STATUS_BITS, TDC_STATUS_BITS);

    return status;
}

std::vector<hptdc_bsr_t> JTAG_Engine_HPTDC::hptdc_get_bsr(size_t ndev)
{
    std::vector<hptdc_bsr_t> bsr(ndev);
    if (ndev == 0)
        return bsr;

    const size_t scanbits = TDC_BSR_BITS * ndev;
    const size_t sz = BYTES_FOR_BITS (scanbits);
    std::vector<unsigned char> buf(sz);

    jchain_irscan_all (HPTDC_I_SAMPLE, ndev);
    jtag_drscan (scanbits, NULL, &buf[0]);

    for (unsigned int i = 0; i < ndev; i++)
        bitcopy (&bsr[i], 0, &buf[0], i * TDC_BSR_BITS, TDC_BSR_BITS);

    return bsr;
}

/**
 * With the BIST instruction all internal memories will be put in a special self test mode. The
 * self test of the memories will be started when entering the JTAG runbist state. A minimum number
 * of 3200 JTAG clocks must be given in the runbist state to finalize the memory tests. After this the
 * memory self test results can be read out on the serial JTAG output.
 */
std::vector<hptdc_bist_t> JTAG_Engine_HPTDC::hptdc_get_bist(size_t ndev)
{
    std::vector<hptdc_bist_t> bist(ndev);
    if (ndev == 0)
        return bist;

    const size_t scanbits = TDC_BIST_BITS * ndev;
    const size_t sz = BYTES_FOR_BITS (scanbits);
    std::vector<unsigned char> buf(sz);

    jchain_irscan_all (HPTDC_I_BIST, ndev);
    jtag_pause(3200);
    jtag_drscan (scanbits, NULL, &buf[0]);

    for (unsigned int i = 0; i < ndev; i++)
        bitcopy (&bist[i], 0, &buf[0], i * TDC_BIST_BITS, TDC_BIST_BITS);

    return bist;
}

/**
 * The JTAG coretest scan path is used to perform extended testing of the TDC chip. The
 * internal scan path gives direct access to the interface between the channel buffers and first level
 * buffer logic. It is used in connection with the test mode select bits in the setup scan path and is only
 * intended for verification and production tests of the TDC chip.
 */
std::vector<hptdc_coretest_t> JTAG_Engine_HPTDC::hptdc_get_coretest(size_t ndev)
{
    std::vector<hptdc_coretest_t> coretest(ndev);
    if (ndev == 0)
        return coretest;

    const size_t scanbits = TDC_CORETEST_BITS * ndev;
    const size_t sz = BYTES_FOR_BITS (scanbits);
    std::vector<unsigned char> buf(sz);

    jchain_irscan_all (HPTDC_I_Coretest, ndev);
    jtag_drscan (scanbits, NULL, &buf[0]);

    for (unsigned int i = 0; i < ndev; i++)
        bitcopy (&coretest[i], 0, &buf[0], i * TDC_CORETEST_BITS, TDC_CORETEST_BITS);

    return coretest;
}

std::vector<hptdc_readout_t> JTAG_Engine_HPTDC::hptdc_get_jtag_readout(size_t ndev)
{
    std::vector<hptdc_readout_t> data(ndev);

    if (ndev == 0)
        return data;

    const size_t scanbits = TDC_READOUT_BITS * ndev;
    const size_t sz = BYTES_FOR_BITS (scanbits);
    std::vector<unsigned char> buf(sz);

    jchain_irscan_all (HPTDC_I_Readout, ndev);
    jtag_drscan (scanbits, NULL, &buf[0]);

    for (unsigned int i = 0; i < ndev; i++) {
        bitcopy(&data[i], 0, &buf[0], i * TDC_READOUT_BITS, TDC_READOUT_BITS);
        log_debug(4, "jtag readout (dev #%d): %08X", i, data[i].bit.data);
    }

    return data;
}

/**
 * The JTAG setup scan path is used to load programming data that can not be changed while
 * the TDC is actively running. No separate update register is implemented on this scan path so
 * reading the information is "destructive".
 */
void JTAG_Engine_HPTDC::hptdc_write_setup (const std::vector<hptdc_setup_t> &setup)
{
    if (setup.empty())
        return;

    const size_t ndev = setup.size();
    const size_t scanbits = ndev * TDC_SETUP_BITS;
    const size_t sz = BYTES_FOR_BITS (scanbits);
    std::vector<unsigned char> in(sz);
    std::vector<unsigned char> out(sz);

    std::vector<hptdc_setup_t> setup_in(setup);
    for (unsigned int i = 0; i < ndev; i++)
        hptdc_setup_set_parity (&setup_in[i]);

    jchain_irscan_all (HPTDC_I_Setup, ndev);

    for (unsigned int dev = 0; dev < ndev; dev++) {
        bitcopy (&in[0], dev * TDC_SETUP_BITS, &setup_in[dev].all[0], 0, TDC_SETUP_BITS);
    }

    jtag_drscan(scanbits, &in[0], &out[0]);
    log_debug(2, "setup Ok");

    if (debug_level_ge(2))
        for (unsigned int dev = 0; dev < ndev; dev++) {
        hptdc_print_brief_setup (setup_in[dev]);
    }
    if (debug_level_ge(5))
        for (unsigned int dev = 0; dev < ndev; dev++) {
        hptdc_print_setup (setup_in[dev]);
    }
}

/**
 * The JTAG control scan path is used to enable/disable channels which can be done while the
 * TDC is actively running. The control scan path is also use to initialize the PLL and DLL after
 * power up. A global reset (equivalent to asserting the reset pin) can be issued to initialize the global
 * state of the TDC after power up and reconfiguring.
 */
void JTAG_Engine_HPTDC::hptdc_write_control (const std::vector<hptdc_control_t> &control)
{
    if (control.empty())
        return;
    const size_t ndev = control.size();
    const size_t scanbits = ndev * TDC_CONTROL_BITS;
    const size_t sz = BYTES_FOR_BITS (scanbits);
    std::vector<unsigned char> in(sz);
    std::vector<unsigned char> out(sz);

    std::vector<hptdc_control_t> control_tmp(control);
    for (unsigned int i = 0; i < ndev; i++)
        hptdc_control_set_parity (&control_tmp[i]);

    jchain_irscan_all (HPTDC_I_Control, ndev);

    for (unsigned int dev = 0; dev < ndev; dev++) {
        bitcopy (&in[0], dev * TDC_CONTROL_BITS, &control_tmp[dev].all[0], 0, TDC_CONTROL_BITS);
    }
    /*
  bitprint(&control_in[0].all[0], TDC_CONTROL_BITS);
  bitprint(&control_in[1].all[0], TDC_CONTROL_BITS);
  bitprint(&control_in[2].all[0], TDC_CONTROL_BITS);
  bitprint(&control_in[3].all[0], TDC_CONTROL_BITS);
*/
    jtag_drscan(scanbits, &in[0], &out[0]);
    log_debug(2, "control Ok");

    if (debug_level_ge(5))
        for (unsigned int dev=0;dev<ndev;dev++) {
        hptdc_print_control(control_tmp[dev]);
    }
}
