
#include "hal_xml/hw_settings.h"

#include <QDebug>
#include <QFile>
#include <QTextStream>
#include <QDomElement>
#include <QDomDocument>
#include <QDateTime>

#include <ctime>
#include <cstdio>
#include <iostream>
#include <sstream>
#include <fstream>
#include <typeinfo>

#include "hal/hw_backend.h"
#include "hal/base_mod.h"
#include "hal_xml/base_mod_impl.h"
#include "daqcommon/daqexc.h"
#include "hal/device_db.h"

void write_hardware_settings(const QString &fn, const hw_backend &hw)
{
    QString ddb_fn(fn);
    ddb_fn.append(".ddb");

    hw.ddb_save_as(ddb_fn);

    QString str(fn);
    str.append(".txt");

    QFile f(str);
    if (!f.open(QIODevice::WriteOnly)) {
        qWarning() << QString("Can't open file %1 for writing: %2").arg(str).arg(f.errorString());
        return;
    }
    QTextStream ts(&f);
    modmap_t mods = hw.get_modmap();

    ts << "# hw_manager summary\n";
    ts << "[summary]\n";
    ts << "modules=" << mods.size() << '\n';

    Q_FOREACH(const base_mod *p, mods) {
        Q_ASSERT(p);
        //      base_mod &p = dynamic_cast<base_mod&>(*(*it).second);
        //	p.print_info();
        ts << "\n[module]\n";
        ts << "class=" << typeid(p).name() << '\n';
        if (p->setup_valid()) {
            ts << QString("id=%1:%2\n").arg(p->get_vend(), 2, 16, QChar('0')).arg(int(p->get_id()), 2, 16, QChar('0'));
            ts << QString("name=%1\n").arg(module_id_str(p->get_id()));
            ts << QString("serial=0x%1\n").arg(p->get_serial(), 2, 16, QChar('0'));
            ts << QString("flags=0x%1\n").arg(p->get_flags(), 4, 16, QChar('0'));
            ts << QString("warning=%1\n").arg((p->get_flags() & STDMOD_FLAG_WARN)?"yes":"no");
            ts << QString("error=%1\n").arg((p->get_flags() & STDMOD_FLAG_ERR)?"yes":"no");
            ts << QString("failure=%1\n").arg((p->get_flags() & STDMOD_FLAG_FAIL)?"yes":"no");
            ts << QString("test_ok=%1\n").arg((p->get_flags() & STDMOD_FLAG_TESTOK)?"yes":"no");
            ts << QString("position=%1\n").arg(p->get_slot());
            ts << QString("readout=%1\n").arg(p->get_cblt_enable()?"on":"off");
        }
    }
    ts << "\n[end]\n";
    f.close();

    QDomDocument document;
    //    document.set_internal_subset("example_xml_doc", "", "example_xml_doc.dtd");
    QDomElement nodeRoot = document.appendChild(document.createElement("daqsetup")).toElement();

    // Date, Time
    const QDateTime timeNow = QDateTime::currentDateTime().toUTC();
    QString s2 = "Date: " + timeNow.toString("ddd, dd MMM yyyy hh:mm:ss UTC");
    nodeRoot.appendChild(nodeRoot.ownerDocument().createComment(s2));

    // Storage
    QDomElement nodeStorage = nodeRoot.appendChild(nodeRoot.ownerDocument().createElement("storage")).toElement();
    nodeStorage.setAttribute("filename", fn);
    nodeStorage.setAttribute("time_preparerun", QString::number(timeNow.toTime_t()));

    // Modules
    Q_FOREACH(const base_mod *p, mods) {

        QDomElement nodeModule = nodeRoot.appendChild(nodeRoot.ownerDocument().createElement("module")).toElement();

        p->bmi->write_xml_attr(&nodeModule);
        p->bmi->write_xml_setup(&nodeModule);
//        p->bmi->write_xml_calibration(&nodeModule);
    }

    str = fn;
    str.append(".xml");
    QFile xmlFile(str);
    if (!xmlFile.open(QIODevice::WriteOnly)) {
        qWarning() << QString("Can't open file %1 for writing: %2").arg(str).arg(xmlFile.errorString());
        return;
    }
    QTextStream stream(&xmlFile);
    stream << document.toString();
    xmlFile.close();
    if (stream.status() != QTextStream::Ok) {
        qWarning() << QString("Error writing file %1: %2").arg(str).arg(xmlFile.errorString());
        return;
    }
}

void read_hardware_settings(const QString &fn, hw_manager &hw)
{
    QString ddb_fn(fn);
    ddb_fn.append(".ddb");
    device_db ddb(ddb_fn);
    ddb.load();

    QString fileName = fn+".xml";

    QFile xmlFile(fileName);
    if (!xmlFile.open(QIODevice::ReadOnly)) {
        qWarning() << QString("Can't open file %1 for reading: %2").arg(fileName).arg(xmlFile.errorString());
        return;
    }
    QDomDocument document;
    QString errorMsg;
    int errorLine = -1;
    int errorColumn = -1;
    if (!document.setContent(&xmlFile, &errorMsg, &errorLine, &errorColumn)) {
        qWarning() << QString("Error in file %1: %2, line %3, column %4").arg(fileName)
                .arg(xmlFile.errorString().arg(errorMsg).arg(errorLine).arg(errorColumn));
        xmlFile.close();
        return;
    }

    // Walk the tree
    QDomElement nodeRoot = document.documentElement();
    if (nodeRoot.tagName() != "daqsetup") {
        qWarning() << QString("Error in file %1: Invalid root node").arg(fileName);
        return;
    }

    QDomNodeList nl = nodeRoot.elementsByTagName("module");
    for(int i=0; i<nl.count(); i++) {
        QDomElement el = nl.at(i).toElement();
        const QDomNamedNodeMap al = el.attributes();
        unsigned int ddb_key = al.namedItem("ddb_key").toAttr().value().toUInt(NULL, 16);
        //        int serial = al.namedItem("serial").toAttr().value().toUInt(NULL, 16);

        const ddb_ent &ent = ddb.find(ddb_key);
        if (!ent.isNull()) {
            size_t setup_size = ent.size();
            common_setup_t *setup = static_cast<common_setup_t *>(operator new(setup_size));
            memcpy(setup, ent.constData(), setup_size);
            //  		bool setup_valid = base_mod::setup_struct_valid("", setup, setup_size);
            hw.insert_module(ddb_key, setup, setup_size);
        } else {
            printf("DDB entry not found for key %08X\n", ddb_key);
        }
        /*

          // check matching attributes
      // find matching module
      base_mod *mod = 0;
      modmap_t mods = hw.get_modmap();
      foreach (mods, jt) {
        base_mod &p = *(*jt).second;
        printf("Search: key %08X %08X, serial %08X %08X\n", p.get_ddb_key(), ddb_key, p.get_serial() , serial);
        if ((p.get_ddb_key() == ddb_key) && (p.get_serial() == serial)) { // found
                  mod = &p;
          break;
        }
          }
          if (!mod || !mod->bmi) continue;
          printf("Found serial %08X, key %08X\n", mod->get_serial(), mod->get_ddb_key());
          mod->bmi->read_xml_calibration(nodeModule);
*/
        /*
          QDomElement nodeCal = 0;
          xmlpp::Node::NodeList nlm = nodeModule->get_children("calibration");
          foreach(nlm, jt) {
                xmlpp::Node *n = (*jt);
                // check matching attributes
                nodeCal = (QDomElement)n;
//		printf("Found calibration\n");
          }
*/
        //	  if (!nodeCal)
        //		nodeCal = nodeModule->add_child("calibration");
    }
    //    hw.print();
    /*
// Modules
        foreach_const (mods, it) {
      const base_mod &p = *(*it).second;

    QDomElement nodeModule = nodeRoot->add_child("module");

    p.bmi->write_xml_attr(nodeModule);
    p.bmi->write_xml_setup(nodeModule);
    p.bmi->write_xml_calibration(nodeModule);
        }
*/
}
