/*
 *
 *  hptdc.cpp -- implementation of the HPTDC chip JTAG operations
 *
 *  High Performance Time to Digital Converter (HPTDC)
 *  http://micdigital.web.cern.ch/micdigital/hptdc.htm
 *
 *  Copyright (C) 2004-2010   Ilja.Slepnev@jinr.ru
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
*/

#include <QtCore/QtGlobal>
//#include <QDebug>
#include <QString>
#include <QStringList>

#include <cerrno>
#include <cstdio>
#include <cstring>

#include "daqcommon/ansi_col.h"
#include "hptdc/hptdc.h"

int HPTDC_STATUS_ERROR_MASK = 0x3A5;

namespace {

enum {
    TDC_RES_25 = 0,
    TDC_RES_100 = 1,
    TDC_RES_200 = 2,
    TDC_RES_400 = 3,
    TDC_RES_800 = 4,
    TDC_RES_1600 = 5
};

const double clock40 = 40e6;
const double tdc_25ps   = 1. / clock40 / 1024.;
const double tdc_100ps  = 1. / clock40 / 256.;
const double tdc_200ps  = 1. / clock40 / 128.;
const double tdc_400ps  = 1. / clock40 / 64.;
const double tdc_800ps  = 1. / clock40 / 32.;
const double tdc_1600ps = 1. / clock40 / 16.;
const double tdc_tbin[6] = { tdc_25ps, tdc_100ps, tdc_200ps, tdc_400ps, tdc_800ps, tdc_1600ps };
const int tdc_fbins[6] = { 1024, 256, 128, 64, 32, 32 };

// ignore "control_parity", "jtag_instr"  error
//#define HPTDC_STATUS_WARNING_MASK 0x1F7

unsigned int check_parity_buf (const void *buf, unsigned int bits)
{
    unsigned int p = 0;
    for (unsigned int i = 0; i < bits; i++)
        if (((unsigned char*)buf)[i >> 3] & (1 << (i & 7)))
            p++;
    return (p & 1);
}
}

void hptdc_set_setup_defaults(hptdc_setup_t * setup)
{
    Q_ASSERT(setup);

    setup->bit.test_select = 0x02;
    setup->bit.enable_error_mark = 1; // 1
    setup->bit.enable_error_bypass = 0;
    //  setup->bit.enable_error = 0x1F7;  // 0x1F7 - ignore "control_parity", "jtag_instr"  error
    //  setup->bit.enable_error = 0x7B7;  // ignore l1_parity, readout_parity
    //  setup->bit.enable_error = 0x7FF;  // 0x7FF - all errors
    //  setup->bit.enable_error = 0x7FF;  // 0x7FF - all errors
    //  setup->bit.enable_error = 0x0;  // 0x7FF - all errors
    status_error_t enerr;
    enerr.all = 0;
    enerr.bit.jtag_ir_parity = 0;
    enerr.bit.control_parity = 1;
    enerr.bit.setup_parity = 1;
    enerr.bit.readout_err = 1;
    enerr.bit.read_fifo_parity = 0;
    enerr.bit.trigger_state_err = 1;
    enerr.bit.trig_parity = 0;
    enerr.bit.l1_parity = 0;
    enerr.bit.synch = 1;
    enerr.bit.coarse = 0;
    enerr.bit.vernier = 1;

    setup->bit.enable_error = enerr.all;
    setup->bit.readout_single_cycle_speed = 0;
    setup->bit.serial_delay = 0;
    setup->bit.strobe_select = 0;
    setup->bit.readout_speed_select = 0;
    setup->bit.token_delay = 1;
    setup->bit.enable_local_trailer = 1; // 1
    setup->bit.enable_local_header = 1;  // 1
    setup->bit.enable_global_trailer = 0;
    setup->bit.enable_global_header = 0;
    setup->bit.keep_token = 1;
    setup->bit.master = 0;
    setup->bit.enable_bytewise = 0;
    setup->bit.enable_serial = 0;
    setup->bit.enable_jtag_readout = 0;
    setup->bit.tdc_id = 0;
    setup->bit.select_bypass_inputs = 0;
    setup->bit.readout_fifo_size = 7; // 7
    setup->bit.match_window = 43; // 1.1 us
    setup->bit.search_window = setup->bit.match_window + 8;
    setup->bit.trigger_count_offset = 0xFFF - 40 + 1;   // 1us
    setup->bit.reject_count_offset = setup->bit.trigger_count_offset - 4;
    setup->bit.leading_resolution = 0;   // 0 - 100ps
    setup->bit.fixed_pattern = 0x5555555u;  // 28 bits
    setup->bit.enable_fixed_pattern = 0;
    setup->bit.max_event_size = 9; // 9==no_limit, 8=128, 7=64, ... 1=1, 0=0
    setup->bit.reject_readout_fifo_full = 0;   // 0=back_propagate
    setup->bit.enable_readout_occupancy = 0;
    setup->bit.enable_readout_separator = 1;   // debug
    setup->bit.enable_overflow_detect = 1;  // should always be 1
    setup->bit.enable_relative = 0;   // only valid if trigger_matching
    setup->bit.enable_automatic_reject = 1; // should be 1 if trigger_matching
    setup->bit.event_count_offset = 0;
    setup->bit.enable_set_counters_on_bunch_reset = 1;  // 1
    setup->bit.enable_master_reset_code = 0;   // must be 0
    setup->bit.enable_master_reset_on_event_reset = 0;
    setup->bit.enable_reset_channel_buffer_when_separator = 0;
    setup->bit.enable_separator_on_event_reset = 1;  //0
    setup->bit.enable_separator_on_bunch_reset = 1;  //0
    setup->bit.enable_direct_event_reset = 1;
    setup->bit.enable_direct_bunch_reset = 1;
    setup->bit.enable_direct_trigger = 1;   // 1
    setup->bit.offset_00 = 0;
    setup->bit.offset_01 = 0;
    setup->bit.offset_02 = 0;
    setup->bit.offset_03 = 0;
    setup->bit.offset_04 = 0;
    setup->bit.offset_05 = 0;
    setup->bit.offset_06 = 0;
    setup->bit.offset_07 = 0;
    setup->bit.offset_08 = 0;
    setup->bit.offset_09 = 0;
    setup->bit.offset_10 = 0;
    setup->bit.offset_11 = 0;
    setup->bit.offset_12 = 0;
    setup->bit.offset_13 = 0;
    setup->bit.offset_14 = 0;
    setup->bit.offset_15 = 0;
    setup->bit.offset_16 = 0;
    setup->bit.offset_17 = 0;
    setup->bit.offset_18 = 0;
    setup->bit.offset_19 = 0;
    setup->bit.offset_20 = 0;
    setup->bit.offset_21 = 0;
    setup->bit.offset_22 = 0;
    setup->bit.offset_23 = 0;
    setup->bit.offset_24 = 0;
    setup->bit.offset_25 = 0;
    setup->bit.offset_26 = 0;
    setup->bit.offset_27 = 0;
    setup->bit.offset_28 = 0;
    setup->bit.offset_29 = 0;
    setup->bit.offset_30 = 0;
    setup->bit.offset_31 = 0;
    setup->bit.coarse_count_offset = 0;
    setup->bit.dll_tap_adjust_00 = 0;
    setup->bit.dll_tap_adjust_01 = 0;
    setup->bit.dll_tap_adjust_02 = 0;
    setup->bit.dll_tap_adjust_03 = 0;
    setup->bit.dll_tap_adjust_04 = 1;
    setup->bit.dll_tap_adjust_05 = 1;
    setup->bit.dll_tap_adjust_06 = 1;
    setup->bit.dll_tap_adjust_07 = 1;
    setup->bit.dll_tap_adjust_08 = 2;
    setup->bit.dll_tap_adjust_09 = 2;
    setup->bit.dll_tap_adjust_10 = 2;
    setup->bit.dll_tap_adjust_11 = 2;
    setup->bit.dll_tap_adjust_12 = 3;
    setup->bit.dll_tap_adjust_13 = 3;
    setup->bit.dll_tap_adjust_14 = 3;
    setup->bit.dll_tap_adjust_15 = 3;
    setup->bit.dll_tap_adjust_16 = 4;
    setup->bit.dll_tap_adjust_17 = 4;
    setup->bit.dll_tap_adjust_18 = 4;
    setup->bit.dll_tap_adjust_19 = 4;
    setup->bit.dll_tap_adjust_20 = 5;
    setup->bit.dll_tap_adjust_21 = 5;
    setup->bit.dll_tap_adjust_22 = 5;
    setup->bit.dll_tap_adjust_23 = 5;
    setup->bit.dll_tap_adjust_24 = 6;
    setup->bit.dll_tap_adjust_25 = 6;
    setup->bit.dll_tap_adjust_26 = 6;
    setup->bit.dll_tap_adjust_27 = 6;
    setup->bit.dll_tap_adjust_28 = 7;
    setup->bit.dll_tap_adjust_29 = 7;
    setup->bit.dll_tap_adjust_30 = 7;
    setup->bit.dll_tap_adjust_31 = 7;
    setup->bit.rc_adjust_0 = 7;
    setup->bit.rc_adjust_1 = 7;
    setup->bit.rc_adjust_2 = 4;
    setup->bit.rc_adjust_3 = 2;
    setup->bit.rc_adjust_4 = 0;   // not used
    setup->bit.low_power_mode = 1;
    setup->bit.width_select = 0x0A;   // 100ns
    setup->bit.vernier_offset = 0;
    setup->bit.dll_control = 1;
    setup->bit.dead_time = 0;     // 0=5ns, 1=10ns, 2=30ns, 3=100ns
    setup->bit.test_invert = 0;
    setup->bit.test_mode = 0;
    setup->bit.enable_trailing = 0;
    setup->bit.enable_leading = 1;
    setup->bit.mode_rc_compression = 0;
    setup->bit.mode_rc = 0;
    setup->bit.dll_mode = 2;      // 1 - med res
    setup->bit.pll_control = 0x04; // 0x04 in ver 1.2, 0x02 in 1.1
    setup->bit.serial_clock_delay = 0;
    setup->bit.io_clock_delay = 0;
    setup->bit.core_clock_delay = 0;
    setup->bit.dll_clock_delay = 0;
    setup->bit.serial_clock_source = 0;
    setup->bit.io_clock_source = 0;
    setup->bit.core_clock_source = 2; // 0->40MHz, 1->80MHz, 2->160MHz?
    setup->bit.dll_clock_source = 3;  // 2 - med res
    setup->bit.roll_over = 4095;
    setup->bit.enable_matching = 0;
    setup->bit.enable_pair = 0;
    setup->bit.enable_ttl_serial = 1;
    setup->bit.enable_ttl_control = 1;
    setup->bit.enable_ttl_reset = 1;
    setup->bit.enable_ttl_clock = 1;
    setup->bit.enable_ttl_hit = 1;
    setup->bit.parity = check_parity_buf (setup->all, TDC_SETUP_BITS - 1);
}

int hptdc_set_control_defaults (hptdc_control_t * control)
{
    Q_ASSERT(control);

    control->bit.enable_pattern = 0x5;
    control->bit.Global_reset = 0;
    control->bit.enable_channel = 0xFFFFFFFFu;
    control->bit.dll_reset = 0;
    control->bit.pll_reset = 0;
    control->bit.parity = (check_parity_buf (control->all, TDC_CONTROL_BITS - 1));
    return 0;
}

//void hptdc_print_status_bits (const hptdc_status_t &status)
//{
//    QString str = "STATUS = ";
//    for (int i = 0; i < TDC_STATUS_BITS; i++) {
//        str += ((status.all[i >> 3] >> (i % 8)) & 1)?"1":"0";
//    }
//    qDebug() << str;
//}

QString hptdc_status_error_toString (int err)
{
    return QString("err=%1 (%2%3%4%5%6%7%8%9%10%11%12)")
            .arg(err, 3, 16, QChar('0'))
            .arg((err & 0x1)?"vernier ":"")
            .arg((err & 0x2)?"coarse ":"")
            .arg((err & 0x4)?"ch_sel ":"")
            .arg((err & 0x8)?"L1_par ":"")
            .arg((err & 0x10)?"trig_fifo_par ":"")
            .arg((err & 0x20)?"trig_match ":"")
            .arg((err & 0x40)?"rdout_fifo_par ":"")
            .arg((err & 0x80)?"rdout_state ":"")
            .arg((err & 0x100)?"setup_par ":"")
            .arg((err & 0x200)?"ctrl_par ":"")
            .arg((err & 0x400)?"jtag_instr ":"")
            ;
}
/*
void hptdc_print_status (const hptdc_status_t &tdc_status)
{
    printf("err=%s%03x", tdc_status.bit.error ? ANSI_RED : ANSI_NORM, tdc_status.bit.error);
    if (tdc_status.bit.error) {
        printf("%s", hptdc_status_error_toString(tdc_status.bit.error).toLocal8Bit());
    }
    printf("%s", ANSI_NORM);
    printf(" tkn=%d", tdc_status.bit.have_token);
    printf(" rdout:%02x", tdc_status.bit.readout_fifo_occupancy);
    printf("%s", tdc_status.bit.readout_fifo_full ? "F" : "_");
    printf("%s", tdc_status.bit.readout_fifo_empty ? "e" : "_");
    printf(" L1:%02x", tdc_status.bit.group3_l1_occupancy);
    printf(",%02x", tdc_status.bit.group2_l1_occupancy);
    printf(",%02x", tdc_status.bit.group1_l1_occupancy);
    printf(",%02x", tdc_status.bit.group0_l1_occupancy);
    printf(" trig:%01x", tdc_status.bit.trigger_fifo_occupancy);
    printf("%s", tdc_status.bit.trigger_fifo_full ? "F" : "_");
    printf("%s", tdc_status.bit.trigger_fifo_empty ? "e" : "_");
    printf(" dll=%s%d", tdc_status.bit.dll_lock ? ANSI_NORM : ANSI_RED, tdc_status.bit.dll_lock);
    printf("%s\n", ANSI_NORM);
}
*/
void hptdc_print_control (const hptdc_control_t &ctrl)
{
    printf("--- CONTROL ---\n");
    printf("dll_reset: %d\n", ctrl.bit.dll_reset);
    printf("pll_reset: %d\n", ctrl.bit.pll_reset);
    printf("Global_reset: %d\n", ctrl.bit.Global_reset);
    printf("enable_pattern: 0x%X\n", ctrl.bit.enable_pattern);
    printf("enable_channel: 0x%08X\n", ctrl.bit.enable_channel);
}

void hptdc_print_setup (const hptdc_setup_t &setup)
{
    printf("--- SETUP ---\n");
    printf("test_select: 0x%X\n", setup.bit.test_select);
    printf("enable_error_mark: %d\n", setup.bit.enable_error_mark);
    printf("enable_error_bypass: %d\n", setup.bit.enable_error_bypass);
    printf("enable_error: 0x%03X\n", setup.bit.enable_error);
    printf("readout_single_cycle_speed: %d\n", setup.bit.readout_single_cycle_speed);
    printf("serial_delay: 0x%X\n", setup.bit.serial_delay);
    printf("strobe_select: %d\n", setup.bit.strobe_select);
    printf("readout_speed_select: %d\n", setup.bit.readout_speed_select);
    printf("token_delay: 0x%X\n", setup.bit.token_delay);
    printf("enable_local_trailer: %d\n", setup.bit.enable_local_trailer);
    printf("enable_local_header: %d\n", setup.bit.enable_local_header);
    printf("enable_global_trailer: %d\n", setup.bit.enable_global_trailer);
    printf("enable_global_header: %d\n", setup.bit.enable_global_header);
    printf("keep_token: %d\n", setup.bit.keep_token);
    printf("master: %d\n", setup.bit.master);
    printf("enable_bytewise: %d\n", setup.bit.enable_bytewise);
    printf("enable_serial: %d\n", setup.bit.enable_serial);
    printf("enable_jtag_readout: %d\n", setup.bit.enable_jtag_readout);
    printf("tdc_id: 0x%X\n", setup.bit.tdc_id);
    printf("select_bypass_inputs: %d\n", setup.bit.select_bypass_inputs);
    printf("readout_fifo_size: %d\n", setup.bit.readout_fifo_size);
    printf("reject_count_offset: 0x%03X\n", setup.bit.reject_count_offset);
    printf("search_window: 0x%03X\n", setup.bit.search_window);
    printf("match_window: 0x%03X\n", setup.bit.match_window);
    printf("leading_resolution: %d\n", setup.bit.leading_resolution);
    printf("fixed_pattern: 0x%07X\n", setup.bit.fixed_pattern);
    printf("enable_fixed_pattern: %d\n", setup.bit.enable_fixed_pattern);
    printf("max_event_size: 0x%X\n", setup.bit.max_event_size);
    printf("reject_readout_fifo_full: %d\n", setup.bit.reject_readout_fifo_full);
    printf("enable_readout_occupancy: %d\n", setup.bit.enable_readout_occupancy);
    printf("enable_readout_separator: %d\n", setup.bit.enable_readout_separator);
    printf("enable_overflow_detect: %d\n", setup.bit.enable_overflow_detect);
    printf("enable_relative: %d\n", setup.bit.enable_relative);
    printf("enable_automatic_reject: %d\n", setup.bit.enable_automatic_reject);
    printf("event_count_offset: 0x%03X\n", setup.bit.event_count_offset);
    printf("trigger_count_offset: 0x%03X\n", setup.bit.trigger_count_offset);
    printf("enable_set_counters_on_bunch_reset: %d\n", setup.bit.enable_set_counters_on_bunch_reset);
    printf("enable_master_reset_code: %d\n", setup.bit.enable_master_reset_code);
    printf("enable_master_reset_on_event_reset: %d\n", setup.bit.enable_master_reset_on_event_reset);
    printf("enable_reset_channel_buffer_when_separator: %d\n", setup.bit.enable_reset_channel_buffer_when_separator);
    printf("enable_separator_on_event_reset: %d\n", setup.bit.enable_separator_on_event_reset);
    printf("enable_separator_on_bunch_reset: %d\n", setup.bit.enable_separator_on_bunch_reset);
    printf("enable_direct_event_reset: %d\n", setup.bit.enable_direct_event_reset);
    printf("enable_direct_bunch_reset: %d\n", setup.bit.enable_direct_bunch_reset);
    printf("enable_direct_trigger: %d\n", setup.bit.enable_direct_trigger);
    printf("offset_31: 0x%03X\n", setup.bit.offset_31);
    printf("offset_30: 0x%03X\n", setup.bit.offset_30);
    printf("offset_29: 0x%03X\n", setup.bit.offset_29);
    printf("offset_28: 0x%03X\n", setup.bit.offset_28);
    printf("offset_27: 0x%03X\n", setup.bit.offset_27);
    printf("offset_26: 0x%03X\n", setup.bit.offset_26);
    printf("offset_25: 0x%03X\n", setup.bit.offset_25);
    printf("offset_24: 0x%03X\n", setup.bit.offset_24);
    printf("offset_23: 0x%03X\n", setup.bit.offset_23);
    printf("offset_22: 0x%03X\n", setup.bit.offset_22);
    printf("offset_21: 0x%03X\n", setup.bit.offset_21);
    printf("offset_20: 0x%03X\n", setup.bit.offset_20);
    printf("offset_19: 0x%03X\n", setup.bit.offset_19);
    printf("offset_18: 0x%03X\n", setup.bit.offset_18);
    printf("offset_17: 0x%03X\n", setup.bit.offset_17);
    printf("offset_16: 0x%03X\n", setup.bit.offset_16);
    printf("offset_15: 0x%03X\n", setup.bit.offset_15);
    printf("offset_14: 0x%03X\n", setup.bit.offset_14);
    printf("offset_13: 0x%03X\n", setup.bit.offset_13);
    printf("offset_12: 0x%03X\n", setup.bit.offset_12);
    printf("offset_11: 0x%03X\n", setup.bit.offset_11);
    printf("offset_10: 0x%03X\n", setup.bit.offset_10);
    printf("offset_09: 0x%03X\n", setup.bit.offset_09);
    printf("offset_08: 0x%03X\n", setup.bit.offset_08);
    printf("offset_07: 0x%03X\n", setup.bit.offset_07);
    printf("offset_06: 0x%03X\n", setup.bit.offset_06);
    printf("offset_05: 0x%03X\n", setup.bit.offset_05);
    printf("offset_04: 0x%03X\n", setup.bit.offset_04);
    printf("offset_03: 0x%03X\n", setup.bit.offset_03);
    printf("offset_02: 0x%03X\n", setup.bit.offset_02);
    printf("offset_01: 0x%03X\n", setup.bit.offset_01);
    printf("offset_00: 0x%03X\n", setup.bit.offset_00);
    printf("coarse_count_offset: 0x%03X\n", setup.bit.coarse_count_offset);
    printf("dll_tap_adjust_00: %d\n", setup.bit.dll_tap_adjust_00);
    printf("dll_tap_adjust_01: %d\n", setup.bit.dll_tap_adjust_01);
    printf("dll_tap_adjust_02: %d\n", setup.bit.dll_tap_adjust_02);
    printf("dll_tap_adjust_03: %d\n", setup.bit.dll_tap_adjust_03);
    printf("dll_tap_adjust_04: %d\n", setup.bit.dll_tap_adjust_04);
    printf("dll_tap_adjust_05: %d\n", setup.bit.dll_tap_adjust_05);
    printf("dll_tap_adjust_06: %d\n", setup.bit.dll_tap_adjust_06);
    printf("dll_tap_adjust_07: %d\n", setup.bit.dll_tap_adjust_07);
    printf("dll_tap_adjust_08: %d\n", setup.bit.dll_tap_adjust_08);
    printf("dll_tap_adjust_09: %d\n", setup.bit.dll_tap_adjust_09);
    printf("dll_tap_adjust_10: %d\n", setup.bit.dll_tap_adjust_10);
    printf("dll_tap_adjust_11: %d\n", setup.bit.dll_tap_adjust_11);
    printf("dll_tap_adjust_12: %d\n", setup.bit.dll_tap_adjust_12);
    printf("dll_tap_adjust_13: %d\n", setup.bit.dll_tap_adjust_13);
    printf("dll_tap_adjust_14: %d\n", setup.bit.dll_tap_adjust_14);
    printf("dll_tap_adjust_15: %d\n", setup.bit.dll_tap_adjust_15);
    printf("dll_tap_adjust_16: %d\n", setup.bit.dll_tap_adjust_16);
    printf("dll_tap_adjust_17: %d\n", setup.bit.dll_tap_adjust_17);
    printf("dll_tap_adjust_18: %d\n", setup.bit.dll_tap_adjust_18);
    printf("dll_tap_adjust_19: %d\n", setup.bit.dll_tap_adjust_19);
    printf("dll_tap_adjust_20: %d\n", setup.bit.dll_tap_adjust_20);
    printf("dll_tap_adjust_21: %d\n", setup.bit.dll_tap_adjust_21);
    printf("dll_tap_adjust_22: %d\n", setup.bit.dll_tap_adjust_22);
    printf("dll_tap_adjust_23: %d\n", setup.bit.dll_tap_adjust_23);
    printf("dll_tap_adjust_24: %d\n", setup.bit.dll_tap_adjust_24);
    printf("dll_tap_adjust_25: %d\n", setup.bit.dll_tap_adjust_25);
    printf("dll_tap_adjust_26: %d\n", setup.bit.dll_tap_adjust_26);
    printf("dll_tap_adjust_27: %d\n", setup.bit.dll_tap_adjust_27);
    printf("dll_tap_adjust_28: %d\n", setup.bit.dll_tap_adjust_28);
    printf("dll_tap_adjust_29: %d\n", setup.bit.dll_tap_adjust_29);
    printf("dll_tap_adjust_30: %d\n", setup.bit.dll_tap_adjust_30);
    printf("dll_tap_adjust_31: %d\n", setup.bit.dll_tap_adjust_31);
    printf("rc_adjust_0: %d\n", setup.bit.rc_adjust_0);
    printf("rc_adjust_1: %d\n", setup.bit.rc_adjust_1);
    printf("rc_adjust_2: %d\n", setup.bit.rc_adjust_2);
    printf("rc_adjust_3: %d\n", setup.bit.rc_adjust_3);
    printf("rc_adjust_4: %d\n", setup.bit.rc_adjust_4);
    printf("low_power_mode: %d\n", setup.bit.low_power_mode);
    printf("width_select: 0x%X\n", setup.bit.width_select);
    printf("vernier_offset: 0x%02X\n", setup.bit.vernier_offset);
    printf("dll_control: 0x%X\n", setup.bit.dll_control);
    printf("dead_time: %d\n", setup.bit.dead_time);
    printf("test_invert: %d\n", setup.bit.test_invert);
    printf("test_mode: %d\n", setup.bit.test_mode);
    printf("enable_trailing: %d\n", setup.bit.enable_trailing);
    printf("enable_leading: %d\n", setup.bit.enable_leading);
    printf("mode_rc_compression: %d\n", setup.bit.mode_rc_compression);
    printf("mode_rc: %d\n", setup.bit.mode_rc);
    printf("dll_mode: %d\n", setup.bit.dll_mode);
    printf("pll_control: 0x%02X\n", setup.bit.pll_control);
    printf("serial_clock_delay: 0x%X\n", setup.bit.serial_clock_delay);
    printf("io_clock_delay: 0x%X\n", setup.bit.io_clock_delay);
    printf("core_clock_delay: 0x%X\n", setup.bit.core_clock_delay);
    printf("dll_clock_delay: 0x%X\n", setup.bit.dll_clock_delay);
    printf("serial_clock_source: %d\n", setup.bit.serial_clock_source);
    printf("io_clock_source: %d\n", setup.bit.io_clock_source);
    printf("core_clock_source: %d\n", setup.bit.core_clock_source);
    printf("dll_clock_source: %d\n", setup.bit.dll_clock_source);
    printf("roll_over: 0x%03X\n", setup.bit.roll_over);
    printf("enable_matching: %d\n", setup.bit.enable_matching);
    printf("enable_pair: %d\n", setup.bit.enable_pair);
    printf("enable_ttl_serial: %d\n", setup.bit.enable_ttl_serial);
    printf("enable_ttl_control: %d\n", setup.bit.enable_ttl_control);
    printf("enable_ttl_reset: %d\n", setup.bit.enable_ttl_reset);
    printf("enable_ttl_clock: %d\n", setup.bit.enable_ttl_clock);
    printf("enable_ttl_hit: %d\n", setup.bit.enable_ttl_hit);
    printf("parity: %d (%s)\n", setup.bit.parity, (setup.bit.parity == check_parity_buf (setup.all, TDC_SETUP_BITS - 1)) ? "Ok" : "BAD");
}

void hptdc_print_brief_setup (const hptdc_setup_t &setup)
{
    printf("--- HPTDC SETUP ---\n");
    printf("enable_error: 0x%03X\n", setup.bit.enable_error);
    printf("enable_local_trailer: %d\n", setup.bit.enable_local_trailer);
    printf("enable_local_header: %d\n", setup.bit.enable_local_header);
    printf("tdc_id: 0x%X\n", setup.bit.tdc_id);
    printf("reject_count_offset: 0x%03X\n", setup.bit.reject_count_offset);
    printf("search_window: 0x%03X\n", setup.bit.search_window);
    printf("match_window: 0x%03X\n", setup.bit.match_window);
    printf("leading_resolution: %d\n", setup.bit.leading_resolution);
    printf("max_event_size: 0x%X\n", setup.bit.max_event_size);
    printf("event_count_offset: 0x%03X\n", setup.bit.event_count_offset);
    printf("trigger_count_offset: 0x%03X\n", setup.bit.trigger_count_offset);
    printf("coarse_count_offset: 0x%03X\n", setup.bit.coarse_count_offset);
    printf("width_select: 0x%X\n", setup.bit.width_select);
    printf("vernier_offset: 0x%02X\n", setup.bit.vernier_offset);
    printf("enable_trailing: %d\n", setup.bit.enable_trailing);
    printf("enable_leading: %d\n", setup.bit.enable_leading);
    printf("enable_pair: %d\n", setup.bit.enable_pair);
    printf("mode_rc_compression: %d\n", setup.bit.mode_rc_compression);
    printf("mode_rc: %d\n", setup.bit.mode_rc);
    printf("dll_mode: %d\n", setup.bit.dll_mode);
    printf("dll_clock_source: %d\n", setup.bit.dll_clock_source);
    printf("roll_over: 0x%03X\n", setup.bit.roll_over);
    printf("enable_pair: %d\n", setup.bit.enable_pair);
    printf("clock: %s\n", setup.bit.enable_ttl_clock?"TTL":"LVDS");
    printf("hit: %s\n", setup.bit.enable_ttl_hit?"TTL":"LVDS");
    printf("parity: %d (%s)\n", setup.bit.parity, (setup.bit.parity == check_parity_buf (setup.all, TDC_SETUP_BITS - 1)) ? "Ok" : "BAD");
}

void hptdc_print_bsr (const hptdc_bsr_t &bsr)
{
    printf("--- BSR ---\n");
    if (bsr.bit.data_ready /* && bsr.bit.parallel_enable */ )
        printf("parallel_data_out=0x%08x\n", bsr.bit.parallel_data_out);
    //    roc_print_dataword (bsr.bit.parallel_data_out);
    else
        printf("parallel_data_out=0x%08x (invalid)\n", bsr.bit.parallel_data_out);

    printf("token_out=%d ", bsr.bit.token_out);
    printf("strobe_out=%d ", bsr.bit.strobe_out);
    //  printf("serial_out=%d ", bsr.bit.serial_out);
    //  printf("test=%d ", bsr.bit.test);
    printf("error=%d ", bsr.bit.error);
    printf("data_ready=%d ", bsr.bit.data_ready);
    printf("parallel_enable=%d ", bsr.bit.parallel_enable);
    //  printf("encoded_control=%d ", bsr.bit.encoded_control);
    printf("trigger=%d ", bsr.bit.trigger);
    printf("event_reset=%d ", bsr.bit.event_reset);
    printf("bunch_reset=%d ", bsr.bit.bunch_reset);
    printf("get_data=%d ", bsr.bit.get_data);
    //  printf("serial_bypass_in=%d\n", bsr.bit.serial_bypass_in);
    //  printf("serial_in=%d ", bsr.bit.serial_in);
    //  printf("token_bypass_in=%d ", bsr.bit.token_bypass_in);
    printf("token_in=%d ", bsr.bit.token_in);
    printf("reset=%d ", bsr.bit.reset);
    printf("aux_clock=%d ", bsr.bit.aux_clock);
    printf("clk=%d ", bsr.bit.clk);
    printf("hit=0x%08x\n", bsr.bit.hit);
}

int hptdc_bist_finished (const hptdc_bist_t &bist)
{
    return (bist.bit.readout_end && bist.bit.group3_end && bist.bit.group2_end && bist.bit.group1_end && bist.bit.group0_end);
}

int hptdc_bist_failed (const hptdc_bist_t &bist)
{
    return (bist.bit.readout_fail || bist.bit.group3_fail || bist.bit.group2_fail || bist.bit.group1_fail || bist.bit.group0_fail);
}

int hptdc_bist_ok (const hptdc_bist_t &bist)
{
    return (hptdc_bist_finished (bist) && !hptdc_bist_failed (bist));
}

void hptdc_print_bist (const hptdc_bist_t &bist)
{
    int finished, failed;

    finished = hptdc_bist_finished (bist);
    failed = hptdc_bist_failed (bist);;

    printf("--- BIST : %s, %s ---\n", finished ? "finished" : "not ready", failed ? "failed" : "Ok");

    if (!finished || failed)
        printf("readout_end=%d readout_fail=%d g3_end=%d g3_fail=%d g2_end=%d g2_fail=%d g1_end=%d g1_fail=%d g0_end=%d g0_fail=%d\n",
               bist.bit.readout_end,
               bist.bit.readout_fail,
               bist.bit.group3_end,
               bist.bit.group3_fail,
               bist.bit.group2_end,
               bist.bit.group2_fail, bist.bit.group1_end, bist.bit.group1_fail, bist.bit.group0_end, bist.bit.group0_fail);
}

void hptdc_print_coretest (const hptdc_coretest_t &coretest)
{
    printf("--- CORETEST ---\n");
    printf("cms: idle=%d hdr=%d lost_hdr=%d err=%d trail=%d lost_trail=%d sep=%d occ=%d matching=%d\n",
           coretest.bit.cms_idle,
           coretest.bit.cms_header,
           coretest.bit.cms_lost_header,
           coretest.bit.cms_error,
           coretest.bit.cms_trailer,
           coretest.bit.cms_lost_trailer, coretest.bit.cms_separator, coretest.bit.cms_occupancy, coretest.bit.cms_matching);
    printf("trigger: bunch=%d event=%d sep=%d lost=%d par=%d ready=%d\n",
           coretest.bit.bunch_id,
           coretest.bit.event_id, coretest.bit.separator, coretest.bit.trigger_lost, coretest.bit.parity, coretest.bit.trigger_ready);
    printf("\n");
}

void hptdc_print_readout (const hptdc_readout_t &data)
{
    if (data.bit.ready)
        printf("0x%08X\n", data.bit.data);
    else
        printf("  empty   \n");
}

QString hptdc_err_str(int err)
{
    int nerr = 0;
    QStringList sl;
    for (int n = 0; n < 4; n++) {
        if (err & (1 << (n * 3))) {
            sl += QString("gr %1: hit lost: readout overflow").arg(n);
            nerr++;
        }
        if (err & (1 << (n * 3 + 1))) {
            sl += QString("gr %1: hit lost: L1 overflow").arg(n);
            nerr++;
        }
        if (err & (1 << (n * 3 + 2))) {
            sl += QString("gr %1: hit error").arg(n);
            nerr++;
        }
    }
    if (err & (1 << 12)) {
        sl += QString("hits reject: progr limit");
        nerr++;
    }
    if (err & (1 << 13)) {
        sl += QString("event lost: trigger fifo overflow");
        nerr++;
    }
    if (err & (1 << 14)) {
        sl += QString("fatal chip error");
        nerr++;
    }
    return sl.join(", ");
}

void hptdc_print_err (int err)
{
    int n, nerr = 0;;
    for (n = 0; n < 4; n++) {
        if (err & (1 << (n * 3))) {
            printf("%sgr %d: hit lost: readout overflow", nerr ? ", " : "", n);
            nerr++;
        }
        if (err & (1 << (n * 3 + 1))) {
            printf("%sgr %d: hit lost: L1 overflow", nerr ? ", " : "", n);
            nerr++;
        }
        if (err & (1 << (n * 3 + 2))) {
            printf("%sgr %d: hit error", nerr ? ", " : "", n);
            nerr++;
        }
    }
    if (err & (1 << 12)) {
        printf("%shits reject: progr limit", nerr ? ", " : "");
        nerr++;
    }
    if (err & (1 << 13)) {
        printf("%sevent lost: trigger fifo overflow", nerr ? ", " : "");
        nerr++;
    }
    if (err & (1 << 14)) {
        printf("%sfatal chip error", nerr ? ", " : "");
        nerr++;
    }
}

QString hptdc_dataword_str (int d)
{
    QString s;
    //  int tdcid = (d >> 24) & 0xF;
    int code = d >> 28;
    unsigned char ch = (d >> 19) & 0x1F;
    //  s += QString("DATA: %08x ", d);
    //  strncat (str,s,size);
    switch (code) {
    case 0x0:
        s += QString("Bunch ID=%1, EvID=%2").arg(d & 0xFFF).arg((d >> 12) & 0xFFF);
        break;
    case 0x1:
        s += QString("EvID=%1, wordcount=%2").arg((d >> 12) & 0xFFF).arg(d & 0xFFF);
        break;
    case 0x2:
        s += QString("Bunch ID=%1, EvID=%2").arg(d & 0xFFF).arg((d >> 12) & 0xFFF);
        break;
    case 0x3:
        s += QString("EvID=%1, wordcount=%2").arg((d >> 12) & 0xFFF).arg(d & 0xFFF);
        break;
    case 0x4:
        s += QString("ch=%1, coarse=%2 fine=%3 ").arg(ch, 2).arg((d >> 7) & 0xFFF).arg(d & 0x7F);
        break;
    case 0x5:
        s += QString("ch=%1, coarse=%2 fine=%3 ").arg(ch, 2).arg((d >> 7) & 0xFFF).arg(d & 0x7F);
        break;
    case 0x6:
        s += QString("(%1) ").arg(d & 0x7FFF, 4, 16, QChar('0'));
        s += hptdc_err_str(d & 0x7FFF);
        break;
    case 0x7:
        s += QString("debug: ");
        switch ((d >> 20) & 0xF) {
        case 0:
            s += QString("separator Bunch ID=%1").arg(d & 0xFFF);
            break;
        case 1:
            s += QString("L1 group %1 occ.: %2").arg((d >> 8) & 3).arg(d & 0xFF);
            break;
        case 2:
            s += QString("readout fifo occ.: %1, trigger fifo occ.: %2, full_while_matching=%3")
                    .arg(d & 0xFF).arg((d >> 9) & 0xF).arg((d >> 8) & 1);
            break;
        }
        break;
    default:
        s += QString("???");
    }
    return s;
}

void hptdc_print_dataword (int d)
{
    //  int tdcid = (d >> 24) & 0xF;
    int code = d >> 28;
    unsigned char ch = (d >> 19) & 0x1F;
    printf("DATA: %s%08x ", d & 0x80000000 ? ANSI_RED : ANSI_NORM, d);
    switch (code) {
    case 0x0:
        printf("grp hdr: Bunch ID=%03x, EvID=%03x", d & 0xFFF, (d >> 12) & 0xFFF);
        break;
    case 0x1:
        printf("grp trl: EvID=%03x, wordcount=%d", (d >> 12) & 0xFFF, d & 0xFFF);
        break;
    case 0x2:
        printf("tdc hdr: Bunch ID=%03x, EvID=%03x", d & 0xFFF, (d >> 12) & 0xFFF);
        break;
    case 0x3:
        printf("tdc trl: EvID=%03x, wordcount=%d", (d >> 12) & 0xFFF, d & 0xFFF);
        break;
    case 0x4:
        //                              printf("x=%05x",d&0x7FFFF);
        printf("lead  meas: ch=%02d, ", ch);
        printf("coarse=%03x ", (d >> 7) & 0xFFF);
        printf("fine=%02x ", d & 0x7F);
        break;
    case 0x5:
        //        printf("trail meas: ch=%02d, x=%05x", (d >> 19) & 0x1F, d & 0x7FFFF);
        printf("trail meas: ch=%02d, ", ch);
        printf("coarse=%03x ", (d >> 7) & 0xFFF);
        printf("fine=%02x ", d & 0x7F);
        break;
    case 0x6:
        printf("error=%04x", d & 0x7FFF);
        break;
    case 0x7:
        printf("debug: ");
        switch ((d >> 20) & 0xF) {
        case 0:
            printf("separator Bunch ID=%03x", d & 0xFFF);
            break;
        case 1:
            printf("L1 group %d occ.: 0x%02x", (d >> 8) & 3, d & 0xFF);
            break;
        case 2:
            printf("readout fifo occ.: 0x%02x, trigger fifo occ.: %01x, full_while_matching=%d", d & 0xFF, (d >> 9) & 0xF, (d >> 8) & 1);
            break;
        }
        break;
    default:
        printf("???");
    }
    printf("%s", ANSI_NORM);
    printf("\n");
}

void hptdc_setup_set_parity(hptdc_setup_t * setup)
{
    Q_ASSERT(setup);
    setup->bit.parity = (check_parity_buf (setup->all, TDC_SETUP_BITS - 1));
}

void hptdc_control_set_parity (hptdc_control_t * control)
{
    Q_ASSERT(control);
    control->bit.parity = (check_parity_buf (control->all, TDC_CONTROL_BITS - 1));
}

void hptdc_setup_set_id(hptdc_setup_t * setup, int tdcid)
{
    Q_ASSERT(setup);
    setup->bit.tdc_id = tdcid;
}

void hptdc_setup_set_readout(hptdc_setup_t * setup, int readout)
{
    Q_ASSERT(setup);
    switch (readout) {
    case 0:
        setup->bit.enable_jtag_readout = 0;
        break;
    case 1:
        setup->bit.enable_jtag_readout = 1;
        break;
    default:
        break;
    }
}

void hptdc_setup_set_edge(hptdc_setup_t * setup, int edge)
{
    Q_ASSERT(setup);
    switch (edge) {
    case 1:
        setup->bit.enable_leading = 1;
        setup->bit.enable_trailing = 0;
        setup->bit.enable_pair = 0;
        break;
    case 2:
        setup->bit.enable_leading = 0;
        setup->bit.enable_trailing = 1;
        setup->bit.enable_pair = 0;
        break;
    case 3:
        setup->bit.enable_leading = 1;
        setup->bit.enable_trailing = 1;
        setup->bit.enable_pair = 1;
        break;
    default:
        break;
    }
}

void hptdc_setup_set_pattern(hptdc_setup_t * setup, int pattern)
{
    Q_ASSERT(setup);
    setup->bit.fixed_pattern = pattern;
    setup->bit.enable_fixed_pattern = 1;
    setup->bit.enable_matching = 0;
}

//1us window
void hptdc_setup_set_trigmode(hptdc_setup_t * setup)
{
    Q_ASSERT(setup);
    setup->bit.enable_fixed_pattern = 0;
    setup->bit.enable_matching = 1;
    setup->bit.match_window = 12;
    setup->bit.search_window = setup->bit.match_window + 8;
    setup->bit.coarse_count_offset = 0;
    setup->bit.trigger_count_offset = (0xFFF - 14 + 1) & 0xFFF;  // no offset
    setup->bit.reject_count_offset = (setup->bit.trigger_count_offset - 4) & 0xFFF;
    setup->bit.enable_relative = 1;   // only valid if trigger_matching
    setup->bit.enable_automatic_reject = 1; // should be 1 if trigger_matching
    setup->bit.roll_over = 0xFFF; // no roll-over
}

/*
//6.25us window
void hptdc_setup_set_trigmode(hptdc_setup_t *setup)
{
  if (!setup) return;
  setup->bit.enable_fixed_pattern=0;
  setup->bit.enable_matching=1;
  setup->bit.match_window=250;
  setup->bit.search_window=setup->bit.match_window+8;
  setup->bit.coarse_count_offset = 0;
// latency = (coarse_count_offset-trigger_count_offset) % 4096
        setup->bit.trigger_count_offset = (0xFFF-230+1) & 0xFFF; // no offset
        setup->bit.reject_count_offset = (setup->bit.trigger_count_offset-4) & 0xFFF;
        setup->bit.enable_relative = 1; // only valid if trigger_matching
        setup->bit.enable_automatic_reject = 1;  // should be 1 if trigger_matching
  setup->bit.roll_over = 0xFFF; // no roll-over
}
*/
/*
large window
void hptdc_setup_set_trigmode(hptdc_setup_t *setup)
{
  if (!setup) return;
  setup->bit.enable_fixed_pattern=0;
  setup->bit.enable_matching=1;
  setup->bit.match_window=2000; // 50us
  setup->bit.search_window=setup->bit.match_window+8;
  setup->bit.coarse_count_offset = 0;
        setup->bit.trigger_count_offset = (0xFFF-2000+1) & 0xFFF; // no offset
        setup->bit.reject_count_offset = (setup->bit.trigger_count_offset-4) & 0xFFF;
        setup->bit.enable_relative = 1; // only valid if trigger_matching
        setup->bit.enable_automatic_reject = 1;  // should be 1 if trigger_matching
  setup->bit.roll_over = 0xFFF; // no roll-over
}
*/
void hptdc_setup_set_resindex(hptdc_setup_t * setup, TdcResIndex resindex)
{
    Q_ASSERT(setup);

    switch (resindex) {
    case TDC_RES_INDEX_25_UNPACKED: // 25ps RC-mode raw
        setup->bit.mode_rc = 1;
        setup->bit.mode_rc_compression = 0;
        break;
    case TDC_RES_INDEX_25: // 25ps RC-mode interpolated
        setup->bit.mode_rc = 1;
        setup->bit.mode_rc_compression = 1;
        break;
    default:
        setup->bit.mode_rc = 0;
        setup->bit.mode_rc_compression = 0;
    }

    switch (resindex) {
    case TDC_RES_INDEX_100:                      // 100ps
        setup->bit.leading_resolution = 0;
        setup->bit.dll_mode = 2;
        setup->bit.dll_clock_source = 3;
        break;
    case TDC_RES_INDEX_200:                      // 200ps
        setup->bit.leading_resolution = 1;
        setup->bit.dll_mode = 1;    // 0->40MHz, 1->160MHz, 2->320MHz
        setup->bit.dll_clock_source = 2;   // 0->clk_40, 1->pll_40, 2->pll_160, 3->pll_320
        break;
    case TDC_RES_INDEX_400:                      // 400ps
        setup->bit.leading_resolution = 2;
        setup->bit.dll_mode = 1;    // 1 - med res
        setup->bit.dll_clock_source = 2;   // 2 - med res
        break;
    case TDC_RES_INDEX_800:                      // 800ps
        setup->bit.leading_resolution = 3;
        setup->bit.dll_mode = 0;    // 1 - med res
        setup->bit.dll_clock_source = 1;   // 2 - med res
        break;
    case TDC_RES_INDEX_1600:                      // 1600ps
        setup->bit.leading_resolution = 4;
        setup->bit.dll_mode = 0;    // 0 - 40MHz DLL
        setup->bit.dll_clock_source = 1;   // 1 - low res
        break;
    default:
        setup->bit.leading_resolution = 0;
        setup->bit.dll_mode = 2;
        setup->bit.dll_clock_source = 3;
    }
    // don't set width select - done separately
    //  if (setup->bit.enable_pair) {
    //    setup->bit.width_select = setup->bit.leading_resolution;
    //  }
    //  printf("hptdc_setup_set_res: res=%d\n", res);
}

void hptdc_setup_get_resindex (const hptdc_setup_t &setup, TdcResIndex *resindex)
{
    TdcResIndex tmpres = static_cast<TdcResIndex>(setup.bit.leading_resolution);

    if (setup.bit.mode_rc == 1) {
        if (setup.bit.mode_rc_compression == 1)
            tmpres = TDC_RES_INDEX_25;
        else
            tmpres = TDC_RES_INDEX_25_UNPACKED;
    }

    if (resindex)
        *resindex = tmpres;
}

double tdc_res_s (TdcResIndex res_index)
{
    int tdc_res = res_index + 1;
    if (tdc_res < TDC_RES_25)
        tdc_res = TDC_RES_25;
    if (tdc_res > TDC_RES_1600)
        tdc_res = TDC_RES_1600;
    return tdc_tbin[tdc_res];
}

double tdc_res_ns (TdcResIndex res_index)
{
    return (tdc_res_s (res_index) * 1e9);
}

bool tdc_res_vh (TdcResIndex res_index)
{
    return res_index <= TDC_RES_INDEX_25;
}

int tdc_rawfbins (TdcResIndex res_index)
{
    int tdc_res = res_index + 1;
    if (tdc_res < TDC_RES_25)
        tdc_res = TDC_RES_25;
    if (tdc_res > TDC_RES_1600)
        tdc_res = TDC_RES_1600;
    return tdc_fbins[tdc_res];
}

int hptdc_pack_rcadj(const int rcadj[4])
{
    return rcadj[0] | (rcadj[1]<<3) | (rcadj[2]<<6) | (rcadj[3]<<9);
}

void hptdc_unpack_rcadj(int packed, int rcadj[4])
{
    rcadj[0] = packed&7;
    rcadj[1] = (packed>>3)&7;
    rcadj[2] = (packed>>6)&7;
    rcadj[3] = (packed>>9)&7;
}

int decode_tdc_rawtime (int d, TdcResIndex resindex)
{
    int time;
    bool vh = tdc_res_vh(resindex);
    if (vh)
        time = ((d & 0x7FFFF) << 2) | ((d >> 19) & 3);
    else
        time = d & 0x7FFFF;

    //  if (tdcresindex == TDC_RES_INDEX_200)
    //    time >>= 1;
    if (resindex == TDC_RES_INDEX_400)
        time >>= 2;
    if (resindex == TDC_RES_INDEX_800)
        time >>= 2;
    if (resindex == TDC_RES_INDEX_1600)
        time >>= 4;

    return (time);
}

//! time bin is 100ps
int decode_tdc_rawwid (int d, TdcResIndex presindex)
{
    //  int time;
    int wid;

    //  time = d & 0xFFF;
    wid = (d>>12) & 0x7F;
    wid = wid << presindex;

    return (wid);
}
