//
//    Copyright 2003-2011 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "histwidget.h"
#include "ui_histwidget.h"

#include <QAbstractTableModel>
#include <QHeaderView>
#include <QMap>
#include <QDebug>

#include <qwt/qwt_plot_curve.h>

#include <cmath>

#include "daqcommon/dsplib.h"
#include "daqcommon/th1.h"
#include "histplot.h"
#include "decoder/inlcorr.h"

class HistInfoTableModel: public QAbstractTableModel
{
public:
    explicit HistInfoTableModel(QObject * parent = 0) : QAbstractTableModel(parent) {}
    virtual int rowCount(const QModelIndex & parent = QModelIndex()) const { Q_UNUSED(parent); return privData.size(); }
    virtual int columnCount(const QModelIndex & parent = QModelIndex()) const { Q_UNUSED(parent); return 1; }
    virtual QVariant data(const QModelIndex & index, int role = Qt::DisplayRole) const {
        if (role == Qt::DisplayRole) {
            if (index.column() == 0)
                return privData[privData.keys().at(index.row())];
            return QVariant();
        }
        return QVariant();
    }
    virtual QVariant headerData(int section, Qt::Orientation orientation, int role = Qt::DisplayRole ) const {
        if (role == Qt::DisplayRole) {
            if (orientation == Qt::Vertical)
                return privData.keys().at(section);
        }
        return QVariant();
    }
    void setPrivData(const QMap<QString, QString> &p) { privData = p; reset(); }

    QMap<QString, QString> privData;
};

HistWidget::HistWidget(QWidget *parent) :
    QWidget(parent),
    ui(new Ui::HistWidget),
    model(new HistInfoTableModel(this)),
    h(new th1())
{
    ui->setupUi(this);
    ui->tableView->setModel(model);
    ui->tableView->horizontalHeader()->hide();
    ui->splitter->setStretchFactor(0, 50);
    plot_modified = true;
    base = 0;
    scaler = 1;
}

HistWidget::~HistWidget()
{
    delete h;
    delete ui;
}

void HistWidget::setData(const th1 &p)
{
    *h = p;
    plot_modified = true;
    update();
}

void HistWidget::show ()
{
    plot_modified = true;
    QWidget::show();
    update();
}

void HistWidget::update ()
{
    if (!plot_modified) return;

    setCaption(h->getname());

    double xmin = h->get_xmin ();
    double xmax = h->get_xmax ();
    unsigned int nbins = h->get_nbins ();

    if ((nbins == 0) || (xmin == xmax))
        return;

    if (px.size() != nbins)
        px.resize(nbins, 0);
    if (py.size() != nbins)
        py.resize(nbins, 0);

    for (unsigned int i = 0; i < nbins; i++) {
        px[i] = (xmin + 1. * (i+1) / nbins * (xmax - xmin) - base) * scaler;
        py[i] = h->get_bin (i);
    }

    bool isDnl = h->getid().contains("_dnl", Qt::CaseInsensitive);
    if (isDnl) {
        const QVector<double> & ct = InlCorr::makeDNLCorrectionTable(QVector<double>::fromStdVector(py));
        ui->histPlot->setpoints (px, ct.toStdVector());
        ui->histPlot->setCurveStyle(QwtPlotCurve::Lines);
        ui->histPlot->setCurvePen(QPen(Qt::black, 1));
        ui->histPlot->setCurveBrush(Qt::transparent);
    } else {
       ui->histPlot->setpoints (px, py);
       ui->histPlot->setCurveStyle(QwtPlotCurve::Steps);
       ui->histPlot->setCurvePen(QPen(Qt::black, 2));
       ui->histPlot->setCurveBrush(Qt::blue);
    }

    plot_modified = false;

    //  if (!isShown()) return;

    // autorange
    int ar_imin = 0;
    int ar_imax = nbins;
    if (!isDnl) {
        ui->histPlot->autorange();
        ar_imin = ui->histPlot->get_ar_imin();
        ar_imax = ui->histPlot->get_ar_imax();
    }
    double x1 = (xmin + ar_imin*1./nbins * (xmax - xmin) - base) * scaler;
    double x2 = (xmin + (ar_imax+1)*1./nbins * (xmax - xmin) - base) * scaler;
//    qDebug() << "autorange(index): " << ar_imin << ar_imax;
//    qDebug() << "autorange(x): " << x1 << x2;

    QMap<QString, QString> t;
    t[tr("Entries")] = QString::number(h->get_nent());
    t[tr("Over")] = QString::number(h->get_over());
    t[tr("Under")] = QString::number(h->get_under());

    bool perform_fit = enableFit && (h->get_nent () > 10);
    bool fit_ok = false;

    if (perform_fit)
        fit_ok = h->fitgaus (&value, &mean, &sigma, ar_imin, ar_imax);

    if (fit_ok) {            // fit ok
        unsigned int fnbins = ar_imax-ar_imin+1;
        if (fnbins < 250) fnbins = 250;
        if (pfx.size() != fnbins)
            pfx.resize(fnbins, 0);
        if (pfy.size() != fnbins)
            pfy.resize(fnbins, 0);

        for (unsigned int i = 0; i < fnbins; i++) {
            //      pfx[i] = (xmin + (double) (i) / fnbins * (xmax - xmin) - base) * scaler;
            //	  double x = h->bincenter(ar_imin+i);
            double x = x1 + i*1./fnbins * (x2 - x1);
            pfx[i] = x;
            if (sigma!=0)
                pfy[i] = value*gauss(x-mean, sigma);
            else
                pfy[i] = 0;
        }
        /*
        unsigned int fnbins = nbins;
        if (fnbins < 250) fnbins = 250;
        if (pfx.size() != fnbins)
          pfx.resize(fnbins, 0);
        if (pfy.size() != fnbins)
          pfy.resize(fnbins, 0);

    for (unsigned int i = 0; i < fnbins; i++) {
      pfx[i] = (xmin + (double) (i) / fnbins * (xmax - xmin) - base) * scaler;
          double x = pfx[i];
          if (sigma!=0)
                pfy[i] = value*gauss(x+.5-mean, sigma);
          else
                pfy[i] = 0;
        }
*/
        ui->histPlot->set_fit_points (pfx, pfy);

        t[tr("Mean")] = QString::number(mean, 'e', 3);
        t[tr("FWHM")] = QString::number(sigma * GausFWHM, 'e', 3);
        t[tr("RMS")] = QString::number(sigma, 'e', 3);
        if (mean != 0)
            t[tr("Res.")] = QString::number(sigma * GausFWHM / mean * 100., 'f', 3) + "%";
    } else {                      // fit failed
        value=0;mean=0;sigma=0;
        ui->histPlot->hide_fit_points ();
    }
    model->setPrivData(t);
    //  ui->tableView->resizeColumnsToContents();
    //  ui->tableView->resizeRowsToContents();

    ui->histPlot->setTitle(h->getname());
    ui->histPlot->setAxisScale (QwtPlot::xBottom, x1, x2, 0);
    ui->histPlot->setAxisTitle(QwtPlot::xBottom, h->getXLabel()); // FIXME
    ui->histPlot->setAxisTitle(QwtPlot::yLeft, h->getYLabel()); // FIXME

    ui->histPlot->replot ();
}

void HistWidget::setLogY(bool state)
{
    enableLogY = state;
    ui->histPlot->set_log_y(state);
    ui->histPlot->replot ();
    plot_modified = true;
    update();
}

void HistWidget::setShowFit(bool state)
{
    enableFit = state;
    plot_modified = true;
    update();
}

void HistWidget::save_to_file()
{
    if (h)
        h->save_to_file ();
}

void HistWidget::on_checkBoxGaussFit_toggled(bool checked)
{
    setShowFit(checked);
}
