//
//    Copyright 2003-2010 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//


#include <qspinbox.h>
#include <qlabel.h>
#include <qcombobox.h>
#include <qradiobutton.h>
#include <QCheckBox>
#include <q3buttongroup.h>
#include <qdebug.h>

#include "daqcommon/daqexc.h"
#include "hptdc/hptdc.h"
#include "hptdc/hptdc_types.h"
#include "tdcctrlgui.h"
#include "hal/hw_manager.h"

#include "ui_tdcctrlguibase.h"

namespace {
    const mfam_t mf = MF_TDC;
    bool ignore_valueChanged = false;
}

TdcCtrlGui::TdcCtrlGui(QWidget* parent)
    : QWidget(parent),
    ui(new Ui::TdcCtrlGuiBase)
{
    hwf = NULL;
    ui->setupUi(this);
    // Hide Pair mode controls - it is not working properly in HPTDC-1.3
    ui->radioButton_Pair->hide();
    ui->cb_tdc_pres->hide();
    ui->cb_tdc_en_hdr->hide();
    ui->cb_tdc_en_trl->hide();
    ui->frame_RCAdj->hide();

    // Time resolution
    ui->cb_tdc_res->clear();
    ui->cb_tdc_res->insertItem ("25 ps", 0);   // -1
    ui->cb_tdc_res->insertItem ("100 ps", 1);  // 0
    ui->cb_tdc_res->insertItem ("200 ps", 2);  // 1
    ui->cb_tdc_res->insertItem ("400 ps", 3);  // 2
    ui->cb_tdc_res->insertItem ("800 ps", 4);  // 3
    ui->cb_tdc_res->insertItem ("1600 ps", 5); // 4
    // Pair mode width resolution
    ui->cb_tdc_pres->clear();
    ui->cb_tdc_pres->insertItem ("100 ps", 0);
    ui->cb_tdc_pres->insertItem ("200 ps", 1);
    ui->cb_tdc_pres->insertItem ("400 ps", 2);
    ui->cb_tdc_pres->insertItem ("800 ps", 3);
    ui->cb_tdc_pres->insertItem ("1.6 ns", 4);
    ui->cb_tdc_pres->insertItem ("3.2 ns", 5);
    ui->cb_tdc_pres->insertItem ("6.25 ns", 6);
    ui->cb_tdc_pres->insertItem ("12.5 ns", 7);
    ui->cb_tdc_pres->insertItem ("25 ns", 8);
    ui->cb_tdc_pres->insertItem ("50 ns", 9);
    ui->cb_tdc_pres->insertItem ("100 ns", 10);
    ui->cb_tdc_pres->insertItem ("200 ns", 11);
    ui->cb_tdc_pres->insertItem ("400 ns", 12);
    ui->cb_tdc_pres->insertItem ("800 ns", 13);
}

TdcCtrlGui::~TdcCtrlGui()
{
    delete ui;
}

//! Name to GuiControl Multiplexer
void TdcCtrlGui::setValue( const QString &name, const QVariant &val )
{
    int i = val.toInt();
    if (name == "hptdc_min_win")
        ui->sb_tdc_win->setMinValue (i);
    if (name == "hptdc_max_win")
        ui->sb_tdc_win->setMaxValue (i);
    if (name == "hptdc_min_lat")
        ui->sb_tdc_lat->setMinValue (i);
    if (name == "hptdc_max_lat")
        ui->sb_tdc_lat->setMinValue (i);

    if (name == "win")
        ui->sb_tdc_win->setValue (i);
    if (name == "lat")
        ui->sb_tdc_lat->setValue (i);
}

int TdcCtrlGui::get_int(const QString &name)
{
    //  qDebug("TdcCtrlGui::get_int name="+name);
    Q_ASSERT(hwf);
    if (hwf->get_nch_byfam(mf) <= 0) return 0;
    return hwf->get_int(name.toStdString(), mf);
}

void TdcCtrlGui::set_int(const QString &name, int i)
{
    //  qDebug("TdcCtrlGui::set_int name="+name+", val="+QString::number(i));
    if (ignore_valueChanged) return;
    Q_ASSERT(hwf);
    if (!hwf->mf_present(mf)) return;
    try {
        hwf->set_int(i, name.toStdString(), mf);
    }
    catch (exc & e) {
        qWarning() << e.what();
    }
    if (name == "win" || name == "lat")
        updateTextLabels();
}

//! Read hwf and set GUI values
void TdcCtrlGui::readHwf()
{
    if (!isEnabled())
        return;
    //  qDebug("TdcCtrlGui::readHwf");
    ignore_valueChanged = true;
    ui->sb_tdc_win->setMinValue(get_int("hptdc_min_win"));
    ui->sb_tdc_win->setMaxValue(get_int("hptdc_max_win"));
    ui->sb_tdc_win->setValue(get_int("win"));
    ui->sb_tdc_lat->setMinValue(get_int("hptdc_min_lat"));
    ui->sb_tdc_lat->setMaxValue(get_int("hptdc_max_lat"));
    ui->sb_tdc_lat->setValue(get_int("lat"));
    //  ui->cb_tdc_lead->setChecked (get_int("edge") & 1);
    //  ui->cb_tdc_trail->setChecked (get_int("edge") & 2);
    int edge = get_int("edge");
    if (edge < 1 || edge > 3) edge = 1; // default
    if (get_int("pair"))
        edge = 4;
    qDebug() << "TdcCtrlGui::readHwf:" << "edge" << edge;
    switch (edge) {
    case 1 :
        ui->radioButton_Leading->setChecked(true);
        break;
    case 2 :
        ui->radioButton_Trailing->setChecked(true);
        break;
    case 3 :
        ui->radioButton_LeadTrail->setChecked(true);
        break;
    case 4 :
        ui->radioButton_Pair->setChecked(true);
        break;
    }
    //  buttonGroup1->setButton(edge);
    ui->cb_tdc_en_hdr->setChecked (get_int("hdr"));
    ui->cb_tdc_en_trl->setChecked (get_int("trl"));
    ui->cb_tdc_res->setCurrentItem (get_int("res") + 1);
    ui->cb_tdc_pres->setCurrentItem (get_int("pres"));
    int rcadj[5];
    hptdc_unpack_rcadj(get_int("rcadj"), rcadj);
    ui->sb_tdc_rc0->setValue (rcadj[0]);
    ui->sb_tdc_rc1->setValue (rcadj[1]);
    ui->sb_tdc_rc2->setValue (rcadj[2]);
    ui->sb_tdc_rc3->setValue (rcadj[3]);
    updateTextLabels();
    updatePermissions();
    ignore_valueChanged = false;
}

void TdcCtrlGui::updateTextLabels()
{
    //  qDebug("TdcCtrlGui::updateTextLabels");
    ui->tl_tdc_win->setText(QString::number((get_int("win"))*0.025, 'g', 3)+" us");
    ui->tl_tdc_lat->setText(QString::number((get_int("lat"))*0.025, 'g', 3)+" us");
}

void TdcCtrlGui::setHwf(hw_manager *p)
{
    hwf = p;
    updatePermissions();
    readHwf();
}

//! Slots

void TdcCtrlGui::sb_tdc_win_valueChanged( int value )
{
    set_int("win", value);
    Q_EMIT resolutionChanged();
}

void TdcCtrlGui::sb_tdc_lat_valueChanged( int value )
{
    set_int("lat", value);
}

void TdcCtrlGui::chg_tdc_rc( unsigned int n, int value )
{
    if (n>=4) return;
    int rcadj[4];
    hptdc_unpack_rcadj(get_int ("rcadj"), rcadj);
    rcadj[n] = value;
    set_int ("rcadj", hptdc_pack_rcadj(rcadj));
}

void TdcCtrlGui::chg_tdc_rc0( int value )
{
    chg_tdc_rc(0, value);
}

void TdcCtrlGui::chg_tdc_rc1( int value )
{
    chg_tdc_rc(1, value);
}

void TdcCtrlGui::chg_tdc_rc2( int value )
{
    chg_tdc_rc(2, value);
}

void TdcCtrlGui::chg_tdc_rc3( int value )
{
    chg_tdc_rc(3, value);
}

void TdcCtrlGui::cb_tdc_edge_changed( int mask, bool state )
{
    int edge = get_int ("edge");
    if (state)
        edge |= mask;
    else
        edge &= ~mask;
    set_int ("edge", edge);
    set_int ("pair", edge==3);
    Q_EMIT dataFormatChanged();
}

void TdcCtrlGui::cb_tdc_lead_toggled( bool state )
{
    cb_tdc_edge_changed(1, state);
}

void TdcCtrlGui::cb_tdc_trail_toggled( bool state )
{
    cb_tdc_edge_changed(2, state);
}

void TdcCtrlGui::cb_tdc_en_hdr_toggled( bool state )
{
    set_int ("hdr", state);
}

void TdcCtrlGui::cb_tdc_en_trl_toggled( bool state )
{
    set_int ("trl", state);
}

void TdcCtrlGui::cb_tdc_res_activated( int item )
{
    int newres = item - 1;
    set_int ("res", newres);
    Q_EMIT resolutionChanged();
}


void TdcCtrlGui::cb_tdc_pres_activated( int index )
{
    set_int ("pres", index);
    Q_EMIT resolutionChanged();
}

void TdcCtrlGui::setEdge()
{
    int edge=1, pair=0;
    if (ui->radioButton_Leading->isChecked()) {edge=1;pair=0;}
    if (ui->radioButton_Trailing->isChecked()) {edge=2;pair=0;}
    if (ui->radioButton_LeadTrail->isChecked()) {edge=3;pair=0;}
    if (ui->radioButton_Pair->isChecked()) {edge=3;pair=1;}
    //  qDebug() << "TdcCtrlGui::buttonGroup1_clicked:" << "edge" << edge << "pair" << pair;
    set_int ("edge", edge);
    set_int ("pair", pair);
    Q_EMIT dataFormatChanged();
    updatePermissions();
}

void TdcCtrlGui::on_radioButton_Leading_toggled(bool)
{
    setEdge();
}

void TdcCtrlGui::on_radioButton_Trailing_toggled(bool)
{
    setEdge();
}

void TdcCtrlGui::on_radioButton_LeadTrail_toggled(bool)
{
    setEdge();
}

void TdcCtrlGui::on_radioButton_Pair_toggled(bool)
{
    setEdge();
}

void TdcCtrlGui::updatePermissions()
{
    bool enable = false;
    bool pair = false;
    if (hwf && hwf->mf_present(mf)) {
        enable = true;
        pair = get_int ("pair");
    }
    ui->cb_tdc_pres->setEnabled(pair);
    setEnabled(enable);
}
