
#include "frame_hw.h"

#include <QDebug>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/poll.h>
#include <sys/ioctl.h>

#include <ctime>
#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <cerrno>

#ifndef _MSC_VER
#include <zlib.h>
#endif

#include "colorlog/log.h"
#include "mlink.h"
#include "mlink_ioctls.h"
#include "mlink/frame_types.h"
#include "mlink/mlink_frame_header.h"

namespace mlink {

using colorlog::debug_level_ge;

#define MLINK_CHECK_FRAME_CRC 0
#define DEBUG_DUMP_RAW_FRAMES 1
#define DEBUG_DUMP_DECODED_FRAMES 0

// define ML_DUMP in project file to dump M-Link packets

namespace {
// ETIMERECV is not a link error, it is normal
void print_link_error(int ret, const QString &text)
{
    bool is_link_err = (-ETIMESEND == ret || -ETIME == ret || -EREMOTEIO == ret);
    if (is_link_err)
        qWarning() << QString("link error: %1").arg(text);
    else
        qDebug() << QString("%1: %2").arg(ml_strerror(-ret)).arg(text);
}

void print_frame_error(int ret, const QString &text)
{
    bool is_link_err = (-ETIMESEND == ret || -ETIME == ret || -EREMOTEIO == ret);
    if (is_link_err)
        qWarning() << QString("frame error: %1").arg(text);
    else
        qDebug() << QString("%1: %2").arg(ml_strerror(-ret)).arg(text);
}

int ml_fd = -1;
int g_frame_seq = 0;            // global frame sequence number

#ifdef ML_DUMP
FILE *ml_rx_dump_file = NULL;
FILE *ml_tx_dump_file = NULL;
#endif

unsigned int rxb = 0, txb = 0, rxf = 0, txf = 0;

int ml_dma_read (void *buf, size_t nbytes, size_t * fnrecv, int timeout);
int ml_dma_write (const void *buf, size_t nbytes, int timeout);
}

QTextStream &operator<<(QTextStream &stream, const frame_hdr_t &fh)
{
    stream << QString("Frame header: sync %1, type %2, len %3, seq %4, src %5, dst %6")
              .arg(fh.sync, 4, 16, QChar('0'))
              .arg(fh.type, 4, 16, QChar('0'))
              .arg(fh.len, 4, 16, QChar('0'))
              .arg(fh.seq, 4, 16, QChar('0'))
              .arg(fh.src, 4, 16, QChar('0'))
              .arg(fh.dst, 4, 16, QChar('0'))
              ;
    return stream;
}

QDebug &operator<<(QDebug &stream, const frame_hdr_t &fh)
{
    QString str;
    QTextStream ts(&str);
    ts << fh;
    stream << str;
    return stream;
}

const char *ml_strerror(int errnum)
{
    if (errnum == ETIMERECV) return "Receive timeout";
    else if (errnum == ETIMESEND) return "Send timeout";
    else if (errnum == EBADCRC) return "Bad CRC";
    else return strerror(errnum);
}

// Print RAW data (hexadecimal dump)
void ml_print_raw_buffer(const void *ptr, size_t size)
{
    Q_ASSERT(ptr);
    if (!ptr) return;
    const unsigned int *data = (const unsigned int *)ptr;
    qDebug() << QString("Raw dump for %1 bytes (%2 words):").arg(size).arg(size/4);
    size /= 4;
    QString s;
    for (size_t i=0;i<size;i++) {
        if (i%8==0)
            s += QString("  [%1]: ").arg(i, 4, 16, QChar('0'));
        s += QString("%1 ").arg(data[i], 8, 16, QChar('0'));
        if (i%8==7) s += "\n";
    }
    s += "\n";
    qDebug() << s;
}

//--------------------------------------------------------------------------
//
// Low-level device functions
//

int ml_open (const char *devname)
{
    if (!devname || strlen(devname)==0) devname = "/dev/mlink0";

    if (ml_fd != -1)
        ml_close ();
    ml_fd = open (devname, O_RDWR);
    if (ml_fd < 0) {
        ml_fd = -1;
        qWarning() << QString("Can't open '%1': %2").arg(devname).arg(ml_strerror (errno));
        return (-errno);
    }
#ifdef ML_DUMP
    ml_rx_dump_file = fopen("rx.txt", "a");
    fprintf(ml_rx_dump_file, "0 0 50bc\n");
    ml_tx_dump_file = fopen("tx.txt", "a");
    fprintf(ml_tx_dump_file, "0 0 50bc\n");
#endif
    return 0;
}

int ml_close ()
{
    if (ml_fd != -1) {
        close (ml_fd);
        ml_fd = -1;
    }
#ifdef ML_DUMP
    if (ml_rx_dump_file)
        fclose(ml_rx_dump_file);
    if (ml_tx_dump_file)
        fclose(ml_tx_dump_file);
#endif
    return 0;
}

int ml_get_caps (mlink_caps_t *caps)
{
    Q_ASSERT(caps);

    if (ml_fd == -1)
        return (-EINVAL);
    memset(caps, 0, sizeof(*caps));
    if ((ioctl (ml_fd, ML_IOC_GET_CAPS, caps))) {
        qWarning() << QString("ML_IOC_GET_CAPS failed: %1").arg(ml_strerror (errno));
        return (-errno);
    }
    return 0;
}

int ml_get_driver_status (mlink_status_t *st)
{
    Q_ASSERT(st);

    if (ml_fd == -1)
        return (-EINVAL);
    if (!st)
        return (-EINVAL);
    memset(st, 0, sizeof(*st));
    if ((ioctl (ml_fd, ML_IOC_GET_STATUS, st))) {
        qWarning() << QString("ML_IOC_GET_STATUS failed: %1").arg(ml_strerror (errno));
        return (-errno);
    }
    return 0;
}

int ml_get_driver_version(unsigned long *n)
{
    Q_ASSERT(n);

    if (ml_fd == -1)
        return (-EINVAL);
    *n = 0;
    if ((ioctl (ml_fd, ML_IOC_GET_DRIVER_VERSION, n))) {
        qWarning() << QString("ML_IOC_GET_DRIVER_VERSION failed: %1").arg(ml_strerror (errno));
        return (-errno);
    }
    return 0;
}

int ml_get_driver_revision(unsigned long *n)
{
    Q_ASSERT(n);

    if (ml_fd == -1)
        return (-EINVAL);
    *n = 0;
    if ((ioctl (ml_fd, ML_IOC_GET_DRIVER_REVISION, n))) {
        qWarning() << QString("ML_IOC_GET_DRIVER_REVISION failed: %1").arg(ml_strerror (errno));
        return (-errno);
    }
    return 0;
}

int ml_reset ()
{
    //  ssp_omb1_t omb1;
    if (ml_fd == -1)
        return (-EINVAL);
    if ((ioctl (ml_fd, ML_IOC_RESET))) {
        qWarning() << QString("ML_IOC_RESET failed: %1").arg(ml_strerror (errno));
        return (-errno);
    }
    qDebug() << "M-Link interface reset";
    ml_set_ext_loop_mode (0);
    ml_set_jtag_mode(0);
    return 0;
}

int ml_remote_reset ()
{
    if (ml_fd == -1)
        return (-EINVAL);
    if ((ioctl (ml_fd, ML_IOC_REMOTE_RESET))) {
        qWarning() << QString("ML_IOC_REMOTE_RESET failed: %1").arg(ml_strerror (errno));
        return (-errno);
    }
#ifdef ML_DUMP
    fprintf(ml_tx_dump_file, "0 1 f7f7\n0 1 f7f7\n0 1 f7f7\n0 1 f7f7\n0 1 f7f7\n0 0 50bc\n");
#endif
    qDebug() << "M-Link remote reset";
    return 0;
}

namespace {

int ml_dma_read (void *buf, size_t nbytes, size_t * fnrecv, int timeout)
{
    ml_dmaop_t dmaop;
    memset(&dmaop, 0, sizeof(dmaop));
    size_t nrecv;
    if (!buf)
        return (-EINVAL);

    dmaop.size = nbytes;
    dmaop.buf = buf;
    dmaop.timeout = timeout;
    //  fprintf(stderr, "dmaop.buf = %p\n", dmaop.buf);
    if ((ioctl (ml_fd, ML_IOC_DMA_RECV, &dmaop))) {
        //    int prio = LOG_ERR;
        //    if (errno == ETIME)
        //      prio = LOG_WARNING;
        if (errno != ETIME)
            qDebug() << QString("frame receive failed: %1").arg(ml_strerror (errno));
        if (errno == ETIME) errno = ETIMERECV;
        return (-errno);
    }
    nrecv = dmaop.size;
    rxb += nrecv;
    rxf++;
    //    qDebug() << QString("received frame with %1 bytes").arg(nrecv);

    if (fnrecv)
        *fnrecv = nrecv;
#ifdef ML_DUMP
    fprintf(ml_rx_dump_file, "0 1 f7f7\n");
    for(size_t i=0; i<nrecv/4; i++) {
        fprintf(ml_rx_dump_file, "1 0 %02x%02x\n", ((unsigned char *)buf)[4*i+3], ((unsigned char *)buf)[4*i+2]);
        fprintf(ml_rx_dump_file, "1 0 %02x%02x\n", ((unsigned char *)buf)[4*i+1], ((unsigned char *)buf)[4*i]);
    }
    fprintf(ml_rx_dump_file, "0 0 50bc\n");
    fflush(ml_rx_dump_file);
#endif
    return 0;
}

int ml_dma_write (const void *buf, size_t nbytes, int timeout)
{
    ml_dmaop_ro_t dmaop;
    memset(&dmaop, 0, sizeof(dmaop));
    if (!buf)
        return (-EINVAL);

    dmaop.size = nbytes;
    dmaop.buf = buf;
    dmaop.timeout = timeout;
    if ((ioctl (ml_fd, ML_IOC_DMA_SEND, &dmaop))) {
        qWarning() << QString("frame send failed: %1").arg(ml_strerror (errno));
        if (errno == ETIME) errno = ETIMESEND;
        return (-errno);
    }
    txb += nbytes;
    txf++;
#ifdef ML_DUMP
    fprintf(ml_tx_dump_file, "0 1 f7f7\n");
    for(size_t i=0; i<nbytes/4; i++) {
        fprintf(ml_tx_dump_file, "1 0 %02x%02x\n", ((unsigned char *)buf)[4*i+3], ((unsigned char *)buf)[4*i+2]);
        fprintf(ml_tx_dump_file, "1 0 %02x%02x\n", ((unsigned char *)buf)[4*i+1], ((unsigned char *)buf)[4*i]);
    }
    fprintf(ml_tx_dump_file, "0 0 50bc\n");
    fflush(ml_tx_dump_file);
#endif
    //    qDebug() << QString("sent frame with %1 bytes").arg(nbytes);
    return 0;
}
}

//--------------------------------------------------------------------------
//
// Frame operations
//
namespace {
/* Next frame sequence number evaluation */
int next_frame_seq (int s)
{
    return ((s + 1) & 0xFFFF);
}
}
/* Current frame sequence number */
int ml_get_frame_seq ()
{
    return (g_frame_seq);
}

/* Advance frame sequence number */
int ml_adv_frame_seq ()
{
    g_frame_seq = next_frame_seq (g_frame_seq);
    return (g_frame_seq);
}

int ml_make_crc (const void *buf, int len)
{
    uLong crc = crc32(0L, Z_NULL, 0);
    crc = crc32 (crc, (Bytef *)buf, len);
    return crc;
}

int swap_endian (unsigned int *buf, int len)
{
    Q_ASSERT(buf);
    Q_ASSERT(len >= 0);
    int i;

    /*
  for(i=0;i<len;i++) buf[i] =
  ((buf[i] & 0xFF) << 24 ) |
  (((buf[i]>>8) & 0xFF) << 16 ) |
  (((buf[i]>>16) & 0xFF) << 8 ) |
  (((buf[i]>>24)));
*/
    for (i = 0; i < len; i++)
        buf[i] = ((buf[i] & 0xFFFF) << 16) | (((buf[i] >> 16) & 0xFFFF));
    return (0);
}

namespace {
int ml_write_raw_frame (const int *buf, int len, int timeout)
{
    Q_ASSERT(buf);
    Q_ASSERT(len > 0);

#if DEBUG_DUMP_RAW_FRAMES
    {
        if (debug_level_ge(5)) {
            //  printf(" > SEND RAW 0x%08X 0x%08X 0x%08X\n", buf[0], buf[1], buf[2]);
            fprintf (stderr, "\n > SEND RAW:  ");
            for (int i = 0; i < len; i++) {
                fprintf (stderr, "%08X ", buf[i]);
                if (i%8 == 7)
                    fprintf(stderr, "\n              ");
            }
            fprintf (stderr, "\n");
        }
    }
#endif
    //   printf("ml_write_raw_frame: sending %d words...\n", len);
    return ml_dma_write (buf, 4 * len, timeout);
}
}

int ml_send_data_frame (const int *data, size_t len, int type, int src, int dst, int seq, int timeout)
{

    int f[ML_MAX_RAW_FRAME_WORDS];
    frame_hdr_t *fh = (frame_hdr_t *) f;
    int ret = -1;

    //  printf("ml_send_data_frame(%p, %d, %d, %d, %d, %d, %d)\n", data, len, type, src, dst, seq, timeout);
    Q_ASSERT(len <= ML_MAX_FRAME_DATA_WORDS);
    if (len > ML_MAX_FRAME_DATA_WORDS) {
        return (-EINVAL);
    }
    if (!data)
        len = 0;
    type &= 0xFFFF;
    src &= 0xFFFF;
    dst &= 0xFFFF;
    if (seq == -1)
        seq = ml_adv_frame_seq ();
    fh->sync = ML_FRAME_SYNC;
    fh->type = type;
    fh->len = len + 4;
    fh->seq = seq & 0xFFFF;
    fh->dst = dst;
    fh->src = src;

    if (data && len)
        memcpy (&f[3], data, len * 4);
    f[len + 3] = ml_make_crc (f, (len + 3)*4);
#if DEBUG_DUMP_DECODED_FRAMES
    {
        int dl;
        get_debug_level (&dl);
        if (dl >= 3) {
            frame_print (f, len + 4, 0);
        }
    }
#endif
    ret = ml_write_raw_frame (f, len + 4, timeout);
    if (ret)
        return (ret);
    /*
  if (wait)
    ret = wait_dma_complete (0);
*/
    return (ret);
}

int ml_recv_data_frame (int *fdata, size_t *flen, int *ftype, int *fsrc, int *fdst, int *fseq, int timeout)
{
    int rawbuf[ML_RX_BUFFER_WORDS];
    memset(rawbuf, 0, sizeof(rawbuf));
    size_t rawbufsize = sizeof(rawbuf);
    // initialize frameset
    struct frameset_t frameset;// frameset.maxdesc = ML_FRAMESET_NUMDESC;
    //  size_t fsize = 0;
    //  if (flen) fsize = *flen;
    int ret = ml_recv_data_frame_multi(rawbuf, rawbufsize, &frameset, timeout);
    // copy frame data
    for(size_t i=0; i<frameset.count; i++) {
        const struct framedesc_t *fdesc = &frameset.fdesc[i];
        const size_t datawords = fdesc->len - (ML_FRAME_HEADER_WORDS+ML_FRAME_TRAILER_WORDS);
        const int *frame_data = &fdesc->rawdata[ML_FRAME_HEADER_WORDS];
        // check for available destination space
        size_t copywords = 0;
        if (fdata && flen) {
            copywords = datawords;
            if (copywords > *flen) {
                qWarning() << QString("Not enough space for received frame #%1: %2 truncated to %3").arg(i).arg(copywords).arg(*flen);
                copywords = *flen;
            }
            //  	  debug(0, "Copying frame #%zu", i);
            memcpy(fdata, frame_data, copywords*4);
        }
        if (flen) *flen = copywords;
        const frame_hdr_t *fh = (const frame_hdr_t *) fdesc->rawdata;
        if (ftype)
            *ftype = fh->type;
        if (fsrc)
            *fsrc = fh->src;
        if (fdst)
            *fdst = fh->dst;
        if (fseq)
            *fseq = fh->seq;
        // copy only first frame
        break;
    }
    return ret;
}

// fdata is raw buffer with size fsize words. Pointers to frames are returned in framedesc.
int ml_recv_data_frame_multi(int *rawbuf, size_t rawbufsize, struct frameset_t *frameset, int timeout)
{
    frameset->count = 0;
    size_t received_rawbytes = 0;
    int ret = ml_dma_read (rawbuf, rawbufsize, &received_rawbytes, timeout);

    if (ret == -ETIMERECV) return ret;

    if (received_rawbytes == 0) {
        print_link_error(ret, "no received frame available");
        ret = -EREMOTEIO;
        return (ret);
    }

    const size_t received_rawwords = received_rawbytes/4;

    //    qDebug() << QString("Received frameset of %1 words").arg(received_rawwords);

    size_t p = 0;
    size_t frame_count = 0;
    // int frame_warning_count = 0;
    while (p < received_rawwords) {
        const int *raw_frame = &rawbuf[p];
        const frame_hdr_t *fh = (const frame_hdr_t *) raw_frame;
        const size_t rawwords = received_rawwords - p;
        int frame_error = 0;
        int frame_warning = 0;

        int syncval = fh->sync;
        //        int type = fh->type;
        if (syncval != ML_FRAME_SYNC) {
            print_frame_error(ret, QString("Invalid frame #%1: bad sync 0x%2, expected 0x%3")
                              .arg(frame_count).arg(syncval, 4, 16, QChar('0')).arg(ML_FRAME_SYNC, 4, 16, QChar('0')));
            frame_error = 1;
            ml_print_raw_buffer (rawbuf, received_rawbytes);
            break;
        }

        if (rawwords < ML_MIN_RAW_FRAME_WORDS) {
            print_frame_error(ret, QString("Invalid frame #%1 size: %2, minimum is %3")
                              .arg(frame_count).arg(rawwords).arg(ML_MIN_RAW_FRAME_WORDS));
            ml_print_raw_buffer (raw_frame, rawwords*4);
            frame_error = 1;
            break;
        }


        size_t hdrlen = fh->len;
        //        int seq = fh->seq;
        //        int src = fh->src;
        //        int dst = fh->dst;

        size_t rawlen = hdrlen;
        if (rawlen > rawwords) {
            ret = -EREMOTEIO;
            print_frame_error(ret, QString("Invalid frame #%1: header len %2, remaining %3 words, frame truncated")
                              .arg(frame_count).arg(hdrlen).arg(rawwords));
            frame_error = 1;
            rawlen = rawwords;
        }

        if (rawlen < ML_MIN_RAW_FRAME_WORDS) {
            //    ret = -EREMOTEIO;
            print_frame_error(ret, QString("Invalid frame #%1: invalid size %2").arg(frame_count).arg(rawlen));
            //    rawlen = ML_MAX_RAW_FRAME_WORDS;
            frame_error = 1;
            rawlen = ML_MIN_RAW_FRAME_WORDS;
        }

        if (rawlen > ML_MAX_RAW_FRAME_WORDS) {
            //    ret = -EREMOTEIO;
            print_frame_error(ret, QString("Invalid frame #%1: size too large: %2, maximum is %3")
                              .arg(frame_count).arg(rawlen).arg(ML_MAX_RAW_FRAME_WORDS));
            //    rawlen = ML_MAX_RAW_FRAME_WORDS;
            frame_warning++;
        }

#if MLINK_CHECK_FRAME_CRC
        size_t crcoff = rawlen - 1;

        unsigned int crc = raw_frame[crcoff];
        unsigned int calcrc = ml_make_crc(raw_frame, crcoff*4);

        //  if (crc != calcrc && crc != ML_CRC_HACK) {
        if (crc != calcrc) {
            ret = -EBADCRC; /*-EREMOTEIO;*/
            print_frame_error(ret, QString("Invalid frame #%1: bad CRC 0x%2, calculated 0x%3")
                              .arg(frame_count).arg(crc, 8, 16, QChar('0')).arg(calcrc, 8, 16, QChar('0')));
            frame_warning++;
            //    return (ret);
        }
#else
        //        unsigned int crc = 0;
        //        unsigned int calcrc = 0;
#endif

        if (frame_error || frame_warning) {
            qWarning() << QString("Received broken frame #%1:").arg(frame_count) << *fh;
        } else {
            //            qDebug() << QString("Received frame #%1:").arg(frame_count) << *fh;
        }
        int need_raw_dump = 0;
        if ((frame_error||frame_warning) && debug_level_ge(1))
            need_raw_dump = 1;
#if DEBUG_DUMP_RAW_FRAMES
        if (debug_level_ge(5))
            need_raw_dump = 1;
#endif

        if (need_raw_dump) {
            //	debug(0, "Received raw frame #%zu:", frame_count);
            //  	ml_print_raw_buffer (raw_frame, rawwords*4);
            qDebug() << "Received raw data:";
            ml_print_raw_buffer (rawbuf, received_rawbytes);
        }

#if DEBUG_DUMP_DECODED_FRAMES
        if (debug_level_ge(3)) {
            frame_print (raw_frame, rawwords, 1);
        }
#endif

        // fill frame descriptor, if available
        if (frameset->count < ML_FRAMESET_NUMDESC/*frameset->maxdesc*/) {
            frameset->fdesc[frameset->count].rawdata = raw_frame;
            frameset->fdesc[frameset->count].len = rawlen;
            frameset->count++;
        } else {
            qWarning() << "Not enough frame descriptors";
            break;
        }

        // frame_warning_count += frame_warning;
        frame_count++;
        p += hdrlen;
    } // while()
    return 0;
}

// event-driven interface
int ml_get_fd ()
{
    return (ml_fd);
}

int ml_get_rxb ()
{
    return (rxb);
}

int ml_get_txb ()
{
    return (txb);
}

int ml_get_rxf ()
{
    return (rxf);
}

int ml_get_txf ()
{
    return (txf);
}

void ml_clear_rxtx_counters ()
{
    rxb = 0;
    txb = 0;
    rxf = 0;
    txf = 0;
}

int ml_set_jtag_mode (int mode)
{
    int arg = mode;
    if (ml_fd == -1)
        return (-EINVAL);
    if ((ioctl (ml_fd, ML_IOC_LOCAL_JTAG_MODE, &arg))) {
        qWarning() << QString("ML_IOC_LOCAL_JTAG_MODE failed: %1").arg(ml_strerror (errno));
        return (-errno);
    }
    qDebug() << "Internal jtag mode" << (mode?"enabled":"disabled");
    return 0;
}

int ml_set_ext_loop_mode (int mode)
{
    int arg = mode;
    if (ml_fd == -1)
        return (-EINVAL);
    if ((ioctl (ml_fd, ML_IOC_EXT_LOOP_MODE, &arg))) {
        qWarning() << QString("ML_IOC_EXT_LOOP_MODE failed: %1").arg(ml_strerror (errno));
        return (-errno);
    }
    qDebug() << "External loopback" << (mode?"enabled":"disabled");
    return 0;
}

//------------------------------------------------------------------------------
int ml_io_wr (const int *txbuf, int *rxbuf, size_t txlen, size_t rxlen)
{
    int ret, rxtype=0, exprxtype, rxsrc=0, exprxsrc, rxdst=0, exprxdst;
    int txseq, rxseq=0;
    //  int exprxseq;
    const int src = 1, dst = 1;
    const int timeout = 1000; // 1 second
    size_t nrecv=0;
    /*
  if (count > ML_MAX_FRAME_DATA_WORDS) {
        count=ML_MAX_FRAME_DATA_WORDS;
        logm(L_EXP, LOG_WARNING, "too much data to transfer, trunc to %d", count);
  }
*/
    txseq = ml_adv_frame_seq ();
    ret =
            ml_send_data_frame (txbuf, txlen, FRAME_TYPE_LOOPTEST, src, dst, txseq, timeout);
    if (ret) {
        //    print_link_error(ret, "frame send failed");
        return (ret);
    }
    //  printf("Frame sent, data_len=%d\n", txlen);
    nrecv = rxlen;
    ret = ml_recv_data_frame (rxbuf, &nrecv, &rxtype, &rxsrc, &rxdst, &rxseq, timeout);
    if (ret) {
        //    print_link_error(ret, "frame recv failed");
        return (ret);
    }

    exprxtype = FRAME_TYPE_LOOPTEST;
    //  exprxseq = txseq;
    exprxsrc = src;
    exprxdst = dst;
    //  exprxlen = 1;
    if (rxtype != exprxtype) {
        ret = -EREMOTEIO;
        print_frame_error(ret, QString("bad ack frame type: 0x%1, expected 0x%2")
                          .arg(rxtype, 4, 16, QChar('0')).arg(exprxtype, 4, 16, QChar('0')));
        return (ret);
    }
    if (rxsrc != exprxsrc) {
        ret = -EREMOTEIO;
        print_frame_error(ret, QString("bad ack frame src: 0x%1, expected 0x%2")
                          .arg(rxsrc).arg(exprxsrc));
        return (ret);
    }
    if (rxdst != exprxdst) {
        print_frame_error(ret, QString("bad ack frame dst: 0x%1, expected 0x%2")
                          .arg(rxdst).arg(exprxdst));
        ret = -EREMOTEIO;
        return (ret);
    }
    /*
  if (rxseq != exprxseq) {
        ret = -EREMOTEIO;
    frame(ret, "bad ack frame seq: 0x%04X, expected 0x%04X", rxseq, exprxseq);
    return(ret);
  }
*/
    if (nrecv != rxlen) {
        ret = -EREMOTEIO;
        print_frame_error(ret, QString("bad ack data len: 0x%1, expected 0x%2")
                          .arg(nrecv).arg(rxlen));
        return (ret);
    }
    //  log_debug(6, "memcpy: %p, %p, %d", dr, rxbuf, rxlen);
    return (0);
}

int ml_loopback_test (int nloops, unsigned int datasize, int pattern)
{
    int loop;
    int nok = 0, nerr = 0;
    int lasterr = 0;

    if (datasize>ML_MAX_FRAME_DATA_WORDS)
        datasize = ML_MAX_FRAME_DATA_WORDS;

    //printf("MAX_DATA_WORDS=%d\n", MAX_DATA_WORDS);

    for (loop = 0; loop < nloops; loop++) {
        int nwords = rand () % datasize;
        //    nwords = ML_MAX_FRAME_DATA_WORDS;
        //    nwords = 123;
        //    nwords = 251+256-128;
        int sendbuf[nwords];
        int recvbuf[nwords];
        int i, xfer_err = 0;

        int this_pattern = pattern;
        if (pattern == 0)
            this_pattern = 1 + (loop % 3);

        if (this_pattern == 1) {
            for (i=0;i<nwords;i++)
                sendbuf[i] = ((i*2)<<16) | ((i*2+1)&0xFFFF);
        } else if (this_pattern == 2) {
            for (i=0;i<nwords;i++)
                sendbuf[i] = i;
        } else if (this_pattern == 3) {
            for (i=0;i<nwords;i++)
                sendbuf[i] = rand();
        }
        //    fill_buf (sendbuf, nwords);

        int ret = ml_io_wr (sendbuf, recvbuf, nwords, nwords);
        if (ret) lasterr = ret;
        //// HACK
        //      for (i = 0; i < nwords; i++) sendbuf[i] = 0x5a5a5a5a;

        if (ret && ret != -EBADCRC) {
            //      logm (L_ALL, LOG_ERR, "ml_adc_io_wr failed: %s", ml_strerror (-ret));
            //        return (ret);
            xfer_err++;
        } else {
            int differ = memcmp(sendbuf, recvbuf, nwords*4);
            if (differ) {
                //	    if (printerr) {
                //	  	  frame_print (sendbuf, nwords*4, 0);
                //	  	  frame_print (recvbuf, nwords*4, 1);
                //	    }
                unsigned int az=0, ao=0;
                for (i = 0; i < nwords; i++) {
                    if (sendbuf[i] != recvbuf[i]) {
                        qWarning() << QString("[%1]: sent %2, recv %3, xor %4")
                                      .arg(i)
                                      .arg(sendbuf[i], 8, 16, QChar('0'))
                                      .arg(recvbuf[i], 8, 16, QChar('0'))
                                      .arg(sendbuf[i] ^ recvbuf[i], 8, 16, QChar('0'));
                        xfer_err++;
                    }
                    unsigned int id = recvbuf[i];
                    unsigned int od = sendbuf[i];
                    unsigned int x = id ^ od;
                    ao |= id& ~x;
                    az |= ~id& ~x;
                    //      log_debug(0, "Zeroes: %08X, Ones: %08X", az, ao);
                }
                qDebug() << QString("Zeroes: %1, Ones: %2")
                            .arg(az, 8, 16, QChar('0'))
                            .arg(ao, 8, 16, QChar('0'));
            }
        }
        if (xfer_err) {
            nerr++;
            //      printf ("Test failed for data size %d words\n", nwords);
            qWarning() << QString("Test failed for data size %1 words (raw %2 words): %3")
                          .arg(nwords).arg(nwords+4).arg(ml_strerror(-ret));
            break;
        } else {
            nok++;
        }
    }

    if (nerr)
        qWarning() << QString("Loopback test FAILED: %1 Ok, %2 failed cycles").arg(nok).arg(nerr);
    else
        qDebug() << QString("Loopback test Ok: %1 cycles").arg(nok);

    return nerr ? (lasterr) : 0;
}

int ml_get_event (mlink_event_t *ev, int timeout_ms)
{
    if (ml_fd == -1)
        return (-EINVAL);
    ml_poll_ev_op_t op;
    memset(&op, 0, sizeof(op));
    if (ml_fd == -1)
        return (-EINVAL);
    op.timeout_ms = timeout_ms;
    op.mask=0;
    if ((ioctl (ml_fd, ML_IOC_GET_EVENT, &op))) {
        qWarning() << QString("ML_IOC_GET_EVENT failed: %1").arg(ml_strerror (errno));
        return (-errno);
    }
    if (ev) *ev = op.mask;
    return 0;
}

}
